--- Managing members in call groups
-- 
-- @file	cgroups.lua 
-- @author	Vyacheslav Khudan
-- @created 2014/10/03

local channel = channel

module("cgroups", package.seeall)

--- Add member to a call group
-- @param cg_id	 	A call group id using for adding
-- @param member_id	An added member 
-- @return boolean TRUE if member added or FALSE for error 
function add_member (cg_id, member_id)
	if not cg_id or "" == cg_id then
		return false
	end
	
	local member_interface = "Local/" .. member_id .. "@internalcalls"
	app.AddQueueMember(cg_id, member_interface, 1)
	local rv = channel.AQMSTATUS:get() or ""
	if "ADDED" == rv or "MEMBERALREADY" == rv then
		return true
	end 	
	
	return false
end

--- Remove member from the call group
-- @param cg_id	 	A call group id using for removing
-- @param member_id	A removed member 
-- @return boolean TRUE if member removed or FALSE for error
function rm_member (cg_id, member_id)
	if not cg_id or "" == cg_id then
		return false
	end
	
	if not member_id then
		member_id = gSpace["cidNum"]
	end
	
	local member_interface = "Local/" .. member_id .. "@internalcalls"
	app.RemoveQueueMember(cg_id, member_interface)
	local rv = channel.RQMSTATUS:get() or ""
	if "REMOVED" == rv or "NOTINQUEUE" == rv then
		return true
	end
	
	if "NOTDYNAMIC" == rv then
		print(string.format("User %q is not dynamic member of the call group", member_id))
	end
	
	return false
end

--- Count members in the call group
-- @param cg_id	A counted call group
-- @return number of the members in the queue
function count_members(cg_id)
	if not cg_id or "" == cg_id then
		return -1
	end 
	
	local cnt = channel.QUEUE_MEMBER(cg_id, "count"):get() or ""
	return tonumber(cnt) or 0 
end

--- Regrieve list of the members from the queue
-- @param cg_id     A checked call group id
-- @return table with members and their parameters 
function get_members_list(cg_id)
    local rv = {}
    if not cg_id or "" == cg_id then
        return rv
    end
    
    local orderTable  = {}
    local membersList = channel.QUEUE_MEMBER_LIST(cg_id):get() or ""
    for tech, ext, ctx in membersList:gmatch("(%w+)/(%w+)@([^,]+)") do
        if tech and ext and ctx then
            rv[ext] = {
                 technology = tech
                ,exten   = ext
                ,context = ctx
            }
            table.insert(orderTable, ext)
        end
    end
    
    rv["order"] = orderTable
    
    return rv
end

--- Check member in the queue
-- @param cg_id	 	A checked call group id 
-- @param member_id	A searched member
-- @return boolean TRUE if member found or FALSE if not found
function is_member(cg_id, member_id)
	if not cg_id or "" == cg_id or not member_id or "" == member_id then
		return false
	end
	
	local cg_list = get_members_list(cg_id)
	local member  = cg_list[member_id] 
	if member and "Local" == member.technology and "internalcalls" == member.context then
		return true
	end
	
	return false
end

