--- Device state processor with cache implementation
--
-- @file    deviceState.lua
-- @author  Vyacheslav Khudan
-- @created 2016/10/05

--- @module deviceState
local deviceState = {
    cache = nil         -- cache data space
}

--- List of service statuses when we think that device available
local serviceStateTable = {
     COMMON     = {"NOT_INUSE"}
    ,INTRUSION  = {"!NOT_INUSE", "!RINGING"}
    ,MOBILITY   = {"!INUSE", "!RINGING", "!RINGINUSE", "!ONHOLD"}
    ,SILENTRING = {"!NOT_INUSE", "!RINGING"}
}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local state = {
     mPeer     = nil           -- destination peer name
    ,mTech     = nil           -- peer technology
    ,mLogs     = false         -- flag to show logs
    ,mIntState = nil           -- native device state, as internal state
    ,mExtState = nil           -- extended device state by some logic
    ,mDefState = "NOT_INUSE"
}

--- Set up logs to print device state
function state:setLogs(val)
    self.mLogs = val
end

--- Request device state value in parent system and threading by configuration
function state:reqValue()
    if not self.mPeer or "" == self.mPeer then
        return self
    end

    local peerWithTech = self.mTech .. "/" .. self.mPeer
    local devState     = channel.DEVICE_STATE(peerWithTech):get() or ""
    if "" == devState then
        devState = self.mDefState
    elseif "UNAVAILABLE" == devState then  -- workaround for WMS-8002, to do: revise all logic
        devState = self.mDefState
    end
    self.mIntState = devState

    -- checking extended state by configuration
    local devUsedByMobility = false
    if "yes" == channel.EXTENDED_DEVSTATE:get() then
        -- retrieveing status about mobility peer type
        local isRealPeer = channel.SIPPEER(self.mPeer, "check_devstate"):get() or ""
        if "false" == isRealPeer and "NOT_INUSE" ~= devState then
            -- reset device state for mobility peers (real peer is free)
            devState          = self.mDefState
            devUsedByMobility = true
        end
    end
    self.mExtState = devState

    if self.mLogs then
        if devUsedByMobility then
            print(string.format("Device state for '%s' is INUSE by mobility/office number", peerWithTech))
        else
            print(string.format("Device state for '%s' is %s", peerWithTech, devState))
        end
    end

    return self
end

--- Check device availability for specified service
function state:isAvailable4Service(servName)
    local rv = false
    if not servName then
        return rv
    end

    local statesList = serviceStateTable[servName:upper()]
    if not statesList then
        return rv
    end

    rv = true -- set up default value before comparing
    for _, val in ipairs(statesList) do
        if "!" == val:sub(1,1) then
            val = val:sub(2)
            rv  = rv and val ~= self.mExtState
        else
            rv = rv and val == self.mExtState
        end
    end

    return rv
end

--- Check common device availability
function state:isAvailable()
    return self:isAvailable4Service("COMMON")
end


-- Module constructor
function deviceState:new(peer, tech)
    local peer = peer or ""
    local tech = tech or "SIP"
    
    -- retrieve cache space
    if not self.cache or not self.cache.initialized then
        self.cache             = gSpace and gSpace:get4Module() or {}
        self.cache.initialized = true
    end
    
    -- checking cached peer object
    local peerWithTech = tech .. "/" .. peer
    local obj = self.cache[peerWithTech]
    if obj then
        obj:setLogs(false) -- disable logs by default
        return obj
    end

    -- initialize object
    obj = setmetatable({}, {__index = state})

    -- set up initial state
    obj.mPeer = peer
    obj.mTech = tech or "SIP"
    obj.mIntState = obj.mDefState
    obj.mExtState = obj.mDefState

    -- save object to cache
    self.cache[peerWithTech] = obj

    return obj
end

return deviceState

