--- Module to make request to external service for number look up
--
-- @file    externalContactLookUp.lua
-- @author  Vyacheslav Khudan
-- @created 2017/02/15

--- @module externalContactLookUp
local externalContactLookUp = {}
local function __getLookupServiceSettings(params)
    if not params then
        return
    end

    local pbxutils = require"core.pbxutils"
    local backend
    local uri
    local api
    params.timeout = tonumber(params.timeout) or 1.5 -- seconds
    if params.service == "wim" then
        backend = "socket"
        uri     = pbxutils.getCachedEnvKey("SERVICE_WIM_URI")
        api     = "/lookup?phone="
        params.authMethod     = "bearer"
        params.authToken      = pbxutils.getIdToken()
        params.secureProtocol = "tlsv1_2"
    else
        uri = pbxutils.getCachedEnvKey("SERVICE_NAMELOOKUP_URI")
        api = "/v1/phone/"
        params.authMethod = "basic"
        params.authUser   = pbxutils.pbxSerial()
        params.authPass   = pbxutils.pbxKey()
    end

    if uri and "" ~= uri then 
        return {
            backend = backend,
            reqURL  = uri..api,
            params  = params
        }
    end
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Method to make request to retrieve lookup data
-- @param number    A target number that required lookup
-- @param params    A parameters for the request
-- @return found contact name otherwise error string 
function externalContactLookUp.exec(number, params)
    local rv  = nil
    if not number or "" == number then
        return rv, "lookup number not found"
    end
    if "table" ~= type(params) then
        params = {}
    end

    local lookupService = __getLookupServiceSettings(params)
    if not lookupService then
        return nil, string.format("lookup service settings not found")
    end
          
    -- retrieve data via http client
    local hc         = require"core.httpClient":new(lookupService["backend"])
    local reqURL     = lookupService.reqURL .. hc.urlEncode(number)
    local rData, err = hc:get(reqURL, lookupService.params)
    repeat
        if not rData or not rData.body then
            err = "no response"
            break
        end

        local body = require"dkjson".decode(rData.body)
        if 200 ~= rData.code then
            logger:debug("externalContactLookUp response code: ", rData.code)
            err = body and (body.reason or body.message) or "unknown error"
            break
        end

        local result = body and body.result
        if result and "table" == type(result.customer) then
            rv = result.customer
            if result.manager then
                rv.manager = result.manager
            end
        else
            rv = {name = ("string" == type(result) and result or "")}
        end
        if not rv.name or "" == rv.name or "unknown" == tostring(rv.name):lower() then
            rv.name = "unknown"
        end
    until true
    
    return rv, err
end

return externalContactLookUp

