--- Procedures that can be used with external (from/to trunk calls) numbers 
--
-- @file    externalNumber.lua
-- @author  Vyacheslav Khudan
-- @created 2015/12/07

--- @module externalNumber
local externalNumber = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Perform number back normalization
-- @param number    A number for back normalization
-- @param peerName  A peer name with used for this procedure
-- @return unnormalized number
function externalNumber:unnormalize(number, peerName)
    local channel  = channel
    local peerName = peerName or ""
    local rv       = number
    if not number or "" == number then
        return rv
    end
    
    local numberObj = require"common.numberDetect.numberDetectImpl":new()
    if not numberObj:isNormalized(number) then
        return rv
    end
    
    -- retrieving peer country code
    local peerCountryCode = ""
    if peerName and "" ~= peerName then
        peerCountryCode = channel.SIPPEER(peerName, "country_code"):get() or ""
    end
    if "" == peerCountryCode then
        peerCountryCode = channel.GLOBAL("GCOUNTRY_CODE"):get() or ""
    end
    local peerCountryCodeLen = peerCountryCode:len()
    
    -- perform back normalization
    rv = string.sub(rv, 2)  -- skip '+'
    if 0 < peerCountryCodeLen and string.sub(rv, 1, peerCountryCodeLen) == peerCountryCode then -- processing number as national
        if "39" == peerCountryCode or peerName:find("^bri%d") or peerName:find("^pri%d") then
            rv = string.sub(rv, peerCountryCodeLen + 1)
        else
            local nationalPrefix = channel.GLOBAL("GNATIONAL_PREFIX"):get() or ""
            rv                   = nationalPrefix .. string.sub(rv, peerCountryCodeLen + 1)
        end
    else                                                                                 -- processing number as international
        local internationalPrefix = channel.GLOBAL("GINTERNATIONAL_PREFIX"):get() or ""  
        rv                        = internationalPrefix .. rv
    end
    
    if rv ~= number then
        print(string.format("Number back normalization: %s --> %s", number, rv))
    end
    
    return rv
end

--- Custom number normalization for US WMS-4589: fix incoming caller ids for us
-- @param number    A number that must be processed
-- @return          normalized US number
function externalNumber:normalizeUsNumber(number)
    local rv = number or ""
    if #number == 11 and number:sub(1, 1) == "1" then
        rv = "+" .. number
    elseif #number >= 11 and number:sub(1, 1) ~= "1" then
        if number:sub(1, 3) == "011" then
            rv = "+" .. number:sub(4)
        elseif number:sub(1, 2) ~= "00" then
            rv = "+" .. number
        end
    elseif #number == 10 then
        rv = "+1" .. number
    end
    return rv
end

--- Perform number normalization
-- @param number    A number that must be processed
-- @param peerName  A peer name which used for this procedure
-- @return normalized number
function externalNumber:normalize(number, peerName, direction)
    local channel = channel
    local rv      = number
    if not number or "" == number then
        return rv
    end
    
    local numberObj = require"common.numberDetect.numberDetectImpl":new()
    if numberObj:isNormalized(number) then
        return rv
    end
    
    -- processing peer name (for incoming calls)
    if not peerName or "" == peerName then
        local chanName = channel.CHANNEL:get("name") or ""
        if chanName:match("^Local/.*") then
            -- denied executing core normalization function for local channels
            -- core can't find peername for local channel
            return rv
        end
        peerName = channel.CHANNEL("peername"):get() or ""
    end

    local orig_number = number
    local peerCountryCode = channel.SIPPEER(peerName, "country_code"):get() or ""
    if "1" == peerCountryCode and "inc" == direction then
        number = self:normalizeUsNumber(number)
    end

    -- number normalization
    local normalizedNumber = channel.SIPNORMALIZATION(peerName, number):get() or ""
    if "" ~= normalizedNumber then
        rv = normalizedNumber
        if orig_number ~= rv then
            print(string.format("Number normalization: %s --> %s", orig_number, rv))
        end
    end
    
    -- retrieving peer country code
    if "" ~= peerCountryCode then
        print(string.format("Country code for peer(%s): %q", peerName, peerCountryCode))
    end
    
    return rv
end

--- Perform number normalization with short numbers (112, etc.)
-- @param number    A number that must be processed
-- @param peerName  A peer name which used for this procedure
-- @return normalized number
-- @note used for calls to trunks (with cases to emergency numbers)
function externalNumber:normilizeWithShortNumbers(number, peerName)
    local channel = channel
    local rv      = number
    if not number or "" == number then
        return rv
    end
    
    if 4 >= number:len() then -- call to short number like emergency, free, etc.
        local localCountryCode = channel.GLOBAL("GCOUNTRY_CODE"):get() or ""
        if 1 <= string.len(localCountryCode) then
            if "+" ~= string.sub(localCountryCode, 1, 1) then
                localCountryCode = "+" .. localCountryCode
            end
        end
        rv = localCountryCode .. number -- concatenation for correct number detection 
    else
        rv = self:normalize(number, peerName)
    end
    
    return rv
end

--- Retrieving data about external number 
-- @param normalizeNumber   An normalized number used for retrieving data
-- @return table with number data otherwise FALSE value
-- @note Emergency number can be processed, because it can't be normalized
function externalNumber:getInfo(normalizeNumber)
    local rv = nil
    if not normalizeNumber or "" == normalizeNumber then
        return rv
    end
    
    -- Remove country code from short numbers
    local function _fixupShortNumber(result)
        if "table" ~= type(result) then
            return nil
        end
        local numberTypeIdent = result:getTypeIdent()
        local rv              = result:getNumber()
        if "EMER"   == numberTypeIdent or "FREE" == numberTypeIdent or 
            "PREM"  == numberTypeIdent or "PREM2" == numberTypeIdent or
            "PREM3" == numberTypeIdent or "PREM4" == numberTypeIdent or
            "SPS"   == numberTypeIdent then
            local countryCode = result:getCountryCode()
            rv                = rv:sub(countryCode:len() + 1 + 1)
        end
        return rv
    end
    
    local numberObj = require"common.numberDetect.numberDetectImpl":new()
    if not numberObj:isNormalized(normalizeNumber) then
        return rv
    end
    
    local result = numberObj:search(normalizeNumber)
    if not result then
        return rv
    end
    rv           = result:getData()
    
    -- show info about normalized number
    if "yes" == channel.GLOBAL("EXTNUMBERINFO"):get() then
        local numberType   = result:getTypeName()
        local countryName  = result:getCountryName()
        local areaDescr    = result:getAreaDescription() 
        local locationData = {}
        if "" ~= countryName then
            table.insert(locationData, countryName)
            if "" ~= areaDescr then
                table.insert(locationData, areaDescr)
            end
        end
        local locationStr = table.concat(locationData, ", ")
        locationStr       = locationStr and string.format("(%s)", locationStr) or ""
        
        print(string.format(">>> %q is %s %s", _fixupShortNumber(result), numberType, locationStr))
    elseif "EMER" == result:getTypeIdent() then
        print(string.format("Number '%s' is emergency", _fixupShortNumber(result)))
    end
    
    return rv
end

return externalNumber