--- Missed call notification module.
-- Send notification about missed call via Email/SMS/SIP 
--
-- @file    missedCallNotify.lua
-- @author  Vyacheslav Khudan
-- @created 2016/07/13

--- @module missedCallNotify
local missedCallNotify = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local notifierObj = {             -- meta Class
     srcData    = nil   -- caller user data
    ,dstData    = nil   -- called user data
    ,callStatus = nil   -- dial status
}

--- Send notification via SIP message NOTIFY
function notifierObj:SIPNotifyEvent()
    local src     = self.srcData.id
    local dst     = self.dstData.id
    local srcName = self.srcData.attrs.cn or self.srcData.chanvar.cid or src
    
    -- create notification event
    logger:debug("Execute SIPNotifyEvent(%s, %s, %s)", src, dst, srcName)
    return channel.NOTIFY_MISSCALL(src, dst, srcName):set(1)
end

--- Send notification via SIP message MESSAGE
function notifierObj:SIPMessageEvent()
    --TODO: temporary disable sms notification (CONVENTION-417)
    -- Send sip message to phone (for queue call notification) if in moment of the hangup, manager has status not inuse
--  if queue_dialstatus and queue_dialstatus ~= "NOTIFIED" then
--      channel.MESSAGE("body"):set(text)
--      smartApp("MessageSend", "sip:" .. (dst_attrs.telephoneNumber or ""), cid_name .. " <" .. cid .. ">")
--  end
end

--- Send notification via SMS/Email method
function notifierObj:textMsgEvent()
    if self.srcData.id == self.dstData.id then
        print("Missed call notification - same caller and dest caller id")
        return false
    end
    
    -- preparing notification text
    local cidNum   = self.srcData.id
    local cidName  = self.srcData.attrs.cn or self.srcData.chanvar.cid or cidNum
    local dstAttrs = self.dstData.attrs
    local text = _t("string.MISSED_CALL", "L_" .. dstAttrs.preferredLanguage, cidName, cidNum, os.date("%H:%M"), os.date("%d/%m/%y"))
    
    -- Send notification via SMS
    if self.dstData.feat:is_active("SMSMCN") then
        logger:debug("Trying to send missed call notification via SMS.")
        if dstAttrs.mobile then
            local res, err = require"common.smsNotifications".send(dstAttrs.telephoneNumber, dstAttrs.mobile, text)
            if res then
                print("Missed call notification SMS for %s is in queue", dstAttrs.cn)
            elseif err then
                print("Missed call notification SMS for %s error: %s", dstAttrs.cn, err)
                logger:error("Missed call notification SMS for %s error: %s", dstAttrs.cn, err)
            end
        else
            print("Destination caller mobile is not presented in LDAP! Notification about missed call via SMS was not sent.")
        end
    end
    
    -- Send notification via e-mail
    if self.dstData.feat:is_active("EMAILMCN") then
        logger:debug("Trying to send missed call notification via e-mail.")
        if dstAttrs.mail then
            local result, err = require"common.mailNotifications".missedCallNotify(dstAttrs.mail, cidName, cidNum, dstAttrs.preferredLanguage)
            if result then
                logger:debug("Missed call notification e-mail for %s is in queue", dstAttrs.cn)
                print("Missed call notification e-mail for %s is in queue", dstAttrs.cn)
            else
                print("Missed call notification e-mail to '%s' was not sent: %s", dstAttrs.cn, err or "")
                logger:error("Missed call notification e-mail to '%s' was not sent: %s", dstAttrs.cn, err or "")
            end
        else
            print("Destination e-mail is not present in LDAP! Missed call notification via e-mail was not sent.")
            logger:debug("Destination e-mail is not present in LDAP! Missed call notification via e-mail was not sent.")
        end
    end
    
    return true
end


--- Object constructor
-- @param srcData       The caller user data (id/attrs/chanvar)
-- @param dstData       The called user data (id/attrs/feat)
-- @param callStatus    The status of the call
function missedCallNotify:new(srcData, dstData, callStatus)
    local rv = nil
    if "table" ~= type(srcData) then
        return rv, "Invalid format of the src user data"
    end
    if not srcData.id or not srcData.attrs then
        return rv, "Invalid src user data"
    end
    
    if "table" ~= type(dstData) then
        return rv, "Invalid format of the dst user data"
    end
    if not dstData.id or not dstData.attrs or not dstData.feat then
        return rv, "Invalid dst user data"
    end
    
    if not callStatus then
        return rv, "Invalid/not processed call status"
    elseif "ANSWER" == callStatus or "ANSWERED" == callStatus then
        return rv, "call status is ANSWERED"
    end
    
    local obj = {}
    setmetatable(obj, {__index = notifierObj})
    
    -- initialize object result data
    obj.srcData    = srcData
    obj.dstData    = dstData
    obj.callStatus = callStatus

    -- initialize additional variables
    local cidName = channel.CALLERID("name"):get() or ""
    obj.srcData.chanvar = {
        cid = "" ~= cidName and cidName or nil
    }
    
    return obj
end

return missedCallNotify
