--- Dialplan module to processing internal IVR
-- 
-- @file    pbxIVR.lua 
-- @author  Vyacheslav Khudan
-- @created 2017/01/23

local fsutils = require"core.fsutils"
local logger  = require"core.logger":init("[WMS 3.0 Dialplan]", 3)
--- @module pbxIVR
local pbxIVR = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Fixing message by some rules
local function __fixupMessage(val, lang, isTts)
    local rv = val
    isTts = isTts or false
    if "" == val then
        return rv
    end
    
    if "de" == lang then
        if val:find("please enter") then
            if val:find("please enter %P+ you wish to dial") then
                rv = val:gsub("you", "ein you")
            elseif val:find("please enter %P+ or wait") then
                rv = val:gsub("or wait", "ein or wait")
            elseif val:find("please enter .*[\.,]") then
                rv = val:gsub("(please enter [^\.,]+[^\.,])", "%1 ein")
            else
                rv = val:gsub("$", " ein")
            end
        end
        if val:find("the message id %d+") then
            rv = val:gsub("id", "id is")
        end
        if val:find("the phone is in use by") then
            rv = val .. " verwendet"
        end
        if val:find("user status") and not val:find("user status updated") then
            rv = val:gsub("user status", "user status is")
        end
        if val:find("service status") and not val:find("service status updated") then
            rv = val:gsub("service status", "service status is")
        end
        if val:find("leider kann ich keinen kontakt finden") then
            rv = val .. " lautet"
        end
        if val:find("ich rufe") then
            rv = val .. " an"
        end

    elseif "nl" == lang then
        if val:find("please enter") then
            if val:find("to use this phone please enter the password") then
                rv = val:gsub("please enter", "enter")
                rv = rv .. " ein"
            elseif val:find("please enter %P+ you wish to dial") then
                rv = val:gsub("you", "ein you")
            elseif val:find("please enter %P+ or wait") then
                rv = val:gsub("or wait", "ein or wait")
            elseif val:find("welcome to the directory.* please enter part of name or surname for search") then
                rv = val
            elseif val:find("please enter .*[\.,]") then
                rv = val:gsub("(please enter [^\.,]+[^\.,])", "%1 ein")
            else
                rv = val:gsub("$", " ein")
            end
        end
        if val:find("the message id %d+") then
            rv = val:gsub("id", "id is")
        end
        if val:find("user status") and not val:find("user status updated") then
            rv = val:gsub("user status", "user status is")
        end
        if val:find("service status") and not val:find("service status updated") then
            rv = val:gsub("service status", "service status is")
        end
        if val:find("if this is the person you are looking for press 1 otherwise press 0") then
            rv = "press 1 if this is the person you are looking for otherwise press 0"
        end
        
    elseif "it" == lang then
        if "wake up call activated for" == val or "wake up call for" == val then
            rv = val:gsub("for", "for le ore")
        end
        if val:find("please enter the password.* followed by pound") or
           val:find("to set the destination number.* followed by pound") then
            rv = val:gsub("followed by pound", "seguita da cancelletto")
        end
        
    elseif "fr" == lang then
        if val:find("user status") and not val:find("user status updated") then
            rv = val:gsub("user status", "user status is")
        end
        if val:find("service status") and not val:find("service status updated") then
            rv = val:gsub("service status", "service status is")
        end
    end
    
    if val ~= rv then 
        logger:debug("pbxIVR. The message has been fixed. From '%s'. To '%s'", val, rv)
    end
    if not isTts then
        rv = rv and rv:gsub("[\.,]","")
    end

    return rv
end
 
--- Retrieving sound language
local function __getSoundLang(val)
    local utils = require"utils"
    local list = {
         ["en"]     = function() return "en" end
        ,["it"]     = function() 
            return utils:is_lang_pack_present("it", "tvoice") and "it" or "en"
        end
        ,["fr"]     = function()
            return utils:is_lang_pack_present("fr", "tvoice") and "fr" or "en"
        end
        ,["de"]     = function()
            return utils:is_lang_pack_present("de", "tvoice") and "de" or "it"
        end
        ,["nl"]     = function()
            return utils:is_lang_pack_present("nl", "tvoice") and "nl" or "en"
        end
        ,["se"]     = function()
            return utils:is_lang_pack_present("se", "tvoice") and "se" or "en"
        end
        ,["es"]     = function()
            return utils:is_lang_pack_present("es", "tvoice") and "es" or "en"
        end
        ,["es-us"]     = function()
            return utils:is_lang_pack_present("es-us", "tvoice") and "es-us" or "en"
        end
        ,["en-us"]  = function()
            return utils:is_lang_pack_present("en-us", "tvoice") and "en-us" or "en"
        end
    }
    
    return list[val] and list[val]()
end

--- Default failure callback for read mode
local function __defReadFailureFn(obj)
    return obj:tell("No number was entered, please try again")
end

--- Default validate callback for read mode
local function __defReadValidateFn(v)
    return "" ~= tostring(v)
end

--- Executing pre-playback application (sending early media to the device)
local function __preExecApp(doProgress)
    if doProgress then
        if not gSpace["intIVRChanProgress"] then
            smartApp("Progress", "")
            gSpace["intIVRChanProgress"] = true
        end
    else
        if not gSpace["intIVRChanAnswer"] then
            smartApp("Answer", ",nocdr")
            gSpace["intIVRChanAnswer"] = true
        end
    end
end

--- Update options sound sub-directory by defines for some standart phrases
local function __preParsingMessageByOptions(msg, opts)
    local rv = msg
    if not msg or "table" ~= type(opts) then
        return rv
    end
    
    if "beep" == msg or "beeperr" == msg then
        rv = nil
        opts.voiceDir = "/"
    elseif "dialtone" == msg then
        opts.voiceDir = "/"
    elseif "vm-isunavail" == msg then
        opts.voiceDir = "/"
    elseif "vm-enter-num-to-call" == msg then
        opts.voiceDir = "/"
    elseif "vm-sorry" == msg then
        opts.voiceDir = "/"
    end
    
    return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local ivr = {  -- meta class
     mMode          = "speak"                       -- ivr module mode
    ,mSoundDir      = "/var/lib/callweaver/sounds/" -- default sound path
    ,mSoundSubDir   = "tvoice/"                     -- sound sub directory                        
    ,mSoundLang     = "en"                          -- ivr module language
    ,mRuntimeOpts   = nil                           -- list of the runtime options:
--            silent                                   -- uses in case processing BLF (like mute)
--            subDir                                   -- temporary sound sub directory
--            msgRepeat                                -- repeat message some times
--            defLang                                  -- use default the language to search (system sounds)
--           -- read mode options
--            retry                                    -- retry read counter
--            timeout                                  -- timeout to wait for reading
--            maxChars                                 -- limit of the maximum chars for reading
--            wihoutBeep                               -- flag to disable beep sound before reading user choice
--            failureFn                                -- callback to execute in case failure input
--            validateFn                               -- callback to execute for input validation
}

--- Set up mode for ivr class
function ivr:__setMode(mode)
    local mode = ({
         speak = "speak"
        ,play  = "play"
        ,read  = "read"
    })[mode]
    
    if mode then
        self.mMode = mode
        logger:debug("pbxIVR. Set mode to", mode)
    end
    return mode
end

--- Set up ivr class language
function ivr:__setLang(lang)
    local langToSet = __getSoundLang(lang)
    if langToSet then
       self.mSoundLang = langToSet
    end
    if lang ~= langToSet then
        print("Request to set '%s', but '%s' was set", lang, langToSet)
        logger:error("pbxIVR. Request to set '%s', but '%s' was set", lang, langToSet)
    end
    return lang
end

--- Processing runtime options
function ivr:__clearRuntimeOptions(mode)
    if mode then
        self.mRuntimeOpts[mode] = {}
    else    
        self.mRuntimeOpts       = {
             speak = {initialized = false}
            ,play  = {initialized = false}
            ,read  = {initialized = false}
        }
    end
end
function ivr:__initRuntimeOptions(opts)
    opts = "table" == type(opts) and opts or {}
    local runtimeOpts = self.mRuntimeOpts[self.mMode]
    
    if "/" == opts.voiceDir then                                                -- default language for core search
        runtimeOpts.defLang = true
    end
    
    if opts.voiceDir and fsutils.is_dir(self.mSoundDir .. opts.voiceDir) then   -- sub directory
        local dir          = opts.voiceDir
        local finalizer    = dir:find("%/$") and "" or "/"
        dir                = "/" == dir and "" or dir
        runtimeOpts.subDir = dir .. finalizer
    end
    
    runtimeOpts.msgRepeat  = tonumber(opts.iteration) or 1                      -- repeat to play message
    if opts.silent then
        runtimeOpts.msgRepeat = 0                                               -- disable play message
    end
    
    runtimeOpts.appOptions = opts.opts or "s"                                   -- application options
    if "n" == opts.opts then                                                    -- early media
        runtimeOpts.earlyMedia = true
    end
    
    if "read" == self.mMode then
        runtimeOpts.retry       = tonumber(opts.retry) or 3                     -- read reatry counters
        runtimeOpts.wihoutBeep  = opts.nobeep                                   -- flag to disable the beep during read
        runtimeOpts.maxChars    = tonumber(opts.max_chars) or 10                -- maximum characters to input for read application
        runtimeOpts.timeout     = tonumber(opts.timeout)                        -- read timeout parameter (dialplan/variable/default)
        if not runtimeOpts.timeout then
            local systemTimeout = channel.GPBXIVRTIMEOUT:get()
            if tonumber(systemTimeout) then
                runtimeOpts.timeout = systemTimeout
            else
                runtimeOpts.timeout = 5
            end
        end
        
        if "function" == type(opts.validate) then                               -- set up read validator callback
            runtimeOpts.validateFn = opts.validate
        else
            runtimeOpts.validateFn = __defReadValidateFn
        end
        if "function" == type(opts.failure) then                                -- set up read failure callback
            runtimeOpts.failureFn  = opts.failure
        else
            runtimeOpts.failureFn  = __defReadFailureFn
        end
    end
    
    return true
end
function ivr:__getRuntimeOption(name)
    return self.mRuntimeOpts[self.mMode][name]
end

--- Parsing sound message to list
-- @param userMsg   The user message to play back 
-- @param opts      The runtime options for parser
-- @return list with sounds to playback 
-- @note custom voice dir does not support auto build for phrases  
function ivr:__parsingMsgToList(userMsg, opts)
    userMsg = "table" == type(userMsg) and userMsg or {userMsg}
    if (not userMsg[1] and not userMsg["tts"]) or ("" == userMsg[1] and "" == userMsg["tts"]) then
        return nil
    end
    opts = opts or {}

    self:__setLang(channel.CHANNEL("Language"):get())   -- retrieve language

    -- normalize message for processing
    local msg = string.lower(userMsg[1] or "")
    msg       = __fixupMessage(msg, self.mSoundLang)
    
    local isDisplayMsg = __preParsingMessageByOptions(msg, opts)  -- update options & get message to display
    self:__initRuntimeOptions(opts)                               -- processing runtime options
    
    if 0 == self:__getRuntimeOption("msgRepeat") then             -- disable playback
        return nil
    end
    
    -- parsing input message to words with related building type
    -- []   - marker to retrieving correct sound via CW. Basically used for digits
    -- #    - prefix to use digit in general builder
    local wordsList = {}
    --use text to speech if need and available 
    local textToSpeech = require"textToSpeech"
    local ttsOn = userMsg["tts"] and "" ~= userMsg["tts"] and textToSpeech.available()
    if ttsOn and opts.tts_fixup then
        userMsg["tts"] = __fixupMessage(userMsg["tts"]:lower(), self.mSoundLang, true) -- Need for WMS5289
    end
    if isDisplayMsg then print(ttsOn and userMsg["tts"] or userMsg[1]) end             -- display processing message
    if ttsOn then
        local chanState = channel.CHANNEL("state"):get()
        if "Up" ~= chanState and "Ring" ~= chanState then
            smartApp("Ringing", "")
        end
        local ttsLang     = self.mSoundLang
        local synthesized = textToSpeech.getAudio(userMsg["tts"], ttsLang)
        if synthesized then
            wordsList[#wordsList + 1] = {synthesized, "tts"}
            msg = "" --clear standart message
        end
    end
    for word in msg:gmatch("([%w%-_%[%]%#]+)") do
        local buildType = nil
        if word:find("^%[%d+%]$") then    --TODO: using CW possibility
            word      = word:match("%[(%d+)%]")
            buildType = "digit"
        elseif word:find("^%#%d+") then --TODO(sounds): remove after sound aliases
            word      = word:sub(2)
            buildType = "general"
        else
            buildType = self:__getRuntimeOption("subDir") and "custom" or "general"
        end
        wordsList[#wordsList + 1] = {word, buildType}
    end
    
    -- set up sounds path for builders
    local langDir = ""
    if "en" ~= self.mSoundLang and not self:__getRuntimeOption("defLang") then
        langDir = self.mSoundLang .. "/"
    end
    local customSubDir = self:__getRuntimeOption("subDir")
    if customSubDir and not customSubDir:find("^0000") then
        customSubDir = langDir .. customSubDir
    end
    

    local function _digit(v)
        local rv = {}
        for digit in v:gmatch("(%d)") do
            rv[#rv + 1] = langDir .. "digits/" .. digit
        end
        return rv
    end

    -- building sound stack
    local builders   = {
        ["custom"] = function(v)
            customSubDir = customSubDir or ""
            return customSubDir .. v
        end
        ,["general"] = function(t)
            local sExt  = {"alaw", "g729", "opus"}  -- searching file extensons
            local sPath = self.mSoundDir .. langDir .. self.mSoundSubDir
            local rv    = {}

            while 0 < #t do
                -- search sound file
                local s            = require"core.pbxutils".tcopy(t)
                local foundFilePos = 0
                local foundFile    = nil

                for i = #t, 1, -1 do
                    local rvPhrase  = table.concat(s, "_")
                    if fsutils.is_readable(sPath .. rvPhrase, sExt) then
                        foundFile    = langDir .. self.mSoundSubDir .. rvPhrase
                        foundFilePos = #s
                        break
                    end
                    table.remove(s)
                end
                
                if 0 ~= foundFilePos then
                    -- save found file to the result table
                    rv[#rv + 1] = foundFile
                elseif tonumber(t[1]) then
                    for _, digitSound in ipairs(_digit(t[1])) do
                        rv[#rv + 1] = digitSound
                    end
                end

                -- prepare table for next search
                foundFilePos =  0 ~= foundFilePos and foundFilePos or 1
                for i = 1, foundFilePos do
                    table.remove(t, 1)
                end
            end

            return rv
        end
        ,["digit"]   = _digit
        ,["tts"]     = function(v)
            return v
        end
    }
    local soundList = {}
    local generalPool = {}
    for i, data in ipairs(wordsList) do
        local word, bType = data[1], data[2]
        if word and bType and builders[bType] then
            if "general" == bType then  -- general builder
                table.insert(generalPool, word)
                if not wordsList[i + 1] or "general" ~= wordsList[i + 1][2] then
                    for _, digitSound in ipairs(builders[bType](generalPool)) do
                        soundList[#soundList + 1] = digitSound
                    end
                    generalPool = {}
                end
            elseif "custom" == bType then
                soundList[#soundList + 1] = builders[bType](word)
            elseif "digit" == bType then
                for _, digitSound in ipairs(builders[bType](word)) do
                    soundList[#soundList + 1] = digitSound
                end
            elseif "tts" == bType then
                local ttsSound = builders[bType](word)
                if ttsSound then
                    soundList[#soundList + 1] = ttsSound
                end 
            end
        end
    end
    if 0 == #soundList then
        print("Error: empty soundList!")
        logger:error("pbxIVR. Empty soundList!")
    end
    return 0 ~= #soundList and soundList
end

--- Executing applications
-- @param sounds    The string representation of the sound files to playback
-- @param opts      The runtime options for application 
-- @return boolean value or user choice in case read application 
function ivr:__execApp(sounds, opts)
    local rv = false
    if not sounds then
        return rv
    end
    
    self:__initRuntimeOptions(opts)
    
    -- processing sound pathes
    if "table" == type(sounds) then
        sounds = table.concat(sounds, "&")
    end
    local msgRepeat = self:__getRuntimeOption("msgRepeat")
    if 1 < msgRepeat then
        sounds = sounds .. "&"
        sounds = sounds:rep(msgRepeat)
    end
    
    if "speak" == self.mMode then
        local appOpts = self:__getRuntimeOption("appOptions")
        __preExecApp(self:__getRuntimeOption("earlyMedia"))
        smartApp("Read", ".", sounds, nil, appOpts, 1, 0.1)
    elseif "play" == self.mMode then
        smartApp("Playback", sounds)
    elseif "read" == self.mMode then
        -- normilize sounds by 'beep'
        if not self:__getRuntimeOption("earlyMedia") and not self:__getRuntimeOption("wihoutBeep") then   
            sounds = sounds .. "&beep"
        end
        
        local validateFn = self:__getRuntimeOption("validateFn")
        local failureFn  = self:__getRuntimeOption("failureFn")
        local appOpts    = self:__getRuntimeOption("appOptions")
        local maxChars   = self:__getRuntimeOption("maxChars")
        local timeout    = self:__getRuntimeOption("timeout")
        local retry      = self:__getRuntimeOption("retry")
        __preExecApp(self:__getRuntimeOption("earlyMedia"))
        for i = 1, retry do
            smartApp("Read", "IVRReadVal", sounds, maxChars, appOpts, nil, timeout)
            local userVal = channel.IVRReadVal:get()
            print("read value: " .. userVal)
            if not validateFn(userVal) then
                if i ~= retry then
                    failureFn(self, userVal)
                    self:__setMode("read")
                end
            else
                rv = tostring(userVal)
                break
            end
        end
    end
    
    self:__clearRuntimeOptions(self.mMode)
    if check_hangup() then
        smartApp("Hangup", hcause.NORMAL)
    end     
    return rv
end

--- Playback message to user by 'Read' applicaiton
-- @see ivr:__execApp for more details 
function ivr:tell(msg, opts)
    self:__setMode("speak")
    local soundList = self:__parsingMsgToList(msg, opts)
    return self:__execApp(soundList, opts)
end

--- Playback message to user by 'Playback' applicaiton
-- @see ivr:__execApp for more details 
function ivr:play(msg, opts)
    self:__setMode("play")
    local soundList = self:__parsingMsgToList(msg, opts)
    return self:__execApp(soundList, opts)
end

--- Playback message to get user input
-- @see ivr:__execApp for more details 
function ivr:read(msg, opts)
    self:__setMode("read")
    local soundList = self:__parsingMsgToList(msg, opts)
    return self:__execApp(soundList, opts)
end

 
--- Object constructor
function pbxIVR:new()
    
    -- create object
    local obj = setmetatable({}, {__index = ivr} )
    
    -- initialize object
    obj:__clearRuntimeOptions()
    
    return obj
end

return pbxIVR

