--- Speech to text module
--
-- @file    speechToText.lua
-- @author  Gennadiy Ganchev
-- @created 2017/12/18

--- @module speechToText
local speechToText = {}

local languages = {
     ["en"]    = "en-GB"
    ,["it"]    = "it-IT"
    ,["fr"]    = "fr-FR"
    ,["de"]    = "de-DE"
    ,["en-us"] = "en-US"
    ,["nl"]    = "nl-NL"
    ,["ru"]    = "ru-RU"
    ,["es"]    = "es"
    ,["es-us"] = "es-US"
}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Convert speech to text
-- @param  audio An audio file
-- @param  lang A language
-- @return recognized text
function speechToText.recognize(audio,params)
    if not audio then
        return false, "textToSpeech Error! Parameters not valid"
    end
    params          = params or {}
    params.language = params.language or "en"
    params.language = languages[params.language]

    local rv               = nil
    local timeout          = channel.SPEECHTOTEXT_TIMEOUT:get() or ""
          timeout          = "" == timeout and 3 or timeout
    local json             = require"dkjson"
    local httpClient       = require"core.httpClient":new()
    local pbxUtils         = require"core.pbxutils"
    local authUser         = pbxUtils.pbxSerial()
    local authPass         = pbxUtils.pbxKey()
    local reqURL           = "https://stt.wildix.com/api/v1/Common/"
    local socket           = require"socket"
    local recognitionStart = socket.gettime()
    local answer,err       = httpClient:sendFileAsFormData(reqURL, audio, params, {authUser = authUser, authPass = authPass, timeout = timeout})
    if err then
        print("STT error:", err)
        logger:error("STT error:", err)
    end
    local recognitionEnd   = socket.gettime()
    if answer then
        logger:info("Trying to recognize audio from the file %q by external service. Language: '%s'. Recognition lasted %s sec. Answer code: %s", audio, params.language, recognitionEnd - recognitionStart, (answer and answer.scode) or "empty")
    else
        print("Recognition audio file %q failure, no answer from the speech to text service", audio)
        logger:warning("Recognition audio file %q failure, no answer from the speech to text service", audio)
    end
    local body = json.decode(answer and answer.body or "")
    if body and "result" == body["type"] and body["result"] and body["result"][1] and "" ~= body["result"][1]["transcript"] then
        rv = body["result"][1] and body["result"][1]["transcript"]
    end
    logger:debug("Recognition result:", rv or "empty")
    print("Recognition result:", rv or "empty")

    return rv
end

--- Record audio
function speechToText.record(maxduration)
    maxduration   = maxduration or 3
    local fsutils = require"core.fsutils"
    fsutils.rmFilesByPrefix("/tmp/", "stt_", 0.5) -- remove cached files
    math.randomseed(os.time())
    local recFile = "/tmp/stt_" .. math.random(1000,9999) .. ".wav"
    smartApp("Record", recFile, "3", maxduration, "y")
    if not fsutils.is_readable(recFile) then
        print("File %s is not readable.", recFile)
        logger:warning("File %s is not readable.", recFile)
        recFile = nil
    end

    return recFile
end

return speechToText
