--- Module for text to speech feature
--
-- @file    textToSpeech.lua
-- @author  Gennadiy Ganchev
-- @created 2017/10/25

--- @module textToSpeech
local textToSpeech = {}

-- voice  - voice of speech synthesis
-- volume - between -15 and +15 (dB). 0 is normal.
-- speed  - rate between 20% - 100%. 100% is normal speed
-- pitch  - between -50% and +50%. 0% is normal
local paramsList = {
     en    = { voice = "Emma" }
    ,it    = { voice = "Carla" }
    ,fr    = { voice = "Celine", speed = "90"}
    ,de    = { voice = "Vicki"}
    ,en_us = { voice = "Salli"}
    ,nl    = { voice = "Ruben", pitch = "20"}
    ,ru    = { voice = "Tatyana"}
    ,es    = { voice = "Conchita"}
    ,es_us = { voice = "Lupe"}
}

-- lang - language
local function __getVoiceParams(lang)
    local rv     = {}
    local lang   = "" ~= lang and lang or "en"
          lang   = string.gsub(lang, "-","_")
    local voice  = channel.TEXTTOSPEECH_VOICE:get() or ""
    local params = paramsList[lang] or paramsList["en"]
    for k,v in pairs(params) do
        rv[#rv+1] = k.."="..("voice" == k and "" ~= voice and voice or v)
    end

    return rv
end

--- Retrieve sound file by a input text via STT(ASR) service
local function __getSoundByText(URLParams, soundFile, reqParams)
    if "table" ~= type(URLParams) then
        return nil, "invalid URL parameters"
    end
    if not soundFile then
        return nil, "not specified result sound file"
    end
    
    -- set up an authorization data
    local pbxUtils  = require"core.pbxutils"
    local pbxSerial = pbxUtils.pbxSerial()
    if "" == pbxSerial then
        return nil, "not found PBX serial!"
    end
    local pbxKey = pbxUtils.pbxKey()
    if "" == pbxKey then
        return nil, "not found PBX key!"
    end
    
    -- Prepare request params
    if "table" ~= type(reqParams) then
        reqParams = {}
    end
    reqParams.authUser = pbxSerial
    reqParams.authPass = pbxKey
    reqParams.timeout  = tonumber(reqParams.timeout) or 3
    
    -- Prepare URL params
    URLParams = table.concat(URLParams, "&")
        
    local httpClient  = require"core.httpClient":new()
    local reqURL      = "https://tts.wildix.com/Audio/?" .. URLParams
    local answer,err  = httpClient:getFile(reqURL, soundFile, reqParams)
    
    return answer, err
end

--- Convert mp3 to wav
local function __transcode(source, dest)
    if not source or not dest then
        print("Transcode error. Input data not valid.")
        logger:warning("Transcode error. Input data not valid.")
        return nil
    end
    local rv = string.match(dest,"^(.*)%.%w+$")
    local fsutils = require"core.fsutils"
    local err = nil
    -- TODO check on the load test
    local cmd = "nice -n 19 /usr/bin/sox -v 2 -t mp3 %s -t wav -e signed-integer -r8000 -c1 %s 2>&1"
    local res = require"core.pbxutils".run(cmd, source, dest)
    if res and (res:find(" FAIL") or res:find("ERROR:")) then
        err = res
        print("Transcode error:", err)
        logger:warning("Transcode error: " .. err)
    end
    if not fsutils.is_readable(dest) then
        print("File %s is not readable.", dest)
        logger:warning("File %s is not readable.", dest)
        rv = nil
    end
    fsutils.rmfile(source)
    
    return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Convert text to speech
-- @param  text A text to be converted
-- @param  lang A language
-- @return Audio file path or nil
function textToSpeech.getAudio(text,lang)
    local rv = nil
    if not text or "" == text then
        return rv, "textToSpeech Error! Parameters not valid"
    end
    local URLParams         = __getVoiceParams(lang)
    local voiceParams       = table.concat(URLParams, "&")
    local sound   = "/tmp/tts_"..require"md5".digest(voiceParams..":"..text)..".wav"
    require"lfs".touch(sound)
    local fsutils = require"core.fsutils"
    fsutils.rmFilesByPrefix("/tmp/", "tts_", 5)
    if fsutils.is_readable(sound) then
        print("Requested file '%s' found in local cache.", sound)
        logger:info("Requested file '%s' found in local cache.", sound)
        return string.match(sound,"^(.*)%.%w+$")
    end
    URLParams[#URLParams+1] = "text=" .. require"core.httpClient".urlEncode(text)
    
    local tmpFile     = fsutils.tmpname("/tmp/")
    local reqParams   = {
            timeout = channel.TEXTTOSPEECH_TIMEOUT:get()
        }
    
    local synthStart  = require"socket".gettime()
    local answer, err = __getSoundByText(URLParams, tmpFile, reqParams)
    
    local synthDiff   = string.format("%0.3f", require"socket".gettime() - synthStart)
    local cutText     = string.sub(text,1,200)
    if answer then
        logger:info("Trying synthesize a sound from the text %q by extenral TTS service. Language: '%s'. Synthesis took %s sec. TTS Status: %s", cutText, lang, synthDiff, answer.scode)
        rv = __transcode(tmpFile, sound)
    else
        print("TTS error message:", err or "unknown error")
        logger:warning("TTS error message:", err or "unknown error")
    end
    
    return rv
end

--- Check if the text to speech is available
-- @return true if available or false if not
function textToSpeech.available(sType)
    local rv = nil
    sType = sType or "common"
    if channel then -- request from dialplan
        repeat
            -- check configuration variable
            local isServiceEnabled = channel.TEXTTOSPEECH:get()
            if "no" == isServiceEnabled then
                print("Text to speech disabled by global variable")
                logger:debug("Text to speech disabled by global variable")
                break
            end
        
            -- check PBX licensing for this service (CDRVIEW + PERUSER/Subscription)
            if "Enable" == channel.LICENSE("LICENSE_PERUSER"):get() and "Enable" == channel.LICENSE("LICENSE_CDRVIEW"):get() then
                logger:debug("PERUSER + CDRVIEW license found")
                local subscription = channel.LICENSE_SUBSCRIPTION("LICENSE_PERUSER"):get()
                if "Lifetime" == subscription then
                    print("Text to speech disabled for lifetime license")
                    logger:debug("Text to speech disabled for lifetime license")
                    break
                elseif "Expired" == subscription then
                    print("Subscription expired, text to speech disabled")
                    logger:warning("Subscription expired, text to speech disabled")
                    break
                end
            else
                print("Text to speech disabled. Premium license is not found")
                logger:debug("Text to speech disabled. Premium license is not found")
                break
            end

            -- check PBX language
            local lang     = channel.CHANNEL("Language"):get() or ""
            local pbxutils = require"core.pbxutils"
            if ("common" == sType and not pbxutils.in_table(lang, {"en", "it", "de", "fr", "en-us", "nl", "es", "es-us"})) or
                (gSpace["gvars"].dststatusmsg and "" ~= gSpace["gvars"].dststatusmsg and not pbxutils.in_table(lang, {"en", "en-us"})) then
                print("Text to speech disabled by language setting:", lang)
                logger:info("Text to speech disabled by language setting:", lang)
                break
            end

            rv = true
        until true
    end

    return rv
end

return textToSpeech
