--- Authorization class for user lock procedure in case request from device
-- 
-- @file    userAuthorize.lua 
-- @author  Vyacheslav Khudan
-- @created 2017/08/11

--- @module userAuthorize
local userAuthorize = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local authorize = {    -- meta class
     mPinLength      = 5    -- default pin lenth is 5 digits
    ,mInputTimeout   = 5    -- default timeout to wait input is 5 seconds
    ,mMaxAttempts    = 1    -- default value for attemts
    ,mAuthUserObj    = nil  -- link to the authorization object
    
    -- default phrases
    ,mEnterPhonePhrase  = "Please enter your phone number, followed by pound"
    ,mEnterPinPhrase    = "Please enter the password, followed by pound" 
}

--- Setter for an autorization timeout
function authorize:setTimeOut(val)
    local val = tonumber(val)
    if not val then
        return false
    end
    
    self.mInputTimeout = val
    return true
end

--- Setter for an autorization pin length
function authorize:setPinLenth(val)
    local val = tonumber(val)
    if not val then
        return false
    end
    
    val = 3 > val and 3 or val
    self.mPinLength = val
    return true
end

--- Setter for an autorization attempts
function authorize:setMaxAttempts(val)
    local val = tonumber(val)
    if not val then
        return false
    end
    
    self.mMaxAttempts = val
    return true
end

--- Settter for an autorization pin phrase
function authorize:setPinPhrase(val)
    if val and "" ~= val then
        self.mEnterPinPhrase = val
    end
end

--- Main autorization procedure
-- TODO: use error notification on the phone display
function authorize:process(number, pin)
    local ivr    = ivr
    local rv     = false
    local errMsg = nil
    
    -- processing IVR messages
    local attempt   = 1
    local pinViaIVR = false
    repeat
        -- retrieve an user agent for the request
        local userAgent = channel.SIP_HEADER("User-Agent"):get() or ""

        if "" == userAgent or "WildixGW" == userAgent:sub(1, 8) then
            print(string.format("Unsuitable user agent: '%s'", userAgent))
            local trunkAgent = channel["X-TRUNK"]:get() or ""
            userAgent = "" ~= trunkAgent and trunkAgent or "unknown"
            print(string.format("New user agent: '%s'", userAgent))
        end
        
         -- initialize auth lock object and check credentials
        local reqEnv = {
             source = gSpace.cidNum
            ,ua     = userAgent
        }
        local attempt_list = {[5]=30, [10]=30, [15]=60, [20]=60, [25]=60, [30]=60, [35]=60, [40]=120, [45]=120, [50]=120, ["*"]=600}
        local params = {passLen = self.mPinLength, FreezeAttemptList = attempt_list}
        self.mAuthUserObj, errMsg = require"common.userAuth.userAuthLogin":new(reqEnv, params)
        if not self.mAuthUserObj then
            smartApp("Playback", "beeperr")
            print("Authentification failed. See logs for details ...")
            logger:warning("Authentification module error: %s", errMsg)
            break
        end
        
        local freezeTime = self.mAuthUserObj:getSessionFreezeTime()
        if 0 < freezeTime then
            smartApp("Playback", "beeperr")
            print(string.format("An authorization procedure locked for %s seconds", freezeTime))
            break
        end
        
        -- check and retrieve user number/pin
        if not number then
            number = ivr:read(self.mEnterPhonePhrase, {retry = 1, timeout = self.mInputTimeout})
        end
        if not pin then
            local enterPinPhrase = self.enterPinPhrase
            pin = ivr:read(self.mEnterPinPhrase, {retry = 1, max_chars = self.mPinLength, timeout = self.mInputTimeout})
            pinViaIVR = true
        end
        
        -- make the authentification
        rv, errMsg = self.mAuthUserObj:authenticate(number, pin)
        if rv then
            if pinViaIVR then   -- interupt user about successful password
                smartApp("Playback", "beep&beep")
            end
            break
        end
        
        -- some notification
        smartApp("Playback", "beeperr")
        ivr:tell("Authentication failed.")
        print("Authentification error:", errMsg)
        logger:debug("Authentification error: %s", errMsg)
        if attempt < self.mMaxAttempts then
            ivr:tell("Please try again.")
        else
            ivr:tell("Good Bye. Hangup")
            break
        end
        
        -- post clearing entered data
        number = nil
        pin    = nil
        
        attempt = attempt + 1
    until attempt > self.mMaxAttempts
    
    return rv
end

--- Retrieve data about autorized user
function authorize:getUserData()
    local userData = self.mAuthUserObj and self.mAuthUserObj:getUserBackendData()
    return userData, userData and userData.DN
end


--- Object constructor
function userAuthorize:new()
    local obj = setmetatable({}, {__index = authorize})

    return obj 
end

return userAuthorize
