--- Processing operations in case using extended (Hotel) services via API implementation
-- 
-- @file    APIExServiceStrategy.lua 
-- @author  Vyacheslav Khudan
-- @created 2015/07/09

--- @module APIExServiceStrategy
local APIExServiceStrategy = {}

----------------------------------------------------------------
-- Implementation interface
----------------------------------------------------------------
--- Get values from user via IVR service
local function __get_userValue(sounds, opts)
	if "table" ~= type(sounds) or "table" ~= type(opts) then
		return false
	end
	
	local validate = opts.validate
	if not validate then
		return false
	end
	local ivr 		   = ivr
	local val 		   = nil
	local withAnnounce = false
	if 2 == #sounds then
		withAnnounce = true
	end
	
	if withAnnounce then
    	val = ivr:read(table.concat(sounds[1], " "), {
    		 max_chars 	= opts.maxChars
    		,retry		= 1
    		,timeout	= 0.1
    		,nobeep 	= true
    		,validate 	= function() return true end
    	})
    end
	if not validate(val) or not withAnnounce then
		val = ivr:read(table.concat(sounds[withAnnounce and 2 or 1], " "), {
			 max_chars = opts.maxChars
			,validate  = validate
			,failure   = function() ivr:tell(opts.failSound or "The number you entered is incorrect, please try again") end
		})
		if not val then
			ivr:tell("Good bye. Hangup")
		end
	end
	
	return val
end

-- TODO: must be replaced by correct logic in ivr:tell with SayNumber
--- Playback message with correct digit pronunciation
local function __smartPlayback(msg, ...)
	if not msg then
		return false
	end

	local msg 			 = string.format(msg, ...)
	local soundRunner    = {}
	local playbackStack  = {}
	local sayNumberStack = {}
	for word in msg:gmatch("%S+") do
		local number = word:match("%[(%d+)%]")
		if number then
			table.insert(sayNumberStack, number)
			table.insert(playbackStack, 0)
		else
			table.insert(playbackStack, word)
			table.insert(sayNumberStack, 0)
		end
	end
	
	-- @limitation Will not be played first digits at the beginning message
	local msgPartial = {}
	for idx, val in pairs(playbackStack) do
		if 0 == val then
			if 0 ~= #msgPartial then
				table.insert(soundRunner, {"Playback", msgPartial})
				msgPartial = {}
			end
			table.insert(soundRunner, {"SayNumber", sayNumberStack[idx]})
		else
			table.insert(msgPartial, val)
		end
	end
	if 0 ~= #msgPartial then
		table.insert(soundRunner, {"Playback", msgPartial})
		msgPartial = {}
	end
	
	for _, runnerItem in pairs(soundRunner) do
		if "Playback" == runnerItem[1] then
			local strMsg = table.concat(runnerItem[2], " ")
			ivr:play(strMsg)
		elseif "SayNumber" == runnerItem[1] then
			local number = runnerItem[2]
			if "table" == type(number) then
				number = table.concat(number, "")
			end
			smartApp("SayNumber", number)
		end
	end
	
	return true		
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local strategy  	= require"IExServiceStrategy":new() -- meta class
strategy.userAttrs  = nil	-- user attributes for current room
strategy.apiRequest = nil	-- API object

--- @see @see description in IExServiceStrategy
function strategy:serviceMUR(forbidCallback, val)
	if "room" ~= self.userAttrs.userKind then
		return forbidCallback(self.name, "User does not have kind 'room'!")
	end
	
	-- retrieve current MUR value
	local curValue = self.apiRequest:getMURStatus() 
	
	-- validate function for MUR value
	local validate = function(val)
		local val = tonumber(val)
		return 0 == val or 1 == val
	end
	
	-- Playback in case value was not pushed
	if not validate(val) then	
    	local announce 			= {"Service make up the room"}
    	announce[#announce + 1] = "1" == curValue and "activated" or "deactivated"
    	local actions			= {"Press 0 to deactivate and 1 to activate"}
    	
		val = __get_userValue({announce, actions}, {validate = validate, maxChars = 1})
    end
    
    if val then	 -- processing MUR value
    	local rv, errMsg = self.apiRequest:setMURStatus(val) 
    	if rv then
    		ivr:tell("Service make up the room "..("1" == val and " activated" or " deactivated") .. ". Thank you!")
    	else
    		print(string.format("Error during setting MUR status for room %q: %s", self.userAttrs.telephoneNumber or "", errMsg or ""))
    	end
    end
	
	return smartApp("Hangup", hcause.NORMAL)
end

--- @see @see description in IExServiceStrategy
-- @note this service uses in PBX strategy
function strategy:serviceWakeUp(forbidCallback, val)
	-- validate function for WakeUp value
	local validate = function(val)
		local rv 	 = nil
		if not val then
			return rv
		end
		
		local valLen = string.len(val)
		if 1 == valLen then
			val = tonumber(val)
			rv = 0 == val or 1 == val
		elseif 1 < valLen and 4 >= valLen then
			local hour, min = string.match(val, "(%d?%d)(%d%d)")
			hour = tonumber(hour) or -1
			min  = tonumber(min) or -1
			if 0 <= hour and 24 > hour and 0 <= min and 60 > min then
				rv = true
			end
		end
		
		return rv
	end
	
	-- reading first in the queue wake up call
	local hourStr = ""
	local minStr  = ""
	local wkItem  = self.apiRequest:getWakeUp()
	if "table" == type(wkItem) then
		hourStr = string.format("[%s]", wkItem.timeVal.hour)
		minStr  = 0 ~= tonumber(wkItem.timeVal.min) and string.format("[%s]", wkItem.timeVal.min) or "" 
	end
	
	-- retrieve user choice
	if not validate(val) then
		local announce = {"Wake up call"}
		announce[#announce + 1] = wkItem and string.format("activated", hourStr, minStr) or "deactivated"
		local actions  = {(wkItem and "Press 0 to deactivate, 1 for details or p" or "P") .. "lease enter the time for the wake up call in #4 digit format"}    --TODO(sounds): replace sound to 'four_digit_format'
		val 		  = __get_userValue({announce, actions}, {validate = validate, maxChars = 4})	 
	end
	if val then
		if 1 == #val then	-- disable/read nearest alarm clock
			if "table" == type(wkItem) then
    			if "1" == val then
    				__smartPlayback("Wake up call activated for %s %s", hourStr, minStr)
    			elseif "0" == val then
    				local rs, errMsg = self.apiRequest:delWakeUp(wkItem.id) 
    				if rs then
    					__smartPlayback("Wake up call for %s %s deactivated!", hourStr, minStr)
    				else
    					print("Error during alarm clock removing:", errMsg)
    				end
    			end
    		else
    			__smartPlayback("Wake up call deactivated")
    		end
		else 				-- set new alarm clock
			local rv, errMsg = self.apiRequest:setWakeUp(val) 
			if rv then
				local hourStr = string.format("[%s]", rv.hour)
				local minStr  = 0 ~= tonumber(rv.min) and string.format("[%s]", rv.min) or ""
				__smartPlayback("Wake up call for %s %s activated!", hourStr, minStr)
			else
				print(string.format("Error during alarm clock setup for room %q: %s", self.userAttrs.telephoneNumber or "", errMsg or ""))
			end
		end
	end
	
	return smartApp("Hangup", hcause.NORMAL)
end


--- Object constructor
function APIExServiceStrategy:new(userAttrs)
	if "table" ~= type(userAttrs) then
		return false, "Invalid root attributes format"
	end
	
	-- processing room identification
	local cid = userAttrs.telephoneNumber 
	if not cid then
		return false, "Not found room identification"
	end
	
	-- processing authorization data
    local adminData          = require"common.userAuth.adminAuthLogin".getBackendData()
	local apiRequest, errMsg = require"APIWHotelRequest":new(userAttrs, {user = adminData.cn, pass = adminData.userPassword})
	if not apiRequest then
		return false, errMsg
	end
	
	local obj = {}
	setmetatable(obj, {__index = strategy})

	obj.userAttrs  = userAttrs
	obj.apiRequest = apiRequest
	obj.name	   = "API extended service"
	
	return obj
end

return APIExServiceStrategy