--- Module of the Wildix Hotel API requests
-- 
-- @file    APIWHotelRequest.lua 
-- @author  Vyacheslav Khudan
-- @created 2015/07/10

--@module APIWHotelRequest
local APIWHotelRequest   = {}

--- Some local environment
local client  = require"core.httpClient":new()

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Processing responce result
local function __processResponce(resp)
	if "table" ~= type(resp) then
		return false
	end
	
	local rv    = nil
	local err   = "Unknown responce type"
	local json 	= require"dkjson"
	
	if resp and resp.body then
		resp = json.decode(resp.body)
		if "table" == type(resp) then
    		if "error" == resp.type then
    			err = resp.reason
    		elseif "result" == resp.type then
    			rv = resp.result
    			err = rv and nil or "Invalid result format" 
    		end
    	end
	end
	
	return rv, err
end

--- Retrieve user presence
local function __getUserPresence(roomId, authData)
	if not roomId or not authData then
		return false
	end
	
	local reqURL  = string.format("http://127.0.0.1/api/v1/User/%s/Presence/", roomId)
	local resp 	  = client:get(reqURL, {authUser = authData.user, authPass = authData.pass})
	local rv, err = __processResponce(resp)
	
	return "table" == type(rv) and rv.show, err
end

--- Set user presence
local function __setUserPresence(roomId, presence, authData)
	if not roomId or not presence or not authData then
		return false
	end
	
	local reqURL  = string.format("http://127.0.0.1/api/v1/User/%s/Presence/", roomId)
	local data    = {status = presence}
	local resp    = client:put(reqURL, data, {authUser = authData.user, authPass = authData.pass})
	local rv, err = __processResponce(resp)
	if rv and "Success" == rv.result then
		return true
	end
	
	return rv, err
end

--- Get all WakeUps
local function __getAllWakeUp(authData)
	if not authData then
		return {}
	end
	
	local reqURL  = "http://127.0.0.1/api/v1/originate/AlarmClocks/"
	local resp    = client:get(reqURL, {authUser = authData.user, authPass = authData.pass})
	local rv, err = __processResponce(resp)
	if "table" == type(rv) and rv.records then
		return rv.records
	else
		return {}, err or "Invalid result format"
	end
	
	return rv, err
end

--- Remove wakeUp by ID
local function __deleteWakeUp(id, authData)
	if not id or not authData then
		return false
	end
	
	local reqURL  = "http://127.0.0.1/api/v1/originate/AlarmClocks/"..id
	local resp    = client:delete(reqURL, {authUser = authData.user, authPass = authData.pass})
	local rv, err = __processResponce(resp)
	if "Success" == rv then
		return true
	end
	
	return rv, err
end

--- Set up wakeUp
local function __setWakeUp(roomId, wkTime, authData)
	if not roomId or not wkTime then
		return false
	end
	
	local reqURL  = "http://127.0.0.1/api/v1/originate/AlarmClocks/"
	local data    = {number = roomId, postpone = wkTime}
	local resp    = client:post(reqURL, data, {authUser = authData.user, authPass = authData.pass})
	local rv, err = __processResponce(resp)
	if rv then
		return true
	end
	
	return rv, err
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local request = {	-- meta class
	 authData  = nil		-- request authorization data
	,roomId	   = nil		-- caller room identification
	,roomAttrs = nil		-- caller attributes 
}	
--- Retreive user MUR status
function request:getMURStatus()
	local rv 	   		= "0"	
	local presence, err = __getUserPresence(self.roomId, self.authData)
	if "mur" == string.lower(presence or "") then
		rv = "1"
	end
	return rv, err
end

--- Retreive user MUR status
-- @param val		A value of the MUR status
function request:setMURStatus(val)
	if not val then
		return false, "Invalid input data"
	end
	
	local presence = "1" == val and "mur" or "available"	
	return __setUserPresence(self.roomId, presence, self.authData)
end

--- Retreive wake-ups for room
-- @return first by time wake-up of the next day
function request:getWakeUp()
	local function __wkDateParser(date)
		local timeVal = nil
		local rv  	  = 0
		if "string" ~= type(date) then
			return rv
		end
		
		local day, month, year, hour, min = date:match("(%d+)%/(%d+)%/(%d+)%s*(%d+):(%d+)")
		if day and month and year and hour and min then
			local resDateTime = {
				["sec"]		= 0,
				["hour"]	= hour,
				["min"]		= min,
				["year"] 	= year,
				["month"] 	= month,
				["day"]		= day
			}
			rv = os.time(resDateTime) or 0
			if 0 ~= rv then
				timeVal = {hour = hour, min = min}
			end
		end
		
		return rv, timeVal
	end
	
	-- let's search wakeUps for current room
	local rWakeUp		 = nil
	local wakeUpListRoom = nil
	local wakeUpListAll  = __getAllWakeUp(self.authData)
	for _, wk in pairs(wakeUpListAll) do
		if "table" == type(wk) and wk.number then
			if self.roomId == wk.number then
				local wkTime, wkTimeVal = __wkDateParser(wk.postpone)
				if not rWakeUp or rWakeUp.unixTime > wkTime then
					wk.unixTime = wkTime
					wk.timeVal  = wkTimeVal
					rWakeUp  	= wk
				end
			end
		end
	end
	
	return rWakeUp	
end

--- Delete wakeUp from the room by ID
-- @param id	A wakeup identificator
-- @return boolean value
function request:delWakeUp(wkId)
	if not wkId then
		return false
	end
	
	return __deleteWakeUp(wkId, self.authData)
end

--- Set wakwUp in the room
-- @param wkTimeVal		A time of the installed wakeUp
-- @return boolean value
function request:setWakeUp(wkTimeVal)
	if not wkTimeVal then
		return false
	end
	
	local function __getWakeUpDate(wkHour, wkMin)
		if not wkHour or not wkMin then
			return false
		end
		wkHour = tonumber(wkHour)
		wkMin  = tonumber(wkMin)
		
		local curDayData  = os.date("*t", os.time())
		local wkDate      = curDayData
		if wkHour < curDayData.hour or (wkHour == curDayData.hour and wkMin <= (curDayData.min + 1)) then
			local nextDayData = os.date("*t", os.time() + 24*60*60 - 5)
			wkDate 			  = nextDayData
		end
    	wkDate.hour = 0
    	wkDate.min  = 0
    	wkDate.sec  = 0
    	
    	if 1 == string.len(wkDate.day) then
    		 wkDate.day = "0" .. wkDate.day
    	end
    	if 1 == string.len(wkDate.month) then
    		 wkDate.month = "0" .. wkDate.month
    	end
    	
    	return wkDate
    end
    
    local wkHour, wkMin = wkTimeVal:match("(%d?%d)(%d%d)")
    local wkDate 		= __getWakeUpDate(wkHour, wkMin)
    local errMsg		= nil
    if wkDate and wkHour and wkMin then
    	local wkDateTime = string.format("%s/%s/%s %s:%s", wkDate.day, wkDate.month, wkDate.year, wkHour, wkMin)
		local rs		 = false
		rs, errMsg 		 = __setWakeUp(self.roomId, wkDateTime, self.authData)     	
    	if rs then
    		return {hour = wkHour, min = wkMin}
    	end
	end
	
	return false, errMsg
end


--- Object constructor
function APIWHotelRequest:new(roomAttrs, authData)
	if "table" ~= type(roomAttrs) or "table" ~= type(authData) then
		return false, "Invalid input arguments"
	end
	if not authData.pass then
		return false, "Password does not found"
	end
	local roomId  = roomAttrs.telephoneNumber
	authData.user = authData.user or roomId
	
	local obj = {}
	setmetatable(obj, {__index = request})

	obj.authData  = authData
	obj.roomId    = roomId
	obj.roomAttrs = roomAttrs
	
	return obj
end

return APIWHotelRequest