--- Processing operations in case using extended (Hotel) services  via daemon implementation
-- 
-- @file    DaemonExServiceStrategy.lua 
-- @author  Vyacheslav Khudan
-- @created 2015/07/09

--- @module DaemonExServiceStrategy
local DaemonExServiceStrategy = {}

----------------------------------------------------------------
-- Implementation interface
----------------------------------------------------------------
--- Get values from user via IVR service
local function __get_userValue(sounds, opts)
	if "table" ~= type(sounds) or "table" ~= type(opts) then
		return false
	end
	
	local validate = opts.validate
	if not validate then
		return false
	end
	local ivr 		   = ivr
	local val 		   = nil
	local withAnnounce = false
	if 2 == #sounds then
		withAnnounce = true
	end
	
	if withAnnounce then
    	val = ivr:read(table.concat(sounds[1], " "), {
    		 max_chars 	= opts.maxChars
    		,retry		= 1
    		,timeout	= 0.1
    		,nobeep 	= true
    		,validate 	= function() return true end
    	})
    end
	if not validate(val) or not withAnnounce then
		val = ivr:read(table.concat(sounds[withAnnounce and 2 or 1], " "), {
			 max_chars = opts.maxChars
			,validate  = validate
			,failure   = function() ivr:tell(opts.failSound or "The number you entered is incorrect, please try again") end
		})
		if not val then
			ivr:tell("Good bye. Hangup")
		end
	end
	
	return val
end

-- TODO: must be replaced by correct logic in ivr:tell with SayNumber
--- Playback message with correct digit pronunciation
local function __smartPlayback(msg, ...)
	if not msg then
		return false
	end

	local msg 			 = string.format(msg, ...)
	local soundRunner    = {}
	local playbackStack  = {}
	local sayNumberStack = {}
	for word in msg:gmatch("%S+") do
		local number = word:match("%[(%d+)%]")
		if number then
			table.insert(sayNumberStack, number)
			table.insert(playbackStack, 0)
		else
			table.insert(playbackStack, word)
			table.insert(sayNumberStack, 0)
		end
	end
	
	-- @limitation Will not be played first digits at the beginning message
	local msgPartial = {}
	for idx, val in pairs(playbackStack) do
		if 0 == val then
			if 0 ~= #msgPartial then
				table.insert(soundRunner, {"Playback", msgPartial})
				msgPartial = {}
			end
			table.insert(soundRunner, {"SayNumber", sayNumberStack[idx]})
		else
			table.insert(msgPartial, val)
		end
	end
	if 0 ~= #msgPartial then
		table.insert(soundRunner, {"Playback", msgPartial})
		msgPartial = {}
	end
	
	for _, runnerItem in pairs(soundRunner) do
		if "Playback" == runnerItem[1] then
			local strMsg = table.concat(runnerItem[2], " ")
			ivr:play(strMsg)
		elseif "SayNumber" == runnerItem[1] then
			local number = runnerItem[2]
			if "table" == type(number) then
				number = table.concat(number, "")
			end
			smartApp("SayNumber", number)
		end
	end
	
	return true		
end

----------------------------------------------------------------
-- Wake up data helper 
----------------------------------------------------------------
--- Converstion to FIAS date format: HHMMSS YYMMDD
local function __prepare_fiasDate(time, date)
	if not time or not date then
		return false
	end
	
	-- time conversion
	local hour, min = time:match("(%d?%d)(%d%d)")
	local fiasTime  = string.format("%02d%02d00", hour, min)
	
	-- date conversion
	local day   = nil
	local month = nil
	local year  = nil
	
	local curDate = os.date("*t", os.time())
	local dateLen = string.len(date)
	if "0" == date then
		hour = tonumber(hour)
		min  = tonumber(min)
		local wkDayDate = curDate
		if hour < curDate.hour or (hour == curDate.hour and min <= (curDate.min + 1)) then
    		wkDayDate = os.date("*t", os.time() + 24*60*60 - 5)
    	end
		day	  = wkDayDate.day
		month = wkDayDate.month
		year  = string.match(wkDayDate.year, "%d%d(%d%d)")
	elseif 4 >= dateLen then		-- dd/mm
		day, month = string.match(date, "(%d%d)(%d%d)")
		year	   = string.match(curDate.year, "%d%d(%d%d)")
	elseif 6 >= dateLen then		-- dd/mm/yy
		day, month, year = string.match(date, "(%d%d)(%d%d)(%d%d)")
	elseif 8 >= dateLen then		-- dd/mm/yyyy
		day, month, year = string.match(date, "(%d%d)(%d%d)%d%d(%d%d)")
	end
	local fiasDate = string.format("%02d%02d%02d", year, month, day)
	
	return fiasTime, fiasDate
end

--- Collect wake ups
local function __collect_wakeUps(DBKey)
	local rv = {}
	if not DBKey then
		return rv
	end
	
	local function __unix2fias(dateStr)
		if not dateStr or "" == dateStr then
    		return nil
    	end
    	return os.date("%H%M%S", dateStr), os.date("%y%m%d", dateStr)	
    end
		
	local wkItem 	  = nil
	local wkMapString = channel.DB(DBKey.."WAKE-UP-MAP"):get() or ""
	for wkId in wkMapString:gmatch("(%d+)!?") do
		local wkUnixTime = channel.DB(DBKey .."WAKE-UP/"..wkId..""):get() or ""
		if not wkItem or wkUnixTime < wkItem.unix then
			local wkTime, wkDate = __unix2fias(wkUnixTime) 
			if wkTime and wkDate then
				wkItem = {
    				 id   = wkId
    				,time = wkTime
    				,date = wkDate
    				,unix = wkUnixTime
    			}
    		end
		end
	end
	
	return wkItem or {}
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local strategy  = require"IExServiceStrategy":new() -- meta class
strategy.family 	= nil	-- family used by daemon
strategy.userAttrs  = nil	-- user attributes for current room
strategy.ACL	  	= nil	-- user permissions

--- @see description in IExServiceStrategy
function strategy:isCallAllowed(_, callClass)
	-- retrieving ACL value for current caller
	local aclVal = self.ACL
	repeat
        -- check cached value
        if aclVal then
            break
        end
        
        -- checking hotel ACL manage variable
    	local isCheckFIASACL = channel.CHECKFIASACL:get() or ""
    	if "no" == isCheckFIASACL then
    	   aclVal = "PBXACL"
    	   break 
    	end
	
    	-- checking user in callweaver database with family like /WILDIX-HOTEL/100/ACL : ALL
    	local cid = self.userAttrs.telephoneNumber
    	aclVal 	  = channel.DB(self.family .. "/" .. cid .. "/ACL"):get() or ""
    	if "" == aclVal then
    		aclVal = "PBXACL"
    		break
    	end
	until true

    -- caching ACL value
    if not self.ACL then
    	self.ACL = aclVal
    end

	local aclMap = {	-- map of the PBX and HOTEL ACL values
		-- blocked: internal, shortnumber
		 [-1] 	= "BLOCKED"     -- currently used PBX ACL in this case
		,[2] 	= "LOCAL"		-- local
		,[3] 	= "NATIONAL"	-- national
		,[4]	= "NATIONAL"	-- mobile
		,[5]    = "INTERNAL"    -- emergency
		,[6]	= "NATIONAL"	-- free
		,[7]	= "NATIONAL"	-- premium1
		,[8]	= "NATIONAL"	-- premium2
		,[9]	= "NATIONAL"	-- premium3
		,[10]	= "NATIONAL"	-- premium4
		,[11] 	= "ALL"			-- northamerica
		,[12]	= "ALL"			-- africa		
		,[13]	= "ALL"			-- europe1
		,[14]	= "ALL"			-- europe2
		,[15]	= "ALL"			-- southamerica
		,[16]	= "ALL"			-- oceania
		,[17]	= "ALL"			-- russia
		,[18]	= "ALL"			-- asia1
		,[19]	= "ALL"			-- asia2
	}
	local aclOrder = {
	     PBXACL     = -2
    	,BLOCKED	= -1
    	,INTERNAL	= 0
    	,LOCAL		= 1
    	,NATIONAL	= 2
    	,ALL		= 3
	}
	
	local callVal    = aclMap[tonumber(callClass) or -1] or aclMap[-1]
	local callNumVal = aclOrder[callVal]
	local aclNumVal  = aclOrder[aclVal]
	
	local rv = false
	if 0 <= aclNumVal then
		if callClass then
		    if 0 <= callNumVal and aclNumVal >= callNumVal then
                channel.CDR("UserField"):set("WHOTEL-CALL:" .. callVal)	-- set flag for processing PostSimple
                rv = true
            end
		else
            rv = true
		end
	elseif -2 == aclNumVal then
	   if callClass then
            channel.CDR("UserField"):set("WHOTEL-CALL") -- set flag for processing PostSimple calls via PBX ACL
       end
       rv = nil
	end
	
	return rv
end

--- @see description in IExServiceStrategy
function strategy:forbidCall(_, appName, msg)
	-- Do not block some hotel services (like MUR)
	if "WHOTELMUR" == appName or "WHOTELINSPECT" == appName then
		return true
	end
		
	-- update CDR fields
	channel.CDR("userfield"):set("WHOTEL_ACL_BLOCK")
	channel.CDR("to_name"):set("Wildix Hotel ACL")
	
	print(msg or "Blocked by Wildix Hotel ACL")
	return smartApp("Hangup", hcause.REJECT)
end

--- @see description in IExServiceStrategy
function strategy:serviceMUR(forbidCallback, val, inspectVal)
	local channel  = channel
	
	-- set general variables
	local cid 	   = self.userAttrs.telephoneNumber
	local key      = self.family.."/"..cid.."/MUR"
	
	-- validate function for MUR value
	local validate = function(val)
		local val = tonumber(val)
		return 0 == val or 1 == val
	end
	
	-- Playback in case value was not pushed
	if not inspectVal and not validate(val) then	
    	local announce 			= {"Service make up the room"}
    	local actions			= {"Press 0 to set the room status clean and 1 to set the room status dirty"}
    	
		val = __get_userValue({announce, actions}, {validate = validate, maxChars = 1})
    end
    
    -- update user choise by inspected status value
    val = inspectVal and inspectVal or val
    
    if val then	 -- processing MUR value
        channel.CDR("userfield"):set("WHOTEL-MUR:"..val)
        channel.CDR("to_name"):set("WHotel MUR: "..val)
        channel.DB(key):set(val)
    	
    	ivr:tell("Service status updated. Thank you!")
    end
	
	return smartApp("Hangup", hcause.NORMAL)
end

--- @see description in IExServiceStrategy
function strategy:serviceWakeUp(forbidCallback, val)
	--- validate procedures
	local validate_time = function(time)
		local rv 	 = nil
		if not time then
			return rv
		end
		
		local valLen = string.len(time)
		if 1 < valLen and 4 >= valLen then
			local hour, min = string.match(time, "(%d?%d)(%d%d)")
			hour = tonumber(hour) or -1
			min  = tonumber(min) or -1
			if 0 <= hour and 24 > hour and 0 <= min and 60 > min then
				rv = true
			end
		end
		
		return rv
	end
	
	local validate_value = function(val)
		if not val then
			return false
		end
		
		local actLen = string.len(val)
		if 1 == actLen then
			local actNum = tonumber(val)
			return 1 == actNum or 0 == actNum
		else
			return validate_time(val)
		end
	end
	
	--- local set general variables
	local cid 	    = self.userAttrs.telephoneNumber
	local wkDBKey	= string.format("%s/%s/", self.family, cid)
	local wkAction	= false
	local wkTime	= nil
	local wkDate	= nil
	
	-- retreive wake up info
	local isWKExist = channel.DB(wkDBKey.."WAKE-UP-MAP"):get() or ""
	isWKExist		= "" ~= isWKExist and isWKExist
	
	if not validate_value(val) then
		local announce = {"Wake up call " .. (isWKExist and "activated" or "deactivated")}
		local actions  = {(isWKExist and "Press 0 to deactivate, 1 for details or p" or "P") .. "lease enter the time for the wake up call in #4 digit format"} --TODO(sounds): replace sound to 'four_digit_format'
		val 		   = __get_userValue({announce, actions}, {validate = validate_value, maxChars = 4})	 
	end
	if val then
		if 1 == #val then	-- disable/read nearest wake-up
			local wkItem = __collect_wakeUps(wkDBKey)
			if "table" == type(wkItem) then
				wkTime = wkItem.time
				wkDate = wkItem.date
				if "1" == val then
					if wkItem and wkItem.time then
    					wkAction = "Playback"
    					local hour, min = wkItem.time:match("(%d%d)(%d%d)%d+")
    					local hourStr 	= string.format("[%s]", hour)
    					local minStr  	= 0 ~= tonumber(min) and string.format("[%s]", min) or "" 
    					__smartPlayback("Wake up call activated for %s %s", hourStr or "", minStr or "")
    				end
				elseif "0" == val then
					if wkTime and wkDate then
						wkAction = "Clear"
					end
				end
			else
				ivr:tell("Wake up call deactivated")
			end
		else				-- set new wake-up
			wkTime, wkDate = __prepare_fiasDate(val, "0")
			if wkTime and wkDate then
				wkAction = "Request"
			end
		end
	end 
	
	local hangupCause = hcause.FAILURE
	if wkAction then
		if "Playback" ~= wkAction then
			print(string.format("Sending wake-up %s at '%s/%s' !", wkAction:lower(), wkTime, wkDate))
			channel.CDR("userfield"):set(string.format("WHOTEL-WAKEUP-%s:%s!%s!%s", wkAction:upper(), cid, wkTime, wkDate))
        	channel.CDR("to_name"):set(string.format("WHotel W%s: %s/%s", wkAction:sub(1, 1), wkTime, wkDate))
        else
        	channel.CDR("to_name"):set(string.format("WHotel WakeUp: %s/%s", wkTime, wkDate))
        end
		ivr:tell("Thank you!")
		hangupCause = hcause.NORMAL
		
	end
	return smartApp("Hangup", hangupCause)
end

--- @see description in IExServiceStrategy
function strategy:serviceMiniBar(forbidCallback, item, count)
	local channel  = channel
	local ivr	   = ivr
	
	-- validate function for MUR value
	local validate = function(var, val)
		if "countSoft" == var then
			val = tonumber(val) or 1
			val = 2 >= string.len(val) and val or nil
		else
    		val = tonumber(val)
    		val = 0 ~= val and val or nil
    		if val then
    			if "article" == var then
    				val = 4 >= string.len(val) and val or nil
    			elseif "cost" == var then
    				val = 15 >= string.len(val) and val or nil
    			elseif "countStrict" == var then
    				val = 2 >= string.len(val) and val or nil
    			else
    				val = nil
    			end
    		end
    	end
		return val
	end
	local validate_article = function(val)
		return validate("article", val)
	end
	local validate_cost = function(val)
		return validate("cost", val)
	end
	local validate_strictCount = function(val)
		return validate("countStrict", val)
	end
	local validate_softCount = function(val)
		return validate("countSoft", val)
	end
	
	-- set up minibar type flag and validation functions
	local minibarType 	 = channel.MinibarType:get() or ""
	local isArticleType  = minibarType:find("ArticleOnly")
	local validate_item  = validate_cost
	local validate_count = validate_softCount
	if isArticleType then
		validate_item  = validate_article
		validate_count = validate_strictCount
	end
	
	-- processing pushed information
	item  = validate_item(item)
	count = validate_strictCount(count)
	if not item or not count then
		if not item then
			item = __get_userValue({{string.format("Please enter %s, followed by pound", isArticleType and "the product article" or "the item price")}}, {validate = validate_item})
		end
		
		if item and not count then
			count = __get_userValue({{"Please enter the number of items, followed by pound"}}, {validate = validate_count})
		end
	end
	
	-- processing MiniBar values
	item  = validate_item(item)
	count = validate_count(count)
	if item and count then
		channel.CDR("userfield"):set("WHOTEL-MINIBAR:"..item.."#"..count)
		channel.CDR("to_name"):set("WHotel MiniBar: #"..item)

		ivr:tell("Minibar status updated. Thank you.")
		print(string.format("Hotel minibar service requested from room %q: #%d (%d)", self.userAttrs.telephoneNumber, item, count))
	end
	
	return smartApp("Hangup", hcause.NORMAL)
end

--- @see description in IExServiceStrategy
function strategy:serviceINSPECT(forbidCallback, val)
	local inspectVal = 2	-- usual inspect with related room state 
	if "1" == val then
		inspectVal = -2		-- force inspect with room state vacant
	end
	
	return self:serviceMUR(forbidCallback, nil, inspectVal)
end

--- @see description in IExServiceStrategy
function strategy:applicationWakeUp(forbidCallback, mode, arg)
	local channel = channel
	
	local modesTable = {
		 immediate = 0
		,sound	   = 1
		,user	   = 0
	}
	if not modesTable[mode] then
		return forbidCallback(self.name, "Invalid WakeUp application mode: %s", mode or "")
	end
	if #{arg} < modesTable[mode] then
		return forbidCallback(self.name, "Not found WakeUp application argument(s)")
	end
	
	local WakeUpFlag = channel.WHotelWakeup:get() or ""
	if "" == WakeUpFlag then
		return forbidCallback(self.name, "Call made not from WakeUp service")
	end

	if "sound" == mode then
		smartApp("Playback", arg)
	elseif "user" == mode then
		local CheckVal = arg or "5"	-- default value
		local validate = function(val)
			return val == CheckVal 
		end
		local val = __get_userValue({{"Please enter the number "..CheckVal..", followed by pound"}}, {validate = validate})
		if not val then
			return smartApp("Hangup", hcause.FAILURE)
		end
	end
	
	print("WakeUp request was confirmed!")
	channel.CDR("UserField"):set(WakeUpFlag)
	
	if "immediate" == mode and arg and "" ~= arg then
		smartApp("Playback", arg)
	end
	
	return smartApp("Hangup", hcause.NORMAL)
end


--- Object constructor
function DaemonExServiceStrategy:new(userAttrs, family)
	if not family or "table" ~= type(userAttrs) then
		return false
	end
	if not userAttrs.telephoneNumber then
		return false
	end
	
	local obj = {}
	setmetatable(obj, {__index = strategy})

	obj.userAttrs = userAttrs
	obj.family 	  = family
	obj.name	  = "Daemon extended service"
	
	return obj
end

return DaemonExServiceStrategy
