--- Data provider for call out sevice
-- 
-- @file    DataProvider.lua 
-- @author  Vyacheslav Khudan
-- @created 2016/08/30

package.path = package.path .. ";/var/lib/callweaver/dialplan/exService/callOut/dataProvider/?.lua"

--- @module DataProvider
local DataProvider = {
    mData = nil     -- internal module data
                        -- mAccessorObj
                        -- mAccessorType
                        -- mTag
}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local function __accessorRequest(o, m, k, v)
    if "table" ~= type(o) or not m or not k then
        return false
    end
    
    if not o.mData.mTag then
        return nil
    end
    
    if k and "table" ~= type(k) then
        k = o.mData.mTag .. "/" .. k
    end
    return o.mData.mAccessorObj[m](o.mData.mAccessorObj, k, v)
end

local __accessor = {
     put = function(o, k, v)
         v = "boolean" == type(v) and tostring(v) or v
         return __accessorRequest(o, "put", k, v)
     end
    ,puta = function(o, k, v)
        local data  = __accessorRequest(o, "get", k) or ""
              data  = require"dkjson".decode(data) or {}
              v     = "table" == type(v) and v or {v}
        local added = {} 
        for _, val in ipairs(data) do
            added[val] = true
        end
        for key, val in pairs(v) do
            if "number" == type(key) then
                if not added[val] then
                    added[val] = true
                    table.insert(data, val)
                end
            else
                data[key] = val
            end
        end
        data       = require"dkjson".encode(data)
        return __accessorRequest(o, "put", k, data)
     end
    ,get = function(o, k)
        local rv = __accessorRequest(o, "get", k)
        if "true" == rv then
            rv = true
        elseif "false" == rv then
            rv = false
        elseif "nil" == rv then
            rv = nil
        end
        return rv
     end
    ,geta = function(o, k)
        local data = __accessorRequest(o, "get", k) or ""
        data       = require"dkjson".decode(data) or {}
        return data
     end
    ,del = function(o, k)
        return __accessorRequest(o, "del", k)
     end
    ,dela = function(o, k, v)
        local data = __accessorRequest(o, "get", k) or ""
        data       = require"dkjson".decode(data) or {}
        v          = "table" == type(v) or {v}
        local rv   = {}
        for idx, d in pairs(data) do
            for _, val in ipairs(v) do
                if val == d then
                    table.remove(data, idx)
                    break
                end
            end
        end
        table.sort(data, 
            function(a, b) 
                if nil == a then 
                    return false 
                elseif nil == b then 
                    return true
                end
        end)
        
        if 0 == #data then
            return __accessorRequest(o, "del", k)
        else
            data = require"dkjson".encode(data)
            return __accessorRequest(o, "put", k, data)
        end
    end
}

local function __processObjArray(o, m, ptag)
    local rv = false
    if not m or "table" ~= type(o) then
        return rv
    end
    
    local tag  = o.mData.mTag or ""
    local cTag = nil
    local key  = nil
    local val  = nil
    if ptag then
        cTag, key      = ptag:match("^(%w+)%-(%d+%-%d+)$")
    else
        cTag, key, val = tag:match("^(%w+)%-(%d+%-%d+)%-([^%-]+)$")
    end
    if not cTag then
        return rv
    end
    o.mData.mTag = cTag
    if "get" == m then
        local data = __accessor.geta(o, key)
        rv = "table" == type(data) and data or {}
    elseif "add" == m then
        rv = __accessor.puta(o, key, val)
    elseif "del" == m then
        rv = __accessor.dela(o, key, val)
    end
    o.mData.mTag = tag
    
    return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Initialize singletone data provider object
-- @param pType     The provider type (like int/internal or ext/external)
-- @param tag       The provider tag (like CALLOUT-<ext>-<user>)
-- @return object with related methods
function DataProvider:getInstance()
    -- Retrieving provider type by environment variables
    local pType = "ext"
    if "table" == type(app) and "table" == type(channel) then
        pType = "int"
    end
    
    if self.mData and pType == self.mData.mAccessorType and self.mData.mAccessorObj then
        return self
    end
    
    -- pType: internal/external
    local cfg      = { ident = pType }
    local accessor = require"Accessor":new(cfg)
    if not accessor then
        return nil
    end
    
    local o               = self
    o.mData               = gSpace and gSpace:get4Module() or {}
    o.mData.mAccessorObj  = accessor
    o.mData.mAccessorType = pType
    
    --[[ Data structure template
        CALLOUT-<ext>-<user>-<color>
            -- status
            -- color
            -- params
            -- destination[]
                -- active -> active person call 
                -- enabled -> disable from dialplan to remove re-call or by script timeout
                -- pass
                -- id
                -- contact[]
                    -- num
                    -- try
                    -- unavailtry
    --]]

    function o:existByName(n)
        return n and self:getProperty("uid")
    end

    function o:setName(tag)
        if tag then
            self.mData.mTag = tag
            return self:getProperty("uid")
        end
    end
	 
    function o:getIterator(ptag)
        local data = __processObjArray(self, "get", ptag)
        if "table" ~= type(data) then
            return nil
        end
        
        local i    = nil
        local v    = nil
        return function()
            i, v = next(data, i)
            return v and ptag.."-"..v
        end
    end
    
    function o:setProperty(name, val)
        if not name or nil == val then
            return nil
        end
        
        if "uid" == name then
            __processObjArray(self, "add")
        end
        
        val = val or false
        return __accessor.put(self, name, val) 
    end
    
    function o:getProperty(name)
         return __accessor.get(self, name)
    end
     
    function o:setParams(data)
        if "table" == type(data) then
            return __accessor.puta(self, "params", data)
        end
    end
    
    -- Retrieve service params by variable or all data
    function o:getParams(p)
        local rv   = nil
        local data = __accessor.geta(self, "params")
        if "table" == type(data) then
            if p then
                rv = p and data[p] or nil
            else
                rv = data
            end
        end
        
        return rv
    end
    
    
    --- Processing destination data
    function o:addDestination(data, def)
        if "table" == type(data) and data.name then
            --[[ Examle:
                data = {
                    name     = "Jon Doe"
                    pass     = "XXXX"
                    id       = X
                    numbers = {
                        [1] = {
                            num = "+XXXXXX"
                        } 
                    }
                }
                
                def = {     -- optional
                    active     = false
                    enabled    = true
                    try        = 3
                    unavailtry = 30
                }
            --]]
            
            def        = def or {}
            local idef = {
                 active     = nil ~= def.active and tostring(def.active) or "false"
                ,enabled    = nil ~= def.enabled and tostring(def.enabled) or "true"
                ,try        = def.try or 3
                ,unavailtry = def.unavailtry or 30
            }
            
            __accessor.puta(self, "dests", data.destId)
            local destTag = "dest/" .. data.destId
            __accessor.put(self, destTag.."/active", idef.active)
            __accessor.put(self, destTag.."/enabled", idef.enabled)
            for key, val in pairs(data) do
                if val and "table" ~= type(val) then
                    __accessor.put(self, destTag.."/"..key, tostring(val))
                elseif val and key == "numbers" then
                    for i, numData in pairs(val) do
                        if numData.num then
                            __accessor.puta(self, destTag.."/numbers", "num"..i.."="..numData.num)
                            __accessor.put(self, destTag.."/".."contact"..i.."/num", numData.num)
                            __accessor.put(self, destTag.."/".."contact"..i.."/try", idef.try)
                            __accessor.put(self, destTag.."/".."contact"..i.."/unavailtry", idef.unavailtry)
                            __accessor.put(self, destTag.."/".."contact"..i.."/phase", "none")		-- preset with initial value
                        end
                    end
                end
                
            end
        end
    end
    
    function o:getDestCnt()
        local destList = __accessor.geta(self, "dests")
        return #destList
    end
    
    function o:getDestIterator()
        local destList = __accessor.geta(self, "dests")
        local i        = nil
        local val      = nil
        return function()
            i, val = next(destList, i)
            return val
        end
    end
    
    function o:getDestByNumber(number)
        if number then
            return __accessor.get(self, "dest/"..number)
        end
    end
    
    
    --- Processing destination parameters
    function o:setDestParams(ident, p, v)
        if ident and p and nil ~= v then
            v = tostring(v)
            return __accessor.put(self, "dest/"..ident.."/"..p, v)
        end
    end
    
    function o:getDestParams(ident, p)
        if ident and p then
            return __accessor.get(self, "dest/"..ident.."/"..p)
        end
    end    

    --- Processing destination contacts
    function o:getContactsCnt(destId)
        local numbersList = __accessor.geta(self, "dest/"..destId.."/numbers")
        return #numbersList
    end

    function o:getContactsItertator(destId)
        local numbersList = __accessor.geta(self, "dest/"..destId.."/numbers")
        local i           = nil
        local val         = nil
        return function()
            i, val = next(numbersList, i)
            if not val then
            	return nil
            end
            return string.match(val, "^num(%d+)=([+%d]+)$")
        end
    end
    
    function o:setContactParam(destId, numId, p, v)
        if destId and numId and p and nil ~= v then
            v = tostring(v)
            return __accessor.put(self, "dest/"..destId.."/".."contact"..numId.."/"..p, v)
        end
    end
    
    function o:getContactParam(destId, numId, p)
        if destId and numId and p then
            return __accessor.get(self, "dest/"..destId.."/".."contact"..numId.."/"..p)
        end
    end

    --- Processing contact parameters
    function o:setContactTry(destId, numId, v, t)
        t = t or "try"
        if "try" == t or "unavailtry" == t then
            return self:setContactParam(destId, numId, t, v)
        end
    end

    function o:getContactTry(destId, numId, t)
        t        = t or "try"
        local rv = self:getContactParam(destId, numId, t)
        return rv and tonumber(rv)
    end

    function o:incrContactTry(destId, numId, t)
        local rv = self:getContactTry(destId, numId, t)
        if rv then
            rv = rv + 1
            return self:setContactTry(destId, numId, rv, t)
        end
    end 

    function o:decrContactTry(destId,numId, t)
        local rv = self:getContactTry(destId, numId, t)
        if rv and rv > 0 then
            rv = rv - 1
            return self:setContactTry(destId, numId, rv, t)
        end
    end
    
    
    --- Clear service data
    function o:delAllData()
        __processObjArray(self, "del")
        return __accessor.del(self, {ident = "tree", val = self.mData.mTag})
    end
    
    return o
end

return DataProvider

