--- Module for call out services
-- 
-- @file    callOutService.lua 
-- @author  Vyacheslav Khudan
-- @created 2016/07/08

package.path = package.path .. ";/var/lib/callweaver/dialplan/exService/callOut/?.lua"

--- @module callOutService
local callOutService = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------

-- setup data to CDR fields
local function __setupCDRFields(data)
    local rv = false
    if not data then
        return rv
    end
    
    repeat
        -- setup all fields
        data = "table" == type(data) and data or {}
        if not data.serviceId or not data.serviceName then
            break
        end
        channel.CDR("c_from"):set(data.serviceId)
        channel.CDR("from_name"):set(data.serviceName)
        
        if not data.dstNum or not data.dstName then
            break
        end
        channel.CDR("c_to"):set(data.dstNum)
        channel.CDR("to_name"):set(data.dstName)
        
        if not data.tag then
            break
        end
        local tags = string.match((channel.CDR("tags"):get() or ""),"^(tags:.*)$")
        tags = tags and (tags..",") or "tags:"
        channel.CDR("tags"):set(tags..data.tag)
        
        rv = true
    until true
    
    return rv
end

-- cleanup destinations
local function __cleanupDests(callOut,destId)
   if not callOut or not destId then
      return false
   end
   local disabled = true
   for numId, number in callOut:getContactsItertator(destId) do
      if not (0 == callOut:getContactTry(destId, numId, "try") or 0 == callOut:getContactTry(destId, numId, "unavailtry")) then
         disabled = false
      end
   end
   if disabled then
      print("Last try or user made a choice - disable dest")
      callOut:setDestParams(destId, "enabled", false)
   end
   return false
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Call application uses to starting call out service
-- @param userId            A user identifier (internal value) 
-- @param pbName            A phonebook name for the out calls
-- @param limit             A number of the cuncurrent calls
-- @param eventsCtx         A context with description for event ID/pin(s)
-- @param outCallsCtx       A context for particular calls from PB
-- @param unavailTryCount   A counter used for retry in the case destination channel is unavailable
-- @param confirmCode       A code to make an event confirmation
-- @param refuseCode        A code to refuce an event
-- @param voiceDire         A directory used to search sounds for cuttent procedure
-- @return bolean value TRUE otherwiser FALSE with string error
function callOutService.initCall(userId, pbName, limit, eventsCtx, outCallsCtx, unavailTryCount, confirmCode, refuseCode, voiceDir)
    if not pbName or not userId then
        return false, "Invalid input args"
    end

    local fsutils    = require"core.fsutils"
    local callsLimit = tonumber(limit) or 4
    local ext        = channel.EXTEN:get() or ""
    local user       = channel.CALLERID("num"):get() or ""
    eventsCtx        = eventsCtx or "callout_events"
    outCallsCtx      = outCallsCtx or "callout_calls"
    unavailTryCount  = unavailTryCount or 3
    voiceDir         = voiceDir or "snd"
    if "" ~= voiceDir and fsutils.is_dir("/var/lib/callweaver/sounds/00000/"..voiceDir) then    -- need comment which sounds has to be uploaded
    	voiceDir = "00000/"..voiceDir
    else
    	voiceDir = "00000"
    end

    -- processing contact action codes
    confirmCode = "" ~= confirmCode and confirmCode
    refuseCode  = "" ~= refuseCode and refuseCode
    if confirmCode or refuseCode then   -- clear both for auto confirm by default
        confirmCode = confirmCode or 1
        refuseCode  = refuseCode or 3
    end
    
    -- check events list 
    if not extensions[eventsCtx] then
        return false, string.format("Not found events context - %s", eventsCtx or "")
    end
    
    -- check out calls context
    if not extensions[outCallsCtx] then
        return false, string.format("Not found out calls context - %s", outCallsCtx or "")
    end
    
    -- playback greeting notification and set channel variable
    if "function" == type(extensions[eventsCtx]["s"]) then
        extensions[eventsCtx]["s"]()
    end
    
    -- reading event ident for call out service
    -- SOUND: please enter the number (color)
    local options = {
		 max_chars = 8
		,retry 	  = 1
		,timeout   = 10
		,voiceDir  = voiceDir
		,validate  = function(v)
							return extensions[eventsCtx][v]
						 end
      ,failure   = function() end
	 }
	ivr:play("pin_test_or_real_start", {voiceDir = voiceDir})
    local colorId = ivr:read("beep", options)
    if not colorId or "function" ~= type(extensions[eventsCtx][colorId]) then
		  print("Pin not correct or no pin")
		  ivr:play("wrong_pin", {voiceDir = voiceDir})
    	  return smartApp("Hangup", hcause.NORMAL)
    end

	 -- read user-defined situation type
	 extensions[eventsCtx][colorId]() --execute extension (entered pin) in the event context (set EVENT_NAME variable and play sound "test" or "real")
    local color = channel.EVENT_NAME:get()
    if not color or "" == color then
    	print("Error. EVENT_NAME not set.")
    	return smartApp("Hangup", hcause.NORMAL)
    end

    -- check for cancelation request
    if string.find(color,"call_out_stop") then			-- predefined (by Wildix) color for canceling of started call-outs
		 local which = {}
		 if not string.find(color,"_all") then
             local eventName = channel.EVENT_DETAILS:get() or ""
			which[0] = "CALLOUT-"..ext.."-"..user.."-"..eventName
		 else
			 for name in require"DataProvider":getInstance():getIterator("CALLOUT-"..ext.."-"..user) do
				which[#which+1] = name
			 end
		 end
		 
		 for _, name in pairs(which) do
			local obj = require"callOutImpl":getExisting(name)
				if obj then
					obj:setProperty("status", false)-- stop
				end
		 end
		 ivr:play("procedure_complete", {voiceDir = voiceDir})
    	 return smartApp("Hangup", hcause.NORMAL)
	end

    local callOutName = "CALLOUT-"..ext.."-"..user.."-"..color
	
	if require"DataProvider":getInstance():existByName(callOutName) then						-- do nothing if already running
		print(callOutName, " already exist (possible waiting for end), skipping..")
	else
    	-- pin has validated and accepted. initializing new call-out!!
    	
    	local params = {		 -- configuration parameters
            uid             = userId
        	,pb              = pbName
        	,limit           = callsLimit
        	,ctx             = outCallsCtx
        	,events          = eventsCtx
        	,unavailtrycount = unavailTryCount
        	,confirmcode     = confirmCode
        	,refusecode      = refuseCode
        	,voicedir        = voiceDir
            ,route           = channel.EVENT_ROUTE:get() or ""
        }
        
    	local callOut = require"callOutImpl":createNew(callOutName, params)
    	if callOut then
            channel.CDR("to_name"):set(callOut:name().."-"..callOut:ident())
    
            -- execute runner
            callOut:start() -- start call-out
            require"core.pbxutils".exec("nohup callOutRunner %s &", callOutName)
            print(callOut:name(), " started.")
        else
            print("A creating callout call list error occured!")
        end
	end
	ivr:play("procedure_complete", {voiceDir = voiceDir})
	 
   return smartApp("Hangup", hcause.NORMAL)
end

--- Executing calls to contact numbers
function callOutService.callProcess(exten)
    if not exten then
        print("Internal error (extension not defined)")
        return smartApp("Hangup", hcause.FAILURE)
    end
    
    local callOutExt, callOutCtx = exten:match("(.*):(.*)")
	if not callOutExt or not callOutCtx then
		print("Wrong extension for call-out service (waiting for '<ext>:<context>')")
		return smartApp("Hangup", hcause.FAILURE)
	end
    
    local parentName = channel.PARENTNAME:get() or ""
    local callOut    = require"callOutImpl":getExisting(parentName)
    if not callOut then
        print("Service ".. parentName .."not found")
        return smartApp("Hangup", hcause.FAILURE)
    end

    channel.CALLERID("num"):set(callOut:ident())
    channel.CALLERID("name"):set("CallOut")
    channel.CHANNEL("hangup_handler_wipe"):set("wildix-callout,hHandler,1")
    smartApp("Goto", callOutCtx, callOutExt, 1)
end

--- Interactive menu in case contact make answer
function callOutService.answerIVR()
    smartApp("NoCDR", "") -- disable CDR for current channel
    local rv = false
    repeat

        local callOut = require"callOutImpl":getExisting(channel.PARENTNAME:get() or "")
    	if not callOut then
    		print("callOutService.answerIVR: internal error (service not found)")
    		break
    	end

        -- retrieve channel data
        local destId = channel.DESTID:get()
        local numId  = channel.NUMID:get()
        local name   = callOut:getDestParams(destId, "name")
        local num    = callOut:getContactParam(destId, numId, "num")
        if not destId or not numId or not name or not num then
     		print("callOutService.answerIVR: internal error (wrong contact info)")
            break
        end
        channel.CALLERID("num"):set(callOut:ident())
        channel.CALLERID("name"):set("CallOut")
        callOut:setContactParam(destId, numId, "phase", "answered")

        -- playback notification
        local voiceDir = "00000"
        if "" ~= callOut:color() then
            voiceDir = voiceDir.."/"..callOut:color()
        end
        ivr:play(callOut:color(), {voiceDir = voiceDir})

        --- check contact password
        local contPasswd = callOut:getDestParams(destId, "passwd")
        if contPasswd and "" ~= contPasswd then -- we will skip empty password from checking
            local loginOpts = {
                 max_chars = 5
                ,retry 	   = 3
                ,timeout   = 3600       -- required to stop reading user input
                ,voiceDir  = voiceDir
                ,validate  = function(v)
                    return callOut:getDestParams(destId, "passwd") == v and true
                end
                ,failure = function()
                    callOut:setContactParam(destId, numId, "phase","pin_not_correct")
                    ivr:play("wrong_pin", {voiceDir = voiceDir})
                    ivr:play("enter_personal_pin", {voiceDir = voiceDir})
                end
            }
            ivr:play("enter_personal_pin", {voiceDir = voiceDir})
            local login = ivr:read("beep", loginOpts)
            if not login then
                ivr:play("wrong_pin", {voiceDir = voiceDir})
                break
            end
        end
        callOut:setContactParam(destId, numId, "phase", "correct_pin")

        --- retrieve contact actions
        local action = nil
        local params = callOut:params()
        if not params.confirmcode and not params.refusecode then
            -- processing auto confirm
            action             = 1
            params.confirmcode = action
        else
            local options = {
                 max_chars = 2
                ,retry 	   = 3
                ,timeout   = 3600       -- required to stop reading user input
                ,voiceDir  = voiceDir
                ,validate  = function(v)
                    return params.confirmcode == v or params.refusecode == v and true
                end
                ,failure   = function()
                    print("callOutService.answerIVR:  Invalid confirmation code")
                    ivr:play("confirm_or_refuse", {voiceDir = voiceDir})
                end
            }
            ivr:play("confirm_or_refuse", {voiceDir = voiceDir})
            action = ivr:read("beep", options)
            if not action then
                print("callOutService.answerIVR: no confirmation code")
                break
            end
        end
        if params.confirmcode == action then
            callOut:setContactParam(destId,numId,"phase","confirmed")
        elseif params.refusecode == action then
            callOut:setContactParam(destId,numId,"phase","refused")
        end

        -- disable dest
        print("Contact made a choice, go to the next contact.")
        callOut:setDestParams(destId, "enabled", false) -- disable destination
        ivr:play("procedure_complete", {voiceDir = voiceDir})

        -- post anwser action
        if params.confirmcode == action then
            local routeData = callOut:params("route") or ""
            local routeCtx, routeExt = routeData:match("^(%w+),(%w+)$")
            if routeCtx and routeExt then
                smartApp("Goto", routeCtx, routeExt, 1)
            end
        end

        rv = true
    until true

    return rv
end

function callOutService.hHandler()
    local callOut = require"callOutImpl":getExisting(channel.PARENTNAME:get() or "")
    if not callOut then
        print("callOutService.hHandler: internal error (service not found)")
        return
    end

    local destId = channel.DESTID:get()
    local numId  = channel.NUMID:get()
    local name   = callOut:getDestParams(destId, "name")
    local num    = callOut:getContactParam(destId, numId, "num")
    if not destId or not numId or not name or not num then
        print("callOutService.hHandler: internal error (wrong contact info)")
        return
    end

    local dialstatus = ""
    if 18 == tonumber(channel.HANGUPCAUSE:get() or "") then
        dialstatus = "no_answer"
    else
        dialstatus = channel.DIALSTATUS:get() or dialstatus
    end
    dialstatus = dialstatus:lower()

    -- decrement contact try counter
    if "chanunavail" == dialstatus then
        print("Decrease contact unavailable call try.")
        callOut:decrContactTry(destId, numId, "unavailtry")
        __cleanupDests(callOut, destId) -- disable dest if no more try
    else
        print("Decrease contact call try.")
        callOut:decrContactTry(destId, numId, "try")
        __cleanupDests(callOut, destId) -- disable dest if no more try
    end

    local phase = callOut:getContactParam(destId, numId, "phase")
    if not phase or "" == phase or "none" == phase then
        phase = dialstatus
    end

    local contactFlag = nil
    -- define the first/end number in the call-out process
    if "none" == callOut:getProperty("firstDialedContact") then
        callOut:setProperty("firstDialedContact", num)
        contactFlag = "start"
    end
    local enabledDestsCount = 0
    for ident in callOut:getDestIterator() do
        if callOut:getDestParams(ident, "enabled") then
            enabledDestsCount = enabledDestsCount + 1
        end
    end
    if 0 == enabledDestsCount then
        contactFlag = contactFlag and (contactFlag..",stop") or "stop"
    end
    callOut:setDestParams(destId, "active", false)	 -- mark destination as 'inactive'
    callOut:setContactParam(destId,numId,"phase", "none") -- set to default value for next call to current dest

   if 0 == enabledDestsCount then
      contactFlag = contactFlag and (contactFlag..",stop") or "stop" 
   end
   callOut:setDestParams(destId, "active", false)	 -- mark destination as 'inactive'
   callOut:setContactParam(destId, numId, "phase", "none") -- set to default value for next call to current dest	

    __setupCDRFields({
        serviceName = "CallOut"
        ,serviceId   = callOut:ident()
        ,dstName     = name
        ,dstNum      = num
        ,tag         = callOut:color()..",".."number"..numId..","..phase..(contactFlag and ","..contactFlag or "")
    })
end

return callOutService


