--- Module for initialization extended services and excute last one
-- 
-- @file    exService.lua 
-- @author  Vyacheslav Khudan
-- @created 2015/07/09

--- @module exService
local exService = {
--	 strategy = {	-- pool for callers strategies
--	 	 obj = nil		-- strategy object
--	 }	
}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Retreive Wildix Hotel family variable
local function __getWhotelFamily()
	local whotel_family = channel.WHotelFamily:get() or ""
	if "" == whotel_family then
		return false
	end
	
	-- Fix up variable value
	local rv = whotel_family:match('^"(.*)"$')
	if not rv then
		rv = whotel_family 
	end
	
	return rv
end

--- Compare user kind with hotel rooms type
local function __isRoomUser(userKind)
	return "room" == userKind
end

--- Create strategy for room callers
local function __createStrategy(cidAttrs, mName)
	if "table" ~= type(cidAttrs) then
		return false, "Invalid data format"
	end
	
	local cid = cidAttrs.telephoneNumber
	if not cid then
		return false, "User number does not found"
	end
    
	local obj	     = nil
	local errMsg     = nil
    
    local family     = __getWhotelFamily()
    local customList = { applicationCallOut = 1 }
    if mName and customList[mName] then
        obj, errMsg = require"CustomExServiceStragegy":new(cidAttrs)
    elseif family then
		obj, errMsg = require"DaemonExServiceStrategy":new(cidAttrs, family)
	elseif "room" == cidAttrs.userKind then
		obj, errMsg = require"APIExServiceStrategy":new(cidAttrs)
	elseif "fax" ~= cidAttrs.userKind then
		obj, errMsg = require"PBXExServiceStrategy":new(cidAttrs)
	end
	if not obj then
		return false, errMsg or "Not supported user kind"
	end
	
	return obj 
end

--- Run strategy method after object loading
local function __callStrategyMethod(mName, callback, ...)
	if not mName then
		return callback
	end
    
	local exServiceObj, errMsg = __createStrategy(gSpace["cidAttrs"], mName)
	if not exServiceObj then
		return "function" == type(callback) and callback(nil, errMsg) or callback
	end
    
	return exServiceObj[mName](exServiceObj, callback, ...)
end

--- Callbask for blocked Wildix Hotel Service
local function __forbidServiceRules(strategyName, errStr, ...)
	local initStr = string.format("%s is blocked!", strategyName and strategyName or "Service")
	local errMsg  = nil
	if errStr and "" ~= errStr then
		errMsg = string.format(errStr, ...)
	end
	print(string.format("%s%s", initStr, errMsg and " ("..errMsg..")" or ""))
	to_logfile("Hangup", hcause.REJECT)
	return app.Hangup(hcause.REJECT)
end

--- Decline call by some error
local function __declineCallByError(errStr, ...)
    print(string.format(errStr or "", ...))
    return smartApp("Hangup", hcause.REJECT)
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Check call permission for current room caller
-- @return boolean value
function exService:is_callAllowed(callClass)
	return __callStrategyMethod("isCallAllowed", true, callClass)
end

--- Execute default rules for forbidden call
-- @param appName	A called application
-- @param msg		A message printed to logs
-- @return some dialplan actions, like Hangup, etc.
function exService:forbidden_call(appName, msg)
	return __callStrategyMethod("forbidCall", false, appName, msg)
end

--- Executing MUR service for room
-- @param val	A value of the MUR service
-- @return execution of some applications
function exService:service_mur(val)
	return __callStrategyMethod("serviceMUR", __forbidServiceRules, val)
end

--- Executing minibar service for room
-- @param article	A article of the MiniBar item
-- @param count		A count if the MiniBar items
-- @return execution of some applications
function exService:service_minibar(article, count)
	return __callStrategyMethod("serviceMiniBar", __forbidServiceRules, article, count)
end

--- Executing service for processing wakeUp's
-- @param val	A value of the WakeUp service
-- @return execution of some applications 
function exService:service_wakeup(...)
	return __callStrategyMethod("serviceWakeUp", __forbidServiceRules, ...)
end

--- Executing room inspection Hotel service
-- @param val	A value of the INSPECT service
-- @return execution of some applications
function exService:service_inspect(val)
	return __callStrategyMethod("serviceINSPECT", __forbidServiceRules, val)
end

--- Application WakeUp as Hotel implementation
-- @param mode	An application mode value
-- @param arg	An application argument mode
function exService:applicationWakeUp(mode, arg)
	print("Executing 'WakeUp': mode -", mode, ",arg -", arg)
	return __callStrategyMethod("applicationWakeUp", __forbidServiceRules, mode, arg)
end

--- Application call out as extended service
-- @param mixed    The application parameters 
function exService:applicationCallOut(...)
    print("Executing 'CallOut': args ->", table.concat({...}, ", "))
    return __callStrategyMethod("applicationCallOut", __declineCallByError, ...)
end

return exService