--- Creating report about sent fax in PDF format
-- 
-- @file	faxreport.lua 
-- @author	Vyacheslav Khudan
-- @created 2014/07/30
-- 

local require = require
local string = string
local math = math
local io = io
local type = type
local tonumber = tonumber

local hpdf = require"hpdf"

module("faxreport")

--- Add one page of the fax to report as PNG file
local function __add_fax_image(report_pdf, pdf_page, png_file, xpos, ypos, img_width, img_height)
	if not report_pdf or not pdf_page or not png_file then
		return false, "Invalid parameters for adding PNG file"
	end

	if not require"core.fsutils".is_readable(png_file) then
		return false, string.format("File %q not found", png_file)
	end
	
	local faximage = hpdf.LoadPngImageFromFile(report_pdf, png_file)
	if "userdata" == type(faximage) then
		hpdf.Page_DrawImage(pdf_page, faximage, xpos, ypos, img_width - 40, img_height)
		return true
	end
	return false, "Loaded PNG file is not userdata"
end

--- Add text information about fax sending
local function __add_info_header(report, page, page_height, page_width, lang, info_t)
	if not report or not page or not page_height or not page_width or not info_t then
		return 0
	end

	-- load translations
  	local tr = require"common.translation".translation
  	lang = tr:set_language(lang)
	
	-- get font and text encoding for the information message

	local font_type = nil
	local font_name = hpdf.LoadTTFontFromFile(report, "/usr/share/fonts/arial.ttf", 1);
	if font_name then
		font_type = hpdf.GetFont(report, font_name, "UTF-8")  		
  	end
  	
	hpdf.Page_BeginText(page)
	
	local rpt_title	    = tr:_t("faxrep.Title")
	local rpt_status    = tr:_t("faxrep.TXStatus")
	local rpt_title_txt = string.format("----------- %s -----------\n%s %s", rpt_title, rpt_status, info_t.fax_status)
	hpdf.Page_SetFontAndSize(page, font_type, 10)
	hpdf.Page_TextRect(page, 40, page_height - 10, page_width - 40, page_height - 40, rpt_title_txt, "HPDF_TALIGN_CENTER")
	
	local text_left = 200   -- left offset for text in px
	local value_left = 320  -- left offset for value in px
	local top_raw = 50      -- top offset for info raws px
	local shift = 10		-- shift for report strings
	hpdf.Page_SetFontAndSize(page, font_type, 8)
	hpdf.Page_TextOut(page, text_left, page_height - top_raw, tr:_t("faxrep.JobNO"))          -- JOB text
	hpdf.Page_TextOut(page, value_left, page_height - top_raw, info_t.job_id)                 -- JOB value
	
	hpdf.Page_TextOut(page, text_left, page_height - top_raw - shift, tr:_t("faxrep.RemoteId"))  -- REMOTE ID text
	hpdf.Page_TextOut(page, value_left, page_height - top_raw - shift, info_t.remote_id)         -- REMOTE ID value
	shift = shift + 10;
	
	hpdf.Page_TextOut(page, text_left, page_height - top_raw - shift, tr:_t("faxrep.DstNumber")) -- DESTINATION NUMBER text
	hpdf.Page_TextOut(page, value_left, page_height - top_raw - shift, info_t.to_num)            -- DESTINATION NUMBER value
	shift = shift + 10;
	
	hpdf.Page_TextOut(page, text_left, page_height - top_raw - shift, tr:_t("faxrep.SrcNumber")) -- SOURCE NUMBER text
	hpdf.Page_TextOut(page, value_left, page_height - top_raw - shift, info_t.from_num)          -- SOURCE NUMBER value
	shift = shift + 10;
	
	hpdf.Page_TextOut(page, text_left, page_height - top_raw - shift, tr:_t("faxrep.SrcHeader")) -- FAX HEADER text
	hpdf.Page_TextOut(page, value_left, page_height - top_raw - shift, info_t.fax_header)        -- FAX HEADER value
	shift = shift + 10;
	
	if "" ~= info_t.from_company then
		hpdf.Page_TextOut(page, text_left, page_height - top_raw - shift, tr:_t("faxrep.SrcCompany")) -- SENDER COMPANY text
		hpdf.Page_TextOut(page, value_left, page_height - top_raw - shift, info_t.from_company)        -- SENDER COMPANY value
		shift = shift + 10;
	end
	
	hpdf.Page_TextOut(page, text_left, page_height - top_raw - shift, tr:_t("faxrep.Date"))     -- DATE text
	hpdf.Page_TextOut(page, value_left, page_height - top_raw - shift, info_t.date)             -- DATE value
	shift = shift + 10;
	
	hpdf.Page_TextOut(page, text_left, page_height - top_raw - shift, tr:_t("faxrep.Pages"))                             -- PAGES text
	hpdf.Page_TextOut(page, value_left, page_height - top_raw - shift, string.format("%s/%s", info_t.sent_pages, info_t.total_pages)) -- PAGES value
	shift = shift + 10;
	
	hpdf.Page_TextOut(page, text_left, page_height - top_raw - shift, tr:_t("faxrep.AdvStatus")) -- STATUS text
	hpdf.Page_TextOut(page, value_left, page_height - top_raw - shift, info_t.fax_statusstr)     -- STATUS value
	
	local text_position = hpdf.Page_GetCurrentTextPos(page)
	hpdf.Page_EndText(page)
	
	return math.floor(tonumber(text_position.y - 2) or 0)
end

--- Create report PDF from fax sending status and first user fax attachment
-- @param png_image				PNG image used in PDf report
-- @param lang					A language of the report
-- @param job_id				Fax incremental id
-- @param from_num				From user number
-- @param to_num				Destination fax number
-- @param remote_id				Remote id
-- @param fax_status			Fax status code
-- @param fax_statusstr			Fax status in string format 
-- @param sent_pages			Sent fax pages
-- @param total_pages			Total fax pages
-- @param date					Date of the sending fax
-- @param postfix				Postfix for creating PDF report name
-- @return True with created PDF report file
-- @return False with error message
function create_report(png_image, lang, job_id, from_num, fax_header, to_num, from_company, remote_id, fax_status, fax_statusstr, sent_pages, total_pages, date, postfix)
	if not png_image or "" == png_image then
		return false, "PNG image for report was not found"
	end
	
	local report = hpdf.New()
	if report then
		hpdf.UseUTFEncodings(report)
		hpdf.SetCompressionMode(report, hpdf.COMP_ALL)
		
		-- create new page		
		local page = hpdf.AddPage(report)
		local page_height = hpdf.Page_GetHeight(page)
  		local page_width = hpdf.Page_GetWidth(page)
  		
		-- prepare info for header
		local info_t = {
			["job_id"]        = job_id or "",
			["from_num"]      = from_num or "",
			["fax_header"]    = fax_header or "",
			["from_company"]  = from_company or "",
			["to_num"]        = to_num or "",
			["remote_id"]     = remote_id or "",
			["fax_status"]    = fax_status or "",
			["fax_statusstr"] = fax_statusstr or "",
			["sent_pages"]    = sent_pages or 0,
			["total_pages"]   = total_pages or 0,
			["date"]          = date or ""
		}
		
		local err_msg = nil
		local height_offset = __add_info_header(report, page, page_height, page_width, lang, info_t)
		if 0 ~= height_offset then
			_, err_msg = __add_fax_image(report, page, png_image, 30, 0, page_width, height_offset)
		else
			err_msg = "Information header wasn't added"
		end
		
		local rc = false
		if not err_msg then
			local report_file = png_image:match("(.*)%.png$")
			report_file = report_file .. (postfix or "") .. ".pdf"
			
			-- check write permission to report file
			local pdf_file, err = io.open(report_file, "w")
			if not pdf_file then
				rc = false
				err_msg = err
			else
				pdf_file:close()
    			hpdf.SaveToFile(report, report_file)
    			rc = true
    			err_msg = report_file
    		end
		end
		hpdf.Free(report)
		
		return rc, err_msg
	else
		return false, "Error creating PDF object"
	end
end