module ("internalcalls", package.seeall)

--- Show features in the compact format
local function __showActiveUserFeatures(features)
    if "table" ~= type(features) then
        return false
    end
    
    local actClass    = features:getActiveClass()
    local actSubClass = features:getActiveSubClass()
    local data        = features:getSubClassData()
    
    local fmtData = ""
    for name in pairs(data) do
        if name and not name:match("^_.*") then
            local featVal = features:get(name)
            local status  = featVal and 1 or 0
            local value   = 1 == status and "boolean" ~= type(featVal) and string.format("(%s)", featVal) or ""
            fmtData       = string.format("%s %s:%s%s", fmtData, name, status, value) 
        end
    end
    
    -- add global settings
    local road = features:get("ROAD") and 1 or 0
    fmtData    = string.format("%s ROAD:%s", fmtData, road)
    
    print(string.format("Dst features for active calltype/status - %s/%s:\n\t%s", actClass, actSubClass, fmtData))
end

--- Check channel available for specified user 
local function __getUserState(userAttrs, userFeatures, printLogs)
    local rv = {
         isDevAvailable  = false
        ,isChanAvailable = false
    }
    if "table" ~= type(userAttrs) or "table" ~= type(userFeatures) then
        return rv
    end

    local userNumber = userAttrs.telephoneNumber
    if not userNumber then
        return rv
    end

    -- loading device state object
    local userStateObj = require"deviceState":new(userNumber)
    userStateObj:setLogs(printLogs)
    userStateObj:reqValue()
    rv.isDevAvailable = userStateObj:isAvailable()

    -- processing dialplan channel status
    local userCallWaitFeature = userFeatures:get("CW")
    rv.isChanAvailable        = true
    if not rv.isDevAvailable and ("fax" == userAttrs.userKind or not userCallWaitFeature) then
        rv.isChanAvailable = false 
    end

    return rv
end

--- Check status of the return from transfer feature
local function __isAvailableReturnFromTransfer(ctx)
    local rv = true
    local valueFromGlobalVar = channel.RETURNFROMTRANSFER:get() or "" 
    if "no" == valueFromGlobalVar then
        logger:info("Return From Transfer feature disabled by dialplan variable")
        rv = false
    elseif "yes" == valueFromGlobalVar then
        logger:info("Return From Transfer feature enabled by dialplan variable")
    elseif "1" == channel["GCOUNTRY_CODE"]:get() then
        if "parkOrbit" ~= ctx then
            print("Return From Transfer feature blocked by country settings")
            rv = false
        end
    end
    
    return rv
end

--- Make dial to wms user with special options
local function call_to_wms(wms_user, dst_attrs, dial_timeout, dial_options, last_member, fwd_opts, retTransfer, announceOpt)
    local gvars        = gSpace["gvars"]
    local dial_options = dial_options or ""
    local announceOpt  = announceOpt or ""
    
    local wms_calltype = gvars.calltype and gvars.calltype ~= "internal" and gvars.calltype or ""
    local wms_appl     = gvars.application or ""
    local wms_fwd      = fwd_opts or ""
    
    local wms_trans    = channel["TRANSFER_NUMBER"]:get() or ""
    if wms_trans ~= "" then
        if not __isAvailableReturnFromTransfer() then
            wms_trans = "" -- Remove TRANSFER_NUMBER for disabling Return from Truansfer at remote PBX
        end
    end
    
    local wms_trunk = gvars.trunk_call or ""
    local wms_data  = dial_timeout or ""
    if last_member then
        wms_data = wms_data.."|"..last_member
    elseif wms_data ~= "" then
        wms_data = wms_data.."|"
    end

    local wms_failover_user = channel.FAILOVER_USER:get() or ""
    local wms_opts = require"wmsNetCall".makeArgs({
             pbxid        = require"core.strConverter".toB64(dst_attrs.pbxowner)
            ,ctx          = "internalcalls"
            ,calltype     = wms_calltype
            ,trunkname    = wms_trunk
            ,app          = wms_appl
            ,appdata      = wms_data
            ,transfer     = wms_trans
            ,forward      = wms_fwd
            ,failoveruser = wms_failover_user
            ,linkedid     = channel.LINKEDID:get() or ""
    })

    dial_options = require"utils".extend_dial_options(dial_options)
    smartApp("Dial", "SIP/" .. wms_user, nil, dial_options..wms_opts..announceOpt)  -- timeout "nil" for work preanswer service at remote pbx

    local unavailable_pbx = channel["X-PBX_UNAVAILABLE"]:get() or ""
    if unavailable_pbx ~= "" then
        -- get destination user PBX failover configuration
        local failoverObj = require"common.failover":new(dst_attrs.pbxowner)
        if failoverObj and failoverObj:localPbxIsFailover() then -- call via current pbx
            if not dial_timeout or dial_timeout == "" or dial_timeout == "0" then
                dial_timeout = channel["GINTDIAL_TIMEOUT"]:get() or ""
            end
            smartApp("Dial", "SIP/" .. wms_user, dial_timeout, dial_options .. "b(predial^failover^1())")
            if retTransfer then -- use for recall after return from transfer
                return true
            else
                local failover_trans = channel["TRANSFER_NUMBER"]:get() or ""
                if "" ~= failover_trans and __isAvailableReturnFromTransfer() then
                    return return_from_transfer(self, failover_trans, wms_user, "CALL", "failoverUser")
                end
            end
        elseif failoverObj and failoverObj:getPbxDN() then   -- call via remote pbx
            require"services":app_remote_call("Call to failover user at", wms_user, failoverObj:getPbxDN(), "internalcalls")
        end
    end
    return true
end

--- Make a call with mobility options
-- @param dst_attrs     Destination LDAP attributes
-- @param dst_features  Destination features
-- @param optsSet       Set of the dial options
-- @param cid           Caller identification
-- @return Dial execution or false in case error
local function __extendDial(dst_attrs, dst_features, optsSet, cid)
    if "table" ~= type(dst_attrs) or "table" ~= type(dst_features) then
        return false
    end
    if not dst_attrs.telephoneNumber then
        return false
    end

    local optsSet     = "table" == type(optsSet) and optsSet or {}
    local dialOpts    = optsSet["opt"] or ""
    local dialTimeout = tonumber(optsSet["tout"]) or 0
    local ring        = optsSet["ring"] or ""
    local cg          = optsSet["cg"] or ""
    local trunkCall   = optsSet["trunkCall"] or ""
    local dst_uid     = optsSet["dst_uid"] or ""
    
    -- set destination context
    local dstCntx = dst_attrs.userDialplan or "users"
    
    -- get dial timeout options
    if 0 == dialTimeout then
        dialTimeout = tonumber(dst_features:get("TOUT")) or 0
    end
    if 0 == dialTimeout then
        dialTimeout = tonumber(channel["GINTDIAL_TIMEOUT"]:get()) or 65
    end
    
    -- retrieve status of the ring only active device
    local ringOnlyActiveDevice = dst_features:get("ROAD") and "1" or ""
    
    -- check mobility settings and options
    local mobOpts = ""
    local dialStr = "SIP/"..dst_attrs.telephoneNumber
    if dst_features:is_active("MEXT") and dst_attrs.mobile then
        local mobTimeout = dst_features:get("MEXT") or ""
        if "" ~= mobTimeout then
            print("MEXT is active with timeout", mobTimeout)
        end
        local mobOwner   = dst_attrs.owner:gsub(",", ";")    -- also used for indication call to mobility at app_external_call
        local annonceNum = dst_features:is_active("MOBCONFIRMATION") and cid or ""
        mobOpts          = dst_attrs.mobile.."|"..annonceNum.."|".. mobTimeout.."|"..dstCntx.."|"..dst_attrs.preferredLanguage.."|"..mobOwner
        dialStr          = string.format("%s&Local/%s@mobility", dialStr, dst_attrs.telephoneNumber)
        
        -- add fake ring if call to mobility and PAE in ringing
        dialOpts         = dialOpts.."rV" -- option "V" add by Semen in callweaver for setting PAE in ringing
    end

    local tags = channel["TAGS"]:get()
    if tags and "" ~= tags then
        channel.CDR("tags"):set("tags:" .. tags)
    end

    local isqcall = channel["QUEUE_NAME"]:get() or ""
    local to_identity = channel['TO_IDENTITY']:get() or ""
    local from_identity = channel["FROM_IDENTITY"]:get() or ""
    if ("" ~= isqcall and "" ~= from_identity) or "" ~= to_identity then
        dialOpts = dialOpts .. "I"
    end

    local to_identity = channel['TO_IDENTITY']:get() or ""
    if "" ~= to_identity and "" == isqcall and cid then
        local identity_attrs = get_obj_attrs(to_identity)
        if identity_attrs.uid then
            local identity_name = identity_attrs.cn or ""
            local b64 = require"core.strConverter"
            local c_from_b64 = b64.toB64(cid) or ""
            local c_to_b64 = b64.toB64(to_identity) or ""
            local to_name_b64 = b64.toB64(identity_name)
            channel.CDR("c_from"):set(string.format("%s.%s", to_identity, cid))

            local fork_params = string.format("vTw(c_from=%s,c_to=%s,to_name=%s)", c_from_b64, c_to_b64, to_name_b64)
            app.ForkCDR(fork_params)
            channel.CALLERID("num"):set(string.format("%s.%s", to_identity, cid))
            channel.CONNECTEDLINE("all"):set(string.format([===["%s" <%s>]===], identity_name, identity_attrs.telephoneNumber))
            app.SipAddHeader(string.format("Caller-Image: http://127.0.0.1/user_avatar/%s.png", cid))

            local dst_gid = string.match(identity_attrs["owner"] or "", "gid=(%d+),.*")
            local group_attrs = get_obj_attrs(dst_gid, "gid", "wGroup")
            cg = group_attrs.wcgrp or ""
            dst_uid = identity_attrs.uid
        end
    end

    -- concatenate dial options with predial string 
    if "" ~= mobOpts or "" ~= ring or "" ~= cg or "" ~= ringOnlyActiveDevice then
        dialOpts = string.format("%sb(predial^internalcall^1(%s,%s,%s,%s,%s,%s,%s))", dialOpts, mobOpts, ring, cg, trunkCall, ringOnlyActiveDevice, dst_uid, tostring(optsSet["callToLocalFailoverUser"]))
    end
    
    dialOpts = require"utils".extend_dial_options(dialOpts)
    return smartApp("Dial", dialStr, dialTimeout, dialOpts)
end
 
--- Processing dial to park orbit user
local function __parkOrbitDial(dstAttrs, dstFeatures, parkTimeout)
    if "table" ~= type(dstAttrs) or "table" ~= type(dstFeatures) then
        return smartApp("Hangup", hcause.FAILURE)
    end
    
    -- set up default dial status (especially for enabled CW)
    channel.DIALSTATUS:set("BUSY")
    
    -- update fields for connected line
    channel.CONNECTEDLINE("all"):set(string.format([===["%s" <%s>]===], dstAttrs.cn or "", dstAttrs.telephoneNumber or ""))
    
    -- processing callwaiting feature
    local parkSlotState = __getUserState(dstAttrs, dstFeatures, true)
    if parkSlotState.isChanAvailable then
        local parkStr     = "SIP/" .. dstAttrs.telephoneNumber
        local parkOpts    = "mb(predial^parkorbit^1())"

        smartApp("Dial", parkStr, parkTimeout, parkOpts)
    else
        print(string.format("Call Waiting is disabled for parking slot '%s'", dstAttrs.telephoneNumber))
    end
    
    -- processing return from parking in case transfer action
    local transferNum = channel["TRANSFER_NUMBER"]:get() or ""
    if "" ~= transferNum then
        return return_from_transfer(self, transferNum, dstAttrs.telephoneNumber, "CALL", "parkOrbit")
    end
    return smartApp("Hangup", hcause.NORMAL)
end

-- Check if the destinatiaon number is an identity number, and if so - go to identities_routing context
local function __forwardIfIdentity(number)
    if "table" == type(extensions["identities_routing"]) then
        for exten, _ in pairs(extensions["identities_routing"]) do
            if "_" == exten:sub(1, 1) then
                exten = exten:gsub("_", "")
                exten = exten:gsub("[xX]", "%%d")
                exten = exten:gsub("[zZ]", "[1-9]")
                exten = exten:gsub("[nN]", "[2-9]")
                exten = "^" .. exten .. "$"
                if string.match(number, exten) then
                    return app.Goto("identities_routing", number, 1)
                end
            elseif exten == number then
                return app.Goto("identities_routing", number, 1)
            end
        end
    end
end

--- Forward user by FCF, UCF, BCF
-- @param number        A destination extension used for forward
-- @param prefix        A prefix used for change caller name (exp. "F500-user_200") (for VM always equal "VM")
-- @param fwd           An extension indicate who done forward
-- @param fwd_context   A context of the extension who done forward
-- @param pbx_src       An ip of our pbx, used for compare with call to remote pbx
-- @param fwd_hops      A number of the forward hops
-- @param last_member   Do indication about order app-dial in dialplan context
-- @param dial_timeout  A timeout for dial after forward to remote pbx
local function forward_user(number, prefix, fwd, fwd_context, pbx_src, fwd_hops, last_member, dial_timeout)
    local channel = channel
    local cid_attrs, gvars = cid_attrs, gSpace["gvars"]
    
    -- go to Voicemail
    if prefix == "VM" then
       local rv = nil
        
        if fwd == "PREANSWER" then
            rv = require"services":app_voicemail(number, "leave", "internalcalls|s")
        elseif fwd:match("^FORWARD") then
            local opts = "u"
            if fwd == "FORWARD_B" then
                opts = "b"
            end
            rv = require"services":app_voicemail(number, "leave", "internalcalls|" .. opts)
        end
        if rv then
            gSpace["execute_missed_call"] = 2
            rv = 2
        end 
        return rv
    end
    
    -- increase fwd hops
    local fwd_hops = fwd_hops + 1
    channel["FORWARDED"]:set(fwd_hops.."|")
    channel["FORWARD_FROM"]:set(fwd)

    -- modify and set callerid name
    local cid_name = channel.CALLERID("name"):get() or ""
    if cid_name == "" then cid_name = gSpace["cidNum"] or "" end
    local fwd_name = prefix .. fwd .. "-" .. cid_name
    channel.CALLERID("name"):set(fwd_name)
    
    -- clear variable for checking type of preanswer sounds
    if gSpace["gvars"].preanswer ~= "run_answer" then
        gSpace["gvars"].preanswer = nil
    end
    
    -- get number_attrs from LDAP and check is_local_dst
    local number_attrs = get_obj_attrs(number)
    if number_attrs.telephoneNumber then
        if pbx_src[1] == number_attrs.pbxowner then
            __forwardIfIdentity(number)
            return require"services":start_call("fwd-internalcall", number, last_member)
        else
            local failoverObj = require"common.failover":new(number_attrs.pbxowner)
            local isRegistered = require"core.sipproxy.sipproxy":new():getOnlineRegistrationByExten(number)
            if failoverObj:localPbxIsFailover() and isRegistered then
                __forwardIfIdentity(number)
                return require"services":start_call("fwd-internalcall", number, last_member)
            else
                local fwd_opts = nil
                if fwd_hops and fwd_hops ~= 0 then
                    fwd_opts = (fwd_hops.."|"..fwd_name)
                end
                return call_to_wms(number, number_attrs, dial_timeout, nil, last_member, fwd_opts)
            end
        end
    else
        local fwd_attrs = get_obj_attrs(fwd)
        if fwd_attrs.uid and require"common.permissions":check(fwd_attrs["owner"], "canuse", "TRUNK_TO_TRUNK_FORWARD", nil, fwd_attrs["licenseType"]) ~= "yes" then
            print("Forward trunk to trunk rejected by group permissions")
            smartApp("Hangup", hcause.ACL_FORBIDDEN)
        end
        if "yes" == channel["BLOCK_TRANSFER"]:get() then
            channel["__BLOCK_TRANSFER"]:set("")
        end
        gSpace["gvars"].forward_number = fwd
        return smartApp("Goto", fwd_context, number, 1)
    end
    return false
end

--- Preanswer service function
-- @param dst_attrs     A attributes of called user
-- @param dialStatus    A status of previous dial
-- @param dial_timeout  A timeout used in the dial in CC procedure
-- @param dial_options  An options used in the dial in CC procedure
-- @return 0 - execute service, 1 - no execute or "STATUS" - causecode
local function preanswer_service(dst_attrs, dialStatus, dial_timeout, dial_options)
    local channel      = channel
    local app          = app
    local ivr          = ivr
    local cid_attrs    = gSpace["cidAttrs"]
    local dialStatus   = dialStatus or ""
    local application  = channel.APPLICATION:get()
    local hangupCause  = tonumber(channel.HANGUPCAUSE:get()) or 0
    local onlyPlayback = false
    local tts          = nil
    local calltype     = channel.CALLTYPE:get() or ""
    local is_trunk = channel["X-TRUNK"]:get() or ""
    is_trunk = is_trunk ~= ""

    -- checkign preanswer service ACL   
    local wmsperm = require"common.permissions"
    local is_allowed = wmsperm:check(cid_attrs["owner"] or "", "canuse", "PRE_ANSWER_SERVICES", nil, cid_attrs["licenseType"]) == "yes"
    if not is_allowed then
        onlyPlayback = true
        print("Preanswer service was disabled by ACL")
    end
    
    -- set up flag in case call from KITE service
    if "KITE" == application then
         onlyPlayback = true
    end
    -- WMS-12595 If calltype="external" that means that custom variable REPORT_USER_STATUS_FOR_EXT_CALLS
    -- was set. In that case we will allow pre answer service, but without voice menu.
    if "external" == calltype or is_trunk then
         onlyPlayback = true
    end
    -- set flab for missed call (enable when not read DTMF)
    gSpace["execute_missed_call"] = 1
    
    -- processing playback message
    local isUseInstr       = "yes" ~= channel.DIALNOINSTR:get()     -- used for disable message preanswer instruction
    local instrMsg         = (isUseInstr and not onlyPlayback) and " Press asterisk for options" or ""
    local preanswerMsgType = gSpace["gvars"].preanswer or ""
    local preanswerMsg     = nil
    local DisableFwdNotif  = "yes" == channel.DISABLE_FORWARD_NOTIFICATION:get()

    if "BUSY" == dialStatus then
        if "cw_noopt_progress" == preanswerMsgType then
            preanswerMsg = "User busy"
        elseif "bcf_noopt_progress" == preanswerMsgType then
            preanswerMsg = not DisableFwdNotif and "User busy. Your call will be forwarded"
        elseif "answer" == preanswerMsgType or "progress" == preanswerMsgType then
            if 21 == hangupCause then
                preanswerMsg = "User busy." .. instrMsg -- call rejected
            else
                preanswerMsg = "User busy." .. instrMsg -- user has busy status
            end
        end
    elseif "NOANSWER" == dialStatus then
        if "ucf_noopt_progress" == preanswerMsgType then
            preanswerMsg = not DisableFwdNotif and "User not available. Your call will be forwarded"
        else
            preanswerMsg = "The dialed number is not answering." .. instrMsg
        end
    elseif "CONGESTION" == dialStatus then
        preanswerMsg = not DisableFwdNotif and ("User not available." .. instrMsg)
    elseif "AWAY_progress" == preanswerMsgType or "DND_progress" == preanswerMsgType then
        if "CANCEL" ~= dialStatus then
            onlyPlayback = true
        end
        local userStatus    = preanswerMsgType:match("(%w+)_progress")
        local dstStatusMsg  = gSpace["gvars"].dststatusmsg or ""
        local statusTbl     = {AWAY = "away", DND = "do not disturb"}
        local ttsUserStatus = statusTbl[userStatus] or ""
        -- if call is local pbx call then we keep the current behaviour, otherwise, 
        -- if call comes from trunk we will analyse REPORT_USER_STATUS_FOR_EXT_CALLS variable. 
        -- Possible values are: "all"-status+message will be played, "status"-only status will be played
        if not is_trunk or "all" == channel["REPORT_USER_STATUS_FOR_EXT_CALLS"]:get() or "KITE" == application then
            if dstStatusMsg and "" ~= dstStatusMsg then
                tts = string.format("User status %s. %s. %s.",ttsUserStatus, dstStatusMsg, "CANCEL" == dialStatus and instrMsg or "")
            end
            preanswerMsg = string.format("User status %s. %s", userStatus, "CANCEL" == dialStatus and instrMsg or "")
        elseif "status" == channel["REPORT_USER_STATUS_FOR_EXT_CALLS"]:get() then
            preanswerMsg = string.format("User status %s. %s", userStatus, "CANCEL" == dialStatus and instrMsg or "")
        end
    end
    if not preanswerMsg then
        channel.DISABLE_VIDEO:set("NULL")    -- enable video by exit from pre-answer service
        return 1 
    end

    local to_identity = channel["TO_IDENTITY"]:get() or ""
    if "" == to_identity then
        -- set callerd name from this point (for correct name and photo in preanswer service)
        if dst_attrs.cn and dst_attrs.telephoneNumber then
            channel.CONNECTEDLINE("name"):set(dst_attrs.cn)
            channel.CONNECTEDLINE("num"):set(dst_attrs.telephoneNumber)
            channel.CONNECTEDLINE("all"):set(string.format([===["%s" <%s>]===], dst_attrs.cn or "", dst_attrs.telephoneNumber or ""))
        end
    end
    
    if onlyPlayback then -- play back for kite users
        ivr:tell({preanswerMsg, tts = tts}, {opts = "n"})
        channel.DISABLE_VIDEO:set("NULL")    -- enable video by exit from pre-answer service
        return 1
    end
    
    -- processing timeout for playbach with choice
    local playbackTimeout = 2   -- default playback timeout
    local playbackOpt     = ""
    local attendTransfer  = channel.TRANSFERERNAME:get() or ""
    if preanswerMsgType:find("_progress") then
        if "CANCEL" == dialStatus then
            playbackTimeout = 4     -- for reject call features
        else
            playbackTimeout = 0.1   -- short preanswer notification
        end
        if "" == attendTransfer then
            playbackOpt = "n"
        end
    else
        if "" == attendTransfer and "progress" == preanswerMsgType then
            playbackOpt = "n"
        end
    end
    
    -- set up presnwer feature codes
    local preanswerFeatList = gl_features and gl_features["PREANSWER"] or {}
    if not next(preanswerFeatList) then
        preanswerFeatList = {gotovm = "1", gotoprivat = "2", gotointercom = "4", gotocc = "5"} -- default codes
    end
    local featList = {}
    for featName, codeVal in pairs(preanswerFeatList) do
        if featName and codeVal then
            codeVal            = tostring(codeVal)
            featList[codeVal]  = featName
            featList[featName] = codeVal
        end
    end
    
    -- retrieve related variables for this service
    local devState       = require"deviceState"
    local destnumber     = dst_attrs.telephoneNumber
    local isCCEnable     = "no" ~= channel.CALLCOMPLETION:get()
    local isMobility     = "" ~= (channel.MOBILITY_OPTS:get() or "") or (application or ""):match("^DISA.*")
    local isFailoverUser = "yes" == channel.FAILOVER_USER:get()
    
    -- run playback service for reading choice
    local options = {
         max_chars = 1
        ,retry     = 1
        ,timeout   = playbackTimeout
        ,opts      = playbackOpt
    }

    local choice = ivr:read({preanswerMsg, tts = tts}, options)
    if "*" == choice and isUseInstr then 
        local instrMsg = {"Press"}
        if featList.gotovm and "yes" == wmsperm:check(dst_attrs["owner"] or "", "canuse", "VOICEMAIL", nil, cid_attrs["licenseType"]) then
            instrMsg[#instrMsg+1] = featList.gotovm .. " for voicemail,"
        end
        if featList.gotoprivat and devState:new(destnumber):reqValue():isAvailable4Service("INTRUSION") then
            if "yes" == wmsperm:check(cid_attrs["owner"] or "", "yesno", "SUPERVISION", dst_attrs["owner"] or "", cid_attrs["licenseType"]) then
                instrMsg[#instrMsg+1] = featList.gotoprivat  .. " for call intrusion,"
            end
        end
        if featList.gotointercom and "CONGESTION" ~= dialStatus and "yes" == wmsperm:check(cid_attrs["owner"] or "", "yesno", "INTERCOM", dst_attrs["owner"] or "", cid_attrs["licenseType"]) then
            instrMsg[#instrMsg+1] = featList.gotointercom .. " for intercom,"
        end
        if featList.gotocc and isCCEnable and not isMobility and not isFailoverUser and ("NOANSWER" == dialStatus or "BUSY" == dialStatus) then
            instrMsg[#instrMsg+1] = featList.gotocc .. " for call completion."
        end
        if instrMsg[2] then
            choice = ivr:read(table.concat(instrMsg," "), options)
        end
    end
    gSpace["execute_missed_call"] = 0   -- disable missed call

    local feature = featList[tostring(choice)]
    if feature then
        -- disable video in case processing pre-answer service
        channel.DISABLE_VIDEO:set(1)
    end
    if "gotocc" == feature then
        if not isCCEnable then
            logger:info("CallCompletion service was disabled by global variable CALLCOMPLETION")
        else
            local CCExec = false
            local reDial = false
            repeat
                if   isMobility                                          -- call from mobility
                  or "CANCEL" == dialStatus                              -- user has reject option
                  or isFailoverUser                                      -- failover user
                  or ("NOANSWER" ~= dialStatus and "BUSY" ~= dialStatus) -- accepted only NOANSER and BUSY dial statuses
                then
                    break                                                -->> service unavailable
                end

                if    "BUSY" == dialStatus                               -- BUSY dial status
                  and 21 ~= hangupCause                                  -- call was not rejected by user
                  and "bcf_noopt_progress" ~= preanswerMsgType           -- BCF disabled
                  and devState:new(destnumber):reqValue():isAvailable()   -- destination device state is AVAILABLE
                then
                    reDial = true                                        -->> processing re-dial to user
                    break
                end

                if not devState:new(destnumber):reqValue():isAvailable()  -- only for BUSY dial status
                  and "BUSY" == dialStatus                               -- destination device state is NOT AVAILABLE
                then
                    CCExec = "fix"                                       -->> hack for call completion
                    break
                end
                CCExec = "std"
            until true
            
            local rv = nil
            if CCExec then
                if "fix" == CCExec then
                    channel.SETCALLLIMIT(destnumber):set(1)
                    smartApp("Dial", "SIP/" .. destnumber, 1)
                    channel.SETCALLLIMIT(destnumber):set(0)
                end
                smartApp("CallCompletionRequest", "")
                ivr:tell("Your request has been recorded")
                rv = 0
            elseif reDial then
                gSpace["gvars"].preanswer = "run_answer"
                channel.DISABLE_VIDEO:set("NULL")    -- enable video by exit from pre-answer service
                run(self, destnumber, "", nil, false, dial_timeout, dial_options)
                rv = 0
            end

            if not rv then
                gSpace["execute_missed_call"] = 1   -- enable missed call
                ivr:tell("Service currently unavailable")
                rv = "FAILURE"
            end

            return rv
        end
    elseif "gotovm" == feature then
        return forward_user(destnumber, "VM", "PREANSWER")
    elseif "gotoprivat" == feature then
        if devState:new(destnumber):reqValue():isAvailable4Service("INTRUSION") then
            require"services":start_call("PREBARGE", destnumber, nil, "PREANSWER", false, dst_attrs)
            return 0
        end
        gSpace["execute_missed_call"] = 1   -- enable missed call
        ivr:tell("Service currently unavailable.")
        return "FAILURE"
    elseif "gotointercom" == feature then
        require"services":start_call("PREINTERCOM", destnumber, nil, nil, dst_attrs)
        return 0
    end
    
    gSpace["execute_missed_call"] = 1   -- enable missed call
    channel.DISABLE_VIDEO:set("NULL")    -- enable video by exit from pre-answer service
    return 1
end

--- Handling return from transfer
-- @param transferer    The user who performed transfer
-- @param trans_dst     The number where was done transfer
-- @param retMod        The return transfer mode (CONF/NONUSER/CALL)
-- @param context       The place where executed return from transfer
function return_from_transfer(self, transferer, trans_dst, retMode, context)
    if not __isAvailableReturnFromTransfer(context) then
        return false
    end

    local strRetMode = "CONF" == retMode and "conference" or "transfer"
    print(string.format("Return from '%s' to %s (%s)", strRetMode, transferer or "", context))
    if transferer then
        if "CONF" ~= retMode then   -- conference application is trusted
            local dst_tr = get_obj_attrs(transferer)
            if not dst_tr.cn then
                dst_tr = get_obj_attrs(transferer, "mobile")
                if not dst_tr.cn then
                    print("Return number not found in LDAP")
                    return smartApp("Hangup", hcause.NOT_FOUND)
                end
            end
        end
        
        local cid_name = channel.CALLERID("name"):get() or ""
        if "" == cid_name then
            cid_name = gSpace["cidNum"] or "" 
        end
        
        -- update some channel variables for this procedure
        if "NOUSER" == retMode then
            cid_name = "X" .. trans_dst .. "-" .. string.gsub(cid_name, "^X","")
        elseif "CONF" == retMode then
            cid_name = "C" .. trans_dst .. "-" .. string.gsub(cid_name, "^C","")
        else
            cid_name = "R" .. trans_dst .. "-" .. string.gsub(cid_name, "^R","")
        end
        
        channel["_RET_TRANS"]:set("1")  -- marker for fix caller id after return from transfer
        channel.CALLERID("name"):set(cid_name)
        channel["TRANSFER_NUMBER"]:set(cid_name)
        channel["DISABLE_DIRECTRTP"]:set(1)
        
        return require"services":start_call("returnFromTransfer", transferer)
    end
end

--- Fix caller name for re-call after return from transfer
function fix_callername_return_transfer()
    if channel["RET_TRANS"]:get() == "1" then   -- used for recall after return from blind transfer
        local tmp_name = channel.CALLERID("name"):get() or ""
        local fix_name = tmp_name:match("^[XRC].*%-(.*)")
        if fix_name then
            channel.CALLERID("name"):set(fix_name)
            channel["RET_TRANS"]:set("NULL")
        end
    end
end

--- Make call to switching extension  
-- @param dst   A destination for this call 
function switchcalls(self, dst)
    if not dst or "" == dst then
        return false
    end
    
    local hcause  = hcause
    local app     = app
    local channel = channel
    
    print(string.format("Making switch call from %q to %q", gSpace["cidNum"], dst))
    
    -- get dst_attrs from LDAP
    local dst_attrs = get_obj_attrs(dst)
    if not dst_attrs.uid then
        dst_attrs = get_obj_attrs(dst, "mobile")
        if not dst_attrs.uid then
            print("Number", dst, "is not found in LDAP")
            return smartApp("Hangup", hcause.NOT_FOUND)
        end
    end
    
    -- loading destination user features
    local dst_features = require"common.userFeatures.userFeatures":new(dst_attrs.uid)
    if not dst_features then
        print("Error loading features of the destination user")
        return smartApp("Hangup", hcause.NOT_FOUND)
    end
    dst_features:load()
    dst_features:setUpActiveClass(gSpace["gvars"].calltype)
    dst_features:setUpActiveClassValuesByActivatedFlag()
    
    local ringer = require"ringTone".getAlertHeader(dst_features:get("RING"), gSpace["gvars"].calltype)
    
    -- fix CDR fields
    local dialed_num = channel.DIALEDPEERNUMBER:get()
    if dialed_num ~= dst then -- call was made by switched extension
        channel.CDR("c_to"):set(gSpace["cidNum"])
        channel.CDR("to_name"):set(channel.CALLERID("name"):get() or "")
        channel.CDR("c_from"):set(dst)
        channel.CDR("from_name"):set(dst_attrs.cn or "")
    end

    local callToLocalFailoverUser = false
    local failoverObj = require"common.failover":new(dst_attrs.pbxowner)
    local isRegistered = require"core.sipproxy.sipproxy":new():getOnlineRegistrationByExten(dst)
    if failoverObj:localPbxIsFailover() and isRegistered then
        callToLocalFailoverUser = true
    end
    
    local optset = {
        ["opt"]    = "m",
        ["tout"]   = channel["GINTDIAL_TIMEOUT"]:get() or "",
        ["ring"]   = ringer,
        ["callToLocalFailoverUser"] = callToLocalFailoverUser,
    }
    
    -- make switch call in cycle
    local retry = 2
    repeat
        __extendDial(dst_attrs, dst_features, optset)
        retry = retry - 1
    until retry <= 0
        
    return smartApp("Hangup", hcause.NORMAL)
end

--- Main function for internalcall
-- @param dst           A destination number for the call
-- @param cid           A source number of the call
-- @param dial_timeout  A timeout for the dial
-- @param dial_options  An options for the dial
-- @param last_member   A last member in list of extension(0) or no(1) (set up when compile wms dialplan)
-- @param retTransfer   An indication of return from transfer
-- @param announce      An announcement to the called party
function run(self, dst, dial_timeout, dial_options, last_member, retTransfer, announce)
    ----------------- Local function definitions -----------------
    --- Get application values from channel variable
    local function get_application_values ()
       local application = channel.APPLICATION:get() or ""
       local appst = {}
       appst["disa"]    = application:match("^DISA.*")
       appst["kite"]    = application == "KITE"
       appst["queue"]   = application == "QUEUE"
       appst["callback"] = application == "CALLBACK"
       appst["fax"]     = application == "FAX"
       return appst
    end
    
    --- Check conditions for use preanswer service and tell "away" status or "user forward".
    -- Re-read channel variable for current channel state
    local function preanswer_condit(cid, cid_attrs, dst, dst_attrs, last_member, calltype)
        -- Get information about report user status notification
        -- "no" is default behaviour when users from trunk will not hear any status or message
        -- "status" when users from trunk will hear only callee status (AWAY or DND)
        -- "all" when users from trunk will hear callee status (AWAY or DND) and additional message from collaboration 
        local function report_user_status()
            local is_trunk = channel["X-TRUNK"]:get() or ""
            local report_user_status = channel["REPORT_USER_STATUS_FOR_EXT_CALLS"]:get() or ""
            if "" == report_user_status or "no" == report_user_status then
                report_user_status = "no"
            elseif "status" == report_user_status then
                report_user_status = "status"
            elseif "all" == report_user_status then
                report_user_status = "all"
            else
                logger:warning("Incorrect value: '%s' for variable REPORT_USER_STATUS_FOR_EXT_CALLS. Possible values are: 'all','status','no'. User status notification for external call is disabled", report_user_status)
                print("Incorrect value: '%s' for variable REPORT_USER_STATUS_FOR_EXT_CALLS. Possible values are: 'all','status','no'. User status notification for external call is disabled", report_user_status)
                report_user_status = "no"
            end
            channel["REPORT_USER_STATUS_FOR_EXT_CALLS"]:set(report_user_status)
            if "all" == report_user_status or "status" == report_user_status then
                logger:info("Report user status notification for external calls is enabled")
                print("Report user status notification for external calls is enabled")
            end
            return report_user_status
        end

        local rv       = false
        local is_trunk = channel["X-TRUNK"]:get() or ""
        local appst    = get_application_values()
        repeat
            if cid == dst then
                break
            end
            
            if last_member and 0 ~= last_member then
                break
            end
            
            if dst_attrs and dst_attrs.userKind == "fax" then
                break
            end
            
            if appst["queue"] or appst["callback"] or appst["fax"] then
                break
            end

            if "no" == report_user_status() then
                if "" ~= is_trunk and not appst["disa"] and not appst["kite"] then
                    break
                end

                if "external" == calltype and not appst["kite"] then
                    break
                end
            end

            local preAnserServiceACL =  gSpace.PREANSER_SERVICE_ACL
            if not preAnserServiceACL then
                preAnserServiceACL      = cid_attrs and require"common.permissions":check(cid_attrs["owner"] or "", "canuse", "PRE_ANSWER_SERVICES_FULL", nil, cid_attrs["licenseType"])
                gSpace.PREANSER_SERVICE_ACL = preAnserServiceACL
            end
            if preAnserServiceACL and "yes" ~= preAnserServiceACL then
                break
            end
            
            rv = true
        until true
        
        return rv
    end
    
    if "true" == channel.SIP_HEADER("X-From-WMSNet"):get() then
        __forwardIfIdentity(dst)
    end
    ----------------- Internal function processing -----------------
    local app, channel = app, channel
    local ivr, hcause  = ivr, hcause
    
    local cid           = gSpace["cidNum"]
    local cid_attrs     = gSpace["cidAttrs"]
    local gvars         = gSpace["gvars"]
    local is_trunk_call = gvars.trunk_call
    local dial_options  = dial_options or ""
    -- application definitions
    local appst = get_application_values()
    
    -- get unique id of the fax sending
    local faxnumberid = nil 
    if appst["fax"] then 
        faxnumberid = channel.faxnumberid:get() or "" 
    end
    
    -- retrieve QUEUE information
    local queueData = { 
        reCallTimeout = 5
       ,strategy      = ""
    }
    if appst["queue"] then
        local reCallTimeout = channel.QUEUE_RECALL_TIMEOUT:get() or ""
        reCallTimeout       = tonumber(reCallTimeout)
        if reCallTimeout and 2 <= reCallTimeout then
            queueData.reCallTimeout = reCallTimeout
        end
        
        local queueName = channel.QUEUE_NAME:get() or ""
        if "" ~= queueName then
            if "0" == channel.QUEUE_VARIABLES(queueName):get() then
                queueData.strategy = channel.QUEUESTRATEGY:get() or ""
            end
        end
    end
    
    -- check timeout and last member getting from remote pbx via app-dial-internalcall
    local rem_opts = channel.APPLICATION_DATA:get() or ""
    local rem_dial_timeout, rem_last_member = rem_opts:match("(%d*)|(%d?)")
    last_member = last_member or tonumber(rem_last_member)
    if not last_member and appst["queue"] then
        local flag4ElswereDial = channel["DIAL_OPTION_E"]:get() or ""
        if "" ~= flag4ElswereDial then
            last_member = 1
        end
    end
    if rem_dial_timeout then
        dial_timeout = rem_dial_timeout
    end
    local dialplan_app = last_member and true or false  
    
    -- processing anouncement options
    local announceOpts = announce and "" ~= announce and string.format("A(%s)", announce) or ""

    -- check number of forward (FCF, BCF, UCF)
    local MAX_FORWARDS = 3
    local fwd_hops, fwd_name = (channel["FORWARDED"]:get() or ""):match("(.*)|(.*)")
    if fwd_name and fwd_name ~= "" then -- used after forward from remote pbx
        channel.CALLERID("name"):set(fwd_name)
    end
    fwd_hops = tonumber(fwd_hops) or 0
    if 0 < fwd_hops then
        dial_timeout = gvars.dial_timeout or dial_timeout or ""
        dial_options = gvars.dial_options or dial_options or ""
        if MAX_FORWARDS > fwd_hops then
            print(string.format("We have forwarded call in %d hop(s)", fwd_hops))
        else
            print(string.format("We have forwarded call reached maximum hops (%d)", MAX_FORWARDS))
        end
    end
    
    -- used for next forwarders (for app-dial-internalcall)
    gSpace["gvars"].dial_options = dial_options
    gSpace["gvars"].dial_timeout = dial_timeout

    channel.CDR("c_to"):set(dst) -- set CDR field for correct history in case number not found, etc.

    local cid_features = require"common.userFeatures.userFeatures":new(cid_attrs.uid)
    if cid_features then
        cid_features:load()
        channel.PRESENCE_STATUS = cid_features:getUserStatus():upper()
        channel.PRESENCE_MESSAGE = require"base64".encode(
                                    require"core.pbxutils".tojson({
                                        location = tostring(cid_features:getUserStatusData("location")),
                                        message = tostring(cid_features:getUserStatusData("message")),
                                        untilTime = tostring(cid_features:getUserStatusData("untilTime"))}))
    end

    -- Retrieving destination attributes from LDAP    
    local dst_attrs = get_obj_attrs(dst)
    
    -- Processing return from transfer procedure
    local needTransferReturn = false
    if not dst_attrs.telephoneNumber then 
        print("Number", dst, "is not found in LDAP")
        if last_member and last_member == 1 then
            return true
        else
            needTransferReturn = true
        end
    elseif not retTransfer then
        -- skip processing recall in case attendant transfer
        local attendantTransferFlag = channel["SIPTRANSFER_REPLACES"]:get() or ""
        if "" ~= attendantTransferFlag then
            needTransferReturn = true
        end
    end
    if needTransferReturn then 
        local transferNum = channel["TRANSFER_NUMBER"]:get() or ""
        if "" ~= transferNum then
            fix_callername_return_transfer()
            local retMode = dst_attrs.telephoneNumber and "CALL" or "NONUSER"
            local rftCtx  = dst_attrs.telephoneNumber and "localUser" or "failExten"
            if not return_from_transfer(self, transferNum, dst, retMode, rftCtx) then
                return smartApp("Hangup", hcause.NOT_FOUND)
            else
                return true
            end
        end
    end
    
    if not dst_attrs.telephoneNumber then
        -- set correct dialstatus and hangup cause for outgoingfax
        if appst["fax"] then    
            channel.GLOBAL(faxnumberid.."_stat"):set("CHANUNAVAIL|1D")
            app.NoCDR()
        end
            
        if not dialplan_app and (not last_member or last_member == 0) then
            return smartApp("Hangup", hcause.NOT_FOUND)
        else
            return false
        end
    end
    
    -- rewrite callerid for call from A to B, when A is mobility callback call. User B will see that call from A extension (before user B had seen that call from A mobility)
    if channel.mobilityCIDRewrite:get() then
        cid = gSpace["cidAttrs"].telephoneNumber or cid
        channel.CALLERID("num"):set(cid)
    end

    -- processing changing identity (call from identity to internal user)
    local isqcall = channel["QUEUE_NAME"]:get() or ""
    if is_trunk_call == "" and cid_features and not require"common.helper".isTableEmpty(cid_features) and isqcall == "" then
        local from_identity = channel["FROM_IDENTITY"]:get()
        from_identity = "" ~= from_identity and from_identity
        local custom_identity = from_identity or  cid_features:getValue("CURRENT_IDENTITY")
        if custom_identity and "" ~= custom_identity then
            local available_identities = cid_features:getCustomIdentities() or {}
            if require"core.pbxutils".in_table(custom_identity, available_identities) then
                if cid_attrs.telephoneNumber and cid_attrs.telephoneNumber ~= custom_identity then
                    print("Call with a changed identity: %s => %s", cid_attrs.telephoneNumber, custom_identity)
                    logger:info("Call with a changed identity: %s => %s", cid_attrs.telephoneNumber, custom_identity)
                    local orig_cid = cid
                    cid = custom_identity
                    cid_attrs = get_obj_attrs(cid)
                    if cid_attrs.uid then
                        cid_features = require"common.userFeatures.userFeatures":new(cid_attrs.uid)
                        if cid_features then
                            cid_features:load()
                        end
                    end
                    if not cid_features then
                        print("Error loading features of the caller user %s", cid)
                        logger:error("Error loading features of the caller user %s", cid)
                        smartApp("Hangup", hcause.NOT_FOUND)
                    end
                    local identity_name = cid_attrs.cn or ""
                    channel.CDR("c_from"):set(cid)
                    channel.CDR("from_name"):set(identity_name)
                    local b64 = require"core.strConverter"
                    local c_from_b64 = b64.toB64(orig_cid) or ""
                    local c_to_b64 = b64.toB64(string.format("%s.%s", custom_identity, dst)) or ""
                    local to_name = dst_attrs.cn or dst
                    local to_name_b64 = b64.toB64(to_name) or ""

                    local fork_params = string.format("vTw(c_from=%s,c_to=%s,to_name=%s)", c_from_b64, c_to_b64, to_name_b64)

                    app.ForkCDR(fork_params)
                    channel.CDR("c_from"):set(orig_cid or "")
                    channel.CALLERID("num"):set(cid)
                    channel.CALLERID("name"):set(identity_name)
                    channel.CONNECTEDLINE("all"):set(string.format([===["%s" <%s>]===], dst_attrs.cn or dst, string.format("%s-%s", custom_identity, dst)))
                    dial_options = dial_options .. "I"
                    app.SipAddHeader(string.format("Caller-Image: http://127.0.0.1/user_avatar/%s.png", cid))
                end
            end
        end
    end

    -- check permissions of "CALL" for local users
    local pbx_src = require"core.wmsldap":get("pbxSerial", require"core.pbxutils".pbxSerial(), "wPbx")[1]
    local is_dst_local = pbx_src and next(dst_attrs) and pbx_src[1] == dst_attrs.pbxowner
    if is_trunk_call == "" then
        print(string.format("Dialing from %s to number %s %s", cid, dst, is_dst_local and "" or "at remote pbx"))
        if not retTransfer then
            local identity_attrs = {}
            local to_identity = channel["TO_IDENTITY"]:get() or ""
            if "" ~= to_identity then
                identity_attrs = get_obj_attrs(to_identity)
            end
            local is_allowed = require"common.permissions":check(cid_attrs.owner or "", "yesno", "CALL", identity_attrs.owner or dst_attrs.owner or "", cid_attrs["licenseType"]) == "yes"
            if not is_allowed then
                if not last_member or last_member == 0 then
                    print("Call is rejected by group permissions")
                    ivr:tell("Destination of call is not allowed")
                end
                if not dialplan_app then
                    return smartApp("Hangup", hcause.ACL_FORBIDDEN)
                else
                    return false
                end
            end
        end
    else
        print(string.format("Dialing from %s/%s to number %s", is_trunk_call, cid, dst))
    end

    if is_trunk_call and "" ~= is_trunk_call then
        if dst_attrs.owner and require"common.permissions":check(dst_attrs.owner, "canuse", "TRUNK_TO_TRUNK_TRANSFER", nil, dst_attrs["licenseType"]) ~= "yes" then
            channel.__BLOCK_TRANSFER = "yes"
        end
    end

    -- fix CDR fields with LDAP data
    if appst["fax"] then 
        app.NoCDR() -- not write CDR for call to fax, write only sending fax
    else
        if appst["queue"] then  -- for correct caller to queue
            channel.CDR("c_from"):set(cid_attrs.telephoneNumber or cid or "")
        end
        if cid_attrs.cn and not appst["disa"] then
            channel.CDR("from_name"):set(cid_attrs.cn)
        end
    end
    if dst_attrs.cn then
        if cid_attrs.cn then
            require"peerCallClass":setUp2Channel("internal")    
        end
        if appst["fax"] then    -- set correct dst name and number in case send local fax from cticonnect
            channel.GLOBAL(faxnumberid.."_pb"):set(dst.."|"..dst_attrs.cn.."|"..(cid_attrs.cn or ""))
        else
            channel.CDR("to_name"):set(dst_attrs.cn)
        end
    end
    
    -- disable video for inherited channel
    if "1" == channel.DISABLE_VIDEO:get() then
        channel._DISABLE_VIDEO:set(1)
    end

    -- disable opus for inherited channel
    if "1" == channel.DISABLE_OPUS:get() then
        channel._DISABLE_OPUS:set(1)
    end

    if not retTransfer then
        fix_callername_return_transfer()
    end

    local callToLocalFailoverUser = false
    if not is_dst_local then -- call to remote pbx
        local failoverObj = require"common.failover":new(dst_attrs.pbxowner)
        local isRegistered = require"core.sipproxy.sipproxy":new():getOnlineRegistrationByExten(dst)
        if failoverObj:localPbxIsFailover() and isRegistered then
            callToLocalFailoverUser = true
        else
            return call_to_wms(dst, dst_attrs, dial_timeout, dial_options, last_member, nil, retTransfer, announceOpts)
        end
    end

    -- call to local pbx
    if retTransfer then   -- clear TRANSFER_NUMBER if return from transfer to local pbx
        channel["TRANSFER_NUMBER"]:set("NULL")
    end

    -- clear variable for checking type of preanswer sounds
    if gvars.preanswer ~= "run_answer" then
        gSpace["gvars"].preanswer = nil
    end

    -- processing destination user data
    local calltype    = gvars.calltype
    local dst_context = dst_attrs.userDialplan or "users"
    if "fax" == dst_attrs.userKind then
        print("Kind of the destination user is 'fax'")
    end

    -- processing destination user features
    local dst_features = require"common.userFeatures.userFeatures":new(dst_attrs.uid)
    if not dst_features then
        print("Error loading features of the destination user")
        return smartApp("Hangup", hcause.NOT_FOUND)
    end
    dst_features:load()
    dst_features:setUpActiveClass(calltype)
    dst_features:setUpActiveClassValuesByActivatedFlag()

    -- retrive destination user call timeout
    local dstUserTimeout = dst_features:get("TOUT") or "0"
    local parkTimeout = tonumber("0" ~= dstUserTimeout and dstUserTimeout or (channel["GPARK_TIMEOUT"]:get() or "0")/1000)
    if "0" == dstUserTimeout then
        dstUserTimeout = channel["GINTDIAL_TIMEOUT"]:get() or "0"
    end
    dstUserTimeout = tonumber(dstUserTimeout)

    -- processing call to park oprbit
    if "park_orbit" == dst_attrs.userKind then
        print("Kind of the destination user is 'park orbit'")
        return __parkOrbitDial(dst_attrs, dst_features, parkTimeout)
    end

    -- show call information
    print("Calltype '"..(calltype or "").."'")
    __showActiveUserFeatures(dst_features)

    if cid ~= dst then
        if dst_features:is_active("FCF") then   -- Check call for feature 'Always call forward'
            local value = dst_features:get("FCF")
            if value == dst then
                print("ERROR! FCF feature for", dst, "is set to the same number!")
                print("Will proceed the call to", dst)
            elseif value == cid then
                print("Processing call from FCF number", cid, "...")
            elseif value == "VOICEMAIL" and not appst["queue"] and dst_attrs.userKind ~= "fax" then
                local fwdrv = forward_user(dst_attrs.telephoneNumber, "VM", "FORWARD_F")
                if fwdrv then
                    setpostaction {
                        ["name"] = "MISSED_CALL",
                        ["dst_attrs"] = dst_attrs,
                        ["dst_features"] = dst_features,
                        ["calltype"] = calltype,
                    }
                    channel.CHANNEL("hangup_handler_wipe"):set("hdlr,internal,1")
                    return fwdrv
                end
            elseif value ~= "" and value ~= "VOICEMAIL" and fwd_hops < MAX_FORWARDS then
                print("FCF is active for", dst)
                return forward_user(value, "F", dst, dst_context, pbx_src, fwd_hops, last_member, dial_timeout)
            end
        elseif dst_features:is_active("UCF") then   -- check dial options in case feature 'Unavailable call forward'
           if not appst["queue"] and fwd_hops < MAX_FORWARDS then
               local value = dst_features:get("UCF")
               if value and cid ~= value and dst ~= value and "VOICEMAIL" ~= value then
                   -- check dialplan rule configurations
                   local dialTimeout = tonumber(dial_timeout)
                   if not (1 == last_member and dstUserTimeout and dialTimeout and dialTimeout < dstUserTimeout) then
                       dial_options = dial_options .. "E"
                   end
               end
           end
        elseif 1 == last_member then                -- check dial options for non last dialplan application
            dial_options = dial_options .. "E"
        end
    end

    -- add default parameters for notify missed call
    setpostaction {
        ["name"] = "MISSED_CALL",
        ["dst_attrs"] = dst_attrs,
        ["dst_features"] = dst_features,
        ["calltype"] = calltype,
    }

    if appst["fax"] then
        dial_timeout = channel["GINTDIAL_TIMEOUT"]:get() or ""
        local addaction = {
            ["name2"] = "SENDFAX",
            ["faxnumberid"] = faxnumberid,
        }
        setpostaction(addaction, "add")
    end

    gSpace["execute_missed_call"] = 1       -- enable missed call notification. Will be disable following actions
    if not appst["queue"] then
        channel.CHANNEL("hangup_handler_wipe"):set("hdlr,internal,1")
    end

    channel.DIALSTATUS:set("BUSY")      -- used in case CW is disable for return correct status
    channel.CC_NAME:set(dst_attrs.cn)   -- used for CC callback (correct cli)
    local causecode   = hcause.NULL

    local userState   = __getUserState(dst_attrs, dst_features, true)

    -- connect to calling user when outgoing call is directed to the same number WMS-401
    if cid ~= dst and "internal" == calltype and not userState.isDevAvailable then
        local dst_chan       = channel.CHANNELS("SIP/" .. dst .. "-.*"):get() or ""
        local dst_chan_exten = channel.IMPORT(dst_chan, "EXTEN"):get() or ""
        local dst_bridgepeer = channel.IMPORT(dst_chan, "BRIDGEPEER"):get() or ""
        if "" == dst_bridgepeer and dst_chan_exten == cid then
            smartApp("Bridge", dst_chan)
            causecode = "SUCCESS" == channel["BRIDGERESULT"]:get() and hcause["NORMAL"] or hcause["FAILURE"]
            return smartApp("Hangup", causecode)
        end
    end

    local reject_call = dst_features:get("REJECT")
    --waiting changing of status for queue call
    if appst["queue"] and "ringall" == queueData.strategy then
        while reject_call do
            app.Wait(queueData.reCallTimeout)
            dst_features = require"common.userFeatures.userFeatures":new(dst_attrs.uid)
            if not dst_features then
                print("Error loading features of the destination user")
                return smartApp("Hangup", hcause.NOT_FOUND)
            end
            dst_features:load()
            dst_features:setUpActiveClass(calltype)
            dst_features:setUpActiveClassValuesByActivatedFlag()
            reject_call = dst_features:get("REJECT")
        end
    end

    -- processing user status notification
    local dst_status  = dst_features:getUserStatus():upper()
    if cid ~= dst and "run_answer" ~= gvars.preanswer and not retTransfer and userState.isChanAvailable and 
      preanswer_condit(cid, cid_attrs, dst, dst_attrs, last_member, calltype) then
        if "AWAY" == dst_status or "DND" == dst_status then   -- notification about special user status
            gSpace["gvars"].preanswer    = dst_status .. "_progress"
            local dstStatusMsg = dst_features:getUserStatusData("message") or ""
            if "" ~= dstStatusMsg and not string.find(dstStatusMsg, "^.*%.$") then -- add a point to the end of the status message
                dstStatusMsg = dstStatusMsg.."."
            end
            local untilOffset  = dst_features:getUserStatusData("untilOffset") or {}
            if untilOffset.difftime and untilOffset.difftime > 0 and untilOffset.difftime < 86400 then
                local hour = untilOffset.hour or 0
                local min  = untilOffset.min or 0
                      hour = (hour > 0 and (" "..hour.."h") or "")
                      min  = (min > 0 and (" "..min.."min ") or "")
                dstStatusMsg  = dstStatusMsg.." User will be available in"..hour..min
            end
            if "" ~= dstStatusMsg then
                gSpace["gvars"].dststatusmsg = dstStatusMsg
            end

            if reject_call then                         -- running preanswer service with call rejecting
                local rv = preanswer_service(dst_attrs, "CANCEL")
                if rv ~= 1 or (rv == 2 and channel.VMSTATUS:get() == "SUCCESS") then
                    return smartApp("Hangup", hcause[rv ~= 0 and rv or "NULL"])
                end
            else                                        -- simple notification about user status
                preanswer_service(dst_attrs)
            end
        end
        userState = __getUserState(dst_attrs, dst_features)
        if not reject_call and not userState.isDevAvailable then    -- check device state BUSY and call waiting enable
            gSpace["gvars"].preanswer = "cw_noopt_progress"
            local rv = preanswer_service(dst_attrs, "BUSY")
            if rv ~= 1 or (rv == 2 and channel.VMSTATUS:get() == "SUCCESS") then
                return smartApp("Hangup", hcause[rv ~= 0 and rv or "NULL"])
            end
        end
    end

    -- processing reject call feature
    if cid ~= dst and not retTransfer and reject_call then
        channel.DIALSTATUS:set("CANCEL")
        print("Executing reject call procedure ...")
        if appst["fax"] then
            setpostaction({"name2", "faxnumberid"}, "rm")
        end
        return false
    end

    -- disable mobility if user already speaking (or state is ringing) on voip device.
    if dst_features:is_active("MEXT") and not require"deviceState":new(dst):reqValue():isAvailable4Service("MOBILITY") then
        dst_features:set("MEXT", {status = 0})
        print("Disabling mobility call because user is available on the VoIP device!")
    end

    local chan_dialstatus = nil
    if userState.isChanAvailable or appst["queue"] or cid == dst then
        -- set channel variable used for transfer (options for transfer add callweaver)
        if dst_attrs.userDialplan then
            channel["__TRANSFER_CONTEXT"]:set(dst_attrs.userDialplan)
        elseif cid_attrs.userDialplan then
            channel["__TRANSFER_CONTEXT"]:set(cid_attrs.userDialplan)
        end

        -- get dst group number from ldap
        local dst_gid     = string.match(dst_attrs["owner"] or "", "gid=(%d+),.*")
        local group_attrs = get_obj_attrs(dst_gid, "gid", "wGroup")
        local callgroup   = group_attrs.wcgrp or "" --user for pickup froup

        -- use for clear timeout after forward via button
        dial_options = dial_options..announceOpts.."z"

        local optset = {
             ["opt"]       = dial_options
            ,["tout"]      = dial_timeout
            ,["ring"]      = ""
            ,["cg"]        = callgroup
            ,["trunkCall"] = gvars.trunk_call
            ,["dst_uid"]       = dst_attrs.uid
            ,["callToLocalFailoverUser"] = callToLocalFailoverUser
        }

        local isReDial = nil
        repeat
            isReDial  = false
            userState = __getUserState(dst_attrs, dst_features)

            -- executing extended dial
            if userState.isChanAvailable or cid == dst then
                -- processing ring tone
                local ringTone = nil
                local silent_waiting = "no" ~= channel.CW_SILENT_RINGTONE:get()
                if silent_waiting and dst ~= cid and require"deviceState":new(dst):reqValue():isAvailable4Service("SILENTRING") then
                    -- set silent ring, because user has already spoken
                    ringTone = "silent"
                else -- processing ring tone by some logic
                    local userRingTone             = nil
                    local dialplanVariableRingtone = channel.RINGTONE:get() or ""
                    if "" ~= dialplanVariableRingtone then
                        print("Using ringtone from dialplan variable:", dialplanVariableRingtone)
                        ringTone = dialplanVariableRingtone
                    else
                        userRingTone = dst_features:get("RING")
                        if "silent" ~= userRingTone then
                            local pbRingTone = channel.PBRINGTONE:get() or ""
                            if "" ~= pbRingTone then
                                print("Using ring tone from phonebook contact data:", pbRingTone)
                                ringTone = pbRingTone
                            end
                        end
                    end
                    if not ringTone or "" == ringTone then
                        ringTone = userRingTone
                    end
                end
                if dst_attrs.userKind and "fax" ~= dst_attrs.userKind then
                    optset.ring = require"ringTone".getAlertHeader(ringTone, calltype)
                end

                __extendDial(dst_attrs, dst_features, optset, cid)
            end
            chan_dialstatus = channel["DIALSTATUS"]:get() or "UNKNOWN"

            -- set up redial flag in case queue call
            if appst["queue"] and "ringall" == queueData.strategy then
                if "BUSY" == chan_dialstatus or "NOANSWER" == chan_dialstatus then
                    local hangupCause = channel.HANGUPCAUSE:get() or ""
                    if "17" == hangupCause then
                      -- update timeout after BUSY cause from device
                      queueData.reCallTimeout = 20
                    end
                    if "21" ~= hangupCause and "26" ~= hangupCause then   -- disable redial in case REJECT from device
                        isReDial = true
                        app.Wait(queueData.reCallTimeout)   -- timeout before recall to queue member
                    else
                        print("Quick redial disabled by reject code " .. hangupCause)
                    end
                end
            end
        until false == isReDial
    end
    if not chan_dialstatus then -- we got dialstatus only after dial
        print("Call Waiting is disabled for", dst)
    print("Setting", dst, "as BUSY")
        causecode = hcause.BUSY
    end

    -- disable for handling next actions in preanswer service
    if preanswer_condit(cid, cid_attrs, dst, dst_attrs, last_member, calltype) then
        gSpace["execute_missed_call"] = 0
    end

    -- return from transfer
    local transferNum = channel["TRANSFER_NUMBER"]:get() or ""
    if retTransfer then -- use for recall after return from transfer
        if extensions["receptionist"] and extensions["receptionist"]["s"] then
            return require"services":start_call("app-goto-general","s","receptionist","1","")
        else
            return true
        end
    elseif "" ~= transferNum then
        local rem_opts = channel.APPLICATION_DATA:get() or ""
        local rem_dial_timeout, rem_last_member = rem_opts:match("(%d*)|(%d?)")
        if tonumber(rem_last_member) == 1 or last_member == 1 then
            return true
        else
            local rft = return_from_transfer(self, transferNum, dst, "CALL", "localUser")
            if rft then
                return rft
            end
        end
    end

    chan_dialstatus = channel["DIALSTATUS"]:get() or "UNKNOWN"
    if appst["fax"] then -- push to outgoing fax correct dialstatus
        local hcausecode = channel.HANGUPCAUSE:get() or ""
        channel.GLOBAL(faxnumberid.."_stat"):set(chan_dialstatus.."|"..hcausecode)
        setpostaction({"name2", "faxnumberid"}, "rm")   -- we have dialstatus, remove action from hangup handler
    end

    -- reboot for w01fxs block
    if dst_attrs.userKind == "fax" and chan_dialstatus == "BUSY" and userState.isChanAvailable then
        setpostaction({["name3"] = "RESET_FAX", ["fax_num"] = dst}, "add")
    end

    if cid ~= dst and ("CONGESTION" == chan_dialstatus or "BUSY" == chan_dialstatus) then
        if dst_features:is_active("BCF") then
            local value = dst_features:get("BCF")
            if value == dst then
                print("ERROR! BCF feature for", dst, "is set to the same number!")
                print("Call will be proceed ...")
            elseif value == cid then
                print("Processing call from BCF number", cid, "...")
            elseif value == "VOICEMAIL" and not appst["queue"] and dst_attrs.userKind ~= "fax" then
                local fwdrv = forward_user(dst_attrs.telephoneNumber, "VM", ("CONGESTION" == chan_dialstatus and "FORWARD_U") or "FORWARD_B")
                if fwdrv then
                    return fwdrv
                end
            elseif value ~= "" and value ~= "VOICEMAIL" and fwd_hops < MAX_FORWARDS then
                -- enable preanswer service during "User busy"
                if preanswer_condit(cid, cid_attrs, dst, dst_attrs, last_member, calltype) then
                    gSpace["gvars"].preanswer = "bcf_noopt_progress"
                    local res = preanswer_service(dst_attrs, chan_dialstatus)
                    if res ~= 1 or (res == 2 and channel.VMSTATUS:get() == "SUCCESS") then
                        local causecode = hcause[res ~= 0 and res or "NULL"]
                        return smartApp("Hangup", causecode)
                    end
                end
                print("BCF is active for", dst)
                return forward_user(value, "B", dst, dst_context, pbx_src, fwd_hops, last_member, dial_timeout)
            end
        end
    end

    if cid ~= dst and ("NOANSWER" == chan_dialstatus or "CHANUNAVAIL" == chan_dialstatus or "CONGESTION" == chan_dialstatus) then
        if dst_features:is_active("UCF") then
            local value = dst_features:get("UCF")
            if value == dst then
                print("ERROR! UCF feature for", dst, "is set to the same number!")
                print("Call will be proceed ...")
            elseif value == cid then
                print("Processing call from UCF number", cid, "...")
            else
                -- checking user timeout with dial timeout
                local dialTimeout = tonumber(dial_timeout)
                if 1 == last_member and dstUserTimeout and dialTimeout and dialTimeout < dstUserTimeout then
                    print("Skipping UFC to execute next dialplan application ...")
                    return true
                end

                if value == "VOICEMAIL" and not appst["queue"] and dst_attrs.userKind ~= "fax" then
                    local fwdrv = forward_user(dst_attrs.telephoneNumber, "VM", "FORWARD_T")
                    if fwdrv then
                        return fwdrv
                    end
                elseif value ~= "" and value ~= "VOICEMAIL" and fwd_hops < MAX_FORWARDS then
                    -- enable preanswer service during "User busy"
                    if preanswer_condit(cid, cid_attrs, dst, dst_attrs, last_member, calltype) then
                        gSpace["gvars"].preanswer = "ucf_noopt_progress"
                        local res = preanswer_service(dst_attrs, chan_dialstatus)
                        if res ~= 1 or (res == 2 and channel.VMSTATUS:get() == "SUCCESS") then
                            local causecode = hcause[res ~= 0 and res or "NULL"]
                            return smartApp("Hangup", causecode)
                        end
                    end
                    print("UCF is active for", dst)
                    return forward_user(value, "T", dst, dst_context, pbx_src, fwd_hops, last_member, dial_timeout)
                end
            end
        end
    end

    -- if last in list then last_member = 0
    if not last_member or last_member == 0 then -- check for recall after return from transfer
        if preanswer_condit(cid, cid_attrs, dst, dst_attrs, last_member, calltype) then
            if gSpace["gvars"].preanswer then
                gSpace["gvars"].preanswer = "answer"
            else
                gSpace["gvars"].preanswer = "progress"
            end
            preanswer_service(dst_attrs, chan_dialstatus, dial_timeout, dial_options)
        end
    end

    -- return from transfer in case attendant transfer
    fix_callername_return_transfer()
    local transferNum = channel["TRANSFER_NUMBER"]:get() or ""
    if "" ~= transferNum then
        return return_from_transfer(self, transferNum, dst, "CALL", "localUser")
    end

    return true
end

--- Preparing for messed call notification
-- if execute = 0 it means not send notification (by default)
-- if execute = 1 it means that preanswer service failed and send notification
-- if execute = 2 it means that send voicemail via preanswer service and need check vm status
function hangup(self)
    local channel = channel
    
    --check if bridged channel
    if "DONE" == channel["BRIDGING_CHANNEL_APP"]:get() or "SUCCESS" == channel["BRIDGERESULT"]:get() then
        return false
    end
    
    local execute    = gSpace["execute_missed_call"]
    local dialStatus = channel["DIALSTATUS"]:get() or ""
    local action     = getpostaction()
    if not action then 
        return false 
    end
    
    if "MISSED_CALL" == action.name and not "ANSWER" ~= dialStatus then
        local isProcessing = false
        if 1 == execute then
            local isRedirected = channel.REDIRECT:get() or ""
            if "" == isRedirected then
                isProcessing = true
            end
        elseif 2 == execute then
            local vmstatus = channel.VMSTATUS:get() or ""
            if vmstatus ~= "" and vmstatus ~= "SUCCESS" then
                isProcessing = true
            end
        end
        
        if isProcessing then
            local srcData = {
                 id    = gSpace["cidNum"]
                ,attrs = gSpace["cidAttrs"]
            }

            local custom_identity
            local from_identity = channel["FROM_IDENTITY"]:get() or ""
            from_identity = "" ~= from_identity and from_identity
            if from_identity then
                custom_identity = from_identity
            else
                local cid_features = gSpace["cidFeatures"]
                if cid_features and not require"common.helper".isTableEmpty(cid_features) then
                    identity = cid_features:getValue("CURRENT_IDENTITY") or ""
                    custom_identity = "" ~= identity and identity
                end
            end
            if custom_identity then
                    srcData = {
                         id    = custom_identity
                        ,attrs = get_obj_attrs(custom_identity) or {}
                    }
            end

            local dstData = {
                 id    = action.dst_attrs.telephoneNumber
                ,attrs = action.dst_attrs
                ,feat  = action.dst_features
            }
            local notifier, errMsg = require"missedCallNotify":new(srcData, dstData, dialStatus)
            if notifier then
                notifier:SIPNotifyEvent()
                notifier:textMsgEvent()
            else
                print("Missed call notification was not sent:", errMsg)
                logger:error("Missed call notification was not sent:", errMsg)
            end
        end
    end
    
    if action.name2 == "SENDFAX" and "ANSWER" ~= dialStatus then
        -- executed in case wrong DIALSTATUS in outgoingfax
        app.MacroExclusive("faxcounter", "--")
        local faxnumberid = action.faxnumberid
        channel.GLOBAL(faxnumberid.."_pb"):set("NULL") --remove global variable
        channel.GLOBAL(faxnumberid.."_stat"):set("NULL") --remove global variable
    end
    
    if action.name3 == "RESET_FAX" then
        local msg = string.format("FAX answered BUSY, but it was NOT_INUSE. Sending notify check-sync to %s", action.fax_num)
        logger:debug(msg); print(msg)
        require"utils":notify_peer(action.fax_num, 1)
    end
end
