module("queuecalls", package.seeall)

--- Checking member status by some conditiona and return extension
-- that must be added to queue call 
local function __proceed_member_status(memberNumber, callType)
    local rv  = nil
    local err = nil
    if not memberNumber or "" == memberNumber then
        return rv
    end
    local memberLocation = nil
    
    repeat
        -- loading member data from LDAP
        local memberAttrs = get_obj_attrs(memberNumber)
        if not memberAttrs.uid then
            err = "Not found data in LDAP"
            break
        end
        
        -- check member location
        local localPbxDN = require"core.pbxutils".getPbxDNBySerial()
        if localPbxDN then
            if localPbxDN == memberAttrs.pbxowner then
                memberLocation = "local"
            elseif memberAttrs.pbxowner then
                memberLocation = memberAttrs.pbxowner:match("^pbxid=(%w+)") 
                break
            end
        end
        if not memberLocation then
            err = "Not found member location"
            break
        end
        
        --- loading member preferences
         -- loading destination user features
        local memberFeatures = require"common.userFeatures.userFeatures":new(memberAttrs.uid)
        if not memberFeatures then
            print("Error loading features of the destination user")
            return smartApp("Hangup", hcause.NOT_FOUND)
        end
        memberFeatures:load()
        memberFeatures:setUpActiveClass(callType)
        memberFeatures:setUpActiveClassValuesByActivatedFlag()
        
        -- checking FCF status
        local FCFValue = memberFeatures:get("FCF")
        if FCFValue then
            rv = __proceed_member_status(FCFValue, callType)
            break
        end
        
        -- checking 'Reject call' feature
        local rejectCall = memberFeatures:get("REJECT")
        if rejectCall then
            err = "Reject call is activated in user features"
            break
        end
        
        -- checking channel availability with member status
        local devStateObj     = require"deviceState":new(memberNumber)
        devStateObj:setLogs(true)
        devStateObj:reqValue()
        local isDevAvailable  = devStateObj:isAvailable()
        local memberCallWait  = memberFeatures:get("CW")
        local isChanAvailable = true
        if not isDevAvailable and ("fax" == memberAttrs.userKind or not memberCallWait) then
            isChanAvailable = false 
        end
        if not isChanAvailable then
            err = "Devise is busy at this time"
            break
        end
        
    until true
    
    if not err then
        rv = rv or {
             num = memberNumber
            ,loc = memberLocation 
        }
    end
    
    return rv, err
end

--- Main function for running queue calls procedure
-- @param callGroupId       Call group name
-- @param callGroupMembers  Custom list of the queue members
-- @return executing callweaver application for call group call
function run(callGroupId)
    local membersList       = {}
    local customMembersList = channel.APPLICATION_DATA:get() or ""
    if "" ~= customMembersList then
        callGroupId = ""    -- Call group ID does not present in case processing queuecall on remote PBX
        for member in customMembersList:gmatch("(%d+)") do
            table.insert(membersList, member)
        end
    else
        local callGroupMembers = require"cgroups".get_members_list(callGroupId)
        if callGroupMembers then
            membersList = callGroupMembers.order
        end
    end
    
    if membersList[1] then
        -- checking call type value
        local callType = gSpace["gvars"].calltype
        print("Call type is", callType)
        
        -- processing members list by checking user features and status
        local cgExtensionsList = {}
        for _, memberNum in ipairs(membersList) do
            local member, err = __proceed_member_status(memberNum, callType)
            if member then
                local cgExtensions = nil
                if cgExtensionsList[member.loc] then
                    cgExtensions = cgExtensionsList[member.loc]
                else
                    cgExtensions                 = {}
                    cgExtensionsList[member.loc] = cgExtensions
                end
                if not cgExtensions[member.num] then
                    table.insert(cgExtensions, member.num)
                end
            end
        end
        
        -- build dial request
        local dialStr       = {}
        local dialTimeout   = ""
        for memberLocation, membersList in pairs(cgExtensionsList) do
            if 0 ~= #membersList then
                local cgExts = table.concat(membersList, ";")
                if "local" == memberLocation then
                    dialTimeout           = channel["GINTDIAL_TIMEOUT"]:get() or ""
                    table.insert(dialStr, 1, "SIP/"..membersList[1])
                    if "" ~= customMembersList then
                        print("Executing power call group dial from remote pbx: ", cgExts)
                    else
                        print("Executing power call group dial for local users: ", cgExts)
                    end
                else
                    dialStr[#dialStr + 1] = string.format("Local/WMSNet_%s@queuecalls", memberLocation)
                    print(string.format("Executing power call group dial for remote users (%s): %s", memberLocation, cgExts))
                end
            end
        end
        
        if dialStr[1] then
            local encCGExtensions = require"dkjson".encode(cgExtensionsList)
            encCGExtensions       = require"base64".encode(encCGExtensions)
            local dialOptions     = string.format("b(predial^queuecalls^1(%s,%s))", encCGExtensions, callGroupId)
            return smartApp("Dial", table.concat(dialStr, "&"), dialTimeout, dialOptions)
        else
            print("Call group members list is empty -", callGroupId)
            return smartApp("Hangup", 1)
        end
    end
    
    print("Could not found members in call group -", callGroupId)
    return smartApp("Hangup", 1)
end

--- Running queue call via remote PBX
function runRemoteCall(dstStr)
    dstStr = dstStr or ""
    dstStr = dstStr:match("^WMSNet_(%w+)")
    if not dstStr or "" == dstStr then
        print("Invalid destination string data for remote queue call")
        return smartApp("Hangup", 1)
    end
    
    local cgRemExts = channel.GCREMOTEEXT:get() or ""
    if "" == cgRemExts then
        print("Empty variable with the remote call group users")
        return smartApp("Hangup", 1)
    end
    
    local cgName     = channel.GCREMOTENAME:get() or ""
    if "" == cgName then
        print("Not found call group ident")
        return smartApp("Hangup", 1)
    end
    
    cgRemExts = require"dkjson".decode(cgRemExts)
    cgRemExts = cgRemExts[dstStr]
    if "table" ~= type(cgRemExts) or 0 == #cgRemExts then
        print("Empty list of the remote call group users")
        return smartApp("Hangup", 1)
    end
    
    local remoteMember = cgRemExts[1]
    local remMembAttrs = get_obj_attrs(remoteMember)
    local callType     = channel.CALLTYPE:get() or ""
    local dialOpts     = require"wmsNetCall".makeArgs({
             pbxid        = require"core.strConverter".toB64(remMembAttrs.pbxowner)
            ,ctx          = "queuecalls"
            ,calltype     = callType
            ,appdata      = table.concat(cgRemExts, ";")
            ,cgpickupinfo = "pickup*cg_"..cgName
            ,linkedid     = channel.LINKEDID:get() or ""
        })
    return smartApp("Dial", "SIP/"..remoteMember, "", dialOpts)
end

