--- Search user
-- 
-- @file	lookup.lua 
-- @author  Gennadiy Ganchev
-- @created 2017/12/29

--- @module lookup
local lookup = {}


function getASRChannelVars()
	local asr_patterns = {}
	local asr_extens = {}
	local i = 1
	while channel["ASR_IVR_PATTERN_" .. tostring(i)]:get() and "" ~= channel["ASR_IVR_PATTERN_" .. tostring(i)]:get() do
		local chan_var = channel["ASR_IVR_PATTERN_" .. tostring(i)]:get()
		local tmp_table = {}
		for pattern in string.gmatch(chan_var .. ",", "(.-),") do
			table.insert(tmp_table, pattern:gsub("^%s*(.-)%s*$", "%1"):lower())
		end
		if #tmp_table > 1 then
			asr_extens[#asr_extens + 1] = tmp_table[#tmp_table]
			tmp_table[#tmp_table] = nil
			table.insert(asr_patterns, tmp_table)
		end
		i = i + 1
	end
	return asr_patterns, asr_extens
end

--- Search user by name
-- @param rData  A table with found users
-- @param sVal   A value for search
-- @return true or false
function lookup.byName(rData, sVal, sFilter, sSource)
	if "table" ~= type(rData) or not sFilter or not sVal then
		return false
	end
	sVal = string.gsub(sVal, "[.,!?:;]", "")

	local pbxutils = require"core.pbxutils"
	local sSource  = sSource or "LDAP"

	-- search procedure
	local searchedData = {}
	if "LDAP" == sSource then
		local ldapData = require"core.wmsldap":get("cn", "*")
		if ldapData and ldapData[1] then
			for _, data in pairs(ldapData) do
				if data[2] and data[2].uid and data[2].owner then
					searchedData[#searchedData + 1] = {
						 uid       = data[2].uid
						,owner     = data[2].owner
						,ext       = data[2].telephoneNumber
						,commName  = data[2].cn
						,firstName = data[2].givenName
						,fnHash    = data[2].givenName and pbxutils.chars2number(data[2].givenName)
						,lastName  = data[2].sn
						,lnHash    = data[2].sn and pbxutils.chars2number(data[2].sn)
						,license   = data[2].licenseType
					}
				end
			end
		end
	end

	local function __lookup_find(str, sval)
		local escaped = string.gsub(sval, "%-", "[%%- ]")
		return string.find(string.lower(str), string.lower(escaped))
	end

	-- apply filter options and process requested data
	noResults = true
	for idx, data in pairs(searchedData) do
		if data.commName and data.ext then
			if "f" == sFilter or "b" == sFilter then
				if data.fnHash and data.fnHash:find("^"..sVal) then
					searchedData[idx]["accepted"] = true
				end
			end
			if "l" == sFilter or "b" == sFilter then
				if data.lnHash and data.lnHash:find("^"..sVal) then
					searchedData[idx]["accepted"] = true
				end
			end
			if "a" == sFilter or "i" == sFilter then
				if __lookup_find(data.commName, "^"..sVal.."$") then
					searchedData[idx]["accepted"] = true
					noResults = false
				elseif data.lastName and __lookup_find(data.lastName, "^"..sVal.."$") then
					searchedData[idx]["accepted"] = true
					noResults = false
				elseif data.firstName and __lookup_find(data.firstName, "^"..sVal.."$") then
					searchedData[idx]["accepted"] = true
					noResults = false
				end
			end

			-- check user ACL to update serached data result
			if "yes" ~= require"common.permissions":check(data.owner , "yesno", "DIRECTORY_LOOKUP", nil, data.license) then
				if "i" ~= sFilter then
					searchedData[idx]["accepted"] = false
					logger:debug("Remove user %s from directory lookup results due to acl", data.commName)
				end
			end
		end
	end

	if noResults then
		if "a" == sFilter or "i" == sFilter then
			for w in sVal:gmatch("%S+") do
				for idx, data in pairs(searchedData) do
					if data.commName and data.ext then
						if data.lastName and (("a" == sFilter and __lookup_find(data.lastName, w))
										  or ("i" == sFilter and string.lower(data.lastName) == string.lower(w))) then
							searchedData[idx]["accepted"] = true
						elseif data.firstName and (("a" == sFilter and __lookup_find(data.firstName, w))
										  or ("i" == sFilter and string.lower(data.firstName) == string.lower(w))) then
							searchedData[idx]["accepted"] = true
						end
						-- check user ACL to update serached data result
						if "yes" ~= require"common.permissions":check(data.owner , "yesno", "DIRECTORY_LOOKUP", nil, data.license) then
							if "i" ~= sFilter then
								searchedData[idx]["accepted"] = false
								logger:debug("Remove user %s from directory lookup results due to acl", data.commName)
							end
						end
					end
				end
			end
		end
	end

	if "i" == sFilter then
		local asr_dialbyname = channel["ASR_DIALBYNAME"]:get() or "yes"
		if "no" == asr_dialbyname:gsub("%s",""):lower() then
			searchedData = {}
			print("Skip lookup by name, because ASR_DIALBYNAME = no")
			logger:info("Skip lookup by name, because ASR_DIALBYNAME = no")
		end
		local asr_patterns, asr_extens = getASRChannelVars()
		for ei, patterns in pairs(asr_patterns) do
			for _, pattern in pairs(patterns) do
				searchedData[#searchedData + 1] = {
					 commName = pattern:lower()
					,ext	  = "ei" .. asr_extens[ei]
				}
			end
			for idx, data in pairs(searchedData) do
				if __lookup_find(data.commName, "^"..sVal.."$") then
					searchedData[idx]["accepted"] = true
				end
			end
		end
	end

	-- build result data
	for _, data in pairs(searchedData) do
		if data and data.accepted then
			rData[#rData + 1] = {
				 name = data.commName
				,ext  = data.ext
			}
		end
	end

	return 0 ~= #rData
end

--- Get all users list
-- @param source - search location ("LDAP" by default)
-- @param limit  - max number of users
-- @return a table with all found users names
function lookup.allUsers(source, limit, sFilter)
	local pbxutils = require"core.pbxutils"
	local source  = source or "LDAP"
	local limit   = tonumber(limit)

	-- search procedure
	local rv    = {}
	if sFilter and "i" == sFilter then
		local asr_patterns, _ = getASRChannelVars()
		for _, patterns in pairs(asr_patterns) do
			for _, pattern in pairs(patterns) do
				rv[#rv + 1] = pattern
			end
		end
	end

	local flags = {} -- additional table for mark already stored items
	local asr_dialbyname = channel["ASR_DIALBYNAME"]:get() or "yes"
	if "no" ~= asr_dialbyname:gsub("%s",""):lower() then
		if "LDAP" == source then
			local ldapData = require"core.wmsldap":get("cn", "*")
			if ldapData and ldapData[1] then
				for _, data in pairs(ldapData) do
					if limit and #rv >= limit then
						break
					end
					if not data[2] or not data[2].uid or not data[2].owner then
						break
					end
					if data[2].sn and not flags[data[2].sn] then
						rv[#rv + 1] = data[2].sn
						flags[data[2].sn] = true
					end
					if data[2].givenName and not flags[data[2].givenName] then
						rv[#rv + 1] = data[2].givenName
						flags[data[2].givenName] = true
					end
					rv[#rv + 1] = data[2].cn
				end
			end
		end
	end

	return rv
end

return lookup