--- Directory service processing
-- 
-- @file	processing.lua 
-- @author  Gennadiy Ganchev
-- @created 2018/01/09

--- @module processing
local processing = {}

local function __setSearchMethod(callback)
	if "function" == type(callback) then
		return callback
	end
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local node = {  -- meta class
		 mFilter     = nil
		,mLimit	     = 12
		,mSource     = nil
		,mSearchFn   = nil
		,mState	     = nil
		,mRetry	     = 1
		,mRetryLimit = 3
		,mPlayObj    = nil
		,mMaxResults = 5
	}

--- Initialize "dial by name" service.
--- Ask the user for the value for search and search it
-- @param A list - table with a list of users
-- @return processing state - "choice": go to the next step; "welcome": return back and repeat the init procedure again
function node:init(list)
	repeat
		-- check input args
		if "table" ~= type(list) then
			break
		end
	
		-- check retry
		if self.mRetry > self.mRetryLimit then
			self.mState = nil
			logger:info("DialByName: limit of recognition attempts(%s) has been reached.", self.mRetryLimit)
			break
		end
		
		-- playback user input prompt
		local searchVal = self.mPlayObj:welcome(self.mLimit, self.mFilter, self.mRetryLimit)
	
		-- validate inptu data
		if not searchVal then
			self.mState = "welcome"
			self.mRetry = self.mRetry + 1
			break
		end
	
		-- search users list
		if self.mSearchFn(list, searchVal, self.mFilter, self.mSource) then
			if #list > self.mMaxResults then
				self.mPlayObj:tooManyMatches(searchVal)
				self.mState = "welcome"
			else
				self.mState = "choice"
			end
		else
			self.mPlayObj:notFoundEntry(searchVal)
			self.mState = "welcome"
			self.mRetry = self.mRetry + 1
		end
	until true

	return self.mState
end
	
--- Making a decision
--- Ask the user to call the found contact or to search again
-- @param A list - table with a list of users
-- @return processing state - "choice": go to the next step; "welcome": return back and repeat the init procedure again
function node:choice(list)
	repeat
		-- check input args
		if "table" ~= type(list) then
			break
		end

		-- check retry
		if self.mRetry > self.mRetryLimit then
			self.mState = nil
			break
		end

		-- processing actual user index
		if not list.actIdx then
			list.actIdx = 1
		end

		-- check next account for playback
		if not list[list.actIdx] then
			if #list > 1 or (self.mFilter ~= "a" and self.mFilter ~= "i") then
				self.mPlayObj:noNextEntry()
			end
			list.actIdx = 1
			self.mState = "welcome"
			self.mRetry = self.mRetry + 1
			break
		end

		-- playback user data
		self.mPlayObj:userData(list[list.actIdx])

		-- retrieve user choice to next action
		local userChoice = self.mPlayObj:selectAction(#list == list.actIdx and true or false)
		if not userChoice then
			break
		end

		if "cancel" == userChoice then
			self.mState = "welcome"
		elseif "0" == userChoice or "next" == userChoice then
			list.actIdx = list.actIdx + 1
		elseif "1" == userChoice or "call" == userChoice then
			list.dialExt = list[list.actIdx] and list[list.actIdx].ext
			self.mState  = nil
		else
			self.mRetry = self.mRetry + 1
		end

	until true

	return self.mState
end

-- Create processing node instance
function processing:new(opts, searchFnList)
	-- check if a playback is available
	local playbackObj = nil
	local err         = nil

	if "a" == opts.filter or "i" == opts.filter then
		playbackObj, err = require"playbackAI":new(opts.filter)
	else
		playbackObj, err = require"playbackStd":new()
	end

	if not playbackObj then
		return false, err
	end
	local obj = {}
	setmetatable(obj, {__index = node})

	-- init object
	if opts.limit then
		obj.mLimit = opts.limit
	end
	if opts.retry then
		obj.mRetryLimit = opts.retry
	end
	
	obj.mPlayObj  = playbackObj
	obj.mFilter   = opts.filter
	obj.mSource   = opts.source
	obj.mSearchFn = __setSearchMethod(searchFnList[opts.filter])
	
	return obj
end

return processing