--- Service to executing contact look up by different methods
--
-- @file    contactLookUp.lua
-- @author  Vyacheslav Khudan
-- @created 2017/02/09

--- @module contactLookUp
local contactLookUp = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local function __searchByLdap(num)
	local ldapData = get_obj_attrs(num, "mobile") -- search by "mobile
    if not ldapData or not ldapData.uid then
        return nil
    end

    return {
         name      = ldapData.cn
        ,email     = ldapData.mail
        ,callType  = "internal"
        ,orgName   = ldapData.ou
        ,altNumber = ldapData.telephoneNumber
   }
end

local function __searchByLocalPB(num, id, owner)
    if not num then
        return nil
    end

	local pbConn = require"core.wsqlite":new("phonebook")
    local sql    = "SELECT organization, email, username, user_class, ringtone, image_url FROM cti_contacts WHERE ( extension='%s' OR phone='%s' OR office='%s' OR mobile='%s' OR home='%s' OR home_mobile='%s' OR fax='%s' )"
		  sql    = string.format(sql, num, num, num, num, num, num, num)

    if "yes" ~= channel.GLOBAL_PHONEBOOK_SEARCH:get() then
        sql = sql .. " AND "
        if id and "all" ~= id then                      -- lookup in specified phone book
            sql = sql .. "phonebook_id IN (" .. id .. ")"
        else                                            -- lookup in all public phone books/local user phone books
            sql = sql .. "phonebook_id IN (SELECT rid FROM cti_phonebooks WHERE type='2'"
            if owner then
                sql = sql .. string.format(" OR owner_id='%s'", owner)
            end
            sql = sql .. ")"
        end
    end

    sql = sql .. " LIMIT 1"
    local cursor = pbConn:execute(sql)
    if "boolean" == type(cursor) or not cursor then
        pbConn:close()
        logger:warning("DB query failed. Look up is not performed.")
        return nil
    end

    local orgName, email, name, callType, ringTone, image_url = cursor:fetch()
    cursor:close()
    pbConn:close()

    -- fixup organization
    if "LDAP" == orgName then
        orgName = nil
    end

    return {
         name     = name
        ,email    = email
        ,callType = callType
        ,orgName  = orgName
        ,ringTone = ringTone
        ,image_url = image_url
    }
end

local function __setSearchResult(data)
    if "table" ~= type(data) then
        return nil
    end
    
    local rv = {}
    for k,v in pairs(data) do
        if v and "" ~= v then
            rv[k] = v
        end
    end

    return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local lookUp = {    -- meta class
      number        = nil      -- required normilized nuber
     ,callDirection = nil      -- call direction
     ,channelName   = nil      -- channel that execute look up
     ,rSource       = nil      -- source where we get contact data
     ,rData         = nil      -- result contact data
--        name                  -- contact name (mandatory field)
--        email                 -- contact email
--        orgName               -- contact organization
--        callType              -- call type tag
--        ringTone              -- contact ring tone value
--        altNumber             -- contact alternative number
}

--- Set up channel name as owner for this procedure
function lookUp:setChannel(name)
    if not name or "" == name then
        return false
    end
    
    self.channelName = name
    return true
end

--- Set up call direction with validation
function lookUp:setDirection(val)
    if "inc" ~= val and "out" ~= val then
        return false
    end
    
    self.callDirection = val
    return true
end

--- Retrieve result source value
function lookUp:getResultSource()
    return self.rSource
end

--- Retrieve looked up contact name
function lookUp:getContactName()
    return self.rData and self.rData["name"] or ""
end

--- Retrieve looked up contact extended fields
function lookUp:getContactByField(name)
    return self.rData and self.rData[name]
end

--- Executing look up by internal mobile user
function lookUp:searchByLdap()
    local rv = nil
    if self.rData then
        return rv
    end
    
    logger:debug("Executing LDAP contact lookup for number", self.number)
    rv = __searchByLdap(self.number)
    rv = __setSearchResult(rv)
    if rv and rv.name then
        self.rData   = rv
        self.rSource = "LDAP"
    end

    return rv
end

--- Executing look up at local collaboration book
function lookUp:searchByLocalPB(id, owner)
    local rv = nil
    if self.rData then
        return rv
    end

    id = "" ~= id and id
    logger:debug("Executing phonebook contact lookup for number '%s'", self.number)
    rv = __searchByLocalPB(self.number, id, "out" == self.callDirection and owner)
    rv = __setSearchResult(rv)
    if rv and rv.name then
        self.rData = rv
        self.rSource = "PBXPB"
    end
    
    return rv
end

--- Executing look up in the external service (async request)
function lookUp:searchByExternalService(credData)
    local rv = nil
    if self.rData then
        logger:debug("External lookup. Data processed by other method")
        return rv
    end
    
    if channel then -- request from dialplan
        repeat
            local numberObj = require"common.numberDetect.numberDetectImpl":new()
            if not numberObj:isNormalized(self.number) then
                logger:debug("External lookup. Input number not normilized: ", self.number)
                break
            end
            
            -- checking configuration variable
            local isServiceEnabled = channel.EXTCONTACTLOOKUP:get()
            if "no" == isServiceEnabled then
                logger:debug("External contact lookup disabled by configuration variable")
                break
            end
            
            -- check PBX licensing for this service (CDRVIEW + PERUSER/Subscription = PREMIUM)
            local perUserLincese = channel.LICENSE("LICENSE_PERUSER"):get()
            if "Enable" ~= perUserLincese then
                logger:debug("External lookup. The license 'PERUSER' disabled for PBX")
                break
            end
            local perUserLicenseSubscription = channel.LICENSE_SUBSCRIPTION("LICENSE_PERUSER"):get()
            if "Expired" == perUserLicenseSubscription then
                logger:warning("External lookup. The license 'PERUSER' subscription is expired")
                break
            end
            if "Lifetime" == perUserLicenseSubscription then
                logger:debug("External lookup. The license 'PERUSER' is lifetime")
                break
            end
            local CDRViewLicense = channel.LICENSE("LICENSE_CDRVIEW"):get()
            if "Enable" ~= CDRViewLicense then
                logger:debug("External lookup. The license 'CDRVIEW' disabled for PBX")
                break
            end
            
            -- executing lookup
            local rData = nil
            local cache = gSpace["contactLookUp"] or {} 
            if cache[self.number] then
                logger:debug("Retrieving external contact lookup result for number '%s' from channel cache", self.number)
                rData = cache[self.number]
            else
                -- execute sync request
                logger:debug("Executing external contact lookup for number:", self.number)
                local err  = nil  
                rData, err = require"externalContactLookUp".exec(self.number, {
                    timeout = channel.EXTCONTACTLOOKUP_TIMEOUT:get()
                })
                if err then
                    logger:warning("Contact external lookup error:", err)
                    break
                end
                
                cache[self.number]      = rData
                gSpace["contactLookUp"] = cache
                
                if rData then
                    logger:debug(rData, "Contact lookup data for number:", self.number)
                    if "unknown" == rData.name then
                        logger:debug("Not found any data for number", self.number)
                        rData = nil
                    end
                end
            end
            
            rv = __setSearchResult(rData)
            if rv and rv.name then
                self.rData   = rv
                self.rSource = "ExtService"
            end
        until true
    end
    
    return rv
end

--- Executing look up in the WIM (integration middleware) service
function lookUp:searchByIntegrationMiddleware()
    local rv = nil
    if self.rData then
        logger:debug("WIM contact lookup. Data processed by other method")
        return
    end
    
    -- request is not from dialplan
    if not channel then
        return
    end

    if tostring(require"common.permissions":getPbxLicense("xbees")) ~= "1"  then
        logger:debug("Skip WIM contact lookup. The license type is not x-bees")
        return
    end

    local numberObj = require"common.numberDetect.numberDetectImpl":new()
    if not numberObj:isNormalized(self.number) then
        logger:debug("WIM contact lookup. Input number not normilized: ", self.number)
        return
    end

    -- executing lookup
    repeat
        local rData = nil
        local cache = gSpace["contactLookUp"] or {}
        if cache[self.number] then
            logger:debug("Retrieving WIM contact lookup result for number '%s' from channel cache", self.number)
            rData = cache[self.number]
        else
            -- execute sync request
            logger:debug("Executing WIM contact lookup for number:", self.number)
            local err  = nil

            rData, err = require"externalContactLookUp".exec(self.number, {
                service = "wim",
                timeout = channel.EXTCONTACTLOOKUP_TIMEOUT:get()
            })
            if err then
                logger:warning("WIM contact lookup error:", err)
                break
            end

            if rData then
                logger:debug(rData, "Contact lookup data for number:", self.number)
                if "unknown" == rData.name then
                    logger:debug("Not found any data for number", self.number)
                    rData = nil
                end
            end

            cache[self.number]      = rData
            gSpace["contactLookUp"] = cache
        end

        rv = __setSearchResult(rData)
        if rv and rv.name then
            self.rData   = rv
            self.rSource = "WIMService"
        end
    until true

    return rv
end

--- Processing result data after successful search
function lookUp:postProcess()
    local rv = false

    if channel then -- dialplan thread

        -- processing searched number
        require"services".set_call_id("num", self.number, self.callDirection)
        if "inc" == self.callDirection then
            channel.CDR("c_from"):set(self.rData and self.rData.altNumber or self.number)
        else
            channel.CDR("c_to"):set(self.number)
        end

        -- processing contact name
        local contName = self.rData and self.rData.name
        if contName then
            print("Try save %s: %s", "inc" == self.callDirection and "from_name" or "to_name", contName)
        end
        local incomingCallerName = "inc" == self.callDirection and gSpace["gvars"].origCidName
        local contName           = contName or incomingCallerName or ""
              contName           = "" == contName and self.number or contName

        -- processing organization data
        local orgName = self.rData and (self.rData.orgName or self.rData.company)
        if orgName then
            print("Try save organization:", orgName)
        end
        orgName = orgName or ""
        channel.CDR("organization"):set(orgName)

        app.CELGenUserEvent("LookUp", string.format("%s,%s,%s,%s,%s", contName or "", orgName, self.number, (self.rData and self.rData.email) or "", self.callDirection))

        -- processing call type
        local callType = self.rData and self.rData.callType
        if callType then
            callType = callType:lower()
			channel["_CALLTYPE"]:set(callType)
			gSpace["gvars"].calltype = callType
            print("Set call type:", callType)
        end
        
        -- processing caller id data by some smart logic
        local callerInfo = contName
        local withoutDisplayInfoProcessing = "no" == channel.DISPLAYINFOUSELOOKUP:get()
        if not withoutDisplayInfoProcessing and "inc" == self.callDirection then
            local extraDisplayInfo = channel.EXTRADISPLAYINFO:get() or ""
            if "" ~= extraDisplayInfo then
                channel["extra"]:set(extraDisplayInfo)
            end
            -- use channel data by formatting pattern
            channel["name"]:set(contName)
            channel["company"]:set(orgName)
            
            local displayInfoPattern = channel.DISPLAYINFOPATTERN:get() or ""
            if "" == displayInfoPattern then
                displayInfoPattern = "" ~= extraDisplayInfo and "{name} {extra}" or "{name} ({company})"
            end
            -- normalize pattern by variables values
            if "" == contName then
                displayInfoPattern = displayInfoPattern:gsub("([^%s]*{name}[^%s]*)", "")
            end
            if "" == orgName then
                displayInfoPattern = displayInfoPattern:gsub("([^%s]*{company}[^%s]*)", "")
            end
            if "" == extraDisplayInfo then
                displayInfoPattern = displayInfoPattern:gsub("([^%s]*{extra}[^%s]*)", "")
            end
            displayInfoPattern = displayInfoPattern:gsub("^%s+", ""):gsub("%s+$", "")
            displayInfoPattern = displayInfoPattern:gsub("{", "${")
            callerInfo         = channel.EVAL(displayInfoPattern):get()
        end
        require"services".set_call_id("name", callerInfo, self.callDirection)
        channel.CDR("inc" == self.callDirection and "from_name" or "to_name"):set(contName)

        -- processing avatar
        if "inc" == self.callDirection and self.rData and self.rData.image_url and "" ~= self.rData.image_url and string.find(self.rData.image_url, "http") then
            local ci_header = string.format("Caller-Image: %s", self.rData.image_url)
            app.SipAddHeader(ci_header)
            logger:info("The following sip-header was added during contact lookup: '%s'", ci_header)
        end
    end
    
    return rv
end


--- Object constructor
function contactLookUp:new(num)
    local obj = nil
    if not num or "" == num then
        return obj, "Invalid input arguments"
    end
    
    -- create object
    obj = setmetatable({}, {__index = lookUp})
    
    -- initialize object
    obj.number = num
    obj:setChannel(channel and channel.CHANNEL("NAME"):get())

    return obj
end

return contactLookUp
