--- Service for processing personal/pbx recording/playback sounds
-- 
-- @file    pbxSound.lua 
-- @author  Vyacheslav Khudan
-- @created 2015/07/29

--- @module pbxSound
local pbxSound = {}

--- Method used for dialing to user who makes recording
-- @param ctx       The destination user cotext
-- @param exten     The destination user extension
function pbxSound.dialToUser(ctx, exten)
    if not exten or not ctx then
        return false
    end
    
    -- retrieve extension data
    local extAttrs = get_obj_attrs(exten)
    if not extAttrs or not extAttrs.pbxowner then
        return false
    end
    
    -- set up dial options
    local dialOpts = nil
    local actPBX   = wmsldap:get("pbxSerial", require"core.pbxutils".pbxSerial(), "wPbx")[1]
    if actPBX and actPBX[1] == extAttrs.pbxowner then
        -- user located on local PBX
    else
        -- user located on remote PBX
        dialOpts = require"wmsNetCall".makeArgs({
             pbxid = require"core.strConverter".toB64(extAttrs.pbxowner)
            ,ctx   = ctx
            ,linkedid     = channel.LINKEDID:get() or ""
        })
    end
    
    return smartApp("Dial", "SIP/"..exten, nil, dialOpts)
end

--- Processing record application via IVR requests
-- @param messagecode   
-- @param choice
-- @param rtype
function pbxSound.runRecord(messagecode, choice, rtype)
    local cid       = gSpace["cidNum"]
    local cid_attrs = gSpace["cidAttrs"]
    local fsutils = require"core.fsutils"

    -- check if call from local user		
    if cid_attrs.telephoneNumber == nil then
        print("Cannot define caller")
        return smartApp("Hangup", hcause.CHANUNAVAIL)
    end
    
    -- check permissions
    local is_allowed = require"common.permissions":check(cid_attrs["owner"] or "", "canuse", rtype == "pbx" and "PBXRECORD" or "PERSONALRECORD", nil, cid_attrs["licenseType"]) == "yes"
    if not is_allowed then
        ivr:tell("Destination of call is not allowed")
        return smartApp("Hangup", hcause.ACL_FORBIDDEN)
    end
    
     -- check record flag from WMS (channel var RECORDOWNER)
    local is_wms_record = channel.RECORDOWNER:get() or ""	
    is_wms_record 		= "" ~= is_wms_record and true or false
    if is_wms_record then
        -- wait for channel mascarading 
        app.Wait(1)
    end

    local is_message_exists = nil
    local persondir         = nil
    local pbxfile           = nil
    if "personal" == rtype then
        local diskutils  = require"core.diskutils"
        local soundspath = "/var/spool/callweaver/monitor"
        if not diskutils.is_mount(soundspath) then
            local msg = "No device is mounted for call recoring"
            print(msg); logger:error(msg)

            if not cid_attrs.mail then
                print("No mail address found")
                return smartApp("Hangup", hcause.CHANUNAVAIL)
            end
            if 10240 > diskutils.get_avail_ram_space() then
                return smartApp("Hangup", hcause.CHANUNAVAIL)
            end
            soundspath = "/tmp/session"
            cid        = ""
        end
        persondir = soundspath .. (cid ~= "" and "/" .. cid or "")

        if not messagecode then
            local options = {
                 retry    = 1
                ,timeout  = 4
                ,validate = function(...) return true; end;
            }
            messagecode = ivr:read("Please enter the message id or wait for " .. options.timeout .. " seconds to start the recording", options)
            if "" == messagecode then -- generate id
                local range = 1000
                for i=1,range do
                    messagecode       = math.random (range)
                    is_message_exists = fsutils.is_readable(persondir .. "/"..  messagecode, {"g729","ulaw","gsm","alaw"})
                    if not is_message_exists then
                        break
                    end
                end
                if is_message_exists then
                    ivr:tell("Reached limit ids. Remove unnecessary messages. Good bye. Hangup")
                    return smartApp("Hangup", hcause.CHANUNAVAIL)
                end
		ivr:tell("the message id " .. messagecode)
                choice = "1"
            end
        end
    elseif "pbx" == rtype then
        if not choice and not messagecode then
            messagecode = ivr:read("Please enter the message id", {retry = 1})
            if not messagecode then
                ivr:tell("Good bye. Hangup")
                return smartApp("Hangup", hcause.NORMAL)
            end
        end
    end

    -- validate argument
    if not messagecode or (not string.find(messagecode,"^%d+$") and not is_wms_record) then
        ivr:tell("The value you entered is not valid. Hangup")
        return smartApp("Hangup", hcause.CHANUNAVAIL)
    end

    if rtype == "personal" then
        local messagefile = persondir .. "/"..  messagecode
        is_message_exists = fsutils.is_readable(messagefile, {"g729","ulaw","gsm","alaw","opus"})
        if is_message_exists then
            if not choice then
                smartApp("Read", "empty", pbxfile, 0, nil, 1, 0.1)
            else
                print("message file " .. messagefile .. " is found")
            end
        end
    elseif rtype == "pbx" then
        if is_wms_record then	-- set correct path and choice for reform from WMS
            pbxfile = "/var/lib/callweaver/sounds/" .. messagecode
            choice  = "1"
        else
            pbxfile = "/var/lib/callweaver/sounds/00000/rec"..  messagecode
        end
        is_message_exists = fsutils.is_readable(pbxfile, {"g729","ulaw","gsm","alaw","opus"})

        -- playback if exists
        if is_message_exists and not choice then
            smartApp("Read", "empty", pbxfile, 0, nil, 1, 0.1)
        elseif not choice then
            choice = "1"
        end
    end

    if not choice then
        local announce = {"Press"}
        if is_message_exists then
            announce[#announce+1] = "0 to delete the message, "
        end
        announce[#announce+1] = rtype == "pbx" and "1 to record the message stop the recording with pound" or "1 to record the message stop the recording with pound. You will receive the recording via email"
        if is_message_exists then
            announce[#announce+1] = "or press 2 to playback"
        end

        local options = {
            max_chars = 1,
            validate  = function(choice)
                local t = is_message_exists and {0,1,2} or {1, 2}
                return require"core.pbxutils".in_table(tonumber(choice), t)
            end;
            failure   = function()
		ivr:tell("The number you entered is incorrect, please try again")
            end;
        }
        choice = ivr:read(table.concat(announce, " "), options)
        if not choice then
            ivr:tell("Good bye. Hangup")
            return smartApp("Hangup", hcause.NORMAL)
        end
    end

    -- validate argument
    if not string.find(choice,"^[0-2]*$") then
	ivr:tell("The value you entered is not valid. Hangup")
        return smartApp("Hangup", hcause.NORMAL)
    end

    local messagefile = rtype == "pbx" and pbxfile or persondir .. "/" .. messagecode
    if choice == "0" then -- delete
        fsutils.rmfile("%s.*", messagefile)
    elseif choice == "1" then -- record
        local attach_fmt_rec
        if rtype == "personal" then
            attach_fmt_rec = channel["GATTACH_FORMAT_REC"]:get() or ""
            if attach_fmt_rec == "" then
                attach_fmt_rec = "mp3"
            end
            fsutils.mkdirr(persondir)
        end

        local format         = "alaw"
        local channel_format = string.match(channel.CHANNEL("audionativeformat"):get(),"^%((%w+)%)$")
        if channel_format == "g729" then
            format = "g729"
        end
        setpostaction {
            ["name"]           = "RECORD",
            ["targetfile"]     = messagefile,
            ["origformat"]     = format,
            ["messagecode"]    = messagecode,
            ["cid_attrs_mail"] = rtype == "pbx" and "empty" or (cid_attrs.mail or "empty"),
            ["language"]       = rtype == "pbx" and "" or (cid_attrs.preferredLanguage or ""),
            ["attach_format"]  = rtype ~= "pbx" and attach_fmt_rec or ""
        }
        channel.CHANNEL("hangup_handler_wipe"):set("hdlr,service,1")
        if not is_wms_record then
            ivr:tell("Record your message after the tone ...")
        end
        smartApp("Record", messagefile .. "." .. format, nil, nil, "k")
        -- in case personal rtype transcode and convert moved to /usr/sbin/wms_notify (function send_record_simple), that is called from hangup handler

    elseif choice == "2" and is_message_exists then -- playback
        ivr:tell("Playback")
        smartApp("Read", "empty", messagefile, 0, nil, 1, 0.1)
    end
    if not is_wms_record then
        ivr:tell("Thank you")
    end
    
    return true
end

return pbxSound
