--- Extensions loader from WMS modified files
-- Uses file lock feature to make safe file reading
--
-- @file    dialplanLoader.lua
-- @author  Vyacheslav Khudan
-- @created 2017/02/27

--- @module dialplanLoader
local dialplanLoader = {
    mDialplanFilesPath = "/etc/callweaver/"
    ,mLockFile          = "/tmp/extensions.lock"
    ,mLockHandler       = nil
    ,mIsLocked          = nil
}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Set up shared lock for loader
-- @return boolean value with error message in case error
function dialplanLoader:tryLock()
    local waitTime  = 2 * 1000   -- ms
    local rv        = self.mIsLocked
    local err       = nil
    repeat
        if self.mIsLocked then
            break
        end

        -- processing lock file
        self.mLockHandler, err = io.open(self.mLockFile, "a+")
        if err then
            break
        end

        -- retrieve file descriptor
        local lockDescriptor = nil
        lockDescriptor, err  = require"posix.stdio".fileno(self.mLockHandler)
        if err then
            break
        end

        -- set up shared lock
        local socket = require"socket"
        local pf     = require"posix.flock"
        local bit    = require"bit"
        while not rv and 0 < waitTime do
            rv, err = pf.flock(lockDescriptor, bit.bor(pf.LOCK_SH, pf.LOCK_NB))
            if err and err:find("Resource temporarily unavailable") then
                local sleepTime = 200   -- ms
                socket.sleep(sleepTime/1000)
                waitTime = waitTime - sleepTime
            else
                break
            end
        end
        
        self.mIsLocked = rv
    until true

    if not self.mIsLocked then
        if self.mLockHandler then
            self.mLockHandler:close()
            self.mLockHandler = nil
        end
        err = err or "Unknown error"
    end

    return rv, err
end

--- Disable shared lock for loader
-- @return boolean value with error message in case error
function dialplanLoader:unlock()
    local rv  = nil
    local err = nil
    repeat
        if not self.mIsLocked then
            rv = true
            break
        end

        -- processing lock file
        if not self.mLockHandler then
            err = "Lock handler not found"
            break
        end

        -- retrieve file descriptor
        local lockDescriptor = nil
        lockDescriptor, err  = require"posix.stdio".fileno(self.mLockHandler)
        if err then
            break
        end

        -- remove an existing lock
        local pf = require"posix.flock"
        rv, err = pf.flock(lockDescriptor, pf.LOCK_UN)

        self.mIsLocked = not rv
        self.mLockHandler:close()
        self.mLockHandler = nil
    until true

    if not rv then
        err = err or "Unknown error"
    end

    return rv, err
end

--- Parsing and loading dialplan filename
-- @param fileName  The destination file for loading
-- @return loaded chunk otherwise error message
function dialplanLoader:run(fileName)
    local rv = nil
    if not fileName or "" == fileName then
        return rv, "Invalid filename for loading"
    end

    if not self.mIsLocked then
        return rv, "Can't found shared lock for loader"
    end
    
    local dialplanFileName = self.mDialplanFilesPath .. fileName
    dialplanFileName = fileName:find("%.lua$") and dialplanFileName or dialplanFileName .. ".lua"
    logger:debug("Parsing and loading dialplan file: %s", dialplanFileName)
    return loadfile(dialplanFileName)
end

--- Proceeding loading dialplan file with auto lock/unlock and executing chunk
-- @param fileName  The dialplan file for loading and executing
-- return boolean value
function dialplanLoader:proceed(fileName)
    if not fileName then
        return false
    end

    local chunk = nil
    repeat
        local status, err = self:tryLock()
        if not status then
            logger:error("Dialplan loader lock error: %s", err)
            break
        end

        chunk, err = self:run(fileName)
        if not chunk then
            logger:error("Dialplan loader content error: %s", err)
            break
        end

        status, err = self:unlock()
        if not status then
            logger:error("Dialplan loader unlock error: %s", err)
            chunk = nil
            break
        end
    until true

    if "function" == type(chunk) then -- loading content chunk
        chunk()
    end

    return true
end

return dialplanLoader
