--- Logger for dialplan application execution
--
-- @file    dialplanLogger.lua
-- @author  Vyacheslav Khudan
-- @created 2017/02/27
-- @note Can't be used without 'channel' structure!!!
-- @note Used only to mirror callweaver dialplan logs via syslog/file

--- @module dialplanLogger
local dialplanLogger = {}

----------------------------------------------------------------
-- Syslog stream 
----------------------------------------------------------------
local syslogStream = {  -- meta class
    mMsgData = nil
}

function syslogStream:init(data)
    local rv = false
    if not data then 
        return rv
    end
    
    if "" == data.chan then
        return rv
    end
    if "" == data.exten or "" == data.ctx or "" == data.prio then
        return rv
    end
    
     --- initialize internal values
    self.mMsgData = {
         chan       = data.chan
        ,exten      = data.exten
        ,ctx        = data.ctx
        ,prio       = data.prio
    }
    
    rv = true
    return rv
end

function syslogStream:__buildMsg(appName, ...)
    local rv = nil
    if not appName then
        return rv
    end
    if not self.mMsgData then
        return rv
    end
    
    local appArgs = {}
    for i =1, #arg do
        table.insert(appArgs, arg[i] or "")
    end
    local prefix = "[" .. self.mMsgData.exten .. "@" .. self.mMsgData.ctx .. ":" .. self.mMsgData.prio .. "]"
    local rv     = string.format("-- Executing %s %s(\"%s\", \"%s\")", prefix, appName, self.mMsgData.chan, table.concat(appArgs, ","))
    
    return rv
end

function syslogStream:write(appName, ...)
    local rv  = false
    
    -- make message text
    local msg = self:__buildMsg(appName, ...)
    if not msg then
        return rv
    end
    
    -- write message
    logger:debug(msg)
end

----------------------------------------------------------------
-- File stream
----------------------------------------------------------------
local fileStream = {    -- meta class
     mPath    = "/tmp/calls/"
    ,mName    = nil
    ,mMsgData = nil
}

function fileStream:init(data)
    local rv = false
    if not data then 
        return rv
    end
    if "" == data.path or "" == data.chan then
        return rv
    end
    if "" == data.exten or "" == data.ctx or "" == data.prio then
        return rv
    end
    
    --- initialize internal values
    self.mMsgData = {
         chan       = data.chan
        ,exten      = data.exten
        ,ctx        = data.ctx
        ,prio       = data.prio
    }
    
    -- processing logs filesystem path
    local dirpath, filename = data.path:match("^(.+%/)([^/]+)")
    if not dirpath or not filename then
        return rv, "invalid log path"
    end
    self.mPath = dirpath
    self.mName = filename
    
    -- check file stream directory
    local fsh, err = io.open(self.mPath .. self.mName, "w")
    if not fsh then
        require"core.fsutils".mkdirr(self.mPath)
        fsh, err = io.open(self.mPath .. self.mName, "w")
    end
    if fsh then
        fsh:close()
        rv = true
    end
    
    return rv, err
end

function fileStream:__buildMsg(appName, ...)
    local rv = nil
    if not appName then
        return rv
    end
    if not self.mMsgData then
        return rv
    end
    
    local appArgs = {}
    for i = 1, #arg do
        table.insert(appArgs, arg[i] or "")
    end
    local prefix = "[" .. self.mMsgData.exten .. "@" .. self.mMsgData.ctx .. ":" .. self.mMsgData.prio .. "]"
    local rv     = string.format("-- Executing %s %s(\"%s\", \"%s\")", prefix, appName, self.mMsgData.chan, table.concat(appArgs, ","))
    
    return rv
end

function fileStream:write(appName, ...)
    local rv  = false
    
    -- make message text
    local msg = self:__buildMsg(appName, ...)
    if not msg then
        return rv
    end
    
    -- write data to stream
    local fsh = io.open(self.mPath .. self.mName, "a")
    if fsh then
        fsh:write(msg .. "\n")
        fsh:close()
        rv = true
    end
    
    return rv
end

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local function __backendBuilder(name, env)
    local mt = nil
    if "syslog" == name then
        mt = syslogStream
    elseif "file" == name or "core" == name then
        mt = fileStream
    end
    
    local backend = nil
    local err     = nil
    if mt then
        local stat = false
        backend    = setmetatable({}, {__index = mt})
        stat, err  = backend:init(env)
        if not stat then
            backend = nil
        end
    end
    
    return backend, err
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local dlogger = {    -- meta class
     mMode      = "core"   -- mode for dialplan logger
    ,mBackend   = nil      -- backend object
}   

--- Set up mode for logger
-- @param val   The mode value
-- @return value of the configured mode 
function dlogger:__setMode(val)
    if not val or "" == val then
        return true
    end
    
    local rv = nil
    if "file" == val then
        rv = "file"
    elseif "syslog" == val then 
        rv = "syslog"
    end
    
    if rv then
        self.mMode = rv
    end
    
    return rv
end

--- Initialize logger structure
-- @note allowed to return FALSE without error (for disabled logger)
function dlogger:init()
    local rv  = false
    local err = nil
    
    -- check initialized object
    if self.mBackend and self.mMode then
        rv = true
        return rv
    end
    
    -- check channel structure
    if not channel then
        err = "Not found channel structure"
        return rv, err
    end
    
    -- mode initialization
    local modeVal = channel.DIALPLANLOGGER_MODE:get() or ""
    if not self:__setMode(modeVal) then
        return rv   -- return witout error message
    end
    
    -- initialize channel variables
    local chanName = channel.CHANNEL:get() or ""
    local exten    = channel.EXTEN:get() or ""
    local context  = channel.CONTEXT:get() or ""
    local priority = channel.PRIORITY:get() or ""
    local logpath  = channel.CALLLOG:get() or ""
    if "file" == self.mMode then    -- fixup logpath for "file" mode
        logpath = "/tmp/callweaver_call_" .. require"socket".gettime() .. ".log"
    end
    
    -- initialize backend
    self.mBackend, err  = __backendBuilder(self.mMode, {
                                            chan  = chanName
                                           ,exten = exten
                                           ,ctx   = context
                                           ,prio  = priority
                                           ,path  = logpath
                                       })
    if self.mBackend then
        rv = true
    end
    
    return rv, err
end

--- Write message to backend stream
-- @param appName   An executed application name
-- @param mixed     An additional application arguments 
function dlogger:write(appName, ...)
    local status, err = self:init()
    if not status then
        if err then
            logger:warning("Dialplan logger error: %s", err)
        end
        return false
    end
    
    return self.mBackend:write(appName, ...)
end


--- Object constructor
function dialplanLogger:new()
    local obj = setmetatable({}, {__index = dlogger})
    
    return obj
end

return dialplanLogger
