module("utils", package.seeall)

--- Return unique table
function table_unique (self, t) -- only flat table
	local tb = {}
	for i=1,#t do tb[t[i]] = true end
	t = {}
	for v in pairs(tb) do t[#t+1] = v end
	return t
end

--- Split string with patterns
-- TODO: PRESENT at pbxutils.lua, but need merge with local function
function split (self, str, pat, limit)
	local t = {}
	local fpat = "(.-)" .. pat
	local last_end = 1
	local s, e, cap = str:find(fpat, 1)
	while s and (type(limit) ~= "number" or (type(limit) == "number" and #t < limit)) do
		if s ~= 1 or cap ~= "" then
			table.insert(t,cap)
		end
		last_end = e+1
		s, e, cap = str:find(fpat, last_end)
	end
	if last_end <= #str and (type(limit) ~= "number" or (type(limit) == "number" and #t < limit)) then
		cap = str:sub(last_end)
		table.insert(t, cap)
	end
	return t
end

--- Send a sip notify message to a peer
-- @param cid A string with a peer entry
-- @param timeout A timeout to wait before send notify message
-- @param removecommand Do remove registration from userloc (kamailio)
function notify_peer (self, cid, timeout, removecommand)
	if cid and cid ~= "" then
   		local timeout = timeout or 0
   		local notify_command = "check-sync"
   		
	   	if timeout ~= 0 or removecommand then
	   		local command = [[ 
	   			exec 2>/dev/null; nohup sleep %s && \
	   			lua -e 'require"core.astman".send_command("sip notify %s %s", true)' >/dev/null \
		  		]]
		  	local exec_cmd = string.format(command, timeout, notify_command, cid)
		  	if removecommand and removecommand ~= "" then
		  		exec_cmd = exec_cmd .. string.format(" %s > /dev/null", removecommand)
		  	end
		  	exec_cmd = exec_cmd .. " &"
		  	
		  	return os.execute(exec_cmd)
	   	else
	   		return require"core.astman".send_command(string.format("sip notify %s %s", notify_command, cid))
	   	end
	end
	return false
end

--- Register a device as an unknown account
-- @param mac A string with mac address
-- #param account A string with account
function register (self, mac, account)
	if self:is_unknown (mac) and not account then
		return false
	end
	if not account then
		account = self:get_free_unknown_account ()
	end
    local devconnection = require"core.wsqlite":new("devdb")
	local sql = "insert or replace into autoprov (id,param,value,mac) values((select id from autoprov where param=%q and mac=%q),%q,'%s',%q)"
	devconnection:execute(sql, "username", mac, "username", account, mac)
	devconnection:close()
	return account
end

--- Register FXS, with account and portnumber
-- @param mac A string with mac address
-- @param account A string with account
-- @param portnumber An FXS port number
function registerFXS (self, mac, account, portnumber)
    if not account or not portnumber then
        do return false end
    end
    local devconnection = require"core.wsqlite":new("devdb")
    local sql           = nil
    if account:find("unknown") then
    	sql = [[
	    	select value from autoprov where mac=%q and param='username' and 
	    		value like 'unknown%%' limit 1
	    ]]
	    local cur = devconnection:execute (sql, mac)
	    local res = cur:fetch(); cur:close()
	    account = res and res or account
	end

    sql = [[
		 insert or replace into autoprov (id,param,value,mac,flag) values(
			 (select id from autoprov where param=%q and mac=%q and flag=%q),%q,'%s',%q, %q)
	]]
    devconnection:execute(sql, "username", mac, portnumber,"username", account, mac, portnumber)
    devconnection:close()
    return account
end

--- Return list of unknown users
-- @return A table username = mac
-- @return A table mac = username
function get_unknown_users (self)
	local t, r = {}, {}
    local devconnection = require"core.wsqlite":new("devdb")
	local sql = "select value, mac from autoprov where param=%q and value like %q"
	for username, mac in devconnection:rows (sql, "username", "unknown%%") do
		t[username], r[mac] = mac, username
	end
	devconnection:close()
	return t, r
end

--- Get a new free unknown account name
-- @return A string with an unknown account
function get_free_unknown_account (self)
	local users = self:get_unknown_users()
	local account = "unknown1"
	if users then
		for i=1,10000 do
			account = "unknown" .. i
			if not users[account] then break end
		end
	end
	return account
end

--- Compare @mac with unknow accounts in provisioning db
function is_unknown (self, mac)
	local devconnection = require"core.wsqlite":new("devdb")
	local sql = "select value from autoprov where mac=%q and param=%q and value like %q"
	local cur = devconnection:execute (sql, mac, "username", "unknown%%")
	local res = cur:fetch(); cur:close()
	devconnection:close()
	return res ~= nil
end

--- Cheking sounds pack for @lang language
-- @param #string lang The requested language
-- @param #string dir_name Additional part to main directory
-- @return #boolean 
function is_lang_pack_present(self, lang, dir_name)
	local rv = false
	if not lang or lang == "" then
		return rv
	end

	local connection = require"core.wsqlite":new("wmsdb")
	local sql        = "SELECT value FROM system WHERE variable = 'INSTALLED_LANGUAGES_SOUNDS';"
	local cursor     = connection:execute (sql)

	if cursor and type(cursor) ~= "boolean" then
		local installed_langs = cursor:fetch(); cursor:close()
		connection:close()
		local soundlang = string.gsub(lang, "%-", "%%%-")
		if installed_langs and "" ~= installed_langs then
			if string.find(installed_langs, string.format([["%s"]], soundlang)) then
				rv = true
			else
				print(string.format("Language '%s' is not installed on pbx", lang))
				logger:warning(string.format("Language '%s' is not installed on pbx", lang))
			end
		else
			print("An error occurred while verifying the wmsdb variable 'INSTALLED_LANGUAGES_SOUNDS'")
			logger:error("An error occurred while verifying the wmsdb variable 'INSTALLED_LANGUAGES_SOUNDS'")
		end
	else
		connection:close()
		rv = true
	end

	return rv
end

--- Getting the language for using in dial plan in accordance to existed language pack
-- @param #string lang The suggested language for using
-- @return #string Resulted language for current using
function get_lang_from_pack(self, lang)
	if lang == "it" then
		return "it"
	elseif lang == "de" then
		if is_lang_pack_present(self, "de") then
			return "de"
		else
			return "it"
		end
	elseif lang == "fr" and is_lang_pack_present(self, "fr") then
		return "fr"
	elseif lang == "ru" and is_lang_pack_present(self, "ru") then
		return "ru"
	elseif lang == "nl" and is_lang_pack_present(self, "nl") then
		return "nl"
	elseif lang == "se" and is_lang_pack_present(self, "se") then
		return "se"
	elseif lang == "es" and is_lang_pack_present(self, "es") then
		return "es"
	elseif lang == "es-us" and is_lang_pack_present(self, "es-us") then
		return "es-us"
	elseif lang == "en-us" and is_lang_pack_present(self, "en-us") then
		return "en-us"
	else
		return "en"
	end
end

--- Get silence sound file path
-- @param #number - times to repeat the silence
-- @return #string A silence sound file path
function GetSilence(iterate)
	iterate = iterate or 1
	local silencePath = ""
	for i = 1,iterate do
		silencePath = silencePath .. "tvoice/silence&"
	end
	return silencePath or false
end

--- Add options via channel variable
-- @param #options - existing options
-- @return #string extended options
function extend_dial_options(options)
	local extra_options = channel.DIAL_OPTIONS:get() or ""
	return extra_options .. options
end
