--- Module for wair alarms support
--
-- @file    wair_messaging.lua
-- @author  Yuriy Zveryanskyy
-- @created 2020/11/16

local wair_messaging = {}

local logger = require"core.logger":init("[WMS 3.0 Dialplan]", 3)

local rtx_max = 4096
local default_port = 9300
local template_file = "/var/lib/callweaver/dialplan/alarm.template"
local spool_dir = "/var/spool/callweaver/outgoing/"

-- <datetime>2017-06-20T12:55:55</datetime>
local reply_template = [[<?xml version="1.0" encoding="UTF-8"?>
<response version="1.0.0" type="{{msgtype}}">
<externalid>{{externalid}}</externalid>
<status>1</status>
<statusinfo>Accepted by Wildix PBX</statusinfo>
<systemdata>
<name>{{pbxname}}</name>
<datetime>{{datetime}}</datetime>
<timestamp>{{timestamp}}</timestamp>
<status>1</status>
<statusinfo>System running</statusinfo>
</systemdata>
</response>
]]

local socket = require"socket"
local parser = require"core.cxml"

local messaging = {
     port = nil,
     socket = nil,
     host = nil
}

local function __hostname()
    local host = ""
    local fp = io.open("/etc/hostname")
    if fp then
        host = fp:read("*a")
        fp:close()
    end
    return host
end

local function __datetime()
    local timestamp = os.time()
    local hexstamp = string.format("%x", timestamp)
    local date = os.date("*t", timestamp)
    local datestr = string.format("%d-%02d-%02dT%02d:%02d:%02d", date.year, date.month, date.day, date.hour, date.min, date.sec)
    return hexstamp, datestr
end

-- mostly is copypaste from wair module
local function __check_ip_devdb(ip)
    local connection = require"common.core".connection
    local sql = [[ SELECT mac FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"w-air%%"%%' ]]
    local cursor = connection:execute(sql, "device", ip)
    local mac = cursor:fetch()
    cursor:close()
    connection:close()
    return mac
end

local function __check_ip_localnet(ip)
    local h = require"common.helper"
    return h.isLocalIP(ip) and h.is_local_route(ip)
end

local function __check_ip(ip)
    if __check_ip_devdb(ip) then
        logger:debug("Source %s allowed by devdb", ip)
        return true
    end

    if __check_ip_localnet(ip) then
        logger:debug("Source %s allowed by local network", ip)
        return true
    end

    return false
end

local function __extract(msg)

    local function __process_rssi(rdata)
        local res = {}
        for i=1, #rdata, 2 do
            res[#res + 1] = {["rfpi"] = rdata[i][1], ["rssi"] = rdata[i+1][1]}
        end
        return res
    end

    local function __process_sender(sdata)
        local ext, name
        for i=1, #sdata do
            if "address" == sdata[i]["label"] then
                ext = sdata[i][1]
            elseif "name" == sdata[i]["label"] then
                name = sdata[i][1]
            end
        end
        return ext, name
    end

    local data = msg[2]
    if "table" ~= type(data) then
        return false, "Invalid xml data"
    end

    logger:debug(data)

    if data["label"] ~= "request" then
        return false, "Invalid request"
    end

    local msgtype = data["xarg"] and data["xarg"]["type"]

    local externalid, alarmtype, rssidata, senderext, sendername
    for i=1, #data do
        local label = "table" == type(data[i]) and data[i]["label"]
        if label then
            if "externalid" == label then
                externalid = data[i][1]
            elseif "alarmdata" == label then
                alarmtype = data[i][1][1]
            elseif "rssidata" == label then
                rssidata = __process_rssi(data[i])
            elseif "senderdata" == label then
                senderext, sendername = __process_sender(data[i])
            end
        end
    end

    if not msgtype or not externalid then -- mandatory
        return false, "No mandatory data in request"
    end

    result = {
        ["type"]      = msgtype,
        ["id"]        = externalid,
        ["alarmtype"] = alarmtype,
        ["rssidata"]  = rssidata,
        ["extension"] = senderext,
        ["name"]      = sendername
    }

    return result
end

local function __make_call(extension, name, atype, rssidata)
    local fp = io.open(template_file)
    if not fp then
        logger:error("No template file, cannot process alarm")
        return
    end
    local template = fp:read("*a")
    fp:close()

    local replace = {
        ["EXTENSION"] = extension,
        ["TYPE"]      = atype,
        ["NAME"]      = name
    }
    for i, val in pairs(rssidata) do
        local num = tostring(i)
        local idu = val["rfpi"]:upper()
        replace["ID" .. num]      = idu
        replace["SHORTID" .. num] = string.sub(idu, -2)
        replace["LEVEL" .. num]   = tostring(val["rssi"])
    end
    local replaced = string.gsub(template, "{{(%w+)}}", replace)
    -- set reminder to empty
    replaced = string.gsub(replaced, "{{(%w+)}}", "")

    local callname = "/tmp/alarm-" .. string.sub(tostring(socket.gettime() * 10000), -8) .. ".call"
    local fpt = io.open(callname, "w")
    fpt:write(replaced)
    fpt:close()
    require"core.fsutils".mvfile(callname, spool_dir)
end

local function __process_alarm(params)
    if "alarm" ~= params["type"] then
        logger:debug("Ignore message with type %s", params["type"])
        return
    end
    logger:debug(params)
    local base = params["rssidata"] and params["rssidata"][1] and params["rssidata"][1]["rfpi"]
    if not base then
        logger:error("No rssi info")
        return
    end
    local extension = params["extension"]
    if not extension then
        logger:error("No extension")
        return
    end
    local alarmtype = params["alarmtype"]
    if not alarmtype then
        logger:error("No alarm type")
        return
    end
    local rstr = require"dkjson".encode(params["rssidata"])
    logger:info("Processing alarm from extension '%s' type '%s', rssidata: '%s'", extension, alarmtype, rstr)
    local name = params["name"] or extension
    __make_call(extension, name, alarmtype, params["rssidata"])
end

local function __prepare_reply(params, host)
    local timestamp, datetime = __datetime()
    local replacet = {
        ["msgtype"]    = params["type"],
        ["externalid"] = params["id"],
        ["pbxname"]    = "Wildix Host " .. host,
        ["datetime"]   = datetime,
        ["timestamp"]  = timestamp
    }
    return string.gsub(reply_template, "{{(%w+)}}", replacet)
end

function messaging:serve_messages()
    logger:info("WAIR messaging starting on port %d", self.port)
    while true do
        local data, ip, srcport = self.socket:receivefrom(rtx_max)
        if data then
            logger:info("Data received from %s", ip)
            logger:debug("Input data: %s", data)
            local allow = __check_ip(ip)
            if not allow then
                logger:warning("Skip data from ip %s", ip)
            else
                local ok, xmltable = pcall(parser.collect, data)
                if not ok or "table" ~= type(xmltable) then
                    logger:error("Invalid xml data")
                else
                    local res, params, perr = pcall(__extract, xmltable)
                    if not res then
                        logger:error("Data processing failed")
                    elseif not params then
                        logger:error("Data processing failed, error: %s", perr)
                    else
                        __process_alarm(params)
                        local reply = __prepare_reply(params, self.host)
                        local udpres, uerr = self.socket:sendto(reply, ip, srcport)
                        if not udpres then
                            logger:error("Reply sending failed, error: %s", uerr)
                        end
                    end
                end
            end
        elseif ip ~= "timeout" then
            logger:error("Network error: %s", ip)
        end
    end
end

function wair_messaging:new(p)
    local port = default_port
    if p then
        port = tonumber(p)
        if not port then
            logger:error("Invalid port parameter")
            return
        end
    end
    local udp, udperr = socket.udp()
    if not udp then
        logger:error("Can not create socket, error: %s", udperr)
        return
    end
    udp:settimeout(nil)
    local res, nerr = udp:setsockname('*', port)
    if not res then
        logger:error("setsockname error: %s", nerr)
        return
    end

    local obj = setmetatable({}, {__index = messaging})
    obj.port = port
    obj.socket = udp
    obj.host = __hostname()

    return obj
end

return wair_messaging
