module("services", package.seeall)

local hcause = hcause

--- Function for convert extension or arbitrary variable from wms dialplan (exp. ${EXTEN} or ${CALLERID(NUMBER)})
-- @param exten A extension for conversion
-- @return A converted extension
local function convert_exten (exten)

    local function __convert_exten (exten)
        local channel = channel
        if tostring(exten):find("${.+}") then
            return channel.EVAL(exten):get()
        else
            return exten
        end
    end

    local status, res = pcall(__convert_exten, exten)
    if status then
        return res
    end
    logger:warning("Error attempt to evaluate extention '%s': %s", exten, res)
    return exten
end

--- Search feature application by name
local function feature_in_table(app_name)
    return ({
         ["STATUS"]          = 1
        ,["BCF"]             = 1
        ,["UCF"]             = 1
        ,["FCF"]             = 1
        ,["CW"]              = 1
        ,["TOUT"]            = 1 
        ,["MOBCONFIRMATION"] = 1
        ,["MEXT"]            = 1
        ,["CALLBLOCK"]       = 1
        ,["ROAD"]            = 1
        ,["HIDECID"]         = 1
    })[app_name]
end

--- function for save trunk limits in case call to trunk/HG from fax 
-- and remove records which older than 5 seconds (used by transmitmail2fax)
local function save_trunk_limits (rule)
    if not rule then
        return false
    end
    
    local os, io = os, io
    local waitlimit = 5 -- same value used at transmitmail2fax
    local filename = "/tmp/.trunklimit"
    local limitations, frule, ftime, line = {}
    local cur_time = os.time()
    
    -- read limitations from file
    local file, err = io.open(filename, "r")
    if not file and err then
        print("Error read", err)
        logger:error("Error read", err)
        io.close()
    else
        for line in file:lines() do
            frule, ftime = line:match("%[(.*)%]%s*(%d+)")
            if frule and ftime and (cur_time - waitlimit) < tonumber(ftime) then
                limitations[frule] = ftime
            end
        end
        file:close()
    end
    
    limitations[rule] = cur_time
    -- write updated limitations to file
    file, err = io.open(filename, "w+")
    if not file and err then
        print("Error write to", err)
        logger:error("Error write to", err)
        io.close()
        return false
    end
    for frule, ftime in pairs(limitations) do
        line = "["..frule.."] "..ftime.."\n"
        file:write(line)
    end
    file:close()
end

-- Executing call block procedure
local function __callBlockExecuting()
    local userAuth = require"userAuthorize":new()
          userAuth:setPinPhrase("The telephone is currently blocked. To use this phone please enter the password, followed by pound")
    if userAuth:process(gSpace["cidNum"]) then
        gSpace["gvars"].callBlockLoginChecked = true
        return true
    end
    
    return smartApp("Hangup", hcause.REJECT) 
end

-- Function to return caller id of the current identity
local function __checkIdentity(cid_attrs)
    if "table" == type(cid_attrs) and cid_attrs.uid then
        local features = require"common.userFeatures.userFeatures":new(cid_attrs.uid)
        if features then
            features:load()
            local from_identity = channel["FROM_IDENTITY"]:get()
            from_identity = "" ~= from_identity and from_identity
            local custom_identity = from_identity or features:getValue("CURRENT_IDENTITY")
            if custom_identity and "" ~= custom_identity then
                if cid_attrs.telephoneNumber and cid_attrs.telephoneNumber ~= custom_identity then
                    local available_identities = features:getCustomIdentities() or {}
                    if not require"core.pbxutils".in_table(custom_identity, available_identities) then
                        print("Identities list of the user doesn't contain the requested identity:", custom_identity)
                        logger:warning("Identities list of the user doesn't contain the requested identity:", custom_identity)
                        smartApp("Hangup", hcause.NOT_FOUND)
                    end

                    print("Call with a changed identity: %s => %s", cid_attrs.telephoneNumber, custom_identity)
                    logger:info("Call with a changed identity: %s => %s", cid_attrs.telephoneNumber, custom_identity)
                    local identity_attrs = get_obj_attrs(custom_identity)
                    local identity_features
                    if identity_attrs.uid then
                        identity_features = require"common.userFeatures.userFeatures":new(identity_attrs.uid)
                        if identity_features then
                            identity_features:load()
                        end
                    end
                    if not identity_features then
                        print("Error loading features of the caller user %s", custom_identity)
                        logger:error("Error loading features of the caller user %s", custom_identity)
                        smartApp("Hangup", hcause.NOT_FOUND)
                    end
                    return custom_identity, identity_attrs.cn or custom_identity
                end
            end
        end
    end
end

-- Send copies of faxes to s3 storage
local function __send_to_s3(file, extension)
    local aws = require"common.aws"
    if not aws.enabled() or not extension then
        return
    end

    local s3copy = file .. ".s3copy"
    require"core.fsutils".cpfile(file, s3copy)
    os.execute(string.format("faxtos3 %s %s &", s3copy, extension))
end

--- Main function that call other dialplan application
-- @param application A application which will be executed
-- @param exten A called extension 
-- @param mixed Values for executed application
function start_call(self, application, exten, ...)
    local channel     = channel
    local app         = app
    local hcause      = hcause
    local cid         = ""
    local application = application or ""

    -- execute custom application
    if "app-custom" == application then
        return require"customApplication"(exten, ...)
    end

    -- handling call completion from pbx
    if application == "internalcalls" or application == "app-dial-internalcall" then
        local cc_exten = channel.CC_EXTEN:get()
        if cc_exten then
            smartApp("CallCompletionCancel", "")
            ivr:tell("We are connecting you to the number you requested")
            smartApp("Transfer", "SIP/" .. cc_exten)
            return smartApp("Hangup", hcause.NORMAL)
        end
    end
    
    -- Handling recording pause/unpause
    if "autocallrecpause" == application then
        return app_record_call(nil, "pause")
    end
    -- Handling personal record call
    if "autocallrec" == application or "autocallswitch" == application then
        cid = channel.CDR("c_from"):get() or ""
        if "QUEUE" == channel.APPLICATION:get() then
            cid = channel.CALLERID("num"):get()
        end
        if cid ~= "" then
            local cid_attrs = get_obj_attrs(cid)
            if not cid_attrs.uid then
                cid_attrs = get_obj_attrs(cid, "mobile")
            end
            if cid_attrs.uid then
                gSpace["cidNum"]   = cid
                gSpace["cidAttrs"] = cid_attrs
                if "autocallrec" == application then
                    return app_record_call(nil, "personal")
                elseif "autocallswitch" == application then
                    if "fax" ~= cid_attrs.userKind then
                        return app_continuity_call("incall")
                    end
                end
            end
        end
        return false
    end

    -- Rewrite caller name WMS-2955: support caller id name on w04fxo
    if "anonymous" == channel.CALLERID("name"):get() then
        channel.CALLERID("name"):set("")
    end

    -- Set original caller name
    if not gSpace["gvars"].origCidName then
        gSpace["gvars"].origCidName = channel.CALLERID("name"):get() or ""
    end
    
    -- {{ Get global variable cid_num
    cid = cid == "" and channel.CALLERID("num"):get() or cid

    if cid == "Record" then     -- fix correct cid for Record from WMS
        cid = channel.RECORDOWNER:get() or ""
        if cid == "" then
            print("Invalid RECORDOWNER for application Record from WMS")
            logger:warning("Invalid RECORDOWNER for application Record from WMS")
            return false
        end
    end
    
    -- handle call without cid from trunk as 'Anonymous'
    local is_trunk_call = channel["X-TRUNK"]:get() or ""
    gSpace["gvars"].trunk_call = is_trunk_call
    if is_trunk_call ~= "" and (not cid or cid == "") then
        cid = "Anonymous"
        channel.CALLERID("num"):set(cid)
        channel.CDR("c_from"):set(cid)
    end
    gSpace["cidNum"] = cid
    -- }}
    
    if application ~= "app-set" and application ~= "app-goto-if" then  -- for correct variable pushed to app_set
        exten = convert_exten(exten)
    end
    
    -- Processing application variable
    local appVar         = channel.APPLICATION:get() or ""
    local is_disa        = appVar and appVar:match("^DISA.*")
    local is_kite        = "KITE" == appVar
    local is_queue       = "QUEUE" == appVar
    local is_callback    = "CALLBACK" == appVar
    local isMobilityCall = gSpace["gvars"].mobility_number or appVar:find("^MOBILITY")
    gSpace["gvars"].application = appVar
    
    if is_queue and not gSpace["gvars"].calltype then
        gSpace["gvars"].calltype = channel.CALLTYPE:get() or "internal"
    end
    if is_trunk_call ~= "" then
        if not gSpace["gvars"].calltype then
            channel["_CALLTYPE"]:set("external")
            gSpace["gvars"].calltype = "external"
        end
        
        if not gSpace["gvars"].incExtNumberInfo then
            if "app-lookup-mobility-ext" ~= application and 
              not is_disa and not is_kite and not is_queue and
              "1" ~= channel["RET_TRANS"]:get() then
                
                -- process external number
                local extNumber  = require"externalNumber"    
                local normNumber = extNumber:normalize(cid, nil, "inc")
                extNumber:getInfo(normNumber)
                gSpace["gvars"].incExtNumberInfo = true
                
                -- check call class
                local callClass = require"peerCallClass"
                if not callClass:getFromChannel() then
                    local callClassVal = callClass:evaluate(normNumber)
                    callClass:setUp2Channel(callClassVal)
                end
                
                -- check in phonebook (exception calls from remove PBXes)
                local is_wms_network_call = channel["X-WMS-Network-Call"]:get() or ""
                if is_wms_network_call == "" then
                    app.CELGenUserEvent("TrunkCall", string.format("%s,%s", "incoming", is_trunk_call))
                    number_normalize_with_lookup(normNumber, "inc")
                end
            end
        end
    end

    -- getting calltype from channel variable
    if not gSpace["gvars"].calltype then
        gSpace["gvars"].calltype = channel.CALLTYPE:get() or "internal"
    end
    
    if "switchcalls" == application then
        -- call continuity feature
        return require"internalcalls":switchcalls(exten)
    elseif application == "app-faxdetect" then
        return app_faxdetect(exten, ...)
    elseif application == "app-authenticate-user" then
        return app_authenticate_user(exten, ...)
    elseif application == "LOGIN" then
        return service("LOGIN", nil, exten or 0)
    elseif application == "app-lookup-mobility-ext" then
        return app_mobility_lookup(exten, ...)
    elseif application == "app-mux-mon" then
        return app_record_call(exten, "pbx", ...)
    elseif "app-playback" == application then
        return app_playback(exten, false, ...)
    elseif "app-playback-waitDigits" == application then
        return app_playback(exten, true, ...)
    elseif application == "app-set" then
        -- checking BIG wildix hack (save them for compatibility)
        -- must use application 'Runner' !!!
        if "CallOut" == exten then
            application = "CallOut"
            exten       = require"core.pbxutils".split(select(1, ...), ",")
        else 
            return app_set(exten, ...)
        end
    elseif "queuecalls" == application then
        if exten:find("^WMSNet") then
            return require"queuecalls".runRemoteCall(exten)
        end
        return require"queuecalls".run(exten)
    end
    
    if exten == nil or exten == "" then -- execute service and application without exten and cid attrs
        if application == "app-hangup" then
            local causecode = select(1, ...)
            return app_hangup(causecode)
        elseif application == "ECHO" then
            return service("ECHO")
        elseif application ~= "INTERCOM" and not application:match("^[S,P,C]BARGE") and application ~= "QCALL" and application ~= "CGMANAGER"
            and application ~= "PAGING" and application ~= "app-voicemail-main" and application ~= "VMACCESS" and application ~= "PERSONALRECORD"
            and application ~= "PBXRECORD" and application ~= "FAXRELAY" and application ~= "SENDDOC" and application ~= "app-callback" 
            and application ~= "PICKUP" and  application ~= "app-conference" and application ~= "CONFERENCE" and application ~= "SHORTNUM"
            and application ~= "HIDECID" and application ~= "QPICKUP" and application ~= "GPICKUP" and application ~= "SWITCH" and application ~= "3SWITCH" 
            and application ~= "STIME" and not feature_in_table(application) and application ~= "PSTN_TRUNK_OCCUPATION"
            and application ~= "app-goto-general" and application ~= "app-goto-if-dialstatus" and application ~= "app-goto-if-call-type" 
            and application ~= "app-goto-if-group-count" and application ~= "app-verify-caller" and application ~= "app-timetable"
            and application ~= "app-goto-if-switch-is-on" and application ~= "app-goto-if-three-switch-is-on-or-blinking"
            and "app-directory" ~= application and "DIRECTORY" ~= application and "VMTRANSFER" ~= application
            and "WHOTELMUR" ~= application  and "WHOTELBAR" ~= application and "WHOTELALARM" ~= application and "WHOTELINSPECT" ~= application -- WHotel services
            and "CallOut" ~= application and "CONTINUITYCALL" ~= application and "IDENTITY" ~= application then                                                              -- Callout service
            print("Error(%s): number is empty/not valid!", application)
            logger:error("Error(%s): number is empty/not valid!", application)
            return false
        end
    elseif application == "PREBARGE" then       --> from preanswer services
        return service("BARGE", exten, nil, "PREANSWER", ...)
    elseif application == "PREINTERCOM" then    --> from preanswer services
        return service("INTERCOM", exten, nil, false, ...)
    end
    

    -- chunk of getting cid attrs from LDAP and cid features
    if (cid ~= gSpace["cidAttrs"].telephoneNumber and cid ~= gSpace["cidAttrs"].mobile) or require"common.helper".isTableEmpty(gSpace["cidFeatures"]) then
        if cid == nil or cid == "" then
            print("Caller ID is empty. Exiting.")
            logger:warning("Caller ID is empty. Exiting.")
            smartApp("Hangup", hcause.REJECT)
        end

        -- search in LDAP by telephoneNumber or than by mobile
        local cid_attrs = get_obj_attrs(cid)
        if not cid_attrs.uid then
            cid_attrs = get_obj_attrs(cid, "mobile")
        end

        if cid_attrs.uid and (is_trunk_call == "" or is_disa or isMobilityCall) then
            gSpace["cidAttrs"] = cid_attrs
            if cid_attrs.telephoneNumber == nil then
                print("Caller user is not found in ldap")
                logger:warning("Caller user is not found in ldap")
                return smartApp("Hangup", hcause.FAILURE)
            end
            
            -- check X-Limit for registration or subscriptions
            local limit = channel["X-Limit"]:get() or ""
            if limit:find("max%-contacts") then
                ivr:tell("Max phones number reached", {opts = "n"})
            end
            if limit:find("max%-subscriptions") then
                print("Max number of active subscriptions reached")
                logger:warning("Max number of active subscriptions reached")
            end
            
            local cid_features = require"common.userFeatures.userFeatures":new(cid_attrs.uid)
            if not cid_features then
                print("Error loading features of the caller user with uid %s", cid_attrs.uid)
                logger:error("Error loading features of the caller user with uid %s", cid_attrs.uid)
                smartApp("Hangup", hcause.NOT_FOUND)
            end
            cid_features:load()
            gSpace["cidFeatures"] = cid_features
        end
    end

    -- processing callblock feature
    if gSpace["cidAttrs"].telephoneNumber and gSpace["cidFeatures"]:is_active("CALLBLOCK") and not gSpace["gvars"].callBlockLoginChecked then
        local isCallBlockProcessing = true
        if application:find("internalcall") or "returnFromTransfer" == application then
            isCallBlockProcessing = false
        elseif isMobilityCall then
            isCallBlockProcessing = false
        elseif application:find("^app%-goto%-") or "app-timetable" == application then
            isCallBlockProcessing = false
        elseif "app-dial-hunt-group" == application or "app-dial-trunk" == application then
            isCallBlockProcessing              = false
            gSpace["gvars"].callBlockInProcess = true  -- let's processing call block feature inside external call application    
        end
        
        if isCallBlockProcessing then
            __callBlockExecuting()
        end
    end
    
    -- Fix CDR fields: for call from DISA
    if is_disa then
        channel.CDR("c_from"):set(gSpace["cidAttrs"].telephoneNumber or cid or "")
        channel.CDR("from_name"):set(gSpace["cidAttrs"].cn or "")
    elseif is_callback then -- disable CDR record for call from Local channel 'Callback' to user
        smartApp("NoCDR", "")
    end 

    local cidOfficeNum = gSpace["cidAttrs"] and gSpace["cidAttrs"]["wOfficePhone"] or ""
    -- Block non emergency calls from Collaboration for US if X-Caller-Location header is empty or not present
    if (channel["SKIP_GEOLOCATION_CHECK"]:get() ~= "yes"
        and string.find(channel.SIP_HEADER("User-Agent"):get() or "","Wildix Zero Distance")
        and (channel.GLOBAL("GCOUNTRY_CODE"):get() == "1" or string.find(cidOfficeNum,"^%+?1%d%d%d%d%d%d%d%d%d%d"))
        and require"common.permissions":getPbxLicense("classound") == "1"
        ) then
        local xCallerLocation = channel.SIP_HEADER("X-Caller-Location"):get() or ""
        if xCallerLocation == "" and exten ~= "911" and exten ~= "933" then
            smartApp("Playback", "tvoice/geolocation_must_be_enabled")
            logger:warning("X-Caller-Location header is empty or not present, rejecting the call")
            smartApp("Hangup", hcause.REJECT)
        else
            logger:debug("X-Caller-Location header: %s", xCallerLocation)
        end
    end

    local return_transfer = (channel["TRANSFER_NUMBER"]:get() or ""):match("^(R%w+%-%w+)")
    if not return_transfer and application == "internalcalls" then
        return require"internalcalls":run(exten)
    elseif application == "returnFromTransfer" or (application == "internalcalls" and return_transfer) then
        if return_transfer and application == "internalcalls" then
            print("Return from transfer at remote pbx to", exten)
            channel.CALLERID("name"):set(return_transfer)
            channel["_RET_TRANS"]:set("1")
            channel["TRANSFER_NUMBER"]:set("NULL")
        end
        return require"internalcalls":run(exten, nil, nil, 0, true)
    elseif application == "app-dial-internalcall" then
        local timeout         = select(1, ...)
        local options         = select(2, ...)
        local announce        = select(3, ...)
        local last_member     = select(4, ...)
        local optList         = {"number - " .. exten}
        optList[#optList + 1] = timeout and "timeout - " .. timeout
        optList[#optList + 1] = options and "options - music instead ring"
        optList[#optList + 1] = announce and "announce - " .. announce 
        print(string.format("Executing 'Dial the phone': %s", table.concat(optList, ", ")))
        return require"internalcalls":run(exten, timeout, options, last_member, nil, announce)
    elseif application == "fwd-internalcall" then
        return require"internalcalls":run(exten, nil, nil, ...)
    elseif application == "app-dial-hunt-group" or application == "app-dial-trunk" then
        return app_external_call(exten, ...)
    elseif application == "app-dial-remote-pbx" then
        return app_dialremotepbx(self, "Executing 'Call through remote pbx': ", exten, ...)
    elseif application == "app-lookup-in-phonebook" then
        return number_normalize_with_lookup_via_app(exten, ...)
    elseif "app-directory" == application then
        -- return require"Directory".run(exten, ...)
        return require"Directory".run(exten or (channel.IVR_ASR:get() ~= "no" and "a" or ""), ...)
    elseif application == "app-fax2email" then
        return fax2mail(exten, ...)
    elseif application == "app-service-intercom" then
        local timeout  = select(1, ...)
        local options  = select(2, ...)
        local announce = select(3, ...)
        local optList         = {"number - " .. exten}
        optList[#optList + 1] = timeout and "timeout - " .. timeout
        optList[#optList + 1] = options and "options - music instead ring"
        optList[#optList + 1] = announce and "announce - " .. announce 
        print(string.format("Executing 'Intercom': %s", table.concat(optList, ", ")))
        return service("INTERCOM", exten, "app", true, timeout, options, nil, announce)
    elseif application == "app-service-paging" then
        print ("Executing 'Paging': page id - " .. exten)
        return service("PAGING", exten, "app", true)
    elseif application == "app-voicemail" then
        local opts = select (1, ...)
        print ("Executing 'Go to voicemail': mailbox - " .. exten .. (opts and (", options - " .. opts) or ""))
        return app_voicemail(self, exten, "leave", opts)
    elseif application == "app-queue" then
        return app_queue_call("Executing 'Call group'", exten, ...)
    elseif application == "app-ivr" then
        return application_ivr(exten)
    elseif application == "app-callback" then
        return app_callback()
    elseif application == "app-correct-outgoing-number" then
        return app_correct_outgoing_number(exten, ...)
    elseif application == "app-goto-general" then
        return app_goto_general(exten, ...)
    elseif application == "app-goto-if-dialstatus" then
        return app_goto_dialstatus(exten, ...)
    elseif application == "app-goto-if-call-type" then
        return app_goto_calltype(exten, ...)
    elseif application == "app-goto-if-group-count" then
        return app_goto_group_count(exten, ...)
    elseif application == "app-goto-if-userstatus" then
        return app_goto_userstatus(exten, ...)
    elseif application == "app-goto-if" then
        return app_goto_if_condition(exten, ...)
    elseif application == "app-verify-caller" then
        return app_verify_caller(exten, ...)
    elseif application == "app-goto-if-switch-is-on" then
        return app_switch(exten, ...)
    elseif application == "app-goto-if-three-switch-is-on-or-blinking" then
        return app_three_switch(exten, ...)
    elseif application == "app-timetable" then
        return app_timetable(exten, ...)
    elseif application == "app-conference" then
        return app_conference(exten)
    elseif application == "CONFERENCE" then
        return service("CONFERENCE", nil, exten or 0)
    elseif application == "INTERCOM" then
        return service("INTERCOM", nil, exten or 0, ...)    -- in this case exten is length for strip
    elseif application:match("^[S,P,C]BARGE") then
        return service("BARGE", nil, exten or 0, application, false)
    elseif application == "PAGING" then
        return service("PAGING", nil, exten or 0, false)
    elseif application == "app-voicemail-main" then
        print ("Executing 'Voicemail access'" .. (exten and (": mailbox - " .. exten) or "")  .. (select(1, ...) ~= "" and ((exten and ", " or ": ") .. "options - skip pin check") or ""))
        return app_voicemail(self, exten, "access", ...)
    elseif application == "VMACCESS" then
        return service ("VMACCESS", nil, exten or 0)
    elseif application == "VMTRANSFER" then
        return service ("VMTRANSFER", nil, exten or 0)
    elseif application == "PBXRECORD" then
        return service("RECORDSOUND", nil, exten or 0, nil, "pbx")
    elseif application == "PERSONALRECORD" then
        return service("RECORDSOUND", nil, exten or 0, nil, "personal")
    elseif application == "SENDDOC" then
        return service("SENDDOC", nil, exten or 0)
    elseif application == "FAXRELAY" then
        return service("FAXRELAY", nil, exten or 0)
    elseif application == "PICKUP" then
        return service("PICKUP", nil, exten or 0)
    elseif application == "QPICKUP" then
        return service("QPICKUP", nil, exten or 0)
    elseif application == "GPICKUP" then
        return service("GPICKUP", nil, exten or 0)
    elseif application == "QCALL" then
        return service("QCALL", nil, exten or 0)    -- in this case exten is length for strip
    elseif application == "CGMANAGER" then
        return service("CGMANAGER", nil, exten or 0)
    elseif application == "HIDECID" then
        return service("HIDECID", nil, exten or 0)
    elseif application == "SHORTNUM" then
        return service("SHORTNUM", nil, exten or 0)
    elseif application == "SWITCH" then
        return service("SWITCH", nil, exten or 0)
    elseif application == "3SWITCH" then
        return service("SWITCH", nil, exten or 0, nil, true)
    elseif application == "STIME" then
        return service("STIME", nil, exten or 0)
    elseif application == "PSTN_TRUNK_OCCUPATION" then
        return service("PSTN_TRUNK_OCCUPATION", nil, exten or 0)
    elseif "DIRECTORY" == application then
        return service("DIRECTORY", nil, exten or 0)
    elseif feature_in_table(application) then
        return feature(application, exten, ...)
    elseif application == "OGI" then
        return smartApp("AGI", exten, ...)
    elseif application == "CONTINUITYCALL" then
        return service("CONTINUITYCALL", nil, exten or 0)
    elseif application == "IDENTITY" then
        return service("IDENTITY", nil, exten or 0)
    
    -- {{ WHotel services
    elseif application == "WHOTELMUR" then
        return service("WHOTELMUR", nil, exten or 0)
    elseif application == "WHOTELBAR" then
        return service("WHOTELBAR", nil, exten or 0)
    elseif application == "WHOTELALARM" then
        return service("WHOTELALARM", nil, exten or 0)
    elseif application == "WHOTELINSPECT" then
        return service("WHOTELINSPECT", nil, exten or 0)
    elseif application == "app-wake-up" then
        return require"exService":applicationWakeUp(exten, ...)
    -- WHotel services }}
    
    -- {{ Custom services
    elseif application == "CallOut" then
        if "table" == type(exten) then
            return require"exService":applicationCallOut(unpack(exten))
        else
            return require"exService":applicationCallOut(exten, ...)
        end
    -- Custom services }}
    
    -- {{ Way to execute some unit tests inside callweaver
    elseif application == "RunTests" then
        local unitTestFile = exten
        package.path       = package.path .. ";/tmp/unit-tests/?.lua"
        return require(unitTestFile)
    -- }}
    
    else
        print("Error: please specify correct application or service!")
        logger:error("Error: please specify correct application or service!")
        return smartApp ("Hangup", hcause.FAILURE)
    end
end

--- Function for changing callerid or connectedlineid
function set_call_id(key, value, dsttype)
    if not key or "" == key then
        return false
    end
    if not value then
        return false
    end
    if channel["PRESERVE_CID"]:get() == "yes" then
        logger:debug("Skip setting callerid/connectedline num/name because the variable PRESERVE_CID is set.")
        print("Skip setting callerid/connectedline num/name because the variable PRESERVE_CID is set.")
        return true
    end
    local channel = channel
    if dsttype == "inc" then
        print(string.format("Set Caller %s to %q desttype %q", key, value, dsttype))
        channel.CALLERID(key):set(value)
        if key == "num" or key == "number" then
            gSpace["cidNum"] = value
        end
    elseif dsttype == "out" then
        print(string.format("Set Connected %s to %q desttype %q", key, value, dsttype))
        channel.CONNECTEDLINE(key):set(value)
    end
end

--- Phonebook helpers and public functions ---
--- Common function to make number normalization and lookup contact name
-- @param execType      An type of the executing method
-- @param number        The checked number (can be non-normilized)
-- @param directType    The direction type (incoming/outgoing calls)
-- @param pbID          The available phone book ident
-- @param peer          The peer name used for normalization
local function __number_normalize_with_lookup_common(execType, number, directType, pbID, peer)
    -- processing input arguments
    if not number and "inc" == directType then
        number = channel.CALLERID("num"):get()
    end
    -- processing saved input arguments for outgoing calls
    if "out" == directType then
        gSpace["gvars"].lookup_out = true      -- used to disable re-search
    end
    local pbID = "" ~= pbID and pbID or "all"
    
    -- check for lookup number
    local forceLookUpNumber = channel.FORCE_LOOKUP_NUMBER:get() or ""
    if "" ~= forceLookUpNumber then
        number = forceLookUpNumber
    end

    -- print executing message
    local msg = "Executing 'Lookup "
          msg = msg .. (directType == "inc" and "caller" or "called")
          msg = msg .. " name'"
          if "app" == execType then
              msg = msg .. " (dialplan)"
          end
          msg = msg .. ": "
          msg = msg .. "PBID - " .. pbID
          msg = msg .. ", number - " .. number
          if "" ~= forceLookUpNumber then
              msg = msg .. "(force lookup)"
          end 
    print(msg)

    -- running main procedures
    local normNumber = "" == forceLookUpNumber and require"externalNumber":normalize(number, peer) or number
    local lookupObj  = require"contactLookUp":new(normNumber)
    repeat
        if not lookupObj then
            print("Number not normilized. Continue dialplan proceeding ...")
            break
        end
        lookupObj:setDirection(directType)
        lookupObj:searchByLdap()
        lookupObj:searchByIntegrationMiddleware()
        lookupObj:searchByLocalPB(pbID, gSpace["cidAttrs"].uid)
        lookupObj:searchByExternalService()

         -- print lookup result source
        if "LDAP" == lookupObj:getResultSource() then
            print("Source LDAP, mobility number found")
        elseif "WIMService" == lookupObj:getResultSource() then
            print("Source integration middleware, search number found")
        elseif "PBXPB" == lookupObj:getResultSource() then
            print("Source phonebook, search number found")
        elseif "ExtService" == lookupObj:getResultSource() then
            print("Source external service, search number found")
        else
            print("Search number not found in the contact storages")
        end

        -- post processing look up results
        lookupObj:postProcess()
    until true

    -- processing custom phone book ringtone for incoming trunk calls
    if "inc" == directType then
        local PBContactRingTone = lookupObj and lookupObj:getContactByField("ringTone")
        if PBContactRingTone and "" ~= PBContactRingTone then
            channel["_PBRINGTONE"]:set(PBContactRingTone)
        end
    end

    -- save result data to global variables
    gSpace["gvars"].PBContactNumber = normNumber or ""
    if channel["PRESERVE_CID"]:get() ~= "yes" then
        gSpace["gvars"].PBContactName   = lookupObj and lookupObj:getContactName() or ""
    else
        gSpace["gvars"].PBContactName   = ""
    end
    

    -- processing contact data for outgoing fax calls
    if "out" == directType and "FAX" == channel.APPLICATION:get() then
        local fromName    = gSpace["cidAttrs"].cn or channel.CALLERID("name"):get() or ""
        local contName    = gSpace["gvars"].PBContactName
        local fax_gvar_pb = normNumber.."|"..contName.."|"..fromName
        local faxnumberid = channel.faxnumberid:get() or ""
        channel.GLOBAL(faxnumberid.."_pb"):set(fax_gvar_pb)
    end

    return true
end

--- Search number in user selected phonebooks (User preferences -> Settings -> Phonebooks -> Selected)
-- @param number        number for search
local function __isNumberInSelectPB(number)
    local rv = false
    local cidFeatures  = gSpace["cidFeatures"]
    local userSelectPb = cidFeatures:get("PHONEBOOK-ID")
    local lookupObj    = require"contactLookUp":new(number)
    if lookupObj and "" ~= userSelectPb then
        lookupObj:searchByLocalPB(userSelectPb)
        rv = "PBXPB" == lookupObj:getResultSource()
    end
    print(string.format("Number %s %s in allowed phonebooks.", number, rv and "found" or "not found"))

    return rv
end

function number_normalize_with_lookup_via_app(number, pbID, directType, opts)
    return __number_normalize_with_lookup_common("app", number, directType, pbID)
end

function number_normalize_with_lookup(number, directType, peer, pbID)
    return __number_normalize_with_lookup_common("generic", number, directType, pbID, peer)
end

--- Dialplan hangup function
--@param causecode A code for app.Hangup
function app_hangup(causecode)
    local app, channel = app, channel
    local hcause = hcause
    local q931causes = {1,2,3,6,7,16,17,18,19,21,22,27,28,29,30,31,34,38,41,42,43,44,45,50,52,54,57,58,65,66,69,81,88,95,96,97,98,99,100,101,102,103,111,127}
    local causecode = require"core.pbxutils".in_table(tonumber(causecode), q931causes) and causecode or ""
    
    if causecode == "" then
        local dialstatus = channel["DIALSTATUS"]:get() or "UNKNOWN"
        causecode = ({
        ["BUSY"] = hcause.BUSY,
        ["NOANSWER"] = hcause.NOT_RESPOND,
        ["CONGESTION"]= hcause.NOT_RESPOND,
        ["CHANUNAVAIL"] = hcause.CHANUNAVAIL
        })[dialstatus] or hcause.NORMAL
    end
    return smartApp("Hangup", causecode)
end

--- Setting global channel variable
-- @param variable A psevdo name channel variable
-- @param value A value setted for variable
function app_set (variable, value)
    local channel   = channel
    local variable  = variable or ""
    local value     = value or ""
    local evalValue = convert_exten(value)
    if value ~= evalValue then
        print(string.format("Executing 'Set': %s - %s => (\"%s\" - \"%s\")",  variable, value, variable, evalValue))
    else 
       print(string.format("Executing 'Set': %s - %s", variable, evalValue))
    end
    if variable == "__TRUNK_CALLER_NUM" then
        gSpace["gvars"].tr_caller_num = evalValue
        return true
    elseif variable == "LANGUAGE()" then
        if not evalValue or evalValue == "" then evalValue = "en" end   -- by defautl "en"
        return channel.CHANNEL("Language"):set(evalValue)
    elseif variable == "MUSICCLASS()" then
        if not evalValue or evalValue == "" then evalValue = "default" end
        return channel.CHANNEL("musicclass"):set(evalValue)
    elseif variable == "__CALLTYPE" then
        if not evalValue or evalValue == "" then evalValue = "internal" end
        gSpace["gvars"].calltype = evalValue
        return channel["__CALLTYPE"]:set(evalValue)
    elseif variable == "CALLERID(name)" then
        if not evalValue or evalValue == "" then return false end
        gSpace["gvars"].origCidName = evalValue  -- overwrite "original callerid name" variable
        local cdr_from_name = evalValue
        local company = channel["company"]:get() or ""
        if "" ~= company then 
            company = require"common.helper".escapeLuaString("(" .. company .. ")")
            cdr_from_name = cdr_from_name:gsub(company, "")
        end
        channel.CDR("from_name"):set(cdr_from_name)
        return channel.CALLERID("name"):set(evalValue)
    elseif variable == "CALLERID(number)" then
        if not evalValue or evalValue == "" then return false end
        return channel.CALLERID("num"):set(evalValue)
    elseif variable == "SIP_CODECS" then
        if not evalValue or evalValue == "" then return false end
        channel["SIP_CODEC_INBOUND"]:set(evalValue)
        channel["_SIP_CODEC_OUTBOUND"]:set(evalValue)
        channel["DISABLE_DIRECTRTP"]:set(1)
        return true
    elseif "DIRECTRTP" == variable then
        if "yes" == evalValue then
            channel["FORCE_DIRECTRTP"]:set(1)
        elseif "no" == evalValue then
            channel["DISABLE_DIRECTRTP"]:set(1)
        end
    elseif "CDR(tags)" == variable then
        local evalValue      = evalValue or ""
        local chanTagVal = evalValue:match("tags:(.+)") or ""
        channel.CDR("tags"):set(evalValue)
        channel["__CALLTAGS"]:set(chanTagVal)
        channel["__TAGS"]:set(chanTagVal)
        app.CELGenUserEvent("Tag", chanTagVal)
    elseif "TIMEOUT" == variable then
        evalValue = tonumber(evalValue)
        if not evalValue then return false end
        return channel.TIMEOUT("absolute"):set(evalValue)
    elseif "RINGTONE()" == variable then
        return channel["_RINGTONE"]:set(evalValue)
    elseif variable and variable ~= "" then
        return channel[variable]:set(evalValue or "")
    end
    return false
end

--- Application playback
-- @param soundPath     A related path to sound file
-- @param waitDigits    A flag for executing application with wait for digits
-- @param opts          An options of the playback application
function app_playback(soundPath, waitDigits, opts)
    print(string.format("Executing 'Playback': sound - %q, waitDigits - %s, opts - %s", soundPath or "", waitDigits and "y" or "n", opts or ""))
    if not soundPath or "" == soundPath then
        return false
    end
    local app      = app
    local noAnswer = "noanswer" == opts

    local channel  = channel
    
    local textToSpeech = require"textToSpeech"
    local ttsStatus    = "UNKNOWN"

    -- check if soundPath is a static file or a text
    local glob       = require"posix".glob
    local sFileExist = nil
    for _, ext in pairs({".alaw", ".g729", ".wav", ".ulaw", ".opus"}) do
        if glob("/var/lib/callweaver/sounds/" .. soundPath .. ext) then
            sFileExist = true
            break
        end
    end

    if not sFileExist and textToSpeech.available() then
        if noAnswer then
            app.Ringing()
        end
        local message      = convert_exten(soundPath)
        local language     = channel.CHANNEL("Language"):get()
        local ttsSoundPath = textToSpeech.getAudio(message, language)
        if ttsSoundPath then
            soundPath = ttsSoundPath
            ttsStatus = "SUCCESS"
        else
            ttsStatus = "FAILURE"
        end
    end
    channel["TTSSTATUS"]:set(ttsStatus)
    
    if noAnswer then
        smartApp("Progress")
    end
    if waitDigits then
        smartApp("Background", soundPath, noAnswer and "n" or "")
    else
        smartApp("Playback", soundPath, noAnswer and "noanswer" or "")
    end
    return true
end

--- Checking mobility in LDAP and go to disa context
-- @param mobnumber A mobility for find in LDAP
-- @param opts A options, such as callback
function app_mobility_lookup(mobnumber, opts)
    print ("Executing 'Mobility extension lookup': number - " .. (mobnumber or "") .. (opts == "d" and ", options - callback" or ""))
    if not mobnumber or mobnumber == "" then
        return false
    end
    local app, channel, hcause = app, channel, hcause
    mobnumber = require"externalNumber":normalize(mobnumber)
    
    local user_attrs = get_obj_attrs(mobnumber, "mobile")
    if not user_attrs.uid then
        print ("No user found for mobile number " .. mobnumber)
        return false
    end
    
    if not user_attrs.telephoneNumber then
        print ("No internal number is set for the user with mobile number " .. mobnumber)
        return false
    end
    
    -- check options to get if we need to call back
    if opts == "d" then
        to_logfile ("Ringing", "")
        app.Ringing()
        to_logfile ("Wait", 2)
        app.Wait(2)
        os.execute("nohup wms_disa_callback \"" .. mobnumber .. "\" \"" .. user_attrs.userDialplan .. "\" \"" .. user_attrs.telephoneNumber .. "\" \"" .. user_attrs.cn .. "\" \"" .. user_attrs.preferredLanguage .. "\" &")
        to_logfile ("Hangup", hcause.NORMAL)
        return app.Hangup(hcause.NORMAL)
    end
    
    channel["__TRANSFER_CONTEXT"]:set(user_attrs.userDialplan)
    gSpace["gvars"].disa_context = user_attrs.userDialplan
    channel.CALLERID("name"):set("M - " .. user_attrs.cn)
    channel.CALLERID("num"):set(user_attrs.telephoneNumber)
    channel.CHANNEL("Language"):set(user_attrs.preferredLanguage)
    gSpace["gvars"].disa_lang = user_attrs.preferredLanguage --used in predial for local channel (in case preanswer service)
    return extensions.wmsdisa.s()
end

--- Application for external call such as trunk and hunt groups
-- @param arg1  A extension for the call
-- @param arg2  A trunkname (TRUNK)                 - A trunk group ID (TRUNK GROUP)
-- @param arg3  A timeout for the call (TRUNK)      - A callclass for the call (TRUNK GROUP)
-- @param arg4  A maximum concurrent calls (TRUNK)  - A trunk group strategy (TRUNK GROUP)
-- @param arg5  A dial options (TRUNK)              - A trunk group unique value writing by wms (TRUNK GROUP)
-- @param arg6  A call class (TRUNK)                - Not used (TRUNK GROUP)
-- @param arg7  A last member in list of extension(0) or no(1) (set up when compile wms dialplan)
function app_external_call(arg1, arg2, arg3, arg4, arg5, arg6, arg7)
    local json      = require"dkjson"
    local wmsperm   = require"common.permissions"
    local extNumber = require"externalNumber"
    local callClass = require"peerCallClass"
    
    local smartApp  = smartApp
    local channel   = channel
    
    local cid_attrs, cid_features, gvars = gSpace["cidAttrs"], gSpace["cidFeatures"], gSpace["gvars"]
    local exten, maxcalls, callclass, tr_type
    local trunkname, timeout, dial_opts
    local hg_id, hg_strategy, hg_uniqueval
    local tr_group = "trunk"
    local hg_content = {}
    local hg_orderby, hg_start, t_hg_var, hg_row, number_to_call
    local callerInitName = channel.CALLERID("name"):get()

    if arg2 == "HG/" then
        print ("Executing 'Hunt group': specify correct hungoup name in dialplan rule")
        return false
    elseif not arg2 or arg2 == "" then
        print ("Executing 'Dial the trunk': specify correct trunk name in dialplan rule")
        return false
    end
    
    exten = arg1
    if arg2:find("^HG/%d+") then
        tr_group     = "huntgroup"
        hg_id        = arg2:sub(4)
        callclass    = arg3
        hg_strategy  = arg4
        hg_uniqueval = arg5 -- used for setting global var in case RR-strategy
        print(string.format("Executing 'Trunk group': number - %s, id - %s, callclass - %s, strategy - %s", exten, hg_id, callclass, hg_strategy))
    else
        trunkname    = arg2
        timeout      = tonumber(arg3)
        maxcalls     = arg4 or ""
        dial_opts    = arg5 or ""
        callclass    = arg6 and "" ~= arg6 and arg6
        print(string.format("Executing 'Dial the trunk': number - %s, tr_name - %s%s%s%s%s", exten, trunkname, 
            timeout and ", timeout - "..timeout or "", "" ~= maxcalls and ", maxcalls - "..maxcalls or "", "" ~= dial_opts and ", options - music instead ring" or "", callclass and ", callclass - "..callclass or ""))
    end

    local last_exten = true
    if arg7 and arg7 == 1 then
        last_exten = false
    end
    -- check kamailio limitation for concurrent call to trunk
    if channel.SIPLIMIT():get() == "TRUE" then
        if cid_attrs.uid then
            ivr:tell("Max number of calls reached", {opts = "n"})
        end
        return smartApp("Hangup", hcause.CHANUNAVAIL)
    end
    
    -- parsing mobility args getting via app_remote_call
    local remote_mob_caller, remote_mob_lang, remote_mob_owner, remote_mob_number
    if gvars.application then
        remote_mob_caller, remote_mob_lang, remote_mob_owner, remote_mob_number = gvars.application:match("^MOBILITY|(.*)|(%w*)|(.*)|(.*)")
    end
    
    -- processing trunk/trunk group settings
    local connection = require"core.wsqlite":new("wmsdb")
    local do_iterate, iteration_num
    local trunk_first, trunk_settings

    hg_orderby = "ASC"
    if tr_group == "huntgroup" then
        if hg_strategy == "Decrescent" or hg_strategy == "ReverseRoundRobin" then
            hg_orderby = "DESC"
        end
        local sql = "SELECT outgoing_groups_rules.trunkname AS hg_trunkname, outgoing_groups_rules.prepend AS hg_prepend, outgoing_groups_rules.remove AS hg_remove, outgoing_groups_rules.max_calls AS hg_maxcalls, outgoing_groups_rules.timeout AS hg_timeout FROM outgoing_groups_rules LEFT JOIN outgoing_groups_main ON (outgoing_groups_main.id = outgoing_groups_rules.group_id) WHERE outgoing_groups_main.id=%d ORDER BY outgoing_groups_rules.priority " .. hg_orderby .. ";"
        for hg_trunkname,hg_prepend,hg_remove,hg_maxcalls,hg_timeout in connection:rows(sql, tonumber(hg_id)) do
            hg_content[#hg_content+1] = {
                ['hg_trunkname'] = hg_trunkname,
                ['hg_prepend'] = hg_prepend or "",
                ['hg_remove'] = tonumber(hg_remove) or 0,
                ['hg_maxcalls'] = tonumber(hg_maxcalls),
                ['hg_timeout'] = tonumber(hg_timeout)
            }
        end
        connection:close()
        if 0 == #hg_content then
            print("Group with id %s is empty", hg_id)
            logger:info("Group with id %s is empty", hg_id)
            return false
        end
        
        trunk_first  = string.sub(hg_content[1]["hg_trunkname"], 5)
        if trunk_first then
            -- fix up for PSTN trunk names 
            -- in wms database save only bri/gsm/pri0_0 (fxo0_0_0 - correct)
            local tr_type = trunk_first:sub(1, 3)
            if tr_type == "pri" or tr_type == "gsm" or tr_type == "bri" then
                trunk_first = trunk_first .. "_0"       
            end
        end
        
        do_iterate    = #hg_content
        iteration_num = 1
        hg_start      = 1
        if hg_strategy == "RoundRobin" or hg_strategy == "ReverseRoundRobin" then
            t_hg_var = string.format("hg_%s_%s", hg_id, hg_uniqueval)
            hg_start = tonumber(channel[t_hg_var]:get() or 1)
            if hg_start > #hg_content then hg_start = 1 end
        end
    else
        trunk_first   = trunkname
        do_iterate    = 1
        iteration_num = 0
    end
    
    -- processing extenral number and call class
    local normNumber    = extNumber:normilizeWithShortNumbers(exten, trunk_first)
    local numberInfo    = extNumber:getInfo(normNumber)
    
    local callClassArg  = tonumber(callclass)
    local callClassVal  = callClass:evaluate(normNumber)
    local callClassName = callClass:getNameById(callClassVal)
    if "emergency" ~= callClassName and callClassArg then
        callClassVal    = callClassArg
        callClassName   = callClass:getNameById(callClassVal)
    end
    callClass:setUp2Channel(callClassVal)
    local isEmergencyCall = "emergency" == callClassName
    channel["callclass"]:set((callClassName and callClassName ~= "") and callClassName or "-")
    
    --- disable searching real peer by core
    if isEmergencyCall then
        channel.NOT_REAL_PEER:set(1)
    end
    
    -- processing call block feature
    if gSpace["gvars"].callBlockInProcess then
        local isCallBlockProcessing = true
        if "emergency" == callClassName then
            isCallBlockProcessing = false
        elseif "internal" == callClassName then
            isCallBlockProcessing = false
        end

        if isCallBlockProcessing then
            __callBlockExecuting()
        end
    end

    -- processing changing identity
    if cid_features and not require"common.helper".isTableEmpty(cid_features) then
        local from_identity = channel["FROM_IDENTITY"]:get()
        from_identity = "" ~= from_identity and from_identity
        local custom_identity = from_identity or cid_features:getValue("CURRENT_IDENTITY")
        if custom_identity and "" ~= custom_identity then
            local available_identities = cid_features:getCustomIdentities() or {}
            if require"core.pbxutils".in_table(custom_identity, available_identities) then
                if cid_attrs.telephoneNumber and cid_attrs.telephoneNumber ~= custom_identity then
                    print("Call with a changed identity: %s => %s", cid_attrs.telephoneNumber, custom_identity)
                    logger:info("Call with a changed identity: %s => %s", cid_attrs.telephoneNumber, custom_identity)
                    local cid = custom_identity
                    cid_attrs = get_obj_attrs(cid)
                    if cid_attrs.uid then
                        cid_features = require"common.userFeatures.userFeatures":new(cid_attrs.uid)
                        if cid_features then
                            cid_features:load()
                            channel["_FORCE_CTO"]:set(string.format("%s.%s", custom_identity, normNumber))
                        end
                    end
                    if not cid_features then
                        print("Error loading features of the caller user %s", cid)
                        logger:error("Error loading features of the caller user %s", cid)
                        smartApp("Hangup", hcause.NOT_FOUND)
                    end
                end
            else
                print("Identities list of the user doesn't contain the requested identity:", custom_identity)
                smartApp("Hangup", hcause.NOT_FOUND)
            end
        end
    end

    -- processing call permissions
    if not isEmergencyCall then
        local cid_owner = cid_attrs["owner"] 
        local mob_owner = gvars.mobility_owner or remote_mob_owner or ""
        if mob_owner and mob_owner ~= "" then   -- it's call to mobility
            -- Fix CDR fields: for correct CDR in case call to mobility
            local src_name, scr_num
            local cdr_clid = channel.CDR("clid"):get() or ""
            if cdr_clid ~= "" and cdr_clid:find("<.*>") then
                src_name, scr_num = cdr_clid:match("^\"(.*)\".*<(.*)>$")
            elseif  cdr_clid ~= "" then
                scr_num = cdr_clid
            end
            if scr_num then
                channel.CDR("c_from"):set(scr_num)
                if src_name and src_name ~= "" then
                    channel.CDR("from_name"):set(src_name)
                end
            end
            cid_owner = mob_owner:gsub(";", ",")
        end
        
        -- try to check whotel ACL
        local exService = require"exService"
        local ACLResult = exService:is_callAllowed(callClassVal) 
        if false == ACLResult then
            exService:forbidden_call(nil, "Rejected by Wildix Hotel ACL for current call class!")
        end
        
        if nil == ACLResult or "fax" == cid_attrs["userKind"] then -- WMS-8115 we ignore strange hotels logic for faxes above due to security reasons
            -- check common PBX ACL
            if "yes" ~= wmsperm:check(cid_owner or "", "cancall", callClassVal, nil, cid_attrs["licenseType"]) then
                -- check short number
                if not gvars.shortnumber_call then -- nor short number call
                    -- short number calls not allowed by ACL and called number not found in user selected phonebooks WMS-2845
                    if not ("yes" == wmsperm:check(cid_attrs["owner"] or "", "cancall", "20", nil, cid_attrs["licenseType"]) and __isNumberInSelectPB(normNumber)) then
                        print("Call not allowed. Rejected by ACL for current call class!")
                        return false
                    end
                else -- short number call
                    if "yes" ~= wmsperm:check(cid_attrs["owner"] or "", "cancall", "20", nil, cid_attrs["licenseType"]) then
                        print("Call not allowed. Rejected by ACL for current call class!")
                        return false
                    end
                end
            end
        end
    end

    local in_transfer_block = channel.BLOCK_TRANSFER:get()
    if cid_attrs.owner then
        if wmsperm:check(cid_attrs.owner, "canuse", "TRUNK_TO_TRUNK_TRANSFER", nil, cid_attrs["licenseType"]) ~= "yes" then
            channel.__BLOCK_TRANSFER = "yes"
        else
            channel.__BLOCK_TRANSFER = ""
        end
    end

    if "yes" == channel.BLOCK_TRANSFER:get() then
        if (gvars.trunk_call and "" ~= gvars.trunk_call) or "yes" == in_transfer_block then
            print("Transfer rejected by group permissions")
            return smartApp("Hangup", hcause.ACL_FORBIDDEN)
        end
    end

    -- add mobility language setting setup in internalcall (on local or remote pbx)
    local mob_lang = gvars.mobility_lang or remote_mob_lang or ""  
    -- add mobility options setting setup in internalcall (on local or remote pbx)
    local mob_opts = gvars.mobility_caller or remote_mob_caller or ""
    
    if mob_opts ~= "" then
        mob_opts = mob_opts:gsub("%D", "")
        if mob_opts and mob_opts ~= "" then
            local textToSpeech = require"textToSpeech"
            local play_message = nil
            if not tonumber(callerInitName) and textToSpeech.available() then
                local ttsSoundPath = textToSpeech.getAudio(callerInitName, mob_lang)
                if ttsSoundPath then
                    play_message = "tvoice/wildix_mobility&tvoice/call_from&" .. ttsSoundPath .. "&tvoice/press_asterisk_to_accept_the_call"
                end
            end
            if not play_message then
                local tmp_say_digits = {}
                for i = 1, #mob_opts do
                    tmp_say_digits[#tmp_say_digits+1] = "digits/" .. mob_opts:sub(i,i)
                end
                play_message = "tvoice/wildix_mobility&" .. table.concat (tmp_say_digits , "&") .. "&tvoice/press_asterisk_to_accept_the_call"
            end
            mob_opts = "RA(" .. play_message .. "&" .. play_message .. "&" .. play_message .. ")"
            channel["_MOBILITY_PLAYING_IVR"]:set("yes") -- used for diabling direct RTP during mobility confirmation
        end
    end

    dial_opts = (dial_opts or "") .. (mob_opts or "")
    
    if cid_features and cid_attrs.uid and cid_features:is_active("HIDECID") then
        print("Feature 'Hide number' is active.")
        gSpace["gvars"].anon_num = true
        channel.CALLERID("name"):set("restricted") -- "anonymous"
    end

    -- processing office number. Include some special cases for transfers or mobility calls
    local office_phone     = nil
    local office_num_other = nil
    local tran_num              = channel["TRANSFER_NUMBER"]:get() or ""    -- use for transfer and forward
    if tran_num == "" then tran_num = gvars.forward_number or "" end        -- use for FCF, BCF, UCF
    if not gvars.tr_caller_num then
        local officeNumber = channel["LDAP_OFFICE_NUMBER"]:get() or "wOfficePhone"
        if not gvars.anon_num then
            local tr_attrs = nil
            office_phone = cid_attrs[officeNumber]
            if not office_phone then
                if tran_num ~= "" then  -- set correct caller id number in case blind transfer
                    tr_attrs = get_obj_attrs(tran_num)
                    if tr_attrs[officeNumber] then
                        office_num_other = tr_attrs[officeNumber]
                    end
                end
            end
            if not tr_attrs or not tr_attrs.officeNumber or not tr_attrs.cn then
                require"internalcalls":fix_callername_return_transfer()
            end
        end
        
        -- rewrite source number to destination office number in case call to mobility
        local mobilityOptions = channel.MOBILITY_OPTS:get() or ""
        local mob_num = gvars.mobility_number or remote_mob_number or mobilityOptions:match("^+%d+") or ""
        if mob_num ~= "" and not office_phone then
            local mob_user_attr = get_obj_attrs(mob_num, "mobile")
            office_num_other    = mob_user_attr[officeNumber]
        end
    end
    
    if gvars.application == "DISA_B" then
        if cid_attrs.cn then    -- Fix CDR fields : for DISA callback
            channel.CDR("c_from"):set(cid_attrs.telephoneNumber or "")
            channel.CDR("c_to"):set(cid_attrs.telephoneNumber or "")
            channel.CDR("from_name"):set(cid_attrs.cn or "")
            channel.CDR("to_name"):set(cid_attrs.cn or "")
        end
        if cid_attrs.mobile then
            channel.NORM_NUMBER:set(cid_attrs.mobile)       -- used for mobility presence - DISA callback
        end
    end

    local gtime, faxnumberid
    if gvars.application == "FAX" then
        gtime = channel["GINTDIAL_TIMEOUT"]:get() or ""
        
        -- clear global vars and decrement TXFAXCNT
        faxnumberid = channel.faxnumberid:get() or ""
        setpostaction {
            ["name"] = "SENDFAX",
            ["faxnumberid"] = faxnumberid,
        }
        channel.CHANNEL("hangup_handler_wipe"):set("hdlr,service,1")
    end
    
    channel.CHANNEL("hangup_handler_wipe"):set("hdlr,service,1") -- WMS-8040

    -- set flag about call to external number
    gSpace["gvars"].outTrunkCall = true
    
    -- start iterations for calls
    local do_dial, curcalls, dialstatus, hcausecode, call_limit
    local pricelist, opt_present
    while do_iterate > 0 do
        if tr_group == "huntgroup" then
            hg_row = hg_content[hg_start]
            tr_type, trunkname = string.lower(hg_row["hg_trunkname"]:sub(1, 3)), hg_row["hg_trunkname"]:sub(5)
            number_to_call = hg_row["hg_prepend"] .. string.sub(exten, hg_row["hg_remove"]+1)
            maxcalls = hg_row["hg_maxcalls"] or nil
            timeout = hg_row["hg_timeout"]
            if tr_type == "pri" or tr_type == "gsm" or tr_type == "bri" then
                trunkname = trunkname .. "_0"       -- in wms database save only bri/gsm/pri0_0 (fxo0_0_0 - correct)
            end
        else
            tr_type = string.match(trunkname, "^(%a+)%d+_%d+_%d+") or "sip"
        end
        
        do_dial = 1         -- new iteration have reset do_dial
        if maxcalls and maxcalls ~= "" then
            curcalls = channel.GROUP_COUNT(trunkname.."@trunk"):get() or 0
            if tonumber(curcalls) >= tonumber(maxcalls) then
                do_dial = 0
                channel["DIALSTATUS"]:set("CONGESTION")
                print ("All lines are busy for " .. trunkname .. " (the limit is " .. maxcalls .. ")")
                if gvars.application == "FAX" and call_limit ~= 2 then
                    call_limit = 1 -- set call limit as reachable
                end
            end
        end

        if do_dial == 1 then
            -- let's serach in phonebook (not search in case callback for DISA)
            if gvars.application ~= "DISA_B" then
                local dstContactNumber = "" 
                local dstContactName   = ""
                if not gvars.lookup_out then
                    number_normalize_with_lookup(exten, "out", trunkname)
                    channel.NORM_NUMBER:set(gvars.PBContactNumber)  -- used for mobility presence - call to mobility
                end
                
                if not opt_present then -- fix correct caller name and number in case blind transfer (run only first ones, than use previously information)
                    local pbContactName = require"core.httpClient".urlEncode(gvars.PBContactName)
                    dial_opts   = string.format("%sb(predial^extcall^1(%s,%s,%s,%s))", dial_opts, gvars.PBContactNumber, pbContactName, mob_lang, isEmergencyCall and "1" or "")
                    opt_present = true
                end
            end

            local cid_name            = callerInitName
            local cid_number          = gSpace["cidNum"]
            local force_office_num    = office_num_other or office_phone
            local preserveOrigCid     = channel.SIPPEER(trunkname, "preserve_orig_cid"):get() or ""
            local sip_from_number     = channel.SIPPEER(trunkname, "from_number"):get() or ""
            local mobilityCIDRewrite  = channel.mobilityCIDRewrite:get()
            local mobilityForUser     = channel.MOBILITY_4_USER:get()
            local trunkCall           = channel.trunkCall:get()
            local forwardedCall       = channel.FORWARDED:get() or (tran_num and "" ~= tran_num)
            local externalCall        = not ("internal" == channel.CALLTYPE:get() or "internal" == gSpace["gvars"].calltype) or channel["X-TRUNK"]:get() or trunkCall
            local mobilityCall        = mobilityForUser and (mobilityCIDRewrite or trunkCall)
            if gvars.anon_num then
                if tr_type == "gsm" then
                    exten = "#31#" .. exten
                elseif tr_type == "fxo" then
                    exten = "*67#" .. exten
                elseif tr_type == "pri" or tr_type == "bri" then
                    cid_name = "restricted"
                end
            elseif gvars.tr_caller_num then
                cid_number = gvars.tr_caller_num
            elseif "yes" == preserveOrigCid and (mobilityCall or (externalCall and forwardedCall)) then
                -- rewrite CID to destination number in case call from mobile application in GSM mode. Call back to A mobility (show original caller number is enabled) 
                cid_number = mobilityCIDRewrite or cid_number
                -- fix CDR fields for call back to A mobility (show original caller number is enabled)
                if not trunkCall or "" ~= trunkCall then
                    channel.CALLERID("orig_num"):set(cid_attrs.telephoneNumber or "")
                end
            elseif force_office_num and ("sip" == tr_type or "pri" == tr_type or "bri" == tr_type) then
                -- force processing office number
                cid_number = force_office_num
            elseif office_phone then
                cid_number = office_phone
            elseif "" ~= sip_from_number and not string.match(sip_from_number, "\${.+}") then
                --if "from number" set permanent (not like ${num})
                cid_number = sip_from_number
            end
            local number_in_sip = cid_number

            -- Unnormalize caller number for call to trunk
            cid_number = extNumber:unnormalize(cid_number, trunkname)
            channel.CALLERID("name"):set(cid_name)
            channel.CALLERID("num"):set(cid_number)
            print(string.format("Caller data set to '%s' <%s>", cid_name or "", cid_number or ""))
            
            -- Conversion of the some channel values, like number, name, etc.
            if 'sip' == tr_type and trunkname then
                local sip_privacy_header = channel.SIPPEER(trunkname, "privacy_header"):get() or ""
                local sip_cid_header     = channel.SIPPEER(trunkname, "cid_header"):get() or ""
                local sip_from_name      = channel.SIPPEER(trunkname, "from_name"):get() or ""
                local sip_cid_content    = channel.SIPPEER(trunkname, "cid_content"):get() or ""
                local sip_diversion      = channel.SIPPEER(trunkname, "diversion_header"):get() or ""
                local sip_history_info   = channel.SIPPEER(trunkname, "historyinfo_header"):get() or ""
                local historyinfo_reason = channel.SIPPEER(trunkname, "histinfo_reason"):get() or ""
                
                -- save variable for correct processing inside macro substitute
                channel['num']:set(number_in_sip)
                channel['name']:set(cid_name)

                local function _macros_substitute(str)
                    return channel.EVAL(str):get()
                end

                if sip_from_name and sip_from_name ~= '' then
                    local temp = _macros_substitute(sip_from_name)
                    print(string.format('Substitute name in From to %s',temp))
                    channel.CALLERID("name"):set(temp)
                end
                if sip_from_number and sip_from_number ~= '' then
                    local temp = _macros_substitute(sip_from_number)
                    print(string.format('Substitute number in From to %s',temp))
                    channel.CALLERID("num"):set(temp)
                end

                if sip_cid_header and sip_cid_header ~= '' then
                    if not sip_cid_content or sip_cid_content == '' then
                        sip_cid_content = '"${name}" <sip:${num}@${SIPDOMAIN}>'
                    end
                    local text = _macros_substitute(sip_cid_content)
                    if not string.match(text, "@.+>") then
                        print("SIPDOMAIN is empty, assume 127.0.0.1")
                        text = string.gsub(text, "@>", "@127.0.0.1>")
                    end
                    print(string.format('Add PAI header %q',sip_cid_header..': '.. text))
                    channel._SIPADDHEADER_CID:set('X-Wildix-replaceable:' .. sip_cid_header..': '.. text)
                end

                if sip_privacy_header ~= '' and gvars.anon_num then
                    channel._SIPADDHEADER_PRIVACY:set("Privacy: "..sip_privacy_header)
                end

                if (sip_diversion ~= '' or sip_history_info ~= '') and (mobilityForUser or forwardedCall) then -- add Diversion header and History-Info header in case forward or transfer
                    local div_attrs = get_obj_attrs(("" ~= tran_num and tran_num) or mobilityForUser, "telephoneNumber")
                    channel.div_name:set(div_attrs.cn or "")
                    channel.div_num:set(div_attrs.wOfficePhone or div_attrs.telephoneNumber or "")
                    channel.histinfo_reason:set(require"core.httpClient".urlEncode('SIP;cause=302;text="Moved Temporarily"'))
                    local sipDomain = channel.SIPDOMAIN:get() or ""
                    if "" == sipDomain then
                        channel.SIPDOMAIN:set("127.0.0.1")
                    end
                    if sip_diversion ~= '' then
                        channel._SIPADDHEADER_DIVERSION:set("Diversion: ".._macros_substitute(sip_diversion))
                    end
                    if sip_history_info ~= '' then
                        channel._SIPADDHEADER_HISTORY_INFO:set("History-Info: ".._macros_substitute(sip_history_info))
                    end
                end
                if "classound" == trunkname then
                    local xrealcaller = "system"
                    local xrealcallerlicense = "unknown"
                    local transferNum = channel.TRANSFER_NUMBER:get() or ""
                    local forwardFrom = channel.FORWARD_FROM:get() or ""
                    if mobilityForUser and "" ~= mobilityForUser then
                        xrealcaller = mobilityForUser
                    elseif "" ~= transferNum then
                        xrealcaller = transferNum
                    elseif "" ~= forwardFrom then
                        xrealcaller = forwardFrom
                    elseif cid_attrs and cid_attrs.uid and cid_attrs.telephoneNumber then
                        xrealcaller = cid_attrs.telephoneNumber
                        xrealcallerlicense = cid_attrs.licenseType
                    else
                        local cdr_cfrom = channel.CDR("c_from"):get() or ""
                        local ldapData = get_obj_attrs(cdr_cfrom)
                        if ldapData and ldapData.uid and ldapData.licenseType then
                            xrealcaller = cdr_cfrom
                            xrealcallerlicense = ldapData.licenseType
                        end
                    end
                    if "system" ~= xrealcaller and "unknown" == xrealcallerlicense then
                        local ldapData = get_obj_attrs(xrealcaller)
                        xrealcallerlicense = ldapData.licenseType
                    end

                    app.SipAddHeader("X-Real-Caller: " .. xrealcaller)
                    app.SipAddHeader("X-Real-Caller-License: " .. xrealcallerlicense)

                    local callid = channel["X-Forward"]:get() or ""
                    if "" ~= callid then
                        channel.CALLERID("num"):set(gSpace["cidNum"])
                        app.SipAddHeader("X-Forward: " .. callid)
                    end


                    -- Proxying X-Caller-Location header to an outgoing leg of the call.
                    if (wmsperm:getPbxLicense("classound") == "1"
                        and (exten == "911" or exten == "933")
                        and (channel.GLOBAL("GCOUNTRY_CODE"):get() == "1" or string.find(channel.CALLERID("num"):get() or "","^%+?1%d%d%d%d%d%d%d%d%d%d"))
                        and string.find(channel.SIP_HEADER("User-Agent"):get() or "","Wildix Zero Distance")
                        ) then
                        local xCallerLocation = channel.SIP_HEADER("X-Caller-Location"):get() or ""
                        if xCallerLocation == "" then
                            logger:warning("X-Caller-Location header is empty or not present for emergency call")
                        else
                            app.SipAddHeader("X-Caller-Location: " .. xCallerLocation)
                            logger:debug("Proxying the X-Caller-Location header %s to outgoing leg.", xCallerLocation)
                        end
                    end

                    -- add X-Caller-Device header with MAC address or MAC address and FXS port number
                    -- in case of callng to emergency service (911 or 933)
                    local number = channel.NORM_NUMBER:get() or ""
                    local emerg_number = ("911" == number or "933" == number)
                    if emerg_number then
                        local userAgent = channel.SIP_HEADER("User-Agent"):get() or ""
                        repeat
                            if not userAgent:find("Wildix") then
                                break
                            end

                            local ua_splited = require"core.pbxutils".split(userAgent, " ")
                            local dev_model = ua_splited[2]
                            local mac = ua_splited[4]
                            local ipei = ua_splited[5]

                            if not mac or "" == mac then
                                local pbxSerial = require"core.pbxutils".pbxSerial()
                                app.SipAddHeader("X-Caller-Device: " .. pbxSerial)
                                print("Call to Emergency service %s from PBX with serial number %s", number, pbxSerial)
                                break
                            end

                            if dev_model:find("WP%w+R[234]") then -- WP devices
                                app.SipAddHeader("X-Caller-Device: " .. mac)
                                print("Call to Emergency service %s from %s with mac %s", number, dev_model, mac)
                            elseif dev_model:find("W-AIR") then -- W-AIR devices
                                app.SipAddHeader("X-Caller-Device: " .. mac)
                                print("Call to Emergency service %s from %s with mac %s and IPEI %s", number, dev_model, mac, ipei)
                            elseif dev_model:find("FXS") then -- FXS devices
                                local header_from = channel.SIP_HEADER("From"):get() or ""
                                local port = header_from:match(".*%-(%d+)") or "00"
                                port = string.format("%.2d", port)
                                app.SipAddHeader("X-Caller-Device: " .. mac .. port)
                                print("Call to Emergency service %s from %s with mac %s and port %s", number, dev_model, mac, port)
                            end
                        until true
                    end 
                end
            end

            if hg_strategy == "RoundRobin" or hg_strategy == "ReverseRoundRobin" then
                channel.GLOBAL(t_hg_var):set(hg_start+1)
            end
            
            if cid_attrs.userDialplan then
                channel["__TRANSFER_CONTEXT"]:set(cid_attrs.userDialplan)
            end

            -- set CDR used for billing
            if trunkname and (tr_group == "trunk" and exten or number_to_call) then
                pricelist = channel.SIPPEER(trunkname, "pricelist"):get() or ""
                if pricelist ~= "" then
                    channel.CDR("pricelist"):set(pricelist)
                    channel.CDR("norm_dst"):set(tr_group == "trunk" and exten or number_to_call)
                end
            end
            
            -- ISDN transfer capability per call WMS-366
            if "bri" == tr_type or "pri" == tr_type then
                local itc = channel.ITC:get() or ""
                if "3.1Khz" == itc or "unrestricted" == itc or "speech" == itc then
                    if exten and "" ~= exten then
                        if not exten:match(";phone%-context=") then
                            exten = exten .. ";phone-context=" .. itc
                        end
                    end
                    if number_to_call and "" ~= number_to_call then
                        if not number_to_call:match(";phone%-context=") then
                            number_to_call = number_to_call .. ";phone-context=" .. itc
                        end
                    end
                end
            end
            
            app.CELGenUserEvent("TrunkCall", string.format("%s,%s", "outgoing", trunkname))
            channel.GROUP("trunk"):set(trunkname)
            dial_opts = require"utils".extend_dial_options(dial_opts)
            smartApp("Dial", "SIP/" .. trunkname .. "/" .. (tr_group == "trunk" and exten or number_to_call), timeout or gtime or "", dial_opts)
            channel.CDR("pricelist"):set("") -- empty string will be set if call to trunk or call to huntgroup was unsuccesful to avoid unsuccesful call billing 
            if check_hangup() then
                return smartApp("Hangup", hcause.NORMAL)
            end
        end
        
        if tr_group == "huntgroup" then 
            if iteration_num < #hg_content then
                do_iterate = 1
            else
                do_iterate = 0
            end
            iteration_num = iteration_num + 1
            
            if hg_start < #hg_content then
                hg_start = hg_start + 1
            else
                hg_start = 1
            end
        else
            do_iterate = 0
        end
        
        -- for set call limit as reacheable all trunks must be busy (call_limit = 2 indicate that trunk is free)
        hcausecode = channel.HANGUPCAUSE:get() or ""
        if gvars.application == "FAX" and do_dial == 1 then -- handling call limit
            if hcausecode == "34" then
                if call_limit ~= 2 then
                    call_limit = 1  -- set call limit as reachable
                end
            else
                call_limit = 2  -- set call limit as not reachable
            end
        end
        
        dialstatus = channel.DIALSTATUS:get()
        if dialstatus == "BUSY" or dialstatus == "CANCEL" or dialstatus == "NOANSWER" or
            require"core.pbxutils".in_table(hcausecode, {"17", "18", "19", "20", "21", "22", "57"}) then
            do_iterate = 0
        end
    end
    
    if gvars.application == "FAX" then
        -- send correct dialstatus and causecode to parent channel
        if call_limit == 1 then
            if not hcausecode then
                hcausecode = "34"
            end
            local rule = (channel.EXTEN:get() or "")..'@'..(channel.CONTEXT:get() or "")
            save_trunk_limits(rule)
        end
        local fax_gvar_stat = dialstatus.."|"..hcausecode
        channel.GLOBAL(faxnumberid.."_stat"):set(fax_gvar_stat) -- push dialplan and hangup cause to outgoingfax
        app.NoCDR()         -- not write CDR for call to fax, write only sending fax
        channel.CHANNEL("hangup_handler_wipe"):set("") -- clear hangup handler for removing global variables
    end
    
    local blindtran_num = channel["TRANSFER_NUMBER"]:get() or ""
    if blindtran_num ~= "" then
        channel.CALLERID("num"):set(cid_attrs.telephoneNumber or gSpace["cidNum"])
        if cid_attrs.cn then
            channel.CALLERID("name"):set(cid_attrs.cn)
        end
        if last_exten then
            require"internalcalls":return_from_transfer(blindtran_num, tr_group == "trunk" and exten or number_to_call, "CALL", "externalExten")
        else
            return true
        end
    else -- restore name and number for next dialplan iteration
        channel.CALLERID("name"):set(callerInitName)
        channel.CALLERID("num"):set(gSpace["cidNum"])
    end
    return true
end

--- Call through remote pbx
-- @param prefix_exec A arg used for debug message
-- @param ext A extension for the call to remote pbx
-- @param remote_pbx A dn or serial of the remote pbx (writing by wms)
-- @param context A context for the call through remote pbx
-- @param timeout A timeout for the dial
-- @param opts A dial options
function app_remote_call (self, prefix_exec, ext, remote_pbx, context, timeout, opts)
    local app, channel = app, channel
    if not remote_pbx or remote_pbx == "" then
        print ((prefix_exec and prefix_exec or "Remote call:") .. " please choose remote pbx")
        return false
    elseif not context or context:match("^%s*$") then
        print ((prefix_exec and prefix_exec or "Remote call:") .. " please specify remote context")
        return false
    end
    
    print ((prefix_exec and prefix_exec or "Remote call:") .. " pbx ID - " .. remote_pbx .. ", context - " .. context .. (timeout and (", timeout - " .. timeout) or "") .. (opts == "m" and ", music instead ring" or ""))
    local pbx_domain = ""
    local pbx_result = nil
    local pbxDN = nil

    -- TODO use only dn, serial below is workaround for application
    if string.find(remote_pbx, "pbxid=%x+,") then
        pbx_result = wmsldap:get("dn", remote_pbx, "wPbx")
        pbxDN = remote_pbx
    else
        pbx_result = wmsldap:get("pbxSerial", remote_pbx, "wPbx")
        pbxDN = require"core.pbxutils".getPbxDNBySerial(remote_pbx)     
    end

    if pbx_result[1] ~= nil then
        _, cid_pbx_attrs = unpack(pbx_result[1])
        pbx_domain = cid_pbx_attrs.pbxDomain or ""
    end

    if pbx_domain == "" then
        local msg = "Cannot get PBX IP by given ID " .. remote_pbx
        print(msg)
        logger:error(msg)
        to_logfile ("Hangup", hcause.FAILURE)
        return app.Hangup(hcause.FAILURE)
    end

    require"peerCallClass":setUp2Channel("internal")
    if not timeout or timeout == nil or timeout == "" then
        timeout = channel["GINTDIAL_TIMEOUT"]:get() or ""
    end
    
    -- getting options for call to mobility through remote pbx
    local gvars, mob_opts = gSpace["gvars"], ""
    if gvars.mobility_caller or gvars.mobility_lang or gvars.mobility_owner or gvars.mobility_number then
        mob_opts = "MOBILITY|" .. (gvars.mobility_caller or "") .. "|" .. (gvars.mobility_lang or "") .. "|" .. (gvars.mobility_owner or "") .. "|" .. (gvars.mobility_number or "")
        if gvars.mobility_number then
            channel.NORM_NUMBER:set(gvars.mobility_number)  -- used for mobility presence
        end
    end

    -- processing changing identity
    if gSpace["cidAttrs"].uid then
        local identity_num, identity_name = __checkIdentity(gSpace["cidAttrs"])
        if identity_num and identity_name then
            channel.CALLERID("num"):set(identity_num)
            channel.CALLERID("name"):set(identity_name)
        end
    end

    local is_trunkcall = channel["X-TRUNK"]:get() or ""
    is_trunkcall = is_trunkcall ~= ""
    local wms_opts = require"wmsNetCall".makeArgs({
             pbxid    = require"core.strConverter".toB64(pbxDN) or ""
            ,ctx      = context
            ,calltype = is_trunkcall and gvars.calltype or ""
            ,app      = mob_opts
            ,transfer = channel["TRANSFER_NUMBER"]:get() or ""
            ,linkedid = channel.LINKEDID:get() or ""
        })

    channel["CALL_THROUGH_REMOTE_PBX"]:set("yes")

    return smartApp("Dial", "SIP/" .. ext .. "@" .. pbx_domain, timeout, (opts or "") .. wms_opts)
end

function app_dialremotepbx(self, prefix_exec, ext, remote_pbx, context, timeout, opts)
    app_remote_call(self, prefix_exec, ext, remote_pbx, context, timeout, opts)
    local unavailable_pbx = channel["X-PBX_UNAVAILABLE"]:get() or ""
    if unavailable_pbx ~= "" then
        local pbxDN
        if string.find(remote_pbx, "pbxid=%x+,") then
            pbxDN = remote_pbx
        else
            pbxDN = require"core.pbxutils".getPbxDNBySerial(remote_pbx)
        end
        local failoverObj = require"common.failover":new(pbxDN)
        if failoverObj and "1" == failoverObj:isEnabled() then
            local msg = "Call through remote pbx was unsuccesful, let's try to call through failover pbx"
            print(msg); logger:info(msg)
            app.SipAddHeader("DST-PBX: " .. require"core.pbxutils".getPbxSerialByDN(pbxDN))
            app_remote_call(self, prefix_exec, ext, failoverObj:getPbxDN(), context, timeout, opts)
        end
    end
    return true
end

--- Function for leave voicemail and access to voicemail
-- @param vmbox     A dest mailbox
-- @param vmtype    A type indicate this leave voicemail (leave) or access to voicemail (access)
-- @param opts      A options for voicemail getting from dialplan, supported values:
--                  "" (default) - check authorization (vmbox ~= cid)
--                  "s"          - skip authorization checking
function app_voicemail(self, vmbox, vmtype, opts)
    local app, channel = app, channel
    local hcause, ivr = hcause, ivr
    local diskutils = require"core.diskutils"
    local opts = opts or ""

    -- enable jitter buffer
    app.Set("JITTERBUFFER(adaptive)=500,1000,40")

    -- set to channel correct language
    local language = string.lower(channel.CHANNEL("Language"):get() or "en")
    language = require"utils":get_lang_from_pack(language)
    channel.CHANNEL("Language"):set(language)
    channel.SERVICECALL:set("true")

    local mount_voicemail = diskutils.is_mount("/var/spool/callweaver/voicemail")
    if "access" == vmtype then
        -- check share for voicemail
        if not mount_voicemail then
            local msg = "No device is mounted for voicemails"
            print(msg); logger:error(msg)
            ivr:tell("Service currently unavailable. Please contact administrator.")
            return smartApp("Hangup", hcause.FAILURE)
        end
        
        local cid_features  = gSpace["cidFeatures"]
        local cid           = "userFeatures" == cid_features.type and cid_features:get("SHAREDVM") or gSpace["cidNum"]
        vmbox               = vmbox or cid
        local vmboxUserAttr = {}
        local options       = {
            validate = function(vmbox)
                if not vmbox or not vmbox:find("^(%d+)$") then
                    return false
                end
                if vmbox ~= gSpace["cidNum"] then
                    vmboxUserAttr = get_obj_attrs(vmbox)
                    return vmboxUserAttr.telephoneNumber
                else
                    vmboxUserAttr = gSpace["cidAttrs"]
                    return true
                end
            end;
            failure = function(vmbox)
                ivr:tell("Number is ".. (vmbox ~= "" and "invalid" or "empty") ..", please try again")
            end;
        }

        if not options.validate(vmbox or "") then
            if vmbox and vmbox ~= "" then
                print("Number " .. vmbox .. " is invalid")
            end
            vmbox = ivr:read("Please enter phone number, followed by pound", options)
        end
        
        if channel.FAILOVER_USER:get() == "yes" then -- marker for failover user
            local cid_attr = gSpace["cidAttrs"] or {}
            local failoverObj = require"common.failover":new(cid_attr.pbxowner)
            if not failoverObj:localPbxIsFailover() then
                print("Voicemail access for failover user %q disallowed", vmbox)
                logger:warning("Voicemail access for failover user %q disallowed", vmbox)
                return smartApp("Hangup", hcause.FAILURE)
            end
        end
        if not vmbox then
            ivr:tell("Good bye. Hangup")
            return smartApp("Hangup", hcause.NORMAL)
        end
        
        local permissions = require"common.permissions"
        if opts ~= "s" and (vmbox ~= cid or "yes" ~= permissions:check(vmboxUserAttr["owner"] or "", "canuse", "VOICEMAIL_WITHOUT_PIN", nil, vmboxUserAttr["licenseType"])) then
            local userAuth = require"userAuthorize":new()
            if not userAuth:process(vmbox) then
                return smartApp("Hangup", hcause.NORMAL)
            end
        end
        
        if "yes" ~= permissions:check(vmboxUserAttr["owner"] or "", "canuse", "VOICEMAIL", nil, vmboxUserAttr["licenseType"]) then
            print("Voicemail rejected by group permissions")
            return smartApp("Hangup", hcause.ACL_FORBIDDEN)
        end

        return smartApp("VoiceMailMain", vmbox, opts)
    elseif "leave" == vmtype then
        -- if not from internalcall go to LDAP for checking
        if vmbox == "" or vmbox == nil then
            print("No vmbox for application voicemail")
        end

        local email
        local vmBoxStr = ""
        -- iterate by mailbox in voicemailbox list
        for _, mailbox in pairs(require"core.pbxutils".split(vmbox or "", "[%s,;]+")) do
            local isNotEmailAddr = false
            mailbox = string.gsub(mailbox, "['\"]", "") -- cut quotes and double quotes
            if not mailbox:match("^(.*)@(.*)%.(.*)") then
                isNotEmailAddr = true
                local vm_result = get_obj_attrs(mailbox)
                if vm_result and vm_result.uid then
                    local vm_features = require"common.userFeatures.userFeatures":new(vm_result.uid)
                    if vm_features then
                        vm_features:load()
                        local vm_number = vm_features:get("SHAREDVM")
                        if vm_number then
                            mailbox = vm_number
                            vm_result = get_obj_attrs(mailbox)
                        end
                    end
                end
                if "yes" ~= require"common.permissions":check(vm_result["owner"] or "", "canuse", "VOICEMAIL", nil, vm_result["licenseType"]) then
                    print("Voicemail rejected by group permissions for ", mailbox)
                    return false
                end
            end
            if vm_result then
                local pbxDN = require"core.pbxutils".getPbxDNBySerial()
                if vm_result.pbxowner and pbxDN ~= vm_result.pbxowner then
                    print(string.format("Voicemail of the remote user %q disallowed", mailbox))
                    return smartApp("Hangup", hcause.FAILURE)
                end
                if not vm_result.telephoneNumber then
                    print(string.format("Vmbox %q is not found in LDAP", mailbox))
                    return false
                end
                email = vm_result.mail
            end
            -- prepare mailbox string
            -- add @ to extension in mixed mode (extension and email address used simultaneously) if mailbox is not an email address
            vmBoxStr = vmBoxStr..("" ~= vmBoxStr and "&" or "")..mailbox..(isNotEmailAddr and "@" or "")
        end
        if opts:find("internalcalls|") then
            opts = opts:match("internalcalls|(.*)")
            opts = opts or ""
        elseif opts == "vmtransfer" then 
            opts = "u"
        else
            opts = opts == "" and "s" or opts
        end

        -- check share for voicemail
        if not mount_voicemail then
            if 10240 > diskutils.get_avail_ram_space() then
                local msg = "No device is mounted for voicemails and /dev/shm is full"
                print(msg); logger:error(msg)
                return smartApp("Hangup", hcause.FAILURE)
            end
            if not email or email == "" then
                local msg = "No device is mounted for voicemails and no email for dst user"
                print(msg); logger:error(msg)
                ivr:tell("Voicemail");
                ivr:tell("vm-isunavail"); --TODO(sounds): make alias for this sound in tvoice
                return smartApp("Hangup", hcause.FAILURE)
            end
        end

        if opts == "a" then
            local dialstatus = channel.DIALSTATUS:get() or ""
            if dialstatus == "BUSY" then
                opts = "b"
            else
                opts = "u"
            end
        end
        if opts:find("b") or opts:find("u") then
            opts = opts .. "o"    -- skip by default vm-intro if greeting used
        end

        local orig_c_to = channel.CDR("c_to"):get() or ""
        smartApp("VoiceMail", vmBoxStr, opts)

        if "USEREXIT" == channel.VMSTATUS:get() then  -- this is a fix of a problem in callweaver:
            channel.CDR("c_to"):set(orig_c_to)        -- voicemail application rewrites cdr(c_to) (these are Wildix changes) and it's wrong in some cases
        end

        return true
    else
        print("Invalid type for voicemail application")
        logger:error("Invalid type for voicemail application:", vmtype)
        return false
    end
end

--- Operation with fax file
-- Checking if backup directory is available for the user and
-- moving fax file to it otherwise just to remove fax file
-- @param filename A string with absoulute path to fax file
-- @param userdir A string with user directory (extenNumber)
local function cp_faxfile (filename, userdir)
    if not filename or "" == filename then
        return false
    end
    
    -- check mounted FAX share
    if not require"core.diskutils".is_mount("/mnt/faxes") then
        local err_msg = string.format("Faxes device is not mounted. %s is removed", filename)
        print(err_msg); logger:error(err_msg)
        return false
    end

    -- default received fax save path
    local save_path = (userdir and userdir ~= "") and userdir or "OTHERS"  
    
    -- check global directory settings
    local fax_share_part   = nil
    local fax_receive_fold = nil
    local fax_receive_dir  = channel["FAX_RECEIVE_DIR"]:get() or ""
    if "" ~=  fax_receive_dir then
        fax_share_part, fax_receive_fold = fax_receive_dir:match("(/mnt/faxes)/?(.*)")
        if fax_share_part then
            save_path = fax_receive_fold or ""
        else
            print("FAX_RECEIVE_DIR (%s): not included in '/mnt/faxes/'", fax_receive_dir)
            logger:warning("FAX_RECEIVE_DIR (%s): not included in '/mnt/faxes/'", fax_receive_dir)
        end
    end

    -- let's make copying
    local rc       = false
    local err      = "Unknown status"
    local save_dir = "/mnt/faxes/" .. save_path
    local fsutils  = require"core.fsutils"
    
    -- check destiantion directory
    if not fsutils.is_dir(save_dir) then
        rc, err = fsutils.mkdir(save_dir)
        if not rc then
            local err_msg = string.format("Could not create directory %s", save_dir)
            print(err_msg); logger:error("%s(%s). %s is removed.", err_msg, err, filename)
            return false
        end
    end
    
    -- copy fax file to destination path
    rc, err = fsutils.cpfile(filename, save_dir)
    if not rc then
        local err_msg = string.format("Could not copy %s to %s", filename, save_dir)
        print(err_msg); logger:error("%s(%s). %s is removed.", err_msg, err, filename)
        return false
    end
    
    print(string.format("%s was copied to %s", filename, save_dir))
    
    -- update CDR field
    local lastcdrdata = faxes:getcdr()
    if lastcdrdata ~= "" then
        local file_path = save_path .. filename:match(".*/tmp(/.*)$")
        faxes:setcdr(lastcdrdata:gsub("^([^|]*)||(.*)", "%1|" .. file_path .. "|%2"))
    end
    return true
end

--- Procedure to recieve a fax and send pdf file to mail address
-- @param param A string with either user extenNumber entry or a mail address
-- @param format A string with format to convert recived fax and attach to email. (tiff/pdf)
function fax2mail (param, format, localstationid)
    local util = require"utils"
    local pbxutils = require"core.pbxutils"
    local diskutils = require"core.diskutils"
    local faxes, app, hcause = faxes, app, hcause
    local format = format or "pdf"
    local destNumber, destEmails = {}, {}
    local dst_attrs, user_number = {}, ""
    print ("Executing 'Fax to e-mail': email/user - " .. param .. ", format - " .. format .. (localstationid and (", id fax - " .. localstationid) or ""))
    
    local license_fax = channel.LICENSE("LICENSE_FAX"):get() -- fax license
    if license_fax ~= "Enable" then
        print("Licensing restriction")
        logger:error("Licensing restriction")
        to_logfile("Hangup", hcause.ACL_FORBIDDEN)
        return app.Hangup(hcause.ACL_FORBIDDEN)
    end
    
    local cid = gSpace["cidNum"]
    local cname = channel.CALLERID("name"):get() or ""
    
    parama = util:split(param, "|")
    for i, param in ipairs(parama) do
        -- Look up destination number by e-mail
        if param:find("%@") then
            dst_attrs = get_obj_attrs(param, "mail")
            user_number = dst_attrs.telephoneNumber or ""
            
            destNumber[#destNumber+1] = user_number
            destEmails[#destEmails+1] = param
        else
        -- Look up e-mail number by destination number
            destNumber[#destNumber+1] = param
            dst_attrs = get_obj_attrs(param)
            if dst_attrs.mail and dst_attrs.mail ~= "" then
                destEmails[#destEmails+1] = dst_attrs.mail
            end                 
        end
        -- Fix CDR fields for local fax
        if i == 1 then
            if gSpace["gvars"].FAXRELAY_MODE == "no-mail" then  -- push NoCDR in case FAXRELAY (no-mail)
                to_logfile("NoCDR", "")
                app.NoCDR()
            else
                local chan_name = channel.CHANNEL:get() or ""
                if dst_attrs.cn or chan_name:match("^Local/") then
                    channel.CDR("c_to"):set(destNumber[i] or "")
                    channel.CDR("to_name"):set(dst_attrs.cn or "")  -- set name of the first user from param
                end
                if chan_name:match("^Local/") then  -- from CTI to Fax2mail at local pbx
                    channel.CDR("c_from"):set(cid or "")
                    channel.CDR("from_name"):set(cname)
                end
            end
        end
    end
    
    if dst_attrs.telephoneNumber then
        channel.CONNECTEDLINE("num"):set(dst_attrs.telephoneNumber)
    end
    -- set language for body message
    if dst_attrs.preferredLanguage then
        local language = dst_attrs.preferredLanguage or "en"
        language = util:get_lang_from_pack(language)
        channel.CHANNEL("Language"):set(language)
    end

    -- make destEmails and extenNumber tables with unique values
    destEmails = util:table_unique(destEmails)
    destNumber = util:table_unique(destNumber)
    
    -- RECEIVING FAX
    local fsutils = require"core.fsutils"
    local diskutils = require"core.diskutils"
    local tmpfaxpath = nil
    local tmpfaxname = nil
    local isConfiguredFaxShare = diskutils.is_configured("/mnt/faxes")
    if isConfiguredFaxShare then
        local faxShare = diskutils.is_mount("/mnt/faxes")
        if faxShare then
            tmpfaxpath = "/mnt/faxes/tmp/"
            tmpfaxname = tmpfaxpath .. "lua_" .. math.random(os.time())
            if not fsutils.is_dir(tmpfaxpath) then fsutils.mkdir(tmpfaxpath) end
        else
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            return app.Hangup(hcause.CHANUNAVAIL)
        end
    else
        tmpfaxpath = "/tmp/"
        tmpfaxname = tmpfaxpath .. "lua_" .. math.random(os.time())
        -- check space on /dev/shm before receiving a fax, at least 10M
        if 10240 > diskutils.get_avail_ram_space() then
            local msg = "Fax receive error. No space left on memory device"
            print(msg); logger:error(msg)
            faxes:setcdr("ERR||%s|%s", cid, "No space left on memory device to receive a fax")
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            return app.Hangup(hcause.CHANUNAVAIL)
        end
    end
    
    local localstationid = localstationid or "WILDIXFAX"
    localstationid = string.gsub(string.upper(localstationid), "[^%u%d%+%-_]", "")
    localstationid = localstationid:sub(1,20)
    channel.FAXOPT("localstationid"):set(localstationid)
    
    local start_time = os.time()
    
    -- enable debug (use with enable option fax in logger.conf)
    local faxdebug = channel.FAXDEBUG:get() or ""
    faxdebug = faxdebug == "yes" and "d" or ""
    
    -- receiving a fax
    to_logfile("ReceiveFax", tmpfaxname, faxdebug)
    app.ReceiveFax(tmpfaxname, 'f'..faxdebug)
    
    local duration = os.difftime(os.time(), start_time)
    print("Duration", duration)
    
    local faxpages = tonumber(channel.FAXOPT("pages"):get() or "") or 0
    local faxresolution = tonumber(channel.FAXOPT("resolution"):get() or "") or 0
    local faxstatus = channel.FAXOPT("status"):get()
    local faxstatusstr = channel.FAXOPT("statusstr"):get() or ""
    local remotestationid = channel.FAXOPT("remotestationid"):get() or ""
    
    local msubj, mbody = "", ""
    if cname ~= "" and cname ~= cid then
        cname = '"'..cname..'"' -- update format for notifications
    else
        cname = ""
    end
    local faxname = string.format("%s%s_%s_%s.tiff",tmpfaxpath, os.date('%Y.%m.%d-%H.%M.%S'), faxpages, math.random(os.time()))
    fsutils.mvfile(tmpfaxname, faxname)
    
    -- Check for faxes share mounted if nessesary
    if format == "pdf" and not diskutils.is_mount("/mnt/faxes") then
        print("Fax receive error. Faxes share not mounted. Conversion to PDF are not allowed.")
        logger:error("Fax receive error. Faxes share not mounted. Conversion to PDF are not allowed.")
        msubj = _t("msubj.FAXNOTRECEIVED", cname, cid)
        mbody = _t("mbody.FAXNOSHARE")
        pbxutils.mailto(mbody, msubj, table.concat(destEmails, " "), faxname)
        fsutils.rmfile(faxname)
        to_logfile("Hangup", hcause.CHANUNAVAIL)
        return app.Hangup(hcause.CHANUNAVAIL)
    end
    
    if faxstatus ~= "SUCCESS" then
        print("Fax receive error. Fax transfer was not successful")
        faxes:setcdr("ERR||%s|%s", cid, "Fax transfer was not successful")
        
        if fsutils.filesize(faxname) ~= 0 then
            msubj = _t("msubj.FAXRECEIVEDPARTIAL", cname, cid)
            mbody = mbody .. _t("string.PARTIALRECEIVED") 
            pbxutils.mailto(mbody, msubj, table.concat (destEmails, " "), faxname)
            
            -- make copy for all destinations
            for i, numberEach in ipairs(destNumber) do
                cp_faxfile(faxname, numberEach)
            end
        end
        
        fsutils.rmfile(faxname)
        to_logfile("Hangup", hcause.CHANUNAVAIL)
        return app.Hangup(hcause.CHANUNAVAIL)
    end
    
    print("Faxfile", faxname)
    if fsutils.filesize(faxname) == 0 then
        print("Fax receive error. Zero length file size was received. Fax transfer was not successful")
        logger:warning("Fax receive error. Zero length file size was received. Fax transfer was not successful")
        faxes:setcdr("ERR||%s|%s", cid, "Zero length file size was received")
        fsutils.rmfile(faxname)
        to_logfile("Hangup", hcause.CHANUNAVAIL)
        return app.Hangup(hcause.CHANUNAVAIL)
    end
    
    local attachname
    if faxname:match(".*-.*_%d+\.(.*)$") ~= format then
        local convert_result = pbxutils.run("chrt -o 0 wms_attach_converter %q %s 2>/dev/null", faxname, format)
        if convert_result then
            attachname = convert_result:match("(.-)%s*$")
            print("Converted", attachname)
        else
            print(string.format("Fax %s convert error to format %s", faxname, format))
            attachname = faxname
        end
    else
        attachname = faxname
    end
    local faxpages_bad = tonumber(channel.FAXOPT("bad_pages"):get() or "") or 0
    if faxpages_bad > 0 then
        print(string.format("Fax received successfully (with %d error pages)!", faxpages_bad))
    else
        print("Fax received successfully!")
    end
    
    -- check backup mailbox GBKMAILBOX_0
    local bkmailbox = channel["GBKMAILBOX_0"]:get()
    if bkmailbox and bkmailbox ~= "" then
        destEmails[#destEmails+1] = bkmailbox
    end
    
    local emails = table.concat(destEmails, " ")
    if gSpace["gvars"].FAXRELAY_MODE ~= "no-mail" then
        msubj = _t("msubj.FAXRECEIVED", cname, cid)
        if faxpages_bad > 0 then
            mbody = _t("mbody.FAXRECEIVEDERR", os.date("%Y-%m-%d %X"), remotestationid, cname, cid, faxpages, faxpages_bad, faxstatusstr)
        else
            mbody = _t("mbody.FAXRECEIVED", os.date("%Y-%m-%d %X"), remotestationid, cname, cid, faxpages, faxstatusstr)
        end
        pbxutils.mailto(mbody, msubj, emails, attachname)
    end
    
    if not gSpace["gvars"].FAXRELAY then
        local t = {}
        faxes:setcdr("OK||%s|%s", cid, faxpages)
        __send_to_s3(attachname, destNumber[1])
        for i=1, #destNumber do
            cp_faxfile(attachname, destNumber[i])
            t[#t+1] = '"' .. destNumber[i] .. '"'
        end
        fsutils.rmfile(attachname)
    else
        faxes:setcdr("OK||%s|%s", cid, faxpages)
    end
    if gSpace["gvars"].FAXRELAY or gSpace["gvars"].SENDDOC then
        return attachname, destEmails[1]
    else
        to_logfile("Hangup", hcause.NORMAL)
        return app.Hangup(hcause.NORMAL) 
    end
end

--- Function for execute application Queue
-- @param prefix_exec       A param used for debug message
-- @param call_group_id     A id of call group
-- @param last_member       A last member in list of extension(0) or no(1) (set up when compile wms dialplan)  
----- options for app.Queue (see core show application queue for details)
-- @param opts              A options for queue application
-- @param url               A string <URL> sent to the called party if the channel supports it
-- @param announceoverride  A value allows you to override the announcement specified in queues.conf
-- @param timeout           A Will cause the queue to fail out after this time
-- @param mohclass          A set music on hold
function app_queue_call(prefix_exec, call_group_id, opts, url, announceoverride, timeout, mohclass, last_member)
    local app     = app
    local channel = channel
    if prefix_exec then
        print(prefix_exec .. ": group id - " .. call_group_id .. (announceoverride and (", announceoverride - " .. announceoverride) or "") .. (timeout and (", timeout - " .. timeout) or "") .. (mohclass and (", mohclass - " .. mohclass) or ""))
    end
    
    if channel.QUEUE_EXISTS(call_group_id):get() == "0" then
        print("Incorrect call group id", call_group_id)
        logger:debug("Incorrect call group id", call_group_id)
        return false
    end
    
    local connection = require"core.wsqlite":new("wmsdb")
    local sql = "SELECT call_groupname FROM call_groups WHERE id = %s;"
    local cursor = connection:execute(sql, call_group_id)
    local cgname = cursor:fetch(); cursor:close()
    connection:close()
    app.CELGenUserEvent("CallGroup", string.format("%s,%s", cgname or "", call_group_id or ""))

    -- put managers of queue to postaction space
    local manager = string.match(channel.QUEUE_MEMBER(call_group_id, "manager"):get() or "", "^%w+/(%d+)")
    local mwi_manager = channel.QUEUE_MEMBER(call_group_id, "mwi_manager"):get()
    setpostaction {
        ["name"] = "QUEUE",
        ["manager"] = manager,
        ["mwi_manager"] = mwi_manager,
        ["mwi_email_notify"] = "yes" == channel.CG_MANAGER_MAIL_NOTIFY:get()
    }
    channel.CHANNEL("hangup_handler_wipe"):set("hdlr,service,1")
    
    local mohclass  = mohclass or ""
    local opts      = (opts or "") .. "C"
    if mohclass and mohclass ~= "" then
        smartApp("Answer", ",nocdr")
    else
        mohclass = nil
        opts     = opts .. "r"
    end
    
    -- processing tag about last application in the dialplan rule
    if 1 == last_member then
        opts = opts .. "E"
    end
    
    -- Pre-executing call processing
    local doExecApp = true
    if "0" == channel.QUEUE_VARIABLES(call_group_id):get() then
       -- checking limit of the waiting queue calls
       local queueMaxCalls     = channel.QUEUEMAX:get() or ""
             queueMaxCalls     = tonumber(queueMaxCalls) or 0
       local queueWaitingCalls = channel.QUEUE_WAITING_COUNT(call_group_id):get() or ""
             queueWaitingCalls = tonumber(queueWaitingCalls) or 0
       if queueWaitingCalls >= queueMaxCalls then
           doExecApp = false
           print("Unable to call to queue. Maximum count of the waiting calls to queue is reached:", queueMaxCalls)
           logger:info("Unable to call to queue. Maximum count of the waiting calls to queue is reached:", queueMaxCalls)
       end
    else
        print("Error loading call group environment")
        logger:error("Error loading call group environment")
    end
    
    if doExecApp then

        local ringtone = channel.RINGTONE:get() or ""
        if "" ~= ringtone then
            print(string.format("Using ringtone '%s' from dialplan variable for group %s", ringtone, call_group_id))
            local calltype = gSpace["gvars"].calltype
            local ring = require"ringTone".getAlertHeader(ringtone, calltype)
            channel._SIP_ALL_ADDHEADER_RINGER = ring
        end

        if "yes" == channel.SKIP_PUSH_NOTIFICATION:get() then
            channel.__SIPADDHEADER_SKIP_PUSH:set("X-Skip-Push: yes")
        end

        channel["__DISABLE_DIRECTRTP"]:set(1)


        -- Processing identity feature
        local cid = channel.CALLERID("num"):get() or ""
        local to_identity = channel['TO_IDENTITY']:get() or ""
        local custom_identity, identity_name
        if gSpace["cidAttrs"].uid then
            custom_identity, identity_name = __checkIdentity(gSpace["cidAttrs"])
        end
        if "" ~= to_identity and not custom_identity then -- a call to identity number
            local identity_attrs = get_obj_attrs(to_identity)
            if identity_attrs.uid then
                local identity_name = identity_attrs.cn or ""
                local b64 = require"core.strConverter"
                local c_from_b64 = b64.toB64(cid) or ""
                local c_to_b64 = b64.toB64(to_identity) or ""
                local to_name_b64 = b64.toB64(identity_name)
                channel.CDR("c_from"):set(string.format("%s.%s", to_identity, cid))

                local fork_params = string.format("vTw(c_from=%s,c_to=%s,to_name=%s)", c_from_b64, c_to_b64, to_name_b64)
                app.ForkCDR(fork_params)
                channel.CALLERID("num"):set(string.format("%s.%s", to_identity, cid))
                channel.CONNECTEDLINE("all"):set(string.format([===["%s" <%s>]===], identity_name, identity_attrs.telephoneNumber))
                app.SipAddHeader(string.format("Caller-Image: http://127.0.0.1/user_avatar/%s.png", cid))
            end
        elseif custom_identity and "" == to_identity then -- a call using identity
            local dst = channel.EXTEN:get() or ""
            local orig_cid = cid
            cid = custom_identity
            channel.CDR("c_from"):set(cid)
            channel.CDR("from_name"):set(identity_name or "")
            local b64 = require"core.strConverter"
            local c_from_b64 = b64.toB64(orig_cid) or ""
            local c_to_b64 = b64.toB64(string.format("%s.%s", custom_identity, dst)) or ""
            local to_name = dst
            local to_name_b64 = b64.toB64(to_name) or ""

            local fork_params = string.format("vTw(c_from=%s,c_to=%s,to_name=%s)", c_from_b64, c_to_b64, to_name_b64)

            app.ForkCDR(fork_params)
            channel.CDR("c_from"):set(orig_cid or "")
            channel.CALLERID("num"):set(cid)
            channel.CALLERID("name"):set(identity_name)
            channel.CONNECTEDLINE("all"):set(string.format([===["%s" <%s>]===], dst, string.format("%s-%s", custom_identity, dst)))
            app.SipAddHeader(string.format("Caller-Image: http://127.0.0.1/user_avatar/%s.png", cid))
        elseif custom_identity and "" ~= to_identity then -- a call using identity number to another identity
            local identity_attrs = get_obj_attrs(to_identity)
            local to_identity_name = ""
            if identity_attrs.uid then
                to_identity_name = identity_attrs.cn or ""
            end
            local b64 = require"core.strConverter"
            local c_from_b64 = b64.toB64(cid)
            local c_to_b64 = b64.toB64(string.format("%s.%s", custom_identity, to_identity))
            local to_name_b64 = b64.toB64(to_identity_name)
            local from_name_b64 = b64.toB64(identity_name)

            local fork_params = string.format("vTw(c_from=%s,c_to=%s,to_name=%s,from_name=%s)", c_from_b64, c_to_b64, to_name_b64, from_name_b64)


            channel.CDR("c_from"):set(string.format("%s.%s", to_identity, custom_identity))
            channel.CDR("from_name"):set(identity_name)

            app.ForkCDR(fork_params)

            channel.CALLERID("num"):set(string.format("%s.%s", to_identity, custom_identity))
            channel.CALLERID("name"):set(identity_name)
            channel.CONNECTEDLINE("all"):set(string.format([===["%s" <%s.%s>]===], to_identity_name, custom_identity, to_identity))
            app.SipAddHeader(string.format("Caller-Image: http://127.0.0.1/user_avatar/%s.png", custom_identity))
        end

        return smartApp("Queue", call_group_id, opts, url, announceoverride, timeout, nil, nil, nil, nil, nil, mohclass)
    end
    return true
end

--- Record call
-- @mailaddr A mail or extension for record call
-- @param rtype A type of record (pbx or personal)
-- @param opts A options for MixMonitor
function app_record_call (mailaddr, rtype, opts)
    local app, channel = app, channel
    local diskutils = require"core.diskutils"
            
    print("Executing 'Record call': record type - " .. rtype .. (mailaddr and (", email/user - " .. mailaddr) or "") .. (opts and (", options - " .. opts) or "")) 
    
    --- Send originate to user record_notify context
    local function notification_originate(state)
        if state ~= "start" and state ~= "stop" then
            print("Invalid record notification state -", state)
            return false
        end
       
        local recid = channel.RECID:get() or ""
        if not recid or recid == "" then
            print("Record UID not found")
            return false
        end

        -- Check needed context and extensions
        local notify_ctx = "record_notify"
        if not extensions[notify_ctx] or not extensions[notify_ctx]["start"] or not extensions[notify_ctx]["start"] then
            logger:debug("Record notification context not found or wrong extensions. Feature is disabled")
            return false
        end
       
        local chan_name = channel.CHANNEL:get() or ""
        local chan_spy = chan_name:match(".*%/(.*)") or gSpace["cidNum"]
       
        -- create callfile
        local callfile = "/var/spool/callweaver/outgoing/record_notify_" .. recid .. ".call"
        local cf = require"core.rawlazy.callfile".call_obj
        cf:add("Channel", "Local/" .. chan_spy .. "@connecttorecordcall") 
        cf:add("CallerID", "record notify")
        cf:add("Extension", state)
        cf:add("Context", notify_ctx)
        cf:add("Priority", "1")
        cf:add("Set", "RECID", recid)
        local _, err = cf:writefile(callfile, true) -- create new call file
        if err then 
            print("Error while creating callfile %s: %s", callfile, err)
            logger:error("Error while creating callfile %s: %s", callfile, err)
        end
        logger:debug("Callfile %q was created for record notify", callfile)
    end
    
    -- WMS-5683
    local app_callback = channel["APPCALLBACK"]:get()
    if not app_callback or "" == app_callback then
        app_callback = nil
    end
    if app_callback then
        channel["APPCALLBACK"]:set("")
        local chan_name = channel["CHANNEL"]:get() or ""
        if ";2" == chan_name:sub(-2) then
            print("Skip recording for callback from app")
            return true
        end
    end

    -- Processing feature codes Personal record and Pause recording
    local ami = require"core.AMIClient":new()
    local isCurrentChanRecord = "true" == channel.XAD:get()
    local isAnotherChanRecord = not isCurrentChanRecord and "true" == channel.IS_SYSTEM_RECORD:get() and "true" == ami:getVar("XAD", channel["BRIDGEPEER"]:get())
    local isRecord = isCurrentChanRecord or isAnotherChanRecord

    local function setRecordingVariable(varname, varvalue)
        if isCurrentChanRecord then
            channel[varname] = varvalue
        elseif isAnotherChanRecord then
            local _, err = ami:setVar(varname, varvalue, channel["BRIDGEPEER"]:get())
            if err then
                logger:warning("AMI error:", err)
            end
        end
    end

    local function getRecordingVariable(varname)
        if isCurrentChanRecord then
            local rv = channel[varname]:get()
            return (rv and "" ~= rv) and rv
        elseif isAnotherChanRecord then
            local rv, err = ami:getVar(varname, channel["BRIDGEPEER"]:get())
            if err then
                logger:warning("AMI error:", err)
            end
            return (rv and "" ~= rv) and rv
        end
    end

    local function is_pbx_record()
        local rv = false
        local rtype = getRecordingVariable("IS_SYSTEM_RECORD") or ""
        if "true" == rtype then
            rv = true
        end
        return rv
    end

    if isRecord then
        if rtype == "pbx" then
            if is_pbx_record() then
                local msg = "Record has been already started via dialplan. It can't be started again"
                print(msg); logger:debug(msg)
                return true
            end
        elseif rtype == "personal" then
            if is_pbx_record() then
                local msg = "Record was started via dialplan. It can't be started again"
                print(msg); logger:debug(msg)
                to_logfile("Playback", "beeperr")
                return app.Playback("beeperr")
            end
            setRecordingVariable("XAD", "false")
            setRecordingVariable("MonitorPaused", "false")
            notification_originate("stop")
            to_logfile("Playback", "beep")
            app.Playback("beep")
            to_logfile("StopMonitor", "")
            return app.StopMonitor()
        elseif rtype == "pause" then
            if is_pbx_record() then
                local can_pause = getRecordingVariable("USER_CAN_PAUSE_RECORDING") or "yes"
                if "yes" ~= can_pause then
                    print("Record was started via dialplan. It can't be paused")
                    app.Playback("beeperr")
                    return true
                end
            end
            local MonitorPaused = getRecordingVariable("MonitorPaused") or "false"
            local recChanName
            if isCurrentChanRecord then
                recChanName = channel.CHANNEL("name"):get()
            elseif isAnotherChanRecord then
                recChanName = channel.BRIDGEPEER:get()
            end
            if "true" == MonitorPaused then
                local rv, err = ami:unpauseMonitor(recChanName)
                if not rv then
                    print("Can't execute UnpauseMonitor on channel %s via AMI: %s", recChanName, err)
                    logger:error("Can't execute UnpauseMonitor on channel %s via AMI: %s", recChanName, err)
                    return true
                end
                setRecordingVariable("MonitorPaused", "false")
                app.Playback("recordingstarted")
                return true
            else
                local rv, err = ami:pauseMonitor(recChanName)
                if not rv then
                    print("Can't execute PauseMonitor on channel %s via AMI: %s", recChanName, err)
                    logger:error("Can't execute PauseMonitor on channel %s via AMI: %s", recChanName, err)
                    return true
                end
                setRecordingVariable("MonitorPaused", "true")
                app.Playback("recordingpaused")
                return true
            end
        end
    end
    
    if "pause" == rtype then
        print("There is no active recording.")
        app.Playback("beeperr")
        return true
    end

    -- check permissions for personal type
    local recOwner  = nil
    local recOpp    = nil
    local dst       = nil
    local cid       = gSpace["cidNum"]
    local cid_attrs = gSpace["cidAttrs"]
    local cidName   = gSpace["cidAttrs"].cn or channel.CALLERID("name"):get() or "empty"
    local dstName   = channel.CONNECTEDLINE("name"):get()
    dstName         = "" ~= dstName and dstName or "empty"
    local rec_notify = channel.GREC_NOTIFY:get() or "empty"
    if rtype == "personal" then
        local wmsperm = require"common.permissions"
        if cid_attrs.uid then
            if wmsperm:check(cid_attrs.owner or "", "canuse", "PERSONALRECORD", nil, cid_attrs["licenseType"]) ~= "yes" then
                print(string.format("Call recording for exten %q disallowed by ACL rules.", cid))
                to_logfile("Playback", "beeperr")
                return app.Playback("beeperr")
            end
            if channel.FAILOVER_USER:get() == "yes" then
                print(string.format("Call recording for failover user %q disallowed", cid))
                to_logfile("Playback", "beeperr")
                return app.Playback("beeperr")
            end
            if cid_attrs.mail and wmsperm:check(cid_attrs.owner or "", "yesno", "SEE_CALL_RECORDINGS", nil, cid_attrs["licenseType"]) == "no" then
                rec_notify = "empty"
                print(string.format("Send recording to email for exten %q disallowed by ACL rules.", cid))
            end
        else
            return false
        end
        dst = channel.CONNECTEDLINE("num"):get()
    elseif rtype == "pbx" then
        dst = channel.EXTEN:get() or ""
    end
    
    -- is mail send
    local recipients = {["to"] = "empty", ["bcc"] = "empty"} 
    local lang = "en"
    
    if rtype == "personal" then
        recipients["to"] = cid_attrs.mail or "empty"
        lang = cid_attrs.preferredLanguage or "en"
    elseif rtype == "pbx" then
        local param, user = mailaddr or ""
        if param ~= "" then
            if param:find("@") then
                user = get_obj_attrs(param, "mail")
                user.mail = param
            else
                user = get_obj_attrs(param)
            end
            lang = user.preferredLanguage or channel.CHANNEL("Language"):get() or "en"
            if not user.mail then
                local msg = "Not found email in pbx record. Notification was not sent"
                print(msg); logger:debug(msg)
            end
            if user.telephoneNumber and "" ~= user.telephoneNumber then
                recOwner = user.telephoneNumber
            end
            if recOwner ~= cid then
                recOpp   = cid
            end
            recipients["to"] = user.mail or "empty"
        end
        if app_callback and not recOwner then
            recOwner = app_callback
        end
    end

    local all_recs_email    = channel.GLOBAL("RECORDING_EMAIL"):get() -- copy of all recordings to this email address
    if all_recs_email and "" ~= all_recs_email and all_recs_email:find("@") then
        recipients["bcc"] = all_recs_email 
    end
    
    recipients["str"] = ("empty" ~= recipients["to"] and recipients["to"] or "") .. ("empty" ~= recipients["bcc"] and " " .. recipients["bcc"] or "")
    
    local free_space = diskutils.get_avail_ram_space()
    if 10240 > free_space then
        local logs
        if "yes" ~= rec_notify and not recipients["bcc"] then
            print("Email motification of recorded calls is disabled")
        elseif "empty" ~= recipients["to"] or "empty" ~= recipients["bcc"] then
            local subject
            if rtype == "pbx" then
                subject = _t("msubj.ERRORRAM_" .. rtype, "L_" .. lang, cid, dst)
            else
                subject = _t("msubj.ERRORRAM_" .. rtype, "L_" .. lang, cid)
            end
            local message = _t("mbody.ERRORRAM_" .. rtype, "L_" .. lang, os.date("%Y-%m-%d %X"))
            require"core.pbxutils".mailto(message, subject, recipients, "", 0) 
            logs = string.format("Sent recording error notification to %q", recipients["str"])
        elseif "empty" == recipients["to"] and "empty" == recipients["bcc"] then
            print("No email address")
        end
        local msg = "No available space in RAM for record call."
        print(msg, logs or ""); logger:warning(msg, logs or "")
        if rtype ~= "pbx" then
            to_logfile("Playback", "beeperr")
            return app.Playback("beeperr")
        else
            return false
        end
    end

    local rec_attach = channel.GREC_ATTACH:get() or "empty"
    local rec_share = diskutils.is_mount("/var/spool/callweaver/monitor")
    if not rec_share then
        if "empty" == recipients["to"] and "empty" == recipients["bcc"] then
            local msg = "Email parameter is empty and no device is mounted for call recoring"
            print(msg); logger:error(msg)
            if rtype ~= "pbx" then
                to_logfile("Playback", "beeperr")
                return app.Playback("beeperr")
            else
                return false
            end
        end
        if ("yes" ~= rec_attach or "yes" ~= rec_notify) and "empty" == recipients["bcc"] then
            local msg = "Disabled mail motification or record attachment and no device is mounted for call recoring"
            print(msg); logger:error(msg)
            if rtype ~= "pbx" then
                to_logfile("Playback", "beeperr")
                return app.Playback("beeperr")
            else
                return false
            end
        end
    end
    
    -- preparing for executing record call
    local isFeelingstreamEnabled = require"core.fsutils".is_file("/etc/feelingstream.conf")
    local isGongIoEnabled = require"core.fsutils".is_file("/etc/gong_io.conf")
    local isStereoEnabled = channel["STEREO_RECORDINGS"]:get() or ""
          isStereoEnabled = isStereoEnabled == "yes"
    if isStereoEnabled then
        logger:info("Stereo recordings parameter is enabled")
    end

    local format = "alaw"
    local channel_format = string.match(channel.CHANNEL("audionativeformat"):get(),"^%((%w+)%)$")
    if channel_format == "g729" and not isStereoEnabled then
        format = "g729"
    end
    local manual_codecs = channel["SIP_CODEC_INBOUND"]:get() or ""
    if "" ~= manual_codecs then
        if not string.find(manual_codecs, "g729") then
            format = "alaw"
        end
    end

    local time_refling = free_space < 20480 and 3 or 6
    local att_limit = channel.GATTACH_LIMIT:get() or 5 -- in MegaByte
    local attach_format = channel.GATTACH_FORMAT_REC:get() or "empty"
    local rec_share = rec_share and "yes" or "no"
    local exec_prog = "wms_notify"
    local exec_args = "-rec ".. require"core.strConverter".toB64(require"dkjson".encode({
        rec_notify      = rec_notify,
        rec_attach      = rec_attach,
        attach_format   = attach_format,
        rtype           = rtype,
        email           = recipients["to"],
        all_recs_email  = recipients["bcc"],
        lang            = lang,
        rec_share       = rec_share,
        att_limit       = att_limit,
        time_refling    = time_refling,
        cidName         = cidName,
        cidNum          = cid or "empty",
        dstName         = dstName,
        dstNum          = dst or "empty",
        stereo_enabled  = isStereoEnabled,
        feelingsteam    = isFeelingstreamEnabled,
        gong_io         = isGongIoEnabled,
        timestamp       = os.time()
    }))
    local rec_uniq_id = string.sub(require"md5".digest(math.random(os.time())), 1, 20)

    if rtype == "personal" then
        channel.MONITOR_EXEC:set(exec_prog)
        channel.MONITOR_EXEC_ARGS:set(exec_args)
        channel.XAD:set("true")
        channel.RECID:set(rec_uniq_id)
        local filename = string.format("/tmp/%s-%s-%s-%s", os.date('%Y.%m.%d-%H.%M.%S'), cid, dst, rec_uniq_id)
        to_logfile("Playback", "beep")
        app.Playback("beep")
        print("Call recording has been started.")
        notification_originate("start")
        to_logfile("Monitor", format, filename, "m", time_refling, attach_format, rec_share)
        return app.Monitor(format, filename, "m", time_refling, attach_format, rec_share)
    elseif rtype == "pbx" then
        local filename = string.format("/tmp/%s-%s-%s[pbx]-%s", os.date('%Y.%m.%d-%H.%M.%S'), recOwner or cid, recOpp or dst, rec_uniq_id)
        local otherChanRecord = "true" == channel.RECORD_ON_CALLEE_SIDE:get()
        if not otherChanRecord then
            channel.XAD:set("true")
            channel["__IS_SYSTEM_RECORD"]:set("true")
            channel.MONITOR_EXEC:set(exec_prog)
            channel.MONITOR_EXEC_ARGS:set(exec_args)
            to_logfile("Monitor", format, filename, "m" .. (opts or ""), time_refling, attach_format, rec_share)
            return app.Monitor(format, filename, "m" .. (opts or ""), time_refling, attach_format, rec_share)
        end
        local recParams = require"core.strConverter".toB64(require"dkjson".encode({
            exec_prog       = exec_prog,
            exec_args       = exec_args,
            format          = format,
            filename        = filename,
            opts            = opts,
            time_refling    = time_refling,
            attach_format   = attach_format,
            rec_share       = rec_share,
        }))
        channel._DATA_FOR_RECORDING:set(recParams)
        return true
    end
end

--- Switch call by redirecting from bridged peer (continue the conversation on another device)
function app_continuity_call(mode, callid)
    local cid = gSpace["cidNum"] or ""

    -- processing changing identity
    if gSpace["cidAttrs"] and gSpace["cidAttrs"].uid then
        local identity_num, identity_name = __checkIdentity(gSpace["cidAttrs"])
        if identity_num and identity_name then 
            channel.CALLERID("num"):set(identity_num)
            channel.CALLERID("name"):set(identity_name)
            channel.CDR("from_name"):set(identity_name)
        end
    end
    if "service" == mode then
        local curChanName = channel.CHANNEL("name"):get() or ""
        for _, cidChanName in pairs(require"core.pbxutils".split(channel.CHANNELS("SIP/" .. cid .. "-.*"):get() or "", " ")) do
            local bridgeChanName = channel.IMPORT(cidChanName, "BRIDGEPEER"):get() or ""
            local ChanSipCallid = channel.IMPORT(cidChanName, "SIPCALLID"):get() or ""
            -- Devices has callid like "sqc.6mq6bhum5a", except for W-AIR which has callid like "sqc.6mq6bhum5a@xxxx.wildixin.com"
            -- In case if we've received callid we'll try to match if received callid is the same as callid received from second channel
            local filter_call = (callid and string.match(ChanSipCallid, "^" .. callid .. "@?.*")) or not callid
            local process = curChanName ~= cidChanName and "" ~= bridgeChanName and filter_call
            if process then
                print("Executing 'Call continuity': bridge to - " .. bridgeChanName)
                return smartApp("Bridge", bridgeChanName)
            end
        end

        -- processing a call from W-AIR Headset
        if "*wair*" == channel.EXTEN:get() then
            return require"Directory".run("a")
        end
    elseif "incall" == mode then
        local bridged_peer = channel.BRIDGEPEER:get() or ""
        if "" ~= bridged_peer then
            app.ChannelRedirect(bridged_peer, "switchcalls", cid, 1)
        end
    end
end

--- Detect fax and go to specify exten^context
-- @param timeout A time for wiat fax signal
-- @param extension A go to extension if faxdetect
-- @param context A go to context if faxdetect 
-- @param opts additional option: nowait or empty
function app_faxdetect(timeout, extension, context, _, opts)
    if not extension or not context or not timeout then
        print ("Execute 'Detect fax': faxdetect failed. Incorrect extension, context or timeout.")
        logger:warning("Execute 'Detect fax': faxdetect failed. Incorrect extension, context or timeout.")
        return false
    end
    extension = convert_exten(extension)
    local channel, app = channel, app
    if extension == "default" then extension = "s" end
    print ("Execute 'Detect fax': timeout - " .. timeout .. ", context - " .. context .. ", extension - " .. extension)
    channel.FAX_CONTEXT:set(context)
    channel.FAX_EXTENSION:set(extension)
    channel.FAXOPT("faxdetect"):set("yes," .. timeout)
    channel.FAXDETECT_DIALPLAN:set("true")
    to_logfile ("Answer", ",nocdr")
    app.Answer(",nocdr")
    if "nowait" ~= opts then
        to_logfile ("Wait", timeout)
        app.Wait(timeout)
    else
        print("Continue dial plan execution")
    end
    return true
end

--- Call to conference
-- @param conf_id A id of conference room
function app_conference(conf_id)
    app.Set("JITTERBUFFER(adaptive)=500,1000,40")
    if not conf_id or conf_id == "" then
        conf_id = ivr:read("Please enter conference number", {retry = 1})
        if not conf_id then
            ivr:tell("No number was entered")
            return false
        end
    end
    -- validate argument
    if conf_id ~= "" and not string.find (conf_id,"^%d+$") then
        ivr:tell("The value you entered is incorrect")
        return false
    end
    
    local localUsers = {}
    local ldapData = require"core.wmsldap":get("telephoneNumber", "*")
    if ldapData and ldapData[1] then
        for _, data in pairs(ldapData) do
            if data[2] and data[2].uid and data[2].owner and  data[2].telephoneNumber then
                table.insert(localUsers, data[2].telephoneNumber)
            end
        end
    end

    -- update connected line to correct values before connecting to conference
    channel.CONNECTEDLINE("all"):set(string.format([===["Conf: %s" <conf*%s>]===], conf_id, conf_id))

    --- executing cloud conference
    if 7 <= conf_id:len() and not require"core.pbxutils".in_table(conf_id, localUsers) then
        app.CELGenUserEvent("App", "CloudConference", conf_id)
        channel["BLOCK_LOCAL_MOH"]:set("true")
        channel.CDR("to_name"):set("Conf: " .. conf_id)
        channel["_FORCE_CTO"]:set(string.format("conf*%s", conf_id))
        return smartApp("Dial", "SIP/cloudservice/" .. conf_id, "", "Ib(predial^cloud_srv^1(conference))")
    end
    
    --- executing internal conference
    print("Execute 'Conference': conf id - " .. conf_id)
    -- disable opus for internal conference
    channel._DISABLE_OPUS:set("true")
    channel._SERVICECALL:set("true")
    -- check conference ID as LDAP user
    local exConfMode = get_obj_attrs(conf_id)
    exConfMode       = exConfMode.uid
    local cid        = gSpace["cidNum"]
    local slaveUser  = false
    if exConfMode then
        smartApp("Answer", "")  -- uses confbridge function
        channel.CONFBRIDGE("user", "template"):set("default_user")
        channel.CONFBRIDGE("user", "end_marked"):set("yes")
        slaveUser = conf_id ~= cid
        if not slaveUser then
            channel.CONFBRIDGE("user", "marked"):set("yes")
            channel.CONFBRIDGE("user", "admin"):set("yes")
        end
    end
    
    -- start conference application
    smartApp("ConfBridge", conf_id, "default_bridge", "default_user", "default_menu")
    
    -- processing return from conference procedure
    if exConfMode and slaveUser and not check_hangup() then
        for i=1,3 do
            app.Playback("beep")
            app.Wait(3 ~= i and 1 or 2)
        end
        
        return require"internalcalls":return_from_transfer(conf_id, conf_id, "CONF", "confBridge")
    end
    
    return true
end

-- Authenticate calling user
-- @param number A number for authentication
-- @param is_blocked A option used for check CALLBLOCK
function app_authenticate_user (number, is_blocked)
    local app, channel = app, channel
    local hcause, ivr = hcause, ivr
    print ("Execute 'Authenticate user'" .. (number and (": number - " .. number) or "")  .. (is_blocked and ((number and ", " or ": ") .. "options - only for bloked phone") or ""))
    
    -- validate argument
    if not number or not string.find (number,"^%d+$") then
        ivr:tell("The value you entered is not valid. Hangup")
        return smartApp("Hangup", hcause.FAILURE)
    end
    
    local user = get_obj_attrs(number)
    if not user.uid then
        print("Not found in ldap:", number)
        logger:warning("Not found in ldap:", number)
        return smartApp("Hangup", hcause.NOT_FOUND)
    end
    
    gSpace["cidAttrs"] = user
    local user_features = require"common.userFeatures.userFeatures":new(user.uid)
    if not user_features then
        print("Error loading features of the user")
        logger:error("Error loading features of the user")
        smartApp("Hangup", hcause.NOT_FOUND)
    end
    user_features:load()
    
    -- authorithate only in case number CALLBLOCK
    if is_blocked == "b" then
        if not user_features:is_active("CALLBLOCK") then
            print("CALLBLOCK is disabled")
        end
    else
        local userAuth = require"userAuthorize":new()
              userAuth:setPinPhrase("To use this phone please enter the password, followed by pound")
        if not userAuth:process(number) then
            return smartApp("Hangup", hcause.REJECT)
        end
    end
    
    -- set caller as authenticated user
    gSpace["cidFeatures"]    = user_features
    gSpace["gvars"].calltype = "internal"
    channel.CALLTYPE:set("internal")
    channel["X-TRUNK"]:set("")
    if user.telephoneNumber and "" ~= user.telephoneNumber then
        gSpace["cidNum"] = user.telephoneNumber
        channel.CALLERID("num"):set(user.telephoneNumber)
        channel.CDR("c_from"):set(user.telephoneNumber)
    end
    if user.cn and "" ~= user.cn then
        channel.CALLERID("name"):set(user.cn)
        channel.CDR("from_name"):set(user.cn)
    end
end

--- Correct outgoing number according prefix and global codes (country, national, international)
-- @param exten A exten used for correcting
-- @param prefix A prefix which will add to exten
-- @param context A context used by goto when exnten correct
function app_correct_outgoing_number (exten, prefix, context)
    local app, channel = app, channel
    if not context or context == "" then
        context = "users"
    end
    print ("Executing 'Correct called number': number - " .. exten .. (prefix and (", prefix - " .. prefix) or "") .. ", context - " .. context)
    local exten_filtered = ""
    local prefix = prefix or ""
    local filter_data = {["+"] = true, ["0"] = true, ["1"] = true, ["2"] = true, ["3"] = true, ["4"] = true,
                        ["5"] = true, ["6"] = true, ["7"] = true, ["8"] = true, ["9"] = true }      
    for i = 1, exten:len(), 1 do
        if filter_data[exten:sub(i,i)] then
            exten_filtered = exten_filtered .. exten:sub(i,i)
        end
    end
    local exten = exten_filtered
    if exten:sub(1, 1) == "+" then
        local country_code = channel["GCOUNTRY_CODE"]:get() or ""
        local country_code_len = country_code:len()
        if country_code_len > 0 and exten:sub(2, country_code_len+1) == country_code then
            if country_code == "39" then
                exten = prefix .. exten:sub(country_code_len+2)
            else
                local national_prefix = channel["GNATIONAL_PREFIX"]:get() or ""
                exten = prefix .. national_prefix .. exten:sub(country_code_len+2)
            end
        else
            local international_prefix = channel["GINTERNATIONAL_PREFIX"]:get() or ""
            exten = prefix .. international_prefix .. exten:sub(2)
        end
        to_logfile ("Goto", context, exten, 1)
        app.Goto(context, exten, 1)
    end
end

--- Local function for jump in other @exten and @context
local function jump_general (exten, context, priority, ret_opts, prefix_exec)
    local app = app
    if not exten or exten == "" or not context or context == "" then
        print ("Executing " .. (prefix_exec:match("(\'.*\':)") or "'Jump general':") .. " specify correct context or extension")
        return false
    end
    if exten == "default" then
        exten = "s"
    else
        exten = convert_exten(exten)
    end
    print ("Executing " .. (prefix_exec or "") .. " exten - " .. exten .. ", context - " .. context .. (priority and (", priority - " .. priority) or ""))
    
    if ret_opts == "n" then
        to_logfile ("Dial", "Local/" .. exten .. "@" .. context .. "/n")
        return app.Dial("Local/" .. exten .. "@" .. context .. "/n")
    else
        if context ~= "" then
            to_logfile ("Goto", context, exten, priority or 1)
            return app.Goto(context, exten, priority or 1)
        end
    end
end

--- Function for jump in other @exten and @context
function app_goto_general (exten, context, priority, ret_opts)
    return jump_general (exten, context, priority, ret_opts, "'Jump to':")
end

--- Function for jump if dialstatus equal specified
function app_goto_dialstatus (dialstatus, exten, context, priority, ret_opts)
    local channel = channel
    if not dialstatus then
        print ("Executing 'Jump to if congested/busy': specify dialstatus for comparing")
        return false
    end
    local cur_dialstatus = channel["DIALSTATUS"]:get() or ""
    if dialstatus == cur_dialstatus then
        return jump_general(exten, context, priority, ret_opts, "'Jump to if congested/busy': dialstatus - " .. dialstatus .. ",")
    else
        print ("Executing 'Jump to if congested/busy': dialstatus - " .. dialstatus)
        return true
    end
end

--- Function for jump if calltype equal specified
function app_goto_calltype (calltype, exten, context, priority, ret_opts)
    local channel = channel
    if not calltype then
        print ("Executing 'Jump to if call type is': specify calltype for comparing")
        return false
    end
    local cur_calltype = gSpace["gvars"].calltype or channel["CALLTYPE"]:get() or ""
    if calltype == cur_calltype then
        return jump_general(exten, context, priority, ret_opts, "'Jump to if call type is': calltype - " .. calltype .. ",")
    else
        print ("Executing 'Jump to if call type is': calltype - " .. calltype)
        return true
    end
end

--- Function for jump if count for call for @groupname less then specified @maxcalls
function app_goto_group_count (groupname, maxcalls, exten, context, priority, ret_opts)
    local channel = channel
    if not maxcalls then
        print ("Executing 'Verify number of calls': specify number of max calls")
        return false
    end
    channel.GROUP("app"):set(groupname)
    local cur_calls = channel.GROUP_COUNT(groupname.."@app"):get() or 0
    if tonumber(cur_calls) > tonumber(maxcalls) then
        return jump_general(exten, context, priority, ret_opts, "'Verify number of calls': maxcalls - " .. maxcalls .. ",")
    else
        print ("Executing 'Verify number of calls': maxcalls - " .. maxcalls)
        return true
    end
end

--- Function for jump to depending on user status
function app_goto_userstatus (user, status, exten, context, priority, ret_opts)
    local channel = channel
    if not user or not status then
        print ("Executing 'Jump to if user status': specify user and status for comparing")
        return false
    end
    local userAttrs    = get_obj_attrs(user)
    local userFeatures = require"common.userFeatures.userFeatures":new(userAttrs.uid)
    if not userFeatures then
        print("Error loading features of the destination user")
        return false
    end
    userFeatures:load()
    local currentStatus = userFeatures:getUserStatus()
    if status == currentStatus then
        return jump_general(exten, context, priority, ret_opts, "'Jump to depending on user status': status is - " .. status .. ",")
    else
        print ("Executing 'Jump to if user status': status is - " .. status)
        return true
    end
    
    return false
end

--- Function for jump to depending on arbitrary condition
function app_goto_if_condition (condition, exten, context, priority, ret_opts)
    local conditionConverted = convert_exten(condition or "")
    local conditionEvaluated = channel.EVAL(string.format("$[%s]", condition or "")):get()
    if "1" == conditionEvaluated then
        -- jump: "Executing 'Jump to if $[${CALLERID(number)}=401] - true (401=401)'"
        return jump_general(exten, context, priority, ret_opts, "'Jump to if $[" .. condition .."] - true => ("..conditionConverted..")',")
    else
        -- skip: "Skipping 'Jump to if $[${CALLERID(number)}=401] - false (405=401)'"
        print ("Skipping 'Jump to if $[" .. condition .."] - false => ("..conditionConverted..")'")
        return true
    end
end

--- Verify caller for contanious values such as @verify
function app_verify_caller (verify, exten, context, priority, ret_opts)
    local channel = channel
    if not verify then
        print ("Executing 'Verify caller': specify verify content")
        return false
    end
    local cid_number = gSpace["cidNum"] or ""
    local cid_name = channel.CALLERID("name"):get() or ""
    local is_regex = false
    
    -- our regexps should start with !
    if verify:sub(1,1) == "!" then
        is_regex = true
        verify = verify:sub(2)
    end
    
    -- escaped exclamation sign
    if verify:sub(1,2) == "\\!" then
        is_regex = false
        verify = "!" .. verify:sub(3)
    end

    local we_go
    if is_regex then
        cid_number, cid_name = cid_number:gsub("([,=;\\])","\\%1"), cid_name:gsub("([,=;\\])","\\%1") -- escapes values = , ; \
        we_go = tonumber(channel.REGEX("" .. verify .. "",cid_name,cid_number):get() or 0) or 0
    else
        verify = verify:gsub("(%p)", "%%%1")
        we_go = cid_number:find(verify)
        if not we_go then
            we_go = cid_name:find(verify)
        end
    end
    if we_go and we_go > 0 then
        return jump_general(exten, context, priority, ret_opts, "'Verify caller': verify content - " .. verify .. ",")
    else
        print ("Executing 'Verify caller': verify content - " .. verify)
        return true
    end
end

--- Callback application
function app_callback ()
    require"common.cdrconnection"
    local app, channel = app, channel
    local cid = gSpace["cidNum"]
    local trunkname = gSpace["gvars"].trunk_call
    
    if not trunkname or trunkname == "" then
        print ("Executing 'Automatic Call Back': internal call, skiping")
        return false
    end
    print ("Executing 'Automatic Call Back': for trunkname - " .. trunkname)
    
    local cdr_backend, cdr_table = cdrconnection:init()
    if not cdr_backend or not cdr_table then
        print ("CDR backend or table not found")
        return false
    end
    
    local cbconnection = require"core.wsqlite":new("callback")
    cbconnection:execute("DELETE FROM callbacks WHERE datetime(endtime) < datetime('now','-1 hour');")
    local sql = {
                sqlite3 = [[SELECT src,end from %s where (dst=%q or c_to=%q) and dstchannel LIKE %q and end >= datetime('now', '-1 hour')
                            and (disposition = 'NO ANSWER' or disposition = 'BUSY' or disposition = 'FAILED') ORDER BY end DESC LIMIT 1]],
                mysql =   [[SELECT src,end from %s where (dst=%q or c_to=%q) and dstchannel LIKE %q and end >= now() - interval 1 hour
                            and (disposition = 'NO ANSWER' or disposition = 'BUSY' or disposition = 'FAILED') ORDER BY end DESC LIMIT 1]],
                tds =     [[SELECT TOP 1 src,[end] from %s where (dst='%s' or c_to='%s') and dstchannel LIKE '%s' and [end] >= DATEADD(hour, -1, GETDATE())
                            and (disposition = 'NO ANSWER' or disposition = 'BUSY' or disposition = 'FAILED') ORDER BY [end] DESC]]
            }
            
    -- processing used trunk name
    if trunkname:find("^bri") or trunkname:find("^pri") or trunkname:find("^fxo") then
        local pstrTrunkName = trunkname:match("(%w+_)") 
        if pstrTrunkName then
            trunkname = pstrTrunkName .. "%%_%%"
        end   
    end
    local trunksearch = 'SIP/' .. trunkname .. '-%%'
    
    local cursor = cdrconnection:execute(sql[cdr_backend], cdr_table, cid, cid, trunksearch)
    if not cursor or type(cursor) == "boolean" then
        cbconnection:close(); cdrconnection:close()
        local msg = "Cannot access cdr database"
        print(msg); logger:error(msg)
        return false
    end
    local src, endtime = cursor:fetch(); cursor:close(); cdrconnection:close()
    if src then
        sql = "SELECT count(*) FROM `callbacks` WHERE src=%q and cid=%q and trunk = %q and endtime = %q;"
        local cursor = cbconnection:execute (sql, src, cid, trunkname, endtime)
        if not cursor or type(cursor) == "boolean" then
            cbconnection:close(); cdrconnection:close()
            print("Cannot access callback database")
            logger:error("Cannot access callback database")
            return false
        end
        
        local res = cursor:fetch(); cursor:close()
        if 0 == res then
            local cid_name = channel.CALLERID("name"):get() or ""
            channel.CALLERID("name"):set("Call back " .. cid_name)
            sql = "INSERT INTO callbacks(src, cid, trunk, endtime)VALUES(%q, %q, %q, %q);"
            cbconnection:execute(sql, src, cid, trunkname, endtime)
            cbconnection:close()
            local orig_exten = channel.CDR("c_to"):get() or ""
            channel.CDR("c_to"):set(src)    --set correct CDR field if 'Callback' will be answered
            channel.CDR("to_name"):set("")
            local trunk_lang = channel.CHANNEL("Language"):get() or ""
            to_logfile("Dial", "Local/" .. src .. "@internalcalls", 20, "b(predial^callback^1(" ..trunk_lang.. "))")
            app.Dial("Local/" .. src .. "@internalcalls", 20, "b(predial^callback^1(" ..trunk_lang.. "))")
            channel.CALLERID("name"):set(cid_name)
            channel.CDR("c_to"):set(orig_exten) --'Callback' not answered then set original exension
        end
    else
        cbconnection:close()
    end
end

------- TIMETABLES -----------
--- Check timetable for matching
function verify_timetable(tt_id)
    local connection = require"core.wsqlite":new("wmsdb")
    local sql        = "SELECT state FROM timetable_main WHERE id=%s;"
    local cursor     = connection:execute (sql, tt_id)
    if not cursor or type(cursor) == "boolean" then
        connection:close()
        print("DB query failed. Timetable state is not defined")
        logger:error("DB query failed. Timetable state is not defined")
        return false
    end

    local tt_state = cursor:fetch(); cursor:close()
    local tt_result = false
    if 0 == tt_state then
        -- state off
        tt_result = false
    elseif 1 == tt_state then
        -- state on : perform a query to get the match
        local tt_year, tt_month, tt_day, tt_time, tt_dayweek = string.match (os.date("%Y|%m|%d|%X|%w"), "(.+)|(.+)|(.+)|(.+)|(.+)")
        if tt_dayweek == "0" then tt_dayweek = "7" end
        local sql    = [[SELECT ta.day_of_week_from AS tt_matched FROM timetable_main AS tm, timetable_additional AS ta WHERE ta.parent_id = tm.id AND tm.id = ']].. tt_id ..[[' AND ta.enabled_id = 1 AND ((']].. tt_dayweek ..[[' BETWEEN ta.day_of_week_from AND ta.day_of_week_to) AND (']].. tt_dayweek ..[[' NOT BETWEEN ta.day_of_week_to AND ta.day_of_week_from) OR (ta.day_of_week_from = 0) OR (ta.day_of_week_from = ta.day_of_week_to AND ta.day_of_week_to = ']].. tt_dayweek ..[['))   AND (((']].. tt_time ..[['BETWEEN ta.start_daytime AND ta.end_daytime) AND (ta.start_daytime < ta.end_daytime)) OR ((']].. tt_time ..[[' NOT BETWEEN ta.end_daytime AND ta.start_daytime) AND (ta.start_daytime > ta.end_daytime)) OR (ta.start_daytime = '00:00:00' AND ta.end_daytime = '00:00:00')) AND ((']].. tt_day ..[[' BETWEEN ta.day_id_from AND ta.day_id_to) AND (']].. tt_day ..[[' NOT BETWEEN ta.day_id_to AND ta.day_id_from) OR (ta.day_id_from = '0') OR (ta.day_id_to = ta.day_id_from AND ta.day_id_to = ']].. tt_day ..[[')) AND ((ta.month_id = ']].. tt_month ..[[') OR (ta.month_id = '0')) AND ((ta.year_id = ']].. tt_year ..[[') OR (ta.year_id = '0'));]]
        local cursor = connection:execute(sql)
        if not cursor or type(cursor) == "boolean" then
            connection:close()
            print("DB query failed. Timetable state is not defined")
            logger:error("DB query failed. Timetable state is not defined")
            return false
        end
        local tt_matched = cursor:fetch(); cursor:close()
        if tt_matched and tt_matched ~= nill and tt_matched ~= "" then
            tt_result = true
        end
    elseif 2 == tt_state then
        -- always on (extra)
        tt_result = true
    else
        print("Timetable state is not defined")
        logger:info("Timetable state is not defined")
        tt_result = false
    end
    connection:close()
    
    if tt_result == false then
        do return false end
    end
    return true
end

--- Jump to other @exten and @context if status cheked time or always on
function app_timetable(timetableid, exten, context, priority, ret_opts)
    if not timetableid then
        print ("Executing 'Verify timetable': specify timetable")
        return false
    end
    if verify_timetable(timetableid) then
        return jump_general(exten, context, priority, ret_opts, "'Verify timetable': timetable id - " .. timetableid .. ",")
    else
        print ("Executing 'Verify timetable': timetable id - " .. timetableid)
        return true
    end
end

--- Change state of timetable
-- active - follow time schedue, force on - the condition is always true 
-- force off - the condition is always false
local function change_stime_state (state, timetableid)
    local connection = require"core.wsqlite":new("wmsdb")
    local state = tonumber(state)
    local newstate = state == 2 and 0 or state + 1

    local sql = "UPDATE timetable_main SET state = %q WHERE id = %s;"
    connection:execute(sql, newstate, timetableid)

    local sql = "SELECT timetable,state FROM timetable_main WHERE id = %s;"
    local cursor = connection:execute (sql, timetableid)
    local timetable, s = cursor:fetch(); cursor:close()
    connection:close()
    state = tonumber(s)

    local setstate
    if state == 0 then
        setstate = "NOT_INUSE"
    elseif state == 1 then
        setstate = "RINGING"
    elseif state == 2 then
        setstate = "INUSE"
    end
    
    channel.DEVICE_STATE("Custom:DS/stime*" .. timetableid .. "*1"):set(setstate)
end

--- Getting timetable name and change it status to choose state
-- @param timetableid A id of timetable
-- @param doset A set status of timetable to another
local function stime (timetableid, doset)
    local ivr, hcause = ivr, hcause
    
    local is_allowed = require"common.permissions":check(gSpace["cidAttrs"]["owner"] or "", "canset", "STIME", nil, gSpace["cidAttrs"]["licenseType"]) == "yes"
    if not is_allowed then
        print("Rejected by group permissions")
        to_logfile ("Hangup", hcause.ACL_FORBIDDEN)
        return app.Hangup(hcause.ACL_FORBIDDEN)
    end
    
    if not timetableid then
        timetableid = ivr:read("Please enter number of timetable")
        if not timetableid then
            ivr:tell("No number was entered. Good bye. Hangup")
            to_logfile ("Hangup", hcause.NORMAL)
            return app.Hangup(hcause.NORMAL)
        end
    end
    
    -- validate argument
    if not timetableid or not string.find(timetableid,"^%d+$") or (doset and doset ~= "1") then
        ivr:tell ("The value you entered is not valid. Hangup")
        to_logfile ("Hangup", hcause.CHANUNAVAIL)
        return app.Hangup(hcause.CHANUNAVAIL)
    end
    
    local connection = require"core.wsqlite":new("wmsdb")
    local sql        = "SELECT timetable, state FROM timetable_main WHERE id = %s;"
    local cursor     = connection:execute (sql, timetableid)
    if not cursor or type(cursor) == "boolean" then
        connection:close()
        print("DB query failed. Timetable state is not defined")
        logger:error("DB query failed. Timetable state is not defined")
        return smartApp("Hangup", hcause.CHANUNAVAIL)
    end
    ttimetable, state = cursor:fetch(); cursor:close()
    connection:close()
    if not ttimetable then
        ivr:tell("The number you entered is incorrect")
        return smartApp("Hangup", hcause.CHANUNAVAIL)
    end
    
    if state then
        local options = {
            max_chars = 1,
            validate = function(doset) return doset == "1" end,
            failure = function(doset)
                ivr:tell("The number you entered is incorrect, please try again")
            end
        }
        
        local is_interactive = not options.validate (doset or "")
        if is_interactive then -- read current status and propose to change it
            ivr:tell("Service status " .. ({"off","on","always active"})[tonumber(state)+1])
        end
        
        if not options.validate(doset or "") then
            doset = ivr:read("Press 1 to change", options)
            if not doset then
                ivr:tell("Good bye. Hangup")
                to_logfile ("Hangup", hcause.NORMAL)
                return app.Hangup(hcause.NORMAL)
            end
        end
        
        -- doset = 1
        change_stime_state(state, timetableid)
        if not is_interactive then -- read status after a change and exit
            local newstate = tonumber (state) == 2 and 1 or tonumber (state) + 2
            ivr:tell("Service status " .. ({"off","on","always active"})[newstate])
        end
        if is_interactive then
            return stime(timetableid)
        else
            to_logfile ("Hangup", hcause.NORMAL)
            return app.Hangup(hcause.NORMAL)
        end
    end
    ivr:tell("Good bye. Hangup")
    to_logfile ("Hangup", hcause.NORMAL)
    return app.Hangup(hcause.NORMAL)
end
    
------- SWITCHES -------------
--- Jump to other @exten and @context if swith @switchid is on  
function app_switch (switchid, exten, context, priority, ret_opts)
    if not switchid then
        print ("Executing 'Jump to if switch is on': specify name of the switch")
        return false
    end
    local devstate = channel.DEVICE_STATE("Custom:DS/switch*" .. switchid .. "*1"):get()
    if devstate == "INUSE" then
        return jump_general(exten, context, priority, ret_opts, "'Jump to if switch is on': switch id - " .. switchid .. ",")
    else
        print ("Executing 'Jump to if switch is on': switch id - " .. switchid)
        return true
    end
end

--- Jump to other @exten1 and @context1 if swith @switchid is on or to @exten2 and @context2 if swith is blinking
function app_three_switch (switchid, exten1, context1, priority1, ret_opts2, exten2, context2, priority2, ret_opts2)
    if not switchid then
        print ("Executing 'Jump to if 3 state switch is on': specify name of the 3 state switch")
        return false
    end
    local devstate = channel.DEVICE_STATE("Custom:DS/3switch*" .. switchid .. "*1"):get()
    if devstate == "INUSE" then
        return jump_general(exten1, context1, priority1, ret_opts1, "'Jump to if 3 state switch is on': 3switch id - " .. switchid .. ",")
    elseif devstate == "RINGING" then
        return jump_general(exten2, context2, priority2, ret_opts2, "'Jump to if 3 state switch is on': 3switch id - " .. switchid .. ",")
    else
        print ("Executing 'Jump to if 3 state switch is on': 3switch id - " .. switchid)
        return true
    end
end

--- Change state of switch or switch3
local function change_switch_state(state, switchid, swtype)
    local connection = require"core.wsqlite":new("wmsdb")
    local state = tonumber(state)
    local whint, setstate
    local newstate = 0

    if swtype == 3 then
        whint = "3switch"
        newstate = state == 2 and 0 or state + 1
    else
        whint = "switch"
        newstate = state == 1 and 0 or 1
    end

    local sql = "UPDATE wmsswitches SET state=%q WHERE id=%s AND swtype=%q;"
    connection:execute(sql, newstate, switchid, swtype)

    local sql = "SELECT switchname,state FROM wmsswitches WHERE id=%s AND swtype=%q;"
    local cursor = connection:execute (sql, switchid, swtype)
    local switchname, s = cursor:fetch(); cursor:close()
    connection:close()

    state = tonumber(s)
    if state == 0 then
        setstate = "NOT_INUSE"
    elseif state == 1 then
        setstate = "INUSE"
    elseif state == 2 then
        setstate = "RINGING"
    end

    channel.DEVICE_STATE("Custom:DS/" .. whint .. "*".. switchid .. "*1"):set(setstate)
end

--- Getting switch/swith3 name and change it status to choose state
-- @param swid A id of the swith (associated with id in sqlite)
-- @param doset A set status of switch to another
-- @param sw3type A arg indicate use switch3
local function switch(swid, doset, sw3type)
    local ivr, hcause = ivr, hcause
    
    local is_allowed = require"common.permissions":check(gSpace["cidAttrs"]["owner"] or "", "canset", (sw3type and "3SWITCH" or "SWITCH"), nil, gSpace["cidAttrs"]["licenseType"]) == "yes"
    if not is_allowed then
        print("Rejected by group permissions")
        to_logfile ("Hangup", hcause.ACL_FORBIDDEN)
        return app.Hangup(hcause.ACL_FORBIDDEN)
    end
    
    if not swid then
        swid = ivr:read("Please enter number of switch")
        if not swid then
            ivr:tell("No number was entered. Good bye. Hangup")
            to_logfile ("Hangup", hcause.NORMAL)
            return app.Hangup(hcause.NORMAL)
        end
    end
    
    -- validate argument
    if not swid or not string.find (swid,"^%d+$") or (doset and doset ~= "1") then
        ivr:tell("The value you entered is not valid. Hangup")
        to_logfile ("Hangup", hcause.CHANUNAVAIL)
        return app.Hangup(hcause.CHANUNAVAIL)
    end
    
    local connection = require"core.wsqlite":new("wmsdb")
    local switchname, state
    local sql    = "SELECT switchname, state FROM wmsswitches WHERE id = %s AND swtype= " .. (sw3type and "'3'" or "'0'") .. ";"
    local cursor = connection:execute (sql, swid)
    if not cursor or type(cursor) == "boolean" then
        connection:close()
        print("DB query failed. Switch state is not defined")
        logger:error("DB query failed. Switch state is not defined")
        return smartApp("Hangup", hcause.CHANUNAVAIL)
    end
    switchname, state = cursor:fetch(); cursor:close()
    connection:close()
    
    if not switchname then
        ivr:tell("The number you entered is incorrect")
        to_logfile ("Hangup", hcause.CHANUNAVAIL)
        return app.Hangup(hcause.CHANUNAVAIL)
    end
    
    if state then
        local options = {
            max_chars = 1,
            validate = function(doset) return doset == "1" end,
            failure = function(doset)
                ivr:tell("The number you entered is incorrect, please try again")
            end
        }
        
        local is_interactive = not options.validate (doset or "")
        if is_interactive then              -- read current status and propose to change it
            if sw3type then
                ivr:tell("Service status " .. ({"off","on","extra"})[tonumber(state)+1])
            else
                ivr:tell("Service status " .. ({"off","on"})[tonumber(state)+1])
            end
        end
        
        if not options.validate(doset or "") then
            doset = ivr:read("Press 1 to change", options)
            if not doset then
                ivr:tell("Good bye. Hangup")
                to_logfile ("Hangup", hcause.NORMAL)
                return app.Hangup(hcause.NORMAL)
            end
        end
        
        --doset = 1
        change_switch_state (state, swid, sw3type and 3 or 0)
        if not is_interactive then          -- read status after a change and exit
            if sw3type then
                local newstate = tonumber (state) == 2 and 1 or tonumber (state) + 2
                ivr:tell("Service status " .. ({"off","on","extra"})[newstate])
            else
                local newstate = tonumber (state) == 1 and 1 or tonumber (state) + 2
                ivr:tell("Service status " .. ({"off","on"})[newstate])
            end
        end
        if is_interactive then
            return switch(swid, nil, sw3type)
        else
            to_logfile ("Hangup", hcause.NORMAL)
            app.Hangup(hcause.NORMAL)
        end
    end
    ivr:tell("Good bye. Hangup")
    to_logfile ("Hangup", hcause.NORMAL)
    return app.Hangup(hcause.NORMAL)
end


--------------- SERVICES -------------------------------------------------
local services = {
    --- Echo test service
    ["ECHO"] = function ()
        local app, ivr = app, ivr
        
        channel.SERVICECALL:set("true")

        ivr:tell("The sound test will now start.")
        to_logfile("Playback", "beep")
        app.Playback("beep")
        
        if gSpace["gvars"].application == "KITE" then
            channel.TIMEOUT("absolute"):set(20)
        end
        
        to_logfile("Echo", "")
        app.Echo()

        ivr:tell("Good Bye. Hangup")
        to_logfile("Hangup", hcause.NORMAL)
        app.Hangup(hcause.NORMAL)
    end;
    
    --- Intercome service
    -- @param destnumber     An extension for the intercom
    -- @param timeout        A timeout for the intercom
    -- @param options        An options for the intercom
    -- @param service_no_acl A boolean value indicate call from dialplan application or not
    -- @param dst_attrs      An attributes for destination extension
    -- @param announce       An announcement for called party
    ["INTERCOM"] = function (destnumber, service_no_acl, timeout, options, dst_attrs, announce)
        local app, ivr, hcause = app, ivr, hcause
        local wmsperm  = require"common.permissions"
        local pbxutils = require"core.pbxutils"
        
        if not destnumber then
            destnumber = ivr:read("Please enter extension number", {retry = 1})
            if not destnumber then
                ivr:tell("Good bye. Hangup")
                return smartApp("Hangup", hcause.NORMAL)
            end
        end

        if not string.find (destnumber, "^%d+$") then
            ivr:tell("The value you entered is not valid. Hangup")
            return smartApp("Hangup", hcause.CHANUNAVAIL)
        end
        
        if dst_attrs == nil or type(dst_attrs) ~= "table" then
            dst_attrs = get_obj_attrs (destnumber)
        end
        if dst_attrs.telephoneNumber == nil then
            print("Intercom number not found.")
            logger:warning("Intercom number not found.")
            return smartApp("Hangup", hcause.NOT_FOUND)
        end
        
        
        local is_allowed = service_no_acl or
        wmsperm:check(gSpace["cidAttrs"]["owner"] or "", "yesno", "INTERCOM", dst_attrs["owner"] or "", gSpace["cidAttrs"]["licenseType"]) == "yes"
        if not is_allowed then
            print("Rejected by group permissions")
            return smartApp("Hangup", hcause.ACL_FORBIDDEN)
        end

        -- set up intercom dial timeout
        local timeout = timeout or ""
              timeout = "" == timeout and channel["GINTDIAL_TIMEOUT"]:get() or ""
              timeout = "" == timeout and "20" or timeout  
        
        -- processing announcement option
        local announce    = announce or ""
        local announceOpt = "" ~= announce and string.format("A(%s)", announce) or ""

        local dst_pbx    = pbxutils.getPbxDataByDN(dst_attrs.pbxowner)
        local our_pbx_dn = pbxutils.getPbxDNBySerial()
        if dst_pbx.dn ~= our_pbx_dn then
            local ext = "intercom*" .. destnumber
            app_remote_call(self, nil, ext, dst_pbx.dn, "pbxservices", nil, "r"..announceOpt)
            local unavailable_pbx = channel["X-PBX_UNAVAILABLE"]:get() or ""
            if unavailable_pbx ~= "" then
                -- get destination user PBX failover configuration
                local failoverObj = require"common.failover":new(dst_attrs.pbxowner)
                if failoverObj and "1" == failoverObj:isEnabled() then 
                    -- check if failover call and local PBX is failover for destination user (second check need because of processing after execution dial app)
                    if failoverObj:localPbxIsFailover() then -- call via current pbx
                        print("Trying intercom call to failover user")
                        local options = (options or "").."b(predial^failover^1(intercom))"
                        smartApp("Dial", "SIP/" .. destnumber, timeout or "2", options)
                    elseif failoverObj:getPbxDN() then    -- call via remote pbx
                        local ext = "intercom*" .. destnumber
                        app_remote_call(self, "Making intercom call to failover user at", ext, failoverObj:getPbxDN(), "pbxservices")
                    end
                end
            end
            return smartApp("Hangup", hcause.NORMAL)
        end
        
        local options = (options or "") .. announceOpt .. "b(predial^intercom^1)"
        smartApp("Dial", "SIP/" .. destnumber, timeout or "2", options)
    end;
    
    --- Intrusion services (SBARGE, CBARGE, PBARGE)
    -- @param destnumber A extenstion for intrusion
    -- @param options A options for the intrusion (use for different intrusion)
    -- @param service_no_acl A boolean value indicate call from dialplan application or not
    -- @param dst_attrs A attributes for destination extension
    ["BARGE"] = function (destnumber, stype, service_no_acl, dst_attrs)
        local app, ivr, hcause = app, ivr, hcause
        local wmsperm = require"common.permissions"
        local pbxutils = require"core.pbxutils"
        
        if not destnumber then
            destnumber = ivr:read("Please enter extension number")
            if not destnumber then
                ivr:tell("Good bye. Hangup")
                to_logfile ("Hangup", hcause.NORMAL)
                return app.Hangup (hcause.NORMAL)
            end
        end
        
        if not string.find (destnumber,"^%d+$") then
            ivr:tell("The value you entered is not valid. Hangup")
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            app.Hangup(hcause.CHANUNAVAIL)
        end
        
        if dst_attrs == nil or type(dst_attrs) ~= "table" then
            dst_attrs = get_obj_attrs (destnumber)
        end
        if dst_attrs.telephoneNumber == nil then
            print("Barge number not found.")
            logger:warning("Barge number not found.")
            to_logfile("Hangup", hcause.NOT_FOUND)
            return app.Hangup(hcause.NOT_FOUND)
        end
        
        local is_allowed = service_no_acl or
        wmsperm:check(gSpace["cidAttrs"]["owner"] or "", "yesno", "SUPERVISION", dst_attrs["owner"] or "", gSpace["cidAttrs"]["licenseType"]) == "yes"
        if not is_allowed then
            print("Rejected by group permissions")
            to_logfile("Hangup", hcause.ACL_FORBIDDEN)
            return app.Hangup(hcause.ACL_FORBIDDEN)
        end
        
        -- set correct name for connected line (add PAE)
        set_call_id("name", dst_attrs.cn, "out")
        set_call_id("num", dst_attrs.telephoneNumber, "out")

        local dst_pbx    = pbxutils.getPbxDataByDN(dst_attrs.pbxowner)
        local our_pbx_dn = pbxutils.getPbxDNBySerial()
        if dst_pbx.dn ~= our_pbx_dn then -- destination user belongs another PBX
            local pref = stype:match("^%w"):lower()
            local ext = pref .. "barge*" .. destnumber
            app_remote_call(self, nil, ext, dst_pbx.dn, "pbxservices")
            local unavailable_pbx = channel["X-PBX_UNAVAILABLE"]:get() or ""
            if unavailable_pbx ~= "" then
                -- get destination user PBX failover configuration
                local failoverObj = require"common.failover":new(dst_attrs.pbxowner)
                if failoverObj and "1" == failoverObj:isEnabled() then 
                    if not failoverObj:localPbxIsFailover() and failoverObj:getPbxDN() then -- call via remote pbx
                        app_remote_call(self, "Making intrusion to failover user at", ext, failoverObj:getPbxDN(), "pbxservices")
                    end
                else
                    to_logfile("Hangup", hcause.NORMAL)
                    return app.Hangup(hcause.NORMAL)
                end
            end
        end

        local options
        if stype == "PREANSWER" then    -- at preanwer service check device status
            options = "Ewq"
        else
            local device_available_intrusion = require"deviceState":new(destnumber):reqValue():isAvailable4Service("INTRUSION")
            if device_available_intrusion then
                if stype == "CBARGE" then options = "EBq"
                elseif stype == "SBARGE" then options = "Eq"
                elseif stype == "PBARGE" then options = "Ewq"
                end
            end
        end
        
        local intrusionNotification = channel["INTRUSION_NOTIFICATION"]:get() or "no"

        if options then
            if intrusionNotification == "yes" then 
                options = options .. "P"
                print("call intrusion notification enabled")
                logger:info("call intrusion notification enabled")
            end
            to_logfile("ChanSpy", "SIP/" .. destnumber, options or "")
            channel.SERVICECALL:set("true")
            app.ChanSpy("SIP/" .. destnumber, options or "")
            to_logfile("Hangup", hcause.NORMAL)
            return app.Hangup(hcause.NORMAL)
        else
            to_logfile("Hangup", hcause.FAILURE)
            return app.Hangup(hcause.FAILURE)
        end
    end;
    
    --- Relay a document from a fax to the destination extension and to fax again
    -- @param destnumber A number where send email (or send fax if not specify faxnumber)
    -- @param faxnumber A destination number for sending fax
    ["FAXRELAY"] = function(destnumber, faxnumber)
        local pbxutils       = require"core.pbxutils"
        local fsutils        = require"core.fsutils"
        local app, hcause    = app, hcause
        local subject        = ""
        local is_direct_mode = false
        
        if not destnumber then
            ivr:tell("The value you entered is not valid. Hangup")
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            return app.Hangup(hcause.CHANUNAVAIL) 
        end
        
        -- support <realycode>(*)<faxnumber>
        local faxsender = gSpace["cidNum"]
        if destnumber and not faxnumber then
            -- direct faxing mode
            if not string.find (destnumber,"^[0-9\*\+]*$") then
                ivr:tell("The value you entered is not valid. Hangup")
                to_logfile("Hangup", hcause.CHANUNAVAIL)
                return app.Hangup(hcause.CHANUNAVAIL)
            end
            is_direct_mode         = true
            faxnumber              = destnumber
            subject                = faxsender
            gSpace["gvars"].FAXRELAY_MODE = "no-mail" --used for disable send copy mail to cid
        else
            if not string.find(destnumber,"^%d+$") or not faxnumber or
            not string.find(faxnumber,"^[0-9\*\+]*$") then
                ivr:tell("The value you entered is not valid. Hangup")
                to_logfile("Hangup", hcause.CHANUNAVAIL)
                return app.Hangup(hcause.CHANUNAVAIL)
            end
            subject   = destnumber
            faxsender = destnumber
        end
        if faxnumber then
            gSpace["gvars"].FAXRELAY = faxnumber
        end
        
        -- get faxfile from FAX machine and send them to email
        local faxfile, email = fax2mail(subject)
        if not fsutils.is_readable(faxfile) then
            print("Fax file is not exists")
            logger:warning("Fax file is not exists")
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            return app.Hangup(hcause.CHANUNAVAIL)
        end
        
        local cid_attrs = gSpace["cidAttrs"]
        if faxnumber then   --send fax
            local cmd = "CONTEXT='%s' LANGUAGE='%s' FROMNUMBER='%s' wms_makefax -n '%s' -a '%s' -m '%s'"
            pbxutils.run(cmd, cid_attrs.userFaxDialplan or "users", cid_attrs.preferredLanguage or "en", faxsender or "", faxnumber, faxfile, email or "")
        end
        __send_to_s3(faxfile, destnumber)
        if not is_direct_mode then
            cp_faxfile(faxfile, destnumber or "")
        end
        fsutils.rmfile(faxfile)
        to_logfile("Hangup", hcause.NORMAL)
        return app.Hangup(hcause.NORMAL)
    end;
    
    --- Send a document from a fax to the destination extension
    -- @param destunmber A destination number(email) for send fax
    ["SENDDOC"] = function (destnumber)
        local app, channel = app, channel
        local hcause, dst_attrs = hcause, {}
        if destnumber == nil or destnumber == "" then
            ivr:tell("No number was entered")
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            return app.Hangup(hcause.CHANUNAVAIL)
        end
        if not string.find (destnumber,"^%d+$") then
            ivr:tell("The value you entered is not valid. Hangup")
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            return app.Hangup(hcause.CHANUNAVAIL)
        end
        dst_attrs = get_obj_attrs(destnumber)
        --TODO: check for not found dst number (in case call to trunk or from trunk too!)
        local is_allowed = require"common.permissions":check(gSpace["cidAttrs"]["owner"] or "", "yesno", "IM_FAX", dst_attrs["owner"] or "", gSpace["cidAttrs"]["licenseType"]) == "yes"
        if not is_allowed then
            print("Rejected by group permissions")
            to_logfile("Hangup", hcause.ACL_FORBIDDEN)
            return app.Hangup(hcause.ACL_FORBIDDEN)
        end
        -- send fax2mail
        gSpace["gvars"].SENDDOC = true -- used for return faxfile from fax2mail
        local faxfile, _ = fax2mail(destnumber)
        to_logfile("Hangup", hcause.NORMAL)
        return app.Hangup(hcause.NORMAL) 
    end;
    
    --- Call to call group
    -- @param callgroupid A id of calling callgroup
    ["QCALL"] = function (callgroupid)
        local app, channel = app, channel
        local ivr = ivr
        
        if not callgroupid then
            callgroupid = ivr:read("Please enter group id", {retry = 1})
            if not callgroupid then
                ivr:tell("No number was entered. Good Bye. Hangup")
                to_logfile("Hangup", hcause.CHANUNAVAIL)
                return app.Hangup(hcause.CHANUNAVAIL)
            end
        end
        -- validate argument
        if not string.find(callgroupid,"^%d+$") then
            ivr:tell("The value you entered is not valid. Hangup")
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            return app.Hangup(hcause.CHANUNAVAIL)
        end
        
        app_queue_call(nil, callgroupid, "t")
        to_logfile("Hangup", hcause.CHANUNAVAIL)
        return app.Hangup(hcause.CHANUNAVAIL)
    end;
    
    --- Managing member status in call group
    -- @param action    An action for managing
    -- @param cg_id     An id of the managing call group
    -- @param member_id A managed member
    ["CGMANAGER"] = function (action, cg_id, member_id)
        local app, channel = app, channel
        
        -- get cg_id from action
        if action and 1 < action:len() then
            action, cg_id = action:match("(%d)(%d+)")
        end
        
        -- prompt action for call group manager
        if not action then
            -- get action number
            local options = {
                max_chars = 1,
                validate = function (choice)
                    return require"core.pbxutils".in_table(tonumber(choice), {0,1})
                end;
                failure = function()
                    ivr:tell("The number you entered is incorrect, please try again")
                end;
            }
            action = ivr:read("Press 1 to login or 0 to logout", options)
            if not action then
                ivr:tell("Good bye. Hangup")
                return smartApp("Hangup", hcause.FAILURE)
            end
            
            -- get call group id
            cg_id = ivr:read("Please enter group id", {retry = 1})
            if not cg_id then
                ivr:tell("No group id is defined. Good Bye. Hangup")
                return smartApp("Hangup", hcause.FAILURE)
            end
        end
        
        -- check call group id
        if not cg_id or "" == cg_id or "1" ~= channel.QUEUE_EXISTS(cg_id):get() then
            print("Call group id is not specified")
            logger:warning("Call group id is not specified")
            return ivr:play("beeperr")
        end

        
        if not member_id or "" == member_id then
            member_id = gSpace["cidNum"]
        else
            local can_manage = require"common.permissions":check(gSpace["cidAttrs"]["owner"] or "", "yesno", "MANAGE_CALLCENTER", nil, gSpace["cidAttrs"]["licenseType"]) == "yes"
            if not can_manage then
                print("Rejected by group permissions")
                return ivr:play("beeperr")
            end
        end
        
        local member_attrs = get_obj_attrs(member_id)
        if not member_attrs.uid then
            print("User %q not found in LDAP", member_id)
            logger:warning("User %q not found in LDAP", member_id)
            return ivr:play("beeperr")
        end

        local userFeatures = require"common.userFeatures.userFeatures":new(member_attrs.uid)
        if not userFeatures then
            print("Can not load user features!")
            logger:error("Can not load user features!")
            return ivr:play("beeperr")
        end
        userFeatures:load()
        local cg_wl_enable, allowed_callgroups = userFeatures:getAllowedCallgroups()
        if cg_wl_enable and not require"core.pbxutils".in_table(cg_id, allowed_callgroups) then
            print("Rejected by group permissions")
            return ivr:play("beeperr")
        end
        
        local rv         = false
        local cgm        = require"cgroups"
        local isCGMember = cgm.is_member(cg_id, member_id)
        if isCGMember and ("0" == action or "2" == action) then -- logout action
            if cgm.rm_member(cg_id, member_id) then
                if 1 > cgm.count_members(cg_id) then
                    ivr:tell("Group id " .. cg_id .. " deactivated!")
                else
                    ivr:tell("Logged out")
                end
                rv = true
            end
        elseif not isCGMember and ("1" == action or "2" == action) then -- login action
            if cgm.add_member(cg_id, member_id) then
                ivr:tell("Logged in")
                rv = true
            end
        end
        
        if not rv then
            return ivr:play("beeperr")
        end
        
        return ivr:play("beep")
    end;
    
    --- Activate paging on the defined group
    -- @param groupid A id call group used for paging
    -- @param service_no_acl A boolean value indicate call from dialplan application or not
    ["PAGING"] = function (groupid, service_no_acl)
        
        local function _getPagingChan()
            local chan = channel.CHANNELS("MulticastRTP/.*"):get() or ""
            if "" == chan then
                return false
            end
            
            -- Milticast stream implemented via two channels on different sockets
            -- Make parsing result string with actual multicase channels
            chan = chan:match("^(MulticastRTP%/[^%s]+)%s?")
            if not chan then
                return false
            end

            return chan
        end

        local function _getPagingName(id)
            local rv  = nil
            if not tonumber(id) then
                return rv
            end
            
            local connection = require"core.wsqlite":new("wmsdb")
            local sql    = "SELECT name FROM paging WHERE id='%d';"
            local cursor = connection:execute(sql, id)
            if cursor then
                rv = cursor:fetch(); cursor:close()
            end
            connection:close()

            return "" ~= rv and rv
        end

        local function _getPagingMembers(id)
            local rv = {}
            if not tonumber(id) then
                return rv
            end
            
            local connection = require"core.wsqlite":new("wmsdb")
            local sql        = "SELECT group_member FROM paging_members WHERE paging_id='%d';"
            local pmIterator = connection:rows(sql, id)
            if "function" ~= type(pmIterator) then
                return rv
            end
            for m in pmIterator do
                rv[#rv + 1] = m
                rv[m]       = m
            end

            connection:close()
            return rv
        end

        local function _getMCastGroupsByExt(ext)
            local rv = {}
            if not ext or "" == ext then
                return rv
            end

            local connection = require"core.wsqlite":new("wmsdb")
            local sql        = "SELECT p.id, p.name FROM paging AS p LEFT JOIN paging_members AS pm ON (pm.paging_id=p.id) WHERE pm.group_member=%q order by p.id limit 10;"
            local pgIterator = connection:rows(sql, ext)
            if "function" ~= type(pgIterator) then
                return rv
            end
            for g in pgIterator do
                rv[g] = g
            end

            connection:close()
            return rv
        end

        local function _paging(data)
            local groupId   = data[1]
            local removeACL = data[2]

            -- check premissions
            if not removeACL then
                if require"common.permissions":check(gSpace["cidAttrs"]["owner"] or "", "canuse", "PAGING", nil, gSpace["cidAttrs"]["licenseType"]) ~= "yes" then
                    print("Rejected by group permissions")
                    return smartApp("Hangup", hcause.ACL_FORBIDDEN)
                end
            end

            -- check running paging stream
            if _getPagingChan() then    -- allow only one paging service instance
                ivr:tell("Paging busy. Please try later")
                return smartApp("Hangup", hcause.CHANUNAVAIL)
            end

            -- retieve CPU info and set up limitation
            local slowCPU  = channel.WD_SLOW_CPU:get() or ""
            local actLimit = ({
                 ["1"] = 6      -- geod / via CPU
                ,["0"] = 15     -- atom CPU
            })[slowCPU] or 6

            -- retrieve reg info for paging members
            local pMembers = _getPagingMembers(groupId)
            local regInfo, rerr = require"core.sipproxy.sipproxy":new():getRegistrationByExten(pMembers) 
            if regInfo then
                for regData in regInfo:iterate() do
                    local pExten = regData:getExten()
                    if pMembers[pExten] then
                        if "table" ~= type(pMembers[pExten]) then
                            pMembers[pExten] = {}
                        end
                        table.insert(pMembers[pExten], regData)
                    end
                end
            end
            if rerr then
                local msg = string.format("Get registrations completed with error: %s", rerr)
                print(msg)
                logger:error(msg)
                to_logfile("Page", msg)
            end

            -- check paging members for direct or multicast dial
            local isDirectlyConnected = require"common.helper".host_settings
            local multicastDial       = {}
            local directDial          = {}
            local skipDial            = {}
            for _, num in ipairs(pMembers) do
                if "table" == type(pMembers[num]) then
                    for idx, reg in ipairs(pMembers[num]) do
                        local pUserAgent = reg and reg:getUserAgent() or ""  
                        if pUserAgent:find("Wildix WP%d+") or pUserAgent:find("Wildix W%-PA") then  -- multicast supported devices
                            if isDirectlyConnected(reg:getAddress()) then
                                local mCastGroups = _getMCastGroupsByExt(num)
                                -- group id is in the list of provisioned first 10 paging groups, no need direct dial
                                if mCastGroups[tonumber(groupId)] then
                                    multicastDial[#multicastDial + 1] = num
                                    multicastDial[num]                = num
                                    break
                                end
                            end
                        elseif pUserAgent:find("Wildix W%d+FXS") or
                               pUserAgent:find("Wildix iOS") or
                               pUserAgent:find("Wildix Android") then
                            if 1 == #pMembers[num] then
                                skipDial[num] = num
                            end
                        end
                    end 
                else    -- reg-info was not found (device unregistered)
                    skipDial[num] = num
                end
            end
            local idx = 0
            for _, num in ipairs(pMembers) do   -- processing direct dial paging users by some CPU limitations
                if not multicastDial[num] and not skipDial[num] then
                    if idx < actLimit then
                        local devStateObj = require"deviceState":new(num)
                        devStateObj:reqValue()
                        if devStateObj:isAvailable() then
                            table.insert(directDial, "SIP/" .. num)
                            idx = idx + 1
                        else
                            table.insert(directDial, "Local/" .. num .. "@disable-autoanswer")
                            idx = idx + 1
                        end
                    end
                end
            end

            -- build paging dial
            local dialOpts = ""
            local dialStr  = "MulticastRTP/basic/224.0.0.1:" .. tostring(8999 + groupId) .. "&" .. "MulticastRTP/basic/239.1.1.10:" .. tostring(9999 + groupId)
            if 0 ~= #directDial then
                channel["_VXML_URL"]:set("intercom=true")
                app.SipAddHeader("Call-Info:\ sip:wildixgw;answer-after=0")
                app.SipAddHeader("X-To-Paging-Group: yes")
                dialStr = "" ~= dialStr and dialStr .. "&" or dialStr
                dialStr = dialStr .. table.concat(directDial, "&")
            end

            -- set up variables for pickup procedure
            channel["_PAGEID"]:set(groupId)
            channel["_PAGECPEER"]:set(channel.CHANNEL("name"):get())

            -- set up paging name to caller channel
            local pageName = _getPagingName(groupId)
            if pageName then
                channel.CALLERID("name"):set(pageName)
            end

            app.CELGenUserEvent("Page", string.format("%s,%s", groupId or "", pageName or ""))

            return smartApp("Page", dialStr)
        end

        local function _pickup()
            local pagingChan = _getPagingChan()
            if pagingChan then
                local pagingId = channel.IMPORT(pagingChan, "PAGEID"):get() or ""
                local cidNum   = gSpace["cidAttrs"].telephoneNumber
                if pagingId and cidNum then
                    local pagingMembers = _getPagingMembers(pagingId)
                    if pagingMembers[cidNum] then
                        local remotePagingChan = channel.IMPORT(pagingChan, "PAGECPEER"):get() or ""
                        if "" ~= remotePagingChan then
                            smartApp("Bridge", remotePagingChan)
                            return smartApp("Hangup", hcause.NORMAL)
                        end
                    else
                        print("User does not belong to the active paging group")
                        logging.info("User does not belong to the active paging group")
                        return smartApp("Hangup", hcause.CHANUNAVAIL)
                    end
                end
            end
        end

        local function _dataValidate(method, data)
            local rvMsg    = nil
            local rvPhrase = nil
            local rvCode   = nil
            local rv       = false

            repeat
                -- pickup method
                if "pickup" == method then
                    rv = true
                    break
                end

                -- paging method
                if "table" ~= type(data) then
                    rvCode = hcause.FAILURE
                    break
                end

                local groupId   = data[1]
                local removeACL = data[2]

                -- check group id value
                if not groupId or "" == groupId then
                    rvPhrase = "No number was entered. Good Bye. Hangup"
                    rvCode   = hcause.CHANUNAVAIL
                    break
                end

                local nGroupId = tonumber(groupId)
                if not nGroupId or nGroupId > 100 or nGroupId < 1 then
                    rvPhrase = "The value you entered is not valid. Hangup"
                    rvCode   = hcause.CHANUNAVAIL
                    break
                end

                -- check local user data
                if not removeACL and (not gSpace["cidAttrs"] or not gSpace["cidAttrs"].telephoneNumber) then
                    rvMsg  = "Caller is not found in ldap. Good bye. Hangup"
                    rvCode = hcause.NOT_FOUND
                    break
                end

                rv = true
            until true

            return rv, {c = rvCode, m = rvMsg, p = rvPhrase}
        end

        -- check failover user
        if channel.FAILOVER_USER:get() == "yes" then -- disable paging for filover users
            print(string.format("Paging service for failover user %q disallowed", gSpace["cidNum"]))
            return smartApp("Hangup", hcause.FAILURE)
        end

        -- processing basic functionality
        if not groupid then
            -- suggest pickup
            if _dataValidate("pickup") then
                _pickup()
            end
        end

        -- running paging
        local groupid = groupid or ivr:read("Please enter paging group id", { retry = 1})
        local rv, rData = _dataValidate("paging", {groupid, service_no_acl})
        if not rv then
            if rData.p then
                ivr:tell(rData.p)
            elseif rData.m then
                print(rData.m)
            end
            return smartApp("Hangup", rData.c)
        end

        _paging({groupid, service_no_acl})
    end;
    
    --- Allows to record a personal message, save it on the system and send the recording via email or 
    -- message shared for PBX. If the message is already present it can be played back. Used in combination
    -- whith a call conference allows to put on hold an existing call, call the service, enter the message id.
    -- Pound stops the recording. The recording can be then played back again during a call.
    -- The id of the recording is saved in a sub directory with the name of the calling user.
    -- @param messagecode A sequence number for message
    -- @param choice A choice for activity (remove, record, playback)
    -- @param rtype A type which say what do: send to mail (personal) or use for pbx (pbx)
    ["RECORDSOUND"] = function(messagecode, choice, rtype)
        return require"pbxSound".runRecord(messagecode, choice, rtype)
    end;
    
    --- Go to access to mailbox 
    -- @param vmbox A mail box for access
    ["VMACCESS"] = function (vmbox)
        return app_voicemail(self, vmbox, "access")
    end;

    --- Transfer to voicemail
    -- @param vm_exten Extension for transfer
    ["VMTRANSFER"] = function (vm_exten)
        print ("Transfer to voicemail, extension " .. vm_exten)
        setpostaction({"name"}, "rm")
        return app_voicemail(self, vm_exten, "leave", "vmtransfer")
    end;

    --- Conference call
    -- @param conf_id A id of conference room
    ["CONFERENCE"] = function (conf_id)
        if not app_conference(conf_id) then
            ivr:tell("Good Bye. Hangup")
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            return app.Hangup(hcause.CHANUNAVAIL)
        end
        to_logfile("Hangup", hcause.NORMAL)
        return app.Hangup(hcause.NORMAL)
    end;
    
    -- picks up a ringing call for an extension or a group (no ivr)
    -- @param pickupnumber A number used for pickupchan
    ["PICKUP"] = function (pickupnumber)
        local app, channel = app, channel
        
        if pickupnumber and string.find(pickupnumber,"^cg_") then
            print("Pickup to callgroup, redirecting to QPICKUP")
            return service("QPICKUP",pickupnumber)
        end
        
        local exten = channel.EXTEN:get()
        if string.find(exten, "^(%d+)%*(%w+)$") then
            return service("GPICKUP", pickupnumber)
        end

        if not pickupnumber or pickupnumber == "" then
            print("Pickup number is empty")
            return service("GPICKUP")
        end
        
        -- pickup pri/bri/fxo/gsm trunk should not be executed
        if exten:match("^pickup%*bri") or exten:match("^pickup%*pri") or exten:match("^pickup%*gsm") or exten:match("^pickup%*fxo") then
            return service("PSTN_TRUNK_OCCUPATION")
        end
        
        -- STIME hack
        if string.find(exten, "stime") then
            local t1, t2 = channel.EXTEN:get():match("^pickup%*stime%*(%d+)%*(%d+)$")
            return stime (t1, t2)
        end

        -- 3SWITCH hack
        if string.find(exten, "3switch") then
            local t1, t2 = channel.EXTEN:get():match("^pickup%*3switch%*(%d+)%*(%d+)$")
            return switch(t1, t2, true)
        end

        local wmsperm    = require"common.permissions"
        local pbxutils   = require"core.pbxutils"
        local cid_attrs  = gSpace["cidAttrs"]

        -- processing changing identity
        if cid_attrs.uid then
            local identity_num, identity_name = __checkIdentity(cid_attrs)
            if identity_num and identity_name then
                cid_attrs = get_obj_attrs(identity_num)
                channel.CALLERID("num"):set(identity_num)
                channel.CALLERID("name"):set(identity_name)
                channel.CDR("from_name"):set(identity_name)
                app.SipAddHeader(string.format("Caller-Image: http://127.0.0.1/user_avatar/%s.png", identity_num))
            end
        end

        local dst
        local dst_attrs = get_obj_attrs(pickupnumber)
        if not dst_attrs.cn then
            print(pickupnumber, "is not an user")
            return service("GPICKUP", pickupnumber)
        end

        local remote_channels = {}
        local templates = {string.format("SIP/%s-", pickupnumber), string.format("Local/%s@mobility.*;1", pickupnumber)}
        for _, template in pairs(templates) do
            for _, chan in pairs(pbxutils.split(channel.CHANNELS(template):get() or "", "%s")) do
                table.insert(remote_channels, chan)
            end
        end

        for _, remote_channel in pairs(remote_channels) do
            local isChanBridged = "" ~= (channel.IMPORT(remote_channel, "BRIDGEPEER"):get() or "")
            local isServiceCall = "true" == channel.IMPORT(remote_channel, "SERVICECALL"):get()
            if not isServiceCall and not isChanBridged then
                local to_identity = channel.IMPORT(remote_channel, "TO_IDENTITY"):get() or ""
                local attrs = to_identity ~= "" and get_obj_attrs(to_identity) or dst_attrs
                if attrs.cn then
                    if wmsperm:check(cid_attrs["owner"] or "", "yesno", "PICKUP", attrs["owner"] or "", cid_attrs["licenseType"]) == "yes" then
                        dst = remote_channel
                        dst_attrs = attrs
                        break
                    end
                end
            end
        end

        if not dst and #remote_channels ~= 0 then
            print("Rejected by group permissions")
            smartApp("NoCDR", "")
            return smartApp("Hangup", hcause.ACL_FORBIDDEN)
        end

        if dst then
            smartApp("PickupChan", dst)
        else
            local dst_pbx    = pbxutils.getPbxDataByDN(dst_attrs.pbxowner)
            local our_pbx_dn = pbxutils.getPbxDNBySerial()
            if dst_pbx.dn ~= our_pbx_dn then
                local ext = "pickup*" .. pickupnumber
                app_remote_call(self, nil, ext, dst_pbx.dn, "pbxservices")
                local unavailable_pbx = channel["X-PBX_UNAVAILABLE"]:get() or ""
                if unavailable_pbx ~= "" then
                    -- get destination user PBX failover configuration
                    local failoverObj = require"common.failover":new(dst_attrs.pbxowner)
                    local failoverIsEnabled = failoverObj and "1" == failoverObj:isEnabled()
                    local failoverIsRemote  = failoverObj and not failoverObj:localPbxIsFailover() and failoverObj:getPbxDN()
                    if failoverIsEnabled and failoverIsRemote then
                        app_remote_call(self, "Making pickup to failover user at", ext, failoverObj:getPbxDN(), "pbxservices")
                    end
                end
            end
        end

        smartApp("NoCDR", "")
        return smartApp("Hangup", hcause.NORMAL)
    end;
    
    -- picks up group
    -- @param pickupgroup A id of goup which we want pickup
    ["GPICKUP"] = function(pickupgroup)
        local cid_attrs = gSpace["cidAttrs"]
        local group_dn  = nil
        
        if cid_attrs.uid then
            local features = require"common.userFeatures.userFeatures":new(cid_attrs.uid)
            if features then
                features:load()
                if features:isAlternatePickupMode() then
                    logger:debug("Using named call/pickup groups")
                    local groups = features:getPickupGroups()
                    local pickupgroups = ""
                    if pickupgroup and "" ~= pickupgroup then
                        if require"core.pbxutils".in_table(pickupgroup, groups["pickupgroups"]) then
                            pickupgroups = pickupgroup
                        else
                            print("Pickup group '%s' is not present in pickup groups list of the user.", pickupgroup)
                            smartApp("NoCDR", "")
                            return smartApp("Hangup", hcause.ACL_FORBIDDEN)
                        end
                    else
                        pickupgroups = table.concat(groups["pickupgroups"], ",") or ""
                    end
                    logger:debug("Setting the following pickupgroups:", pickupgroups)
                    channel.CHANNEL("namedpickupgroup"):set(pickupgroups)
                    smartApp("Pickup", "")
                    smartApp("NoCDR", "")
                    return smartApp("Hangup", hcause.NORMAL)
                else
                    logger:debug("Using numbered call/pickup groups")
                end
            else
                logger:error("Can not load user features! Using numbered call/pickup groups.")
            end
        else
            logger:error("UID not found! Using numbered call/pickup groups.")
        end

        -- processing changing identity
        if cid_attrs.uid then
            local identity_num, identity_name = __checkIdentity(cid_attrs)
            if identity_num and identity_name then
                cid_attrs = get_obj_attrs(identity_num)
                channel.CALLERID("num"):set(identity_num)
                channel.CALLERID("name"):set(identity_name)
                channel.CDR("from_name"):set(identity_name)
                app.SipAddHeader(string.format("Caller-Image: http://127.0.0.1/user_avatar/%s.png", identity_num))
            end
        end

        if not pickupgroup or "" == pickupgroup then
            -- find group of cid
            local cid_gid = string.match(cid_attrs["owner"] or "", "gid=(%d+),.*")
            if not cid_gid or cid_gid == "" then
                print("Empty group to pickup")
                smartApp("NoCDR", "")
                return smartApp("Hangup", hcause.NOT_FOUND)
            end
            
            -- find number of group
            local cid_group_attrs     = nil
            cid_group_attrs, group_dn = get_obj_attrs(cid_gid, "gid", "wGroup", true)
            pickupgroup               = cid_group_attrs["wcgrp"] or ""
            if not pickupgroup or "" == pickupgroup then
                print("Empty group to pickup")
                smartApp("NoCDR", "")
                return smartApp("Hangup", hcause.NOT_FOUND)
            end
        end
        
        -- validate argument
        if not pickupgroup:find("^%d+$") or 1024 <= tonumber(pickupgroup) then
            ivr:tell("The value you entered is not valid. Hangup")
            smartApp("NoCDR", "")
            return smartApp("Hangup", hcause.CHANUNAVAIL)
        end
        
        -- get dest group dn by pickupgroup entered by user
        if not group_dn then
            _, group_dn = get_obj_attrs(pickupgroup, "wcgrp", "wGroup", true)
            if not group_dn then
                print("Group is not found")
                smartApp("NoCDR", "")
                return smartApp("Hangup", hcause.NOT_FOUND)
            end
        end
        
        local is_allowed = require"common.permissions":check(cid_attrs["owner"] or "", "yesno", "PICKUP", group_dn or "", cid_attrs["licenseType"]) == "yes"
        if not is_allowed then
            print("Rejected by group permissions")
            smartApp("NoCDR", "")
            return smartApp("Hangup", hcause.ACL_FORBIDDEN)
        end
        channel.CHANNEL("pickupgroup"):set(pickupgroup)
        smartApp("Pickup", "")
        smartApp("NoCDR", "")
        return smartApp("Hangup", hcause.NORMAL)
    end;
    
    -- picks up call group (queue)
    ["QPICKUP"] = function (callgroupid)
        if not callgroupid then
            callgroupid = ivr:read("Please enter group id", {retry = 1})
            if not callgroupid then
                ivr:tell("No number was entered. Good Bye. Hangup")
                smartApp("NoCDR", "")
                return smartApp("Hangup", hcause.CHANUNAVAIL)
            end
        else
            callgroupid = callgroupid:gsub('cg_','')
        end
        
        -- validate argument
        if not callgroupid or not callgroupid:find ("^%d+$") then
            ivr:tell("The value you entered is not valid. Hangup")
            smartApp("NoCDR", "")
            return smartApp("Hangup", hcause.CHANUNAVAIL)
        end
        
        if channel.QUEUE_EXISTS(callgroupid):get() == "0" then
            print("Invalid call group id")
            ivr:tell("The value you entered is not valid. Hangup")
            smartApp("NoCDR", "")
            return smartApp("Hangup", hcause.CHANUNAVAIL)
        end
        
        local is_local_channel = string.find(channel["CHANNEL"]:get(), '^Local%/cg_') ~= nil or
            string.find(channel["CHANNEL"]:get(), '^Local%/qpickup') ~= nil
        if is_local_channel then
            print("Impossible to pickup")
            return false
        end
        
        -- Get a caller which calls to a call group
        local ring_mem = channel.QUEUE_CHAN_RINGING(callgroupid):get() or ""
        if "" ~= ring_mem then
            return smartApp("Bridge", ring_mem)
        end
        
        smartApp("NoCDR", "")
        return smartApp("Hangup", hcause.NORMAL)
    end;

    -- disable callerid visualization for the current call
    ["HIDECID"] = function (destnumber)
        local app, ivr, channel  = app, ivr, channel
        if not destnumber then
            destnumber = ivr:read("Please enter number you wish to dial", {retry = 1, max_chars = 15})
            if not destnumber then
                ivr:tell("Good bye. Hangup")
                to_logfile("Hangup", hcause.NORMAL)
                return app.Hangup(hcause.NORMAL)
            end
        end
        
        -- validate argument
        if not string.find(destnumber,"^%d+$") then
            ivr:tell("The value you entered is not valid. Hangup")
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            return app.Hangup(hcause.CHANUNAVAIL)
        end
        
        local cid                = gSpace["cidNum"]
        local cid_attrs          = gSpace["cidAttrs"]
        local context            = cid_attrs["userDialplan"] or "users"
        gSpace["gvars"].anon_num = true
        channel.CALLERID("name"):set("restricted") -- "anonymous"
        to_logfile("Goto", context, destnumber, 1)
        return app.Goto(context, destnumber, 1)     
    end;
    
    ["SHORTNUM"] = function(shortNumber)
        local ivr = ivr
        
        if not shortNumber then
            shortNumber = ivr:read("Please enter short number you wish to dial", {retry = 1})
            if not shortNumber then
                ivr:tell("Good bye. Hangup")
                to_logfile("Hangup", hcause.NORMAL)
                return app.Hangup(hcause.NORMAL)
            end
        end
        
        -- validate argument
        if not string.find(shortNumber, "^%d+$") then
            ivr:tell("The value you entered is not valid. Hangup")
            to_logfile("Hangup", hcause.CHANUNAVAIL)
            return app.Hangup(hcause.CHANUNAVAIL)
        end
        
        -- set funtion variables
        local permissions = require"common.permissions"
        
        -- retrieve user LDAP attributes (in case blind transfer too)
        local tr_num   = channel["TRANSFER_NUMBER"]:get() or ""
        local cidAttrs = "" ~= tr_num and get_obj_attrs(tr_num) or gSpace["cidAttrs"] 
        local cidOwner = cidAttrs.owner or ""
        
        -- check ACL permissions                
        if "yes" ~= permissions:check(cidAttrs["owner"] or "", "canuse", "SHORTNUM", nil, cidAttrs["licenseType"]) then
            print("Rejected by group permissions")
            to_logfile("Hangup", hcause.ACL_FORBIDDEN)
            return app.Hangup(hcause.ACL_FORBIDDEN)
        end
        
        -- retrieving user available phonebooks
        -- note: the same phonebook search uses in core.lua:get_list_phonebook
        local cidFeatures = gSpace["cidFeatures"]
        if "" ~= tr_num then
            local features = require"common.userFeatures.userFeatures":new(cidAttrs.uid)
            if features then
                features:load()
                cidFeatures = features
            end
        end
        local userSelectPb = cidFeatures:get("PHONEBOOK-ID") or ""
        local isAllowAll   = "yes" == permissions:check(cidOwner, "canset", "PHONEBOOK", nil, cidAttrs["licenseType"]) or "" == userSelectPb
        local sqlPbQuery   = "SELECT rid FROM cti_phonebooks AS pb WHERE "
        if isAllowAll then
            -- select all phonebooks (exception LDAP users)
            sqlPbQuery = string.format("%s (pb.owner_id=%d and pb.type=1) OR pb.type=2", sqlPbQuery, cidAttrs.uid)
        else
            -- only user selected public phonebooks
            sqlPbQuery = string.format("%s rid IN (%s)", sqlPbQuery, userSelectPb or "")
        end
        
        -- select data from phonebooks table
        local pbConn       = require"core.wsqlite":new("phonebook")
        local sqlPbConcats = "SELECT extension, phone, mobile, office, home, home_mobile, username FROM cti_contacts WHERE shortcut='%s' AND phonebook_id IN (%s)"
        local cursor       = pbConn:execute(sqlPbConcats, shortNumber, sqlPbQuery)
        if not cursor or "boolean" == type(cursor) then
            pbConn:close()
            print(string.format("Cannot find the number for shortcut [%s]. DB query failed.", shortNumber))
            logger:error("Cannot find the number for shortcut [%s]. DB query failed.", shortNumber)
            return smartApp("Hangup", hcause.CHANUNAVAIL)
        end
        local contExten, contPhone, contMobile, contOffice, contHome, contHomeMobile, contName = cursor:fetch()
        cursor:close(); pbConn:close()
        
        -- searching short call extension
        local shortExtList = {}
        table.insert(shortExtList, "" ~= contExten and contExten or nil)
        table.insert(shortExtList, "" ~= contPhone and contPhone or nil)
        table.insert(shortExtList, "" ~= contMobile and contMobile or nil)
        table.insert(shortExtList, "" ~= contOffice and contOffice or nil)
        table.insert(shortExtList, "" ~= contHome and contHome or nil)
        table.insert(shortExtList, "" ~= contHomeMobile and contHomeMobile or nil)
        
        local shortExt = shortExtList[1]
        if not shortExt then
            print("Not found contact extension for short number "..shortNumber)
            if "" ~= tr_num then
                return require"internalcalls":return_from_transfer(tr_num, shortNumber, "NOUSER", "failExten")
            end
            return smartApp("Hangup", hcause.NOT_FOUND)
        end
        
        -- finialize short number call service
        if "" == contName then
            return true
        end
        set_call_id("name", contName, "out")
        gSpace["gvars"].shortnumber_call = true
        return smartApp("Goto", cidAttrs.userDialplan, shortExt, 1)
    end;
    
    --- Login system service
    ["LOGIN"] = function (number, wpin)
        local app, channel, ivr = app, channel, ivr
        local pbxutils = require"core.pbxutils"
        local gappl = gSpace["gvars"].application
        local cid   = gSpace["cidNum"]
        local is_failover_user = channel.FAILOVER_USER:get() == "yes"
        if is_failover_user then 
            local cid_attrs = get_obj_attrs(cid)
            local failoverObj = require"common.failover":new(cid_attrs.pbxowner)
            -- Allow login for users on not failover PBX
            if failoverObj and not failoverObj:localPbxIsFailover() then
                is_failover_user = false
            end
        end

        if gappl and gappl:match("^DISA.*") or is_failover_user then
            print(string.format("Service for %s is denied", is_failover_user and "failover user" or "application DISA"))
            ivr:tell("Service currently unavailable.")
            return smartApp("Hangup", hcause.FAILURE)
        end
        
        local src_pbx = pbxutils.getPbxDataBySerial()
        if not src_pbx then
            print("Cannot find our pbx in ldap!")
            logger:error("Cannot find our pbx in ldap!")
            return smartApp("Hangup", hcause.CHANUNAVAIL)
        end
        
        return require"stdService.login.common".run(number,wpin)
    end;
    
    -- activate/disable the conditional behavior of the switch or switch3 (on / off)
    ["SWITCH"] = function (swid, doset, sw3type)
        return switch(swid, doset, sw3type) 
    end;
    
    -- sets the behavior of a time condition
    ["STIME"] = function (timetableid, doset)
        return stime(timetableid, doset)
    end;
    
    -- proceed bri/pri/fxo/gms occupation
    ["PSTN_TRUNK_OCCUPATION"] = function ()
        local channel, app = channel, app
        local exten = channel.EXTEN:get()
        
        if not exten or exten == "" then
            to_logfile ("Hangup", hcause.NORMAL)
            return app.Hangup(hcause.NORMAL)
        end
        
        local trunk_name, number = exten:match("(.*)%*(.*)")
        if trunk_name and number then
            return app_external_call(number, trunk_name, nil, nil, nil)
        end
        
        if exten:find ("pickup") then
            exten = exten:match("^pickup%*(.*)")
        end

        local trunk_state = channel.DEVICE_STATE("SIP/" .. exten):get()
        if trunk_state == "RINGING" then
            print (exten .. " is Ringing")
        elseif trunk_state == "BUSY" then
            print (exten .. " is Busy")
        else
            local options = { nobeep = true, max_chars = 0, retry = 1, timeout = 0.1, --[[opts = "n",]] iteration = 4 } --TODO: after fixing 302 Moved in asterisk
            local number = ivr:read("dialtone", options)
            if not number then
                print("No number was inserted")
                to_logfile ("Hangup", hcause.NORMAL)
                return app.Hangup(hcause.NORMAL)
            end
            to_logfile ("NoCDR", "")
            app.NoCDR()
            to_logfile ("Transfer", "SIP/" .. exten .. "*" .. number)
            return app.Transfer("SIP/" .. exten .. "*" .. number)
        end
    end;

    ["DIRECTORY"] = function(...)
        channel.SERVICECALL:set("true")
        return require"Directory".run(channel.IVR_ASR_SERVICES:get() ~= "no" and "a" or "")
    end;
    
    -- run MUR service for hotels
    ["WHOTELMUR"] = function(...)
        return require"exService":service_mur(...)
    end;
    
    -- run BAR service for hotels
    ["WHOTELBAR"] = function(...)
        return require"exService":service_minibar(...)
    end;
    
    -- execute ALARM service for hotels
    ["WHOTELALARM"] = function(...)
        return require"exService":service_wakeup(...)
    end;
    
    -- run INSPECT service for hotels
    ["WHOTELINSPECT"] = function(...)
        return require"exService":service_inspect(...)
    end;

    --- Change current identity
    -- @param identitynumber    Requested identity
    ["IDENTITY"] = function (identitynumber)
        local ivr = ivr
        local userFeatures, available_identities, current_identity

        local cid_attrs = gSpace["cidAttrs"]
        if cid_attrs.uid then
            userFeatures = require"common.userFeatures.userFeatures":new(cid_attrs.uid)
            if userFeatures then
                userFeatures:load()
                available_identities = userFeatures:getCustomIdentities() or {}
                current_identity = userFeatures:getValue("CURRENT_IDENTITY") or ""
            end
        end
        if not userFeatures then
            print("Can't load user features")
            return smartApp("Hangup", hcause.NORMAL)
        end
        if #available_identities == 0 then
            print("Identities list of the user is empty")
            app.Wait(1); app.Playback("beeperr"); app.Wait(1);
            return smartApp("Hangup", hcause.NORMAL)
        end

        local function changeIdentity()
            if not identitynumber or "" == identitynumber then
                return false
            end
            if not require"core.pbxutils".in_table(identitynumber, available_identities) and cid_attrs.telephoneNumber ~= identitynumber then
                print("Identities list of the user doesn't contain the requested identity:", identitynumber)
                return false
            end
            local identity_attrs = get_obj_attrs(identitynumber)
            if not identity_attrs.uid then
                print("User '%s' not found", identitynumber)
                return false
            end

            if cid_attrs.telephoneNumber == identitynumber then
                userFeatures:setValueData("CURRENT_IDENTITY", "")
            else
                userFeatures:setValueData("CURRENT_IDENTITY", identitynumber)
            end

            local rv, err = userFeatures:saveData()
            if rv then
                require"utils":notify_peer(cid_attrs.telephoneNumber, 0)
            else
                print("Identity has not changed:", err or "")
                app.Playback("beeperr")
            end
            return rv
        end

        if not identitynumber or "" == identitynumber then
            identitynumber = ivr:read("Please enter phone number, followed by pound", {retry = 1})
        end

        local attempts = 3
        repeat
            local rv = changeIdentity()
            if rv then
                app.Wait(1)
                ivr:play("Activated")
                return smartApp("Hangup", hcause.NORMAL)
            end

            if attempts == 0 then
                return smartApp("Hangup", hcause.NORMAL)
            end

            identitynumber = ivr:read("The number you entered is incorrect. Available " .. table.concat(available_identities, " or ") ..
                                      ". Please enter phone number, followed by pound", {retry = 1})
            attempts = attempts - 1
        until false
    end;

    ["CONTINUITYCALL"] = function (exten)
        return app_continuity_call("service", exten)
    end;
}

--- The main procedure to manage services
-- In general the function gets name of services
-- @param name A name of service in upper case
-- @param exten A dst number for service
-- @param length Length of digits to strip from extension
-- @param mixed Values for the call service
function service(name, exten, length, ...)
    local services, hcause = services, hcause
    local channel = channel
    
    --- check number of the supported parameters by service
    local function check_param (srv_name)
        local t = {
            ["RECORDSOUND"] = 2, 
            ["SWITCH"]      = 2,
            ["STIME"]       = 2,
            ["FAXRELAY"]    = 2,
            ["LOGIN"]       = 2,
            ["CGMANAGER"]   = 3,
            
            ["WHOTELBAR"]   = 2,
            ["WHOTELALARM"] = 3 
        }
        
        return t[srv_name] or 1
    end

    if services[name] then
        if length ~= "app" then
            print("Service " .. name)
            logger:info("Service " .. name)
        end
        
        -- disable video for services
        if "ECHO" ~= name then
           channel._DISABLE_VIDEO:set(1)
        end
        
        if length and length ~= "app" then
            local exten = channel.EXTEN:get()
            local parameters = {nil, ...}
            local pbxutils = require"core.pbxutils"
            local srv_param_num = check_param(name)
            if length > 0 then
                exten = exten:sub(length+1)
                for i, value in ipairs(pbxutils.split(exten,"*")) do
                    if i <= srv_param_num then
                        parameters[i] = value ~= "" and value or false
                    end
                end
            else
                for i, value in ipairs(pbxutils.split(exten,"*")) do
                    if i > 1 and i-1 <= srv_param_num then
                        parameters[i-1] = value ~= "" and value or false
                    end
                end
            end
            local action = parameters[1]
            if "HIDECID" == name and 1 >= string.len(action or "") then
                length = length > 0 and length or nil
                return feature("HIDECID", length)
            else
                return services[name](unpack(parameters))
            end
            return services[name](unpack(parameters))
        else
            return services[name](exten, ...)
        end
    else
        print("Service " .. (name or "") .. " not found!")
    end
end

--- Procedure for user status updating
-- @param choice    The choice for managing status
-- @param silent    The flag for disabling sound IVR (in case BLF using) 
local function __statusFeatures(choice, silent)
    --- Describe validators
    local __statusChoiceValidator = function(v)
        return "0" == v or "1" == v or "2" == v
    end
    local function __failure()
        return ivr:tell("The number you entered is incorrect, please try again")
    end
    
    if choice and not __statusChoiceValidator(choice) then
        ivr:tell("The value you entered is not valid. Hangup", {silent = silent})
        return smartApp("Hangup", hcause.NORMAL)
    end
    
    local statusACLName = "STATUS"
    local is_allowed    = require"common.permissions":check(gSpace["cidAttrs"].owner, "canset", statusACLName, nil, gSpace["cidAttrs"]["licenseType"]) == "yes"
    if choice and not is_allowed then
        print("Set up 'User status' is rejected by group permissions")
        return smartApp("Hangup", hcause.ACL_FORBIDDEN)
    end
    
    if not choice then                -- playback IVR menu for set up user status
        -- loading user status
        local cidFeatures = gSpace["cidFeatures"]
        local userStatus  = cidFeatures:getUserStatus()
        local announce    = {"User status " .. userStatus:upper()}
        
        if not is_allowed then  -- Hangup by ACL
            announce[#announce+1] = "Good bye. Hangup"
            ivr:tell(table.concat(announce, " "), {silent = silent})
            print("Set up 'User status' is rejected by group permissions")
            return smartApp("Hangup", hcause.ACL_FORBIDDEN)
        end
        
        local preoptions = {
            max_chars = 1
           ,silent    = silent
           ,retry     = 1
           ,timeout   = 0.1
           ,nobeep    = true
           ,validate  = function() return true end
        }
        choice = ivr:read(table.concat(announce," "), preoptions)
        
        if not __statusChoiceValidator(choice) then
            local announce        = {"Press"}
            announce[#announce+1] = "0 for AVAILABLE"
            announce[#announce+1] = "or 1 for DND"
            announce[#announce+1] = "or 2 for AWAY"
            
            local options = {
                 max_chars = 1
                ,silent   = silent
                ,validate = __statusChoiceValidator
                ,failure  = __failure
            }
            choice = ivr:read(table.concat(announce," "), options)
            if not choice then
                ivr:tell("The value you entered is not valid. Hangup", {silent = silent})
                return smartApp("Hangup", hcause.NORMAL)
            end
        end
    end
    
    -- set up status value by choice
    local statusVal = nil
    if choice then
        statusVal = "available"
        if "1" == choice then
            statusVal = "dnd"
        elseif "2" == choice then
            statusVal = "away"
        end
    end
    
    if statusVal then
        local cidExten = gSpace["cidNum"]
        if require"common.userFeatures.userStatus":setDataToJabber(cidExten, {status = statusVal}) then
            ivr:tell("User status updated", {silent = silent})
        end
    end
    
    return smartApp("Hangup", hcause.NORMAL)
end

--- The main procedure to manage features
-- In general the function gets a name of a feature, load user features table 
-- with current state and gives an ability to change state/value of the feature
-- @param name      An user feature name in upper case
-- @param lenght    The lenght of digits to strip from extension
function feature(name, length)
    local app, channel = app, channel
    local hcause, ivr  = hcause, ivr
    local exten        = channel.EXTEN:get()
    local pbxutils     = require"core.pbxutils"
    
    if not name then
        return smartApp("Hangup", hcause.NORMAL)
    end
    
    local cid       = gSpace["cidNum"]
    local cid_attrs = gSpace["cidAttrs"]
    if not cid_attrs.uid then
        print("User not found in LDAP. Exit.")
        logger:warning("User not found in LDAP. Exit.")
        return smartApp("Hangup", hcause.NOT_FOUND)
    end
    
    if channel.FAILOVER_USER:get() == "yes" then
        local failoverObj = require"common.failover":new(cid_attrs.pbxowner)
        if not failoverObj:localPbxIsFailover() then
            print(string.format("Features for failover user disallowed"))
            return smartApp("Hangup", hcause.FAILURE)
        end
    end

    -- disable video for features
    channel._DISABLE_VIDEO:set(1)
    
    -- processing features parameters
    local params = {}
    if length then -- Not supported class value in this case
        exten             = exten:sub(length+1) -- skip extension number
        local choice      = exten:sub(1,1)      -- get first digit as choice
        params[#params+1] = choice
        if choice == "2" or choice == "3" then  -- second digit as value
            local val         = exten:sub(2)
            params[#params+1] = val and "" ~= val and val or nil
        end
    else
        local i =  0
        for value in exten:gmatch("%w+") do -- skip feature code from parameters
            if i > 0 then params[#params+1] = value end
            i = i + 1
        end
    end
    local arg1, arg2, arg3 = unpack(params)
    
     -- set silent mode (do not tell anything, will use for function keys)
    local silent = exten:find("%*mute")
    
    -- goto status feature
    if "STATUS" == name then
        return __statusFeatures(arg1, silent)
    end
    
    -- check hotel user kind
    if "room" == cid_attrs.userKind then
        print("Set " .. name .. " is rejected for hotel room users")
        return smartApp("Hangup", hcause.ACL_FORBIDDEN)
    end
    
     -- Set up feature type
    local cid_features      = gSpace["cidFeatures"]
    local isGlobalFeature   = not cid_features:isSubClassFeature(name)
    local isTimeoutFeature  = false
    local isForwardFeature  = false
    local isNonValueFeature = false
    if pbxutils.in_table(name, {"TOUT","MEXT"}) then
        isTimeoutFeature = true
    elseif pbxutils.in_table(name, {"BCF", "UCF", "FCF"}) then
        isForwardFeature    = true
    else
        isNonValueFeature = true
    end
    
    -- init function arguments
    local choice           = arg1
    local status           = nil
    local value            = nil
    local class            = nil
    local singleActive     = true
    if "1" == choice or "0" == choice then
        class = arg2
    elseif "2" == choice then
        value = arg2
        class = arg3
    elseif "3" == choice then
        class = arg2
    end
    
    if class then   -- Set defined name for class
        class = ({
             ["0"] = "internal"
            ,["1"] = "external"
            ,["2"] = "blacklist"
            ,["3"] = "whitelist"
        })[class]
    else
        class        = "allactive"
        singleActive = false
    end
    -- fix up processing global feature
    if isGlobalFeature then 
        singleActive = true
        class        = nil
    end

    -- fix up feature choice
    if not choice then
        if "ROAD" == name then
            choice = "1"
        end
    end

    -- set up active class for user feature
    local actSubClass     = cid_features:getActiveSubClass()
    if isGlobalFeature then
         print(string.format("Processing global feature '%s' ...", name))
    elseif singleActive then
        cid_features:setUpActiveClass(class)
        local actClass    = cid_features:getActiveClass()

        if not cid_features:isActivatedClassValues() then
            print(string.format("Class/status (%s/%s) is not activated. Please make activation firstly!", actClass, actSubClass))
            return smartApp("Hangup", hcause.NOT_FOUND)
        end

        print(string.format("Processing feature '%s' for class/status - '%s/%s'", name, actClass, actSubClass))
    elseif "allactive" == class then
        print(string.format("Processing feature '%s' for all active classes with status - '%s'", name, actSubClass))
    else
        print("Processing feature for multi classes...")
    end
    
    --- Describe validators
    local function __choiceValidator(v)
        return "0" == v or "1" == v or "3" == v or ("2" == v and not isNonValueFeature)
    end
    local function __timeoutValidator(v)
        local timeout = tonumber(v)
        return timeout and timeout <= 60 -- max 60 sec
    end
    local function __numberValidator(v)
        return "string" == type(v) and v:find("^%d+$")
    end
    local function __failure()
        return ivr:tell("The number you entered is incorrect, please try again")
    end

    local init_choice = __choiceValidator(choice)
    
    -- Check ACL rules
    local is_allowed = require"common.permissions":check(cid_attrs["owner"], "canset", name, nil, cid_attrs["licenseType"]) == "yes"
    if init_choice and not is_allowed then
        print("Set " .. name .. " is rejected by group permissions")
        return smartApp("Hangup", hcause.ACL_FORBIDDEN)
    end
    
    local function __updateConnectedLineInfo(featureName, statusStr)
        if not featureName or not statusStr then
            return false
        end
        local tr = require"common.translation".translation
        tr:set_language(string.lower(channel.CHANNEL("Language"):get() or "en"))
        local opts = {
                 Channel           = channel.CHANNEL:get() or ""
                ,ConnectedLineNum  = tr:_t("feature.state." .. statusStr)
                ,ConnectedLineName = tr:_t("feature." .. featureName)
            }
        local rs = require"core.astman".send_action("SIPUpdatePAI", opts)
        if not rs then -- show logs from astman module
            print("ConnectedLineNum and ConnectedLineName were not updated")
        else
            print("ConnectedLineNum and ConnectedLineName updated")
        end
    end
    
    -- Reading user input
    local actFeatVal = singleActive and cid_features:get(name)
    if not init_choice then
        local announce        = {"Feature " .. name}
        if singleActive then
            announce[#announce+1] = actFeatVal and "activated" or "deactivated"
            if actFeatVal and not isNonValueFeature then
                if isTimeoutFeature then
                    announce[#announce+1] = "with timeout [" .. actFeatVal .. "] seconds"
                else
                    announce[#announce+1] = "with destination number " .. actFeatVal
                end
            end
        end
        
        if not is_allowed then  -- Hangup by ACL
            announce[#announce+1] = "Good bye. Hangup"
            ivr:tell(table.concat(announce, " "), {silent = silent})
            print("Set " .. name .. " is rejected by group permissions")
            return smartApp("Hangup", hcause.ACL_FORBIDDEN)
        end

        local preoptions = {
             max_chars = 1
            ,silent    = silent
            ,retry     = 1
            ,timeout   = 0.1
            ,nobeep    = true
            ,validate  = function() return true end
        }
        __updateConnectedLineInfo(name, actFeatVal and "activated" or "deactivated")
        choice = ivr:read(table.concat(announce, " "), preoptions)
        if not __choiceValidator(choice) then
            local announce = {"Press"}
            if not singleActive then
                announce[#announce+1] = "0 to deactivate or 1 to activate"
            else
                if actFeatVal then
                    announce[#announce+1] = "0 to deactivate"
                else
                    announce[#announce+1] = "1 to activate"
                end
            end
            if not isNonValueFeature then
                if isTimeoutFeature then
                    announce[#announce+1] = "or press 2 to set timeout followed by pound"
                else
                    announce[#announce+1] = "or press 2 to set the destination number followed by pound"
                end
            end
            local options = {
                max_chars = 1
                ,silent   = silent
                ,validate = __choiceValidator
                ,failure  = __failure
            }
            choice = ivr:read(table.concat(announce," "), options)
            if not choice then
                ivr:tell("The value you entered is not valid. Hangup", {silent = silent})
                return smartApp("Hangup", hcause.NORMAL)
            end
        end
    end

    -- Processing user enetered value
    if "2" == choice then
        local options = {
             max_chars = 12
            ,silent    = silent
            ,failure   = __failure
        }
        local text = {}
        if isTimeoutFeature then
            options.validate = __timeoutValidator

            text[#text+1]    = "Please enter number"
            if value and not __timeoutValidator(value) then
                __failure(); value = nil
            end
        else
            options.validata = __numberValidator
            if isForwardFeature then
                text[#text+1] = "Press 1 for voicemail or"
            end
            text[#text+1] = "please enter phone number"
        end

        -- Reading user value
        if not value then
            value = ivr:read(table.concat(text," "), options)
            if not value then
                ivr:tell("The value you entered is not valid. Hangup", {silent = silent})
                return smartApp("Hangup", hcause.NORMAL)
            end
        end
        
        -- Fixup forward value to voicemail
        if isForwardFeature and "1" == value then
            value = "VOICEMAIL"
        end
    end
    
    -- Processing user choice to modify features
    if "0" == choice or "1" == choice then                          -- Set up feature state
        status = choice
    elseif "3" == choice then                                       -- Set up feature status by switch (with value as option)
        if singleActive then
            status = actFeatVal and "0" or "1"
        else
            -- set up status by value from all classes in following ways: 
            -- set enable for available classes if feature is disabled for at least one class 
            -- set disable or available classes if feature is enabled for available classes
            status = "0"
            for actClass, featStatus in cid_features:getStatusOverActiveClasses(name) do
                if false == featStatus then
                    status = "1"
                    break
                end
            end
        end
    end
    
    -- Updating user features
    if "allactive" == class then
        local updClassed = cid_features:setValueDataOverActiveClasses(name, {
                                 status = status
                                ,value  = value
                            })
        if 0 ~= #updClassed then
            print("Updated features in following classes:", table.concat(updClassed, ", "))
        else
            print("Not found classes for the feature updating!")
            logger:warning("Not found classes for the feature updating!")
        end
    else
        cid_features:set(name, {
            status = status
           ,value  = value
       })
    end
    cid_features:save()
    if status then
        local features_to_sip_sync = {"FCF"}
        if pbxutils.in_table(name, features_to_sip_sync) then
            require"common.userFeatures.syncUserFeaturesViaSip":byExten(cid, name, cid_features)
        end
    end
    if "ROAD" == name and "1" == status then
        local userAgent       = channel.SIP_HEADER("User-Agent"):get() or ""
        local registerContact = channel.SIP_HEADER("X-Wildix-Registered-Contact"):get() or ""
        if "" ~= userAgent and "" ~= registerContact then
            local sipproxy = require"core.sipproxy.sipproxy":new()
            local rv, err  = sipproxy:setActiveDevice({
                 extension = cid
                ,contact   = registerContact
                ,userAgent = userAgent
            })
            
            if rv then
                print(string.format("Active device for '%s' set up to '%s'!", cid, userAgent))
            else
                print(string.format("Could not set up active device for '%s': %s", cid, err or "unknown error"))
                logger:warning("Could not set up active device for '%s': %s", cid, err or "unknown error")
            end
        end
    end
    
    -- actions result notification
    if singleActive and not isNonValueFeature then    
        value = cid_features:getValue(name)
    end
    if status then
        local statusStr = "1" == status and "activated" or "deactivated"
        __updateConnectedLineInfo(name, statusStr)
        ivr:tell("Feature " .. name .. " " .. statusStr, {silent = silent})
        app.wait(2)
    elseif value then
        ivr:tell("Feature " .. name .. " updated", {silent = silent})
    end

    repeat  -- condition to play notification about feature value
        if not value or "" == value then
            break
        end
        if "allactive" == class and "2" ~= choice then
            break   -- only in case force value set up for all active classes
        end
        if "1" ~= status and "2" ~= choice then
            break   -- for activated or in case directly set
        end
        
        if isTimeoutFeature then
            ivr:tell("Set timeout [" .. value .. "] seconds", {silent = silent})
        elseif isForwardFeature then
            ivr:tell("Set destination number " .. value, {silent = silent})
        end
    until true
    
    return smartApp("Hangup", hcause.NORMAL)
end

--- Post actions after hangup the channel
-- @usage
--   setpostaction {...} - set credentials to use here
--   getpostaction       - get credentials to use here
function hangup (self)
    local action = getpostaction()
    
    if action and action.name == "QUEUE" then
        local isProcessing = false
        local dialStatus   = channel["DIALSTATUS"]:get() or ""
        if "ANSWERED" ~= dialStatus then
            local qpikup_done  = channel.BRIDGING_CHANNEL_APP:get() or ""
            if "" == qpikup_done and "SUCCESS" ~= channel.VMSTATUS:get() and "yes" ~= channel.parkorbit:get() then
                isProcessing = true
            end
        end

        local manager = action.manager
        local mwi_manager = action.mwi_manager
        local isProcessing_manager = isProcessing and manager and "" ~= manager
        local isProcessing_mwi = isProcessing and mwi_manager and "" ~= mwi_manager
        local mwi_mail = false

        if isProcessing_mwi then
            local mwi_attrs = get_obj_attrs(mwi_manager)
            local wmsperm   = require"common.permissions"
            if wmsperm:check(mwi_attrs["owner"], "canuse", "MWI_MANAGER", nil, mwi_attrs["licenseType"]) ~= "nouse" then
                local cid = gSpace["cidNum"] or channel.CALLERID("num"):get()
                local cname = gSpace["cidAttrs"].cn or channel.CALLERID("name"):get() or cid
                if action.mwi_email_notify then -- WMS-10797
                    if mwi_attrs.mail then
                        mwi_mail = true
                        local mailnotify = require"common.mailNotifications"
                        local result, err = mailnotify.missedCallNotify(mwi_attrs.mail, cname, cid, mwi_attrs.preferredLanguage)
                        if not result then
                            print("Error, can not send email to mwi manager: ", err)
                            logger:error("Can not send email to mwi manager: ", err)
                        end
                    else
                        print("Error, email is not set fot mwi manager ", mwi_manager)
                        logger:error("Email is not set for mwi manager ", mwi_manager)
                    end
                else
                    local vmApp = require"common.voiceMailApplication"
                    local res, err = vmApp.writeMissedMsg(mwi_manager, cid, cname)
                    if not res then
                        print("Missed call file was not created:", err)
                        logger:error("Missed call file was not created:", err)
                    end
                end
            else
                print(string.format("User %s forbidden for mwi manager due to license type", mwi_manager))
            end
        else
            print("Notification for mwi manager was not sent")
        end

        -- WMS-9878 Fix c_to field. We have to cut off the penalty in case of queue with call all 32 strategy only.
        if channel.QUEUESTRATEGY:get() == "wringall" then
            local cdr_cto = channel.CDR("c_to"):get() or ""
            if cdr_cto:find(",") then
                channel.CDR("c_to"):set(cdr_cto:gsub(",%d+", ""))
            end
        end

        -- loading manager data
        local mng_attrs    = nil
        local mng_features = nil
        if isProcessing_manager then
            isProcessing_manager = false
            mng_attrs    = get_obj_attrs(manager)
            if mng_attrs.uid then
                mng_features = require"common.userFeatures.userFeatures":new(mng_attrs.uid)
                if mng_features then
                    mng_features:load()
                    mng_features:setUpActiveClass(gSpace["gvars"].calltype)
                    mng_features:setUpActiveClassValuesByActivatedFlag()
                    isProcessing_manager = true
                end
            end
        end
        
        
        if isProcessing_manager then
            local srcData = {
                 id    = gSpace["cidNum"]
                ,attrs = gSpace["cidAttrs"]
            }
            local dstData = {
                 id    = mng_attrs.telephoneNumber
                ,attrs = mng_attrs
                ,feat  = mng_features
            }
            local notifier, errMsg = require"missedCallNotify":new(srcData, dstData, dialStatus)
            if notifier then
                if not isProcessing_mwi then
                    notifier:SIPNotifyEvent()
                end
                if not mwi_mail then
                    notifier:textMsgEvent()
                end
            else
                print("Missed call notification was not sent:", errMsg)
                logger:error("Missed call notification was not sent:", errMsg)
            end
        end
    end
    
    if action and action.name == "RECORD" then
        local command = "nohup wms_notify -simplerec %q %q %q %q %q %q> /tmp/notify.out 2>&1 &"
        require"core.pbxutils".exec(command, action.targetfile, action.origformat, action.attach_format, action.messagecode, action.cid_attrs_mail, action.language)
    end
    
    if action and action.name == "LOGIN" then
        require"stdService.login.setUser".sync(action)
    end
    
    if action and action.name == "SENDFAX" and channel.DIALSTATUS:get() ~= "ANSWER" then
        -- executed in case wrong DIALSTATUS in outgoingfax
        app.MacroExclusive("faxcounter", "--")
        local faxnumberid = action.faxnumberid
        channel.GLOBAL(faxnumberid.."_pb"):set("NULL") --remove global variable
        channel.GLOBAL(faxnumberid.."_stat"):set("NULL") --remove global variable
    end

    local callClass = channel["callclass"]:get() or ""
    if channel["DIALSTATUS"]:get() == "ANSWER" and callClass ~= "" then
        local srcip = channel["X-SRC-ADDRESS"]:get() or "-"
        local src = channel.CDR("c_from"):get() or channel.CALLERID("num"):get() or "-"
        local duration = channel["ANSWEREDTIME"]:get() or ""
        duration = "" ~= duration and duration or "0"
        local callid = channel["BRIDGEPVTCALLID"]:get() or "-"
        local userAgent = channel.SIP_HEADER("User-Agent"):get() or ""
        local dialedpeer = channel["DIALEDPEERNUMBER"]:get() or ""
        dialedpeer = "" ~= dialedpeer and dialedpeer or "-"
        local trunkname, dst = string.match(dialedpeer, "(.*)%/(.*)")
        dst = dst or dialedpeer
        if #dst >=4 and #dst <= 6 then
            dst = string.sub(dst, 1, #dst - 3) .. "XX" .. string.sub(dst, -1)
        elseif #dst > 6 then
            dst = string.sub(dst, 1, #dst - 5) .. "XXXX" .. string.sub(dst, -1)
        end

        logger:info("[SECURITY] HangupHandler Src-Addr: [%s]; UA: [%s]; Duration: [%s]; CallId: [%s]; Trunk: [%s]; A-number: [%s]; B-number: [%s]; CallClass: [%s];", 
                    srcip, userAgent, duration, callid, trunkname or dialedpeer, src, dst, callClass)
    end
end;

---- IVR service ----
local function app_ivr_parse_params(params)
    local ret_params = {}
    ret_params["sound"], ret_params["times"], ret_params["timeout"], ret_params["backcode"], ret_params["context"], ret_params["exten"], ret_params["priority"] = params:match("(.-)|(.-)|(.-)|(.-)|(.-)|(.-)|(.-)")
    if ret_params["context"] == " " then ret_params["context"] = "" end
    if ret_params["exten"] == "default" then ret_params["exten"] = "s" end
    if tonumber(ret_params["times"]) == nil then ret_params["times"] = 1 end
    if tonumber(ret_params["timeout"]) == nil then
        ret_params["timeout"] = 0
    end
    return ret_params
end

local function app_ivr_worker(ivr_struct, ivr_sequence, ivr_menu)
    local app, channel = app, channel
    print("Current key sequence: " .. ivr_sequence)
    -- Go to if action
    if ivr_struct["command"] ~= "node" then
        to_logfile ("Goto", ivr_struct["parameters"]["context"], ivr_struct["parameters"]["exten"], "1")
        app.Goto(ivr_struct["parameters"]["context"], ivr_struct["parameters"]["exten"], "1")
        -- Submenu
    else
        local sound = ivr_struct["parameters"]["sound"]
        local tsound = {}
        local i = tonumber(ivr_struct["parameters"]["times"])
        -- Repeat sound specified times
        while i >= 1 do
            table.insert(tsound, sound)
            i = i-1
        end
        sound = table.concat(tsound, "&")
        -- Read pad
        print("Timeout: " .. ivr_struct["parameters"]["timeout"])
        to_logfile ("Read", "choice", sound, 1, "", 0, ivr_struct["parameters"]["timeout"])
        app.Read("choice", sound, 1, "", 0, ivr_struct["parameters"]["timeout"])
        local choice = channel["choice"]:get()
        print("Pressed key: " .. choice)
        
        -- Terminate execution if pressed 'end call'
        if check_hangup() then
            to_logfile ("Hangup", "")
            return app.Hangup()
        end
        
        -- Key pressed
        if choice ~= "" then
            -- If readed digit is not back code
            if choice ~= ivr_struct["parameters"]["backcode"] then
                if ivr_sequence == "root" then ivr_sequence = "" end
                ivr_sequence = ivr_sequence .. choice
                -- Back code
            else
                ivr_sequence = ivr_sequence:sub(1, ivr_sequence:len()-1)
                if ivr_sequence == "" then ivr_sequence = "root" end
            end
            -- Repeat current menu if unassigned key pressed
            if ivr_menu[ivr_sequence] ~= nil then
                app_ivr_worker(ivr_menu[ivr_sequence], ivr_sequence, ivr_menu)
            else
                ivr_sequence = ivr_sequence:sub(1, ivr_sequence:len()-1)
                app_ivr_worker(ivr_struct, ivr_sequence, ivr_menu)
            end
            -- Timeout action
        else
            if ivr_struct["parameters"]["context"] ~= "" and ivr_struct["parameters"]["exten"] ~= "" then
                to_logfile ("Goto", ivr_struct["parameters"]["context"], ivr_struct["parameters"]["exten"], "1")
                app.Goto(ivr_struct["parameters"]["context"], ivr_struct["parameters"]["exten"], "1")
            else
                to_logfile ("Hangup", "")
                app.Hangup()
            end
        end
    end
end

local function app_ivr_get_full_number(app_ivr_full_number, rows, needle)
    if rows[needle]["parent_id"] ~= "0" then
        app_ivr_full_number = rows[needle]["digit"] .. app_ivr_full_number
        app_ivr_full_number = app_ivr_get_full_number(app_ivr_full_number,rows, rows[needle]["parent_id"])
    end
    return app_ivr_full_number
end

function application_ivr(ivrname)
    local connection = require"core.wsqlite":new("wmsdb")
    local sql = "SELECT treeid FROM ivr_tree WHERE number=%q AND parent_id='0';"
    local cursor = connection:execute (sql, ivrname)
    local treeid = cursor:fetch(); cursor:close()

    if not treeid then
        print("IVR does not exist")
        to_logfile("IVR does not exist", hcause.CHANUNAVAIL)
        app.Hangup(hcause.CHANUNAVAIL)
    end

    local tmp = {}
    local app_ivr_full_number = ""
    local ivr_sequence = ""
    local ivr_menu = {}

    local sql = "select id as node_id, parent_id, command, parameters, number, treeid from ivr_tree where treeid=%d order by parent_id, number ASC"
    for node_id, parent_id, command, parameters, number, treeid in connection:rows(sql, tonumber (treeid)) do
        parent_id = tostring(parent_id)
        node_id   = tostring(node_id)
        if number == ivrname and parent_id == "0" then
            tmp[node_id] = {
                parent_id = "0",
                number = ""
            }
            ivr_menu["root"] = {
                ["parent_id"] = "0",
                ["command"] = "node",
                ["parameters"] = app_ivr_parse_params(parameters),
                ["digit"] = number,
                ["number"] = ""
            }
        else
            tmp[node_id] = {
                ["parent_id"] = parent_id,
                ["digit"] = number
            }
            app_ivr_full_number = app_ivr_get_full_number("", tmp, node_id)
            ivr_menu[app_ivr_full_number] = {
                ["parent_id"] = "0",
                ["command"] = command,
                ["parameters"] = app_ivr_parse_params(parameters),
                ["digit"] = number
            }
        end
    end
    connection:close()
    app_ivr_worker(ivr_menu["root"], ivr_sequence, ivr_menu)
end
